/*
 * arch/arm/mach-cxd4132/wdt.c
 *
 * Watchdog timer
 *
 * Copyright 2010 Sony Corporation
 *
 * This code is based on kernel/softlockup.c
 */
/*
 * Detect Soft Lockups
 *
 * started by Ingo Molnar, Copyright (C) 2005, 2006 Red Hat, Inc.
 *
 * this code detects soft lockups: incidents in where on a CPU
 * the kernel does not reschedule for 10 seconds or more.
 */
#include <linux/mm.h>
#include <linux/cpu.h>
#include <linux/nmi.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/freezer.h>
#include <linux/kthread.h>
#include <linux/lockdep.h>
#include <linux/notifier.h>
#include <linux/module.h>
#include <linux/sysctl.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <mach/hardware.h>
#include <asm/hardware/arm_scu.h>
#include <asm/hardware/arm_twd.h>
#include <mach/regs-timer.h>
#include <mach/regs-permisc.h>
#include <mach/regs-misc.h>
#include <mach/pm_export.h>

/* WDT mode */
#define WDT_OFF  0
#define WDT_ON   1
static unsigned int wdt_mode = WDT_OFF;
module_param_named(mode, wdt_mode, uint, S_IRUSR);
static unsigned int wdt_timeout = 0;
module_param_named(timeout, wdt_timeout, uint, S_IRUSR);

/*---------- TIMER IP ----------*/
/* WD clock = 1MHz */
#define CXD4132_WDT_CFG (TMCK_DIV4|TMCS_ONESHOT|TMRST)
static unsigned int cxd4132_wdt_count; /* TIMER IP reload value */

static void cxd4132_wdt_setup(void)
{
	writel(CXD4132_WDT_CFG, VA_TIMER(CXD4132_WDT)+CXD4105_TIMERCTL);
	writel(cxd4132_wdt_count, VA_TIMER(CXD4132_WDT)+CXD4105_TIMERCMP);
}

static void cxd4132_wdt_start(void)
{
	writel(CXD4132_WDT_CFG|TMST, VA_TIMER(CXD4132_WDT)+CXD4105_TIMERCTL);
}

static void cxd4132_wdt_stop(void)
{
	writel(0, VA_TIMER(CXD4132_WDT)+CXD4105_TIMERCTL);
	writel(TMCLR|TMINTCLR, VA_TIMER(CXD4132_WDT)+CXD4105_TIMERCLR);
}

static inline void cxd4132_wdt_restart(void)
{
	writel(TMCLR, VA_TIMER(CXD4132_WDT)+CXD4105_TIMERCLR);
}

/*---------- MPCore WDT ----------*/
static unsigned int cpu_wdt_prescaler;
static unsigned int cpu_wdt_count;

#define WDT_CTRL (((cpu_wdt_prescaler-1)<<TWD_TIMER_CONTROL_PRESCALER_SHIFT)|TWD_WDOG_CONTROL_WDMODE)

static void cpu_wdt_setup(int cpu)
{
	writel(cpu_wdt_count, VA_LOCALTIMER_X(cpu)+TWD_WDOG_COUNTER);
}

static void cpu_wdt_start(int cpu)
{
	writel(WDT_CTRL|TWD_TIMER_CONTROL_ENABLE, VA_LOCALTIMER_X(cpu)+TWD_WDOG_CONTROL);
}

static void cpu_wdt_stop(int cpu)
{
	writel(TWD_WDOG_MAGIC1, VA_LOCALTIMER_X(cpu)+TWD_WDOG_DISABLE);
	writel(TWD_WDOG_MAGIC2, VA_LOCALTIMER_X(cpu)+TWD_WDOG_DISABLE);
	writel(0, VA_LOCALTIMER_X(cpu)+TWD_WDOG_CONTROL);
}

static inline void cpu_wdt_restart(int cpu)
{
	writel(cpu_wdt_count, VA_LOCALTIMER_X(cpu)+TWD_WDOG_LOAD);
}

static int cpu_wdt_stat(int cpu)
{
	return readl(VA_LOCALTIMER_X(cpu) + TWD_WDOG_RESETSTAT) & 1;
}

static void cpu_wdt_clear(int cpu)
{
	writel(1, VA_LOCALTIMER_X(cpu) + TWD_WDOG_RESETSTAT);
}

/*-----------------------------------------------------------*/

/*
 * Touches watchdog from sleep.S
 */
void watchdog_touch(void)
{
	if (WDT_OFF == wdt_mode)
		return;

	cpu_wdt_restart(0);
	cxd4132_wdt_restart();
}

/*
 * This callback runs from the timer interrupt.
 */
void watchdog_tick(void)
{
	int this_cpu;

	if (WDT_OFF == wdt_mode)
		return;

	this_cpu = smp_processor_id();
	cpu_wdt_restart(this_cpu);
	if (!this_cpu) {
		cxd4132_wdt_restart();
	}
}

static int __cpuinit
cpu_callback(struct notifier_block *nfb, unsigned long action, void *hcpu)
{
	int hotcpu = (unsigned long)hcpu;

	switch (action) {
	case CPU_ONLINE:
	case CPU_ONLINE_FROZEN:
		cpu_wdt_stop(hotcpu);
		cpu_wdt_clear(hotcpu);
		cpu_wdt_setup(hotcpu);
		cpu_wdt_start(hotcpu);
		break;
#ifdef CONFIG_HOTPLUG_CPU
	case CPU_DEAD:
	case CPU_DEAD_FROZEN:
		cpu_wdt_stop(hotcpu);
		break;
#endif /* CONFIG_HOTPLUG_CPU */
	}
	return NOTIFY_OK;
}

static struct notifier_block __cpuinitdata cpu_nfb = {
	.notifier_call = cpu_callback
};


#define WDOGINTR_NAME "rstreq"
static int wdog_info;

static irqreturn_t wdt_intr(int irq, void *dev_id)
{
	unsigned long stat;

	stat = readl(VA_PER_RSTREQ_STAT);
	printk(KERN_ERR "WDT:RSTREQ_STAT=0x%lx\n", stat);
	if (stat & (PER_RSTREQ_ST_CPU2|PER_RSTREQ_ST_CPU1|PER_RSTREQ_ST_CPU0)){
		/* clear CPU WDT events */
		writel(PER_RSTREQ_CPUCLR, VA_PER_RSTREQ_SET);
		writel(PER_RSTREQ_CPUCLR, VA_PER_RSTREQ_CLR);
		pm_machine_reset(0);
	}
	BUG();

	return IRQ_HANDLED;
}

#define MAX_PRESCALER 256

static void wdt_setup(void)
{
	unsigned long clk, msec;

	/* CPU WDT load value */
	clk = cxd4132_localtimer_rate();
	cpu_wdt_prescaler = clk / MHZ;
	if (cpu_wdt_prescaler > MAX_PRESCALER)
		cpu_wdt_prescaler = MAX_PRESCALER;
	msec = (clk / cpu_wdt_prescaler) / MSEC_PER_SEC;
	cpu_wdt_count = msec * wdt_timeout;

	/* TIMER WD load value */
	msec = CLOCK_TICK_RATE / MSEC_PER_SEC;
	cxd4132_wdt_count = (msec + 1) * wdt_timeout; /* to fire CPU WDT before TIMER WD */
	cxd4132_wdt_stop();
	cxd4132_wdt_setup();

	/* RSTREQ intr */
	if (request_irq(IRQ_WDOG, wdt_intr, IRQF_DISABLED, WDOGINTR_NAME,
			&wdog_info)) {
		printk(KERN_ERR "WDT:ERROR:request_irq %d failed.\n", IRQ_WDOG);
	}
}

static void wdt_show(void)
{
	unsigned int cpu;

#if 0
	/* CXD4132 does not support this function. */
	printk(KERN_ERR "WDT: checking previous status.\n");
	for (cpu = 0; cpu < NR_CPUS; cpu++) {
		if (cpu_wdt_stat(cpu)) {
			printk(KERN_ERR "WDT: CPU#%d occured.\n", cpu);
			/* clear RESET status */
			cpu_wdt_clear(cpu);
		}
	}
	printk(KERN_ERR "WDT: checking done.\n");
#endif
}

/* pre_smp_initcalls */
static int __init wdt_init(void)
{
	unsigned long modepin;
	void *cpu = (void *)(long)smp_processor_id();

	if (WDT_OFF == wdt_mode) {
		return 0;
	}
	modepin = readl(VA_MISCCTRL(MISCCTRL_MODEREAD));
	if (modepin & MODEREAD_DBG0) {
		wdt_mode = WDT_OFF;
		return 0;
	}

	wdt_show();
	wdt_setup();

	cpu_callback(&cpu_nfb, CPU_ONLINE, cpu);
	cxd4132_wdt_start();

	register_cpu_notifier(&cpu_nfb);

	return 0;
}
early_initcall(wdt_init);

void cxd4132_wdt_resume(void)
{
	if (WDT_OFF == wdt_mode) {
		return ;
	}
	cxd4132_wdt_stop();
	cxd4132_wdt_setup();
	cpu_callback(&cpu_nfb, CPU_ONLINE, (void *)0);
	cxd4132_wdt_start();
}
