/*
 *  linux/arch/arm/mach-realview/localtimer.c
 *
 *  Copyright (C) 2002 ARM Ltd.
 *  All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/smp.h>
#include <linux/jiffies.h>
#include <linux/percpu.h>
#include <linux/clockchips.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/seq_file.h>
#include <linux/snsc_boot_time.h>
#ifdef CONFIG_EJ_USE_TICK_TIMER
#include <linux/profile.h>
#endif

#include <asm/mach/time.h>
#include <asm/hardware/arm_twd.h>
#include <mach/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <mach/gic.h>
#include <mach/time.h>
#include <mach/localtimer.h>

#if NR_CPUS > 4
# error "localtimer.c: NR_CPUS > 4"
#endif

#ifndef CONFIG_EJ_USE_TICK_TIMER
static DEFINE_PER_CPU(struct clock_event_device, local_clockevent);
#endif /* !CONFIG_EJ_USE_TICK_TIMER */

extern unsigned long cxd41xx_tick_timing;

void local_timer_interrupt(void)
{
#ifndef CONFIG_EJ_USE_TICK_TIMER
	struct clock_event_device *clk = &__get_cpu_var(local_clockevent);
#endif /* !CONFIG_EJ_USE_TICK_TIMER */

	if (cxd41xx_tick_timing & 1) {
		BOOT_TIME_ADD1("localtimer");
	}

#ifdef CONFIG_EJ_USE_TICK_TIMER
	update_process_times(user_mode(get_irq_regs()));
	profile_tick(CPU_PROFILING);
#else /* CONFIG_EJ_USE_TICK_TIMER */
	clk->event_handler(clk);
#endif /* CONFIG_EJ_USE_TICK_TIMER */
}

#ifdef CONFIG_LOCAL_TIMERS

#define TWD_BASE		VA_LOCALTIMER

#ifdef CONFIG_EJ_USE_TICK_TIMER
static void local_timer_start_periodic(void)
{
	__raw_writel(TWD_TIMER_CONTROL_ENABLE|TWD_TIMER_CONTROL_IT_ENABLE|TWD_TIMER_CONTROL_PERIODIC, TWD_BASE+TWD_TIMER_CONTROL);
}
#endif /* CONFIG_EJ_USE_TICK_TIMER */

static unsigned long mpcore_timer_rate;

static void __cpuinit twd_calibrate_rate(unsigned int cpu)
{
	const void __iomem __force *twd_base = TWD_BASE;
	unsigned long load;
#ifdef CONFIG_EJ_LOCAL_TIMER_PHASE_SHIFT
	unsigned long pos, offset_abs, offset_rel;
	unsigned long flags;
#endif /* CONFIG_EJ_LOCAL_TIMER_PHASE_SHIFT */
#ifdef CONFIG_QEMU
	unsigned long count;
	u64 waitjiffies;

	/*
	 * If this is the first time round, we need to work out how fast
	 * the timer ticks
	 */
	if (mpcore_timer_rate == 0) {
		printk("Calibrating local timer... ");

		/* Wait for a tick to start */
		waitjiffies = get_jiffies_64() + 1;

		while (get_jiffies_64() < waitjiffies)
			udelay(10);

		/* OK, now the tick has started, let's get the timer going */
		waitjiffies += 5;

				 /* enable, no interrupt or reload */
		__raw_writel(0x1, twd_base + TWD_TIMER_CONTROL);

				 /* maximum value */
		__raw_writel(0xFFFFFFFFU, twd_base + TWD_TIMER_COUNTER);

		while (get_jiffies_64() < waitjiffies)
			udelay(10);

		count = __raw_readl(twd_base + TWD_TIMER_COUNTER);

		mpcore_timer_rate = (0xFFFFFFFFU - count) * (HZ / 5);

		printk("%lu.%02luMHz.\n", mpcore_timer_rate / 1000000,
			(mpcore_timer_rate / 100000) % 100);
	}
#else /* CONFIG_QEMU */
	mpcore_timer_rate = MPCORE_TIMER_RATE;
	printk(KERN_INFO "localtimer: %lu\n", mpcore_timer_rate);
#endif /* CONFIG_QEMU */

	load = mpcore_timer_rate / HZ;

	__raw_writel(load, twd_base + TWD_TIMER_LOAD);

#ifdef CONFIG_EJ_LOCAL_TIMER_PHASE_SHIFT
# ifdef CONFIG_EJ_SEP_TICK_AND_UPDATE
	offset_abs = (load / (NR_CPUS + 1)) * (cpu + 1);
# else
#  if 0
	offset_abs = (load / NR_CPUS) * cpu;
#  else
	offset_abs = (load >> 2) * cpu;
#  endif
# endif /* CONFIG_EJ_SEP_TICK_AND_UPDATE */

	local_irq_save(flags);
	/* current position (convert usec to clock) */
	pos = (system_timer->raw_offset() * (mpcore_timer_rate/10000))/100;
	offset_rel = offset_abs - pos + load;
	offset_rel %= load;
	__raw_writel(offset_rel, twd_base + TWD_TIMER_COUNTER);
	local_irq_restore(flags);
#endif /* CONFIG_EJ_LOCAL_TIMER_PHASE_SHIFT */

#ifdef CONFIG_EJ_USE_TICK_TIMER
	local_timer_start_periodic();
#endif /* CONFIG_EJ_USE_TICK_TIMER */
}

#ifdef CONFIG_EJ_USE_TICK_TIMER
# undef ELAPSE
/*
 * Called with interrupt disabled
 *
 *   ASSUME: do not cpu_up during calibration.
 */
void twd_calibrate_rate_all(unsigned long mhz, unsigned long period)
{
	unsigned long count[NR_CPUS];
	const void __iomem *p;
#ifdef CONFIG_EJ_LOCAL_TIMER_PHASE_SHIFT
	unsigned int pos, shift, offset_abs, offset_rel;
#endif /* CONFIG_EJ_LOCAL_TIMER_PHASE_SHIFT */
	unsigned int cpu, phys;
#ifdef ELAPSE
	unsigned long t1, t2, t3;
#endif /* ELAPSE */

#ifdef ELAPSE
	t1 = mach_read_cycles();
#endif /* ELAPSE */
	if (cxd41xx_tick_timing & 2) {
		BOOT_TIME_ADD1("twd_calibrate_rate_all");
	}

	/* adjust */
	for (cpu = 0; cpu < NR_CPUS; cpu++) {
		count[cpu] = 0;
	}
#ifdef CONFIG_EJ_LOCAL_TIMER_PHASE_SHIFT
# ifdef CONFIG_EJ_SEP_TICK_AND_UPDATE
	shift = period / (NR_CPUS + 1);
# else
#  if 0
	shift = period / NR_CPUS;
#  else
	shift = period >> 2;
#  endif
# endif /* CONFIG_EJ_SEP_TICK_AND_UPDATE */
	/* current position (convert usec to clock) */
	pos = system_timer->raw_offset() * mhz;
	if (pos >= period) {
		pos -= period;
	}
	offset_abs = 0;
# ifdef CONFIG_EJ_SEP_TICK_AND_UPDATE
	cpu = 0;
# else
	cpu = 1;
# endif /* CONFIG_EJ_SEP_TICK_AND_UPDATE */
	while (cpu < NR_CPUS) {
		offset_abs += shift;
		offset_rel = offset_abs - pos + period;
#if 0
		offset_rel %= period;
#else /* avoid divmod */
		if (offset_rel >= period)
			offset_rel -= period;
#endif
		count[cpu] = offset_rel;
		cpu++;
	}
#endif /* CONFIG_EJ_LOCAL_TIMER_PHASE_SHIFT */

#ifdef ELAPSE
	t2 = mach_read_cycles();
#endif /* ELAPSE */

	/* modify local timers */
	for_each_online_cpu(cpu) {
#if !defined(CONFIG_EJ_SEP_TICK_AND_UPDATE)
		if (0 == cpu) {
			/* BootCPU does not use localtimer */
			continue;
		}
#endif /* !CONFIG_EJ_SEP_TICK_AND_UPDATE */
		phys = phys_cpuid(cpu);
		p = VA_LOCALTIMER_X(phys);
		__raw_writel(0,          p + TWD_TIMER_CONTROL); /* stop */
		__raw_writel(period,     p + TWD_TIMER_LOAD);
		__raw_writel(count[cpu], p + TWD_TIMER_COUNTER);
	}
	/* restart local timers */
	for_each_online_cpu(cpu) {
#if !defined(CONFIG_EJ_SEP_TICK_AND_UPDATE)
		if (0 == cpu) {
			/* BootCPU does not use localtimer */
			continue;
		}
#endif /* !CONFIG_EJ_SEP_TICK_AND_UPDATE */
		phys = phys_cpuid(cpu);
		__raw_writel(TWD_TIMER_CONTROL_ENABLE|TWD_TIMER_CONTROL_IT_ENABLE|TWD_TIMER_CONTROL_PERIODIC, VA_LOCALTIMER_X(phys)+TWD_TIMER_CONTROL);
	}
#ifdef ELAPSE
	t3 = mach_read_cycles();
	printk(KERN_ERR "twd_calibrate_rate_all: %lu,%lu [us]\n",
	       mach_cycles_to_usecs(t2-t1), mach_cycles_to_usecs(t3-t2));
#endif /* ELAPSE */
}
#endif /* CONFIG_EJ_USE_TICK_TIMER */

#ifndef CONFIG_EJ_USE_TICK_TIMER
static void local_timer_set_mode(enum clock_event_mode mode,
				 struct clock_event_device *clk)
{
	unsigned long ctrl = 0;
	unsigned long flags;

	switch(mode) {
	case CLOCK_EVT_MODE_PERIODIC:
		/* timer load already set up */
		ctrl = TWD_TIMER_CONTROL_ENABLE | TWD_TIMER_CONTROL_IT_ENABLE
			| TWD_TIMER_CONTROL_PERIODIC;
		break;
	case CLOCK_EVT_MODE_ONESHOT:
		/* period set, and timer enabled in 'next_event' hook */
		ctrl = TWD_TIMER_CONTROL_IT_ENABLE | TWD_TIMER_CONTROL_ONESHOT;
		break;
	case CLOCK_EVT_MODE_RESUME:
		local_irq_save(flags);
		get_irq_chip(IRQ_LOCALTIMER)->unmask(IRQ_LOCALTIMER);
		local_irq_restore(flags);
		twd_calibrate_rate(raw_smp_processor_id());
		break;
	case CLOCK_EVT_MODE_UNUSED:
	case CLOCK_EVT_MODE_SHUTDOWN:
	default:
		break;
	}

	__raw_writel(ctrl, TWD_BASE + TWD_TIMER_CONTROL);
}

static int local_timer_set_next_event(unsigned long evt,
				      struct clock_event_device *unused)
{
	unsigned long ctrl = __raw_readl(TWD_BASE + TWD_TIMER_CONTROL);

	__raw_writel(evt, TWD_BASE + TWD_TIMER_COUNTER);
	__raw_writel(ctrl | TWD_TIMER_CONTROL_ENABLE, TWD_BASE + TWD_TIMER_CONTROL);

	return 0;
}
#endif /* !CONFIG_EJ_USE_TICK_TIMER */

/*
 * local_timer_ack: checks for a local timer interrupt.
 *
 * If a local timer interrupt has occurred, acknowledge and return 1.
 * Otherwise, return 0.
 */
int local_timer_ack(void)
{
	if (__raw_readl(TWD_BASE + TWD_TIMER_INTSTAT)) {
		__raw_writel(1, TWD_BASE + TWD_TIMER_INTSTAT);
		return 1;
	}

	return 0;
}

/*
 * Setup the local clock events for a CPU.
 */
void __cpuinit local_timer_setup(void)
{
	unsigned int cpu = smp_processor_id();
#ifndef CONFIG_EJ_USE_TICK_TIMER
	struct clock_event_device *clk = &per_cpu(local_clockevent, cpu);
#endif /* !CONFIG_EJ_USE_TICK_TIMER */
	unsigned long flags;

	twd_calibrate_rate(cpu);

#ifndef CONFIG_EJ_USE_TICK_TIMER
	clk->name		= "local_timer";
	clk->features		= CLOCK_EVT_FEAT_PERIODIC | CLOCK_EVT_FEAT_ONESHOT;
	clk->rating		= 350;
	clk->set_mode		= local_timer_set_mode;
	clk->set_next_event	= local_timer_set_next_event;
	clk->irq		= IRQ_LOCALTIMER;
	clk->cpumask		= cpumask_of(cpu);
	clk->shift		= 20;
	clk->mult		= div_sc(mpcore_timer_rate, NSEC_PER_SEC, clk->shift);
	clk->max_delta_ns	= clockevent_delta2ns(0xffffffff, clk);
	clk->min_delta_ns	= clockevent_delta2ns(0xf, clk);
#endif /* !CONFIG_EJ_USE_TICK_TIMER */

	/* Make sure our local interrupt controller has this enabled */
	local_irq_save(flags);
	get_irq_chip(IRQ_LOCALTIMER)->unmask(IRQ_LOCALTIMER);
	local_irq_restore(flags);

#ifndef CONFIG_EJ_USE_TICK_TIMER
	clockevents_register_device(clk);
#endif /* !CONFIG_EJ_USE_TICK_TIMER */

}

/*
 * take a local timer down
 */
void __cpuexit local_timer_stop(void)
{
	__raw_writel(0, TWD_BASE + TWD_TIMER_CONTROL);
}

#else	/* CONFIG_LOCAL_TIMERS */

static void dummy_timer_set_mode(enum clock_event_mode mode,
				 struct clock_event_device *clk)
{
}

void __cpuinit local_timer_setup(void)
{
	unsigned int cpu = smp_processor_id();
	struct clock_event_device *clk = &per_cpu(local_clockevent, cpu);

	clk->name		= "dummy_timer";
	clk->features		= CLOCK_EVT_FEAT_DUMMY;
	clk->rating		= 200;
	clk->set_mode		= dummy_timer_set_mode;
	clk->broadcast		= smp_timer_broadcast;
	clk->cpumask		= cpumask_of_cpu(cpu);

	clockevents_register_device(clk);
}

#endif	/* !CONFIG_LOCAL_TIMERS */
