/*--------------------------------------------------------------------------*
 * drivers/usb/scd/scd_bperi.c
 * 
 * Gadget mode operations
 * 
 * Copyright 2011 Sony Corporation
 *
 * This file is part of the HS-OTG Controller Driver SCD.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb/gadget.h>
#include <linux/usb/scd/usb_otg_control.h>
#include <linux/usb/scd/scd_ioctl.h>
#include <linux/dma-mapping.h>
#include <linux/wait.h>

#include "scd.h"
#include "scd_version.h"
#include "scd_register.h"
#include "scd_gadget.h"
#include "scd_debug.h"

/*=============================================================================
	Gadget functions declaration
=============================================================================*/
irqreturn_t scd_gadget_irq(struct scd_control *);
void scd_g_irq_gintsts(struct scd_gadget *, u32);
void scd_g_irq_daint(struct scd_gadget *, u32);
void scd_g_irq_gintsts_resume(struct scd_gadget *);
void scd_g_irq_gintsts_enumedone(struct scd_gadget *);
void scd_g_irq_gintsts_usbrst(struct scd_gadget *);
void scd_g_irq_gintsts_usbsusp(struct scd_gadget *);
void scd_g_irq_gintsts_rxflvl(struct scd_gadget *);
void scd_g_irq_gintsts_otgint(struct scd_gadget *);
void scd_g_irq_gintsts_modemis(struct scd_gadget *);

static struct scd_control *scd_gadget_probe(struct scd_device *);
static int scd_gadget_remove(struct scd_control *);

#ifdef SCD_GADGET_PM
static int scd_gadget_pm_suspend(struct scd_control *, pm_message_t);
static int scd_gadget_pm_resume(struct scd_control *);
#endif

static int scd_gadget_restart(struct scd_gadget *);
static int scd_gadget_start(struct scd_control *);
static int scd_gadget_stop(struct scd_control *);
static void scd_gadget_detect_vbus(struct scd_control *, bool);

/* endpoint ops functions */
static	int	scd_gadget_enable_ep(struct usb_ep *, 
			const struct usb_endpoint_descriptor *);
static	int	scd_gadget_disable_ep(struct usb_ep *);
static	struct usb_request	*scd_gadget_alloc_request(struct usb_ep *, gfp_t);
static	void scd_gadget_free_request(struct usb_ep *, struct usb_request *);
static	int	scd_gadget_queue(struct usb_ep *, struct usb_request *, 
			gfp_t gfp_flags);
static	int	scd_gadget_dequeue(struct usb_ep *, struct usb_request *);
static	int	scd_gadget_set_halt(struct usb_ep *, int);
static	int	scd_gadget_fifo_status(struct usb_ep *);
static	void scd_gadget_fifo_flush(struct usb_ep *);

static struct usb_ep_ops scd_gadget_ep_ops = {
	.enable				=	scd_gadget_enable_ep,
	.disable			=	scd_gadget_disable_ep,
	.alloc_request		=	scd_gadget_alloc_request,
	.free_request		=	scd_gadget_free_request,
	.queue				=	scd_gadget_queue,
	.dequeue			=	scd_gadget_dequeue,
	.set_halt			=	scd_gadget_set_halt,
	.fifo_status		=	scd_gadget_fifo_status,
	.fifo_flush			=	scd_gadget_fifo_flush
};

/* general operations */
static	int	scd_gadget_get_frame(struct usb_gadget *);
static	int	scd_gadget_wakeup(struct usb_gadget *);
static	int	scd_gadget_set_self_powered(struct usb_gadget *, int );
static	int	scd_gadget_vbus_session(struct usb_gadget *, int );
static	int	scd_gadget_vbus_draw(struct usb_gadget *, unsigned );
static	int	scd_gadget_pullup(struct usb_gadget *, int );
static	int	scd_gadget_ioctl(struct usb_gadget *, unsigned , unsigned long );

static struct usb_gadget_ops scd_gadget_ops = {
	.get_frame			=	scd_gadget_get_frame,
	.wakeup				=	scd_gadget_wakeup,
	.set_selfpowered	=	scd_gadget_set_self_powered,
	.vbus_session		=	scd_gadget_vbus_session,
	.vbus_draw			=	scd_gadget_vbus_draw,
	.pullup				=	scd_gadget_pullup,
	.ioctl				=	scd_gadget_ioctl
};

static int scd_ep_tx_enable(struct scd_gadget *, struct usb_ep *);
static int scd_ep_rx_enable(struct scd_gadget *, struct usb_ep *);
static int complete_request_force(struct usb_ep *, struct usb_request *, int);
static int scd_g_get_grxstsp(struct scd_gadget *, struct grxstsr_val *);
static int set_test_mode(struct scd_gadget *, unsigned long);
#ifdef SCD_DEV_REG
static void scdg_dev_release(struct device *);
#endif
/* extern function */
extern void scd_gadget_reset(struct scd_gadget *);
extern void scd_gadget_suspend(struct scd_gadget *);
extern void scd_gadget_resume(struct scd_gadget *);
extern void scd_gadget_dtct_spd(struct scd_gadget *);
extern void scd_gadget_disconnect(struct scd_gadget *);
extern int scd_chg_devstate(int, struct scd_gadget *);
extern void scd_g_ctrl_rx_int(struct scd_gadget *);
extern void scd_g_ctrl_tx_int(struct scd_gadget *);
extern void scd_g_rxtrans_int(struct scd_gadget *, u8);
extern void scd_g_txtrans_int(struct scd_gadget *, u8);
extern u8 scd_get_ep0stage(struct scd_gadget *);
extern int scd_g_ep0_data_out(struct scd_gadget *, gfp_t);
extern int scd_g_ep0_data_in(struct scd_gadget *, gfp_t);
extern int scd_g_ep0_status_in(struct scd_gadget *, gfp_t);
extern int scd_g_general_trns(struct scd_gadget *, u8, gfp_t);
extern int scd_fifo_map_reset(struct scd_gadget *);
extern int scd_fifo_map_config(struct scd_gadget *, unsigned long);
extern int scd_fmapinfo_init(struct scd_gadget *);
extern void scd_g_complete_ep0_request(struct scd_gadget *, int);
extern int scd_set_ctrl_substate(struct scd_gadget *, u8);
extern void scd_g_init_ep0buf(struct scd_dmaep0_buf *);
extern int scd_set_epstate(struct scd_gadget *, u8 , u8 );
extern void scd_doneque_init(void);
extern void scd_fparam_reset(struct scd_gadget *);
extern int scd_chg_ep0stage(struct scd_gadget *, int);
extern int scd_g_fifo_flush_tx(struct scd_gadget *, u8, u8);
extern int scd_g_fifo_flush_rx(struct scd_gadget *, u8, u8);
extern int scd_gadget_complete_request(struct usb_request *, int);
extern int scd_gadget_set_halt_ep0(struct scd_gadget *, int);
extern int scd_gadget_set_halt_in(struct scd_gadget *, u8, int);
extern int scd_gadget_set_halt_out(struct scd_gadget *, u8, int);
#ifdef SCD_POLL_ON
extern int scd_reset_poll_state(struct scd_gadget *, int);
#endif
extern void scd_g_irq_gintsts_goutnakeff(struct scd_gadget *);
extern void dma_unmap_req(struct scd_device *, struct usb_request *);
extern u8 scd_g_get_devstate(struct scd_gadget *);
extern int scd_gadget_complete_request_nolock(struct usb_request *, int);

int scd_g_free_lastbuf(struct scd_gadget *, struct usb_request *, u8);
int scd_g_chk_lastbuf(struct usb_request *);

/* extern variable */
extern struct scd_dmaep0_buf	scd_dma_ep0buf;

/*=============================================================================
	Gadget endpoints
=============================================================================*/
/* Global info */
struct scd_gadget *scd_gadget_info;
struct scd_gadget_fmapinfo *scd_gadget_fminfo;

static char aEpNames[SCD_NUM_ALLEPS][SCD_EP_NAMESZ];
static char aGadgetName[SCD_GADGET_NAMESZ];

/* SCD_INTF for internal interface*/
struct scd_intf scd_gadget_intf;
/* rx status packet register info at the CPU transfer */
struct grxstsr_val grxstsr_info;

/*-------------------------------------------------------
	name	: scd_gadget_flush_ep_requests
	argment	: struct scd_gep *pEp
			  const int status
	return	: none
	description:
			  to operate a FIFO flush and
			   notify a request completion 
	date	: 2011/02/02
-------------------------------------------------------*/
void scd_gadget_flush_ep_requests(struct scd_gep *pEp, const int status)
{
	struct scd_gadget *sgadget = scd_gadget_info;
	struct scd_device *sdev = sgadget->g_device;
	struct scd_gep *gep;
	struct usb_request *req, *tmp;
	const u8 epnum = EP_NUMBER(pEp);
	unsigned long dflags;
	int rslt = 0;
	u8 state;

	DBG(2, "<==\n");

	gep = &sgadget->gep[epnum];
	
	if(epnum==0){	/* EP0 */
		/* EP0 ꥯȴλ */
		/* status-ECONNRESETޤϡ-ESHUTDOWN */
		scd_g_complete_ep0_request(sgadget, status);

#ifdef EP_SKIP_FLUSH2
		state = scd_g_get_devstate(sgadget);
		switch(state){
		 case SCD_GST_DFLT:
		 case SCD_GST_PULLUPON:
		 case SCD_GST_INIT:
			/* ResetFLushΤǡֺ︺Τskip롣 */
		 	break;
		 default:
		 	/* FIFOեå */
		 	scd_g_fifo_flush_tx(sgadget, epnum, SCD_G_FLUSH_NORMAL);
		 	/* FIFOեå */
		 	scd_g_fifo_flush_rx(sgadget, epnum, SCD_G_FLUSH_NORMAL);
		}
#else
		/* FIFOեå */
		scd_g_fifo_flush_tx(sgadget, epnum, SCD_G_FLUSH_NORMAL);
		/* FIFOեå */
		scd_g_fifo_flush_rx(sgadget, epnum, SCD_G_FLUSH_NORMAL);

#endif
	}else{	/* EP0ʳ */
		/* spin_lock_irqsave */
		SCD_G_SPIN_LOCK_IRQ(&scd_ReqDoneQueue.lock, dflags);

		list_for_each_entry_safe(req, tmp, &pEp->req_list, list) {
			if (req->status == -EINPROGRESS){
				/* lastbufȽ */
				rslt = scd_g_chk_lastbuf(req);
				if(rslt == true){
					/* lastbuf unmap */
					scd_g_free_lastbuf(sgadget, req, epnum);
				}
				/* unmap */
				dma_unmap_req(sdev, req);
			}
			if ((req->status == -EINPROGRESS)||(req->status == -EALREADY)) {
				DBG(0,"complete_request\n");
				/* ơؼ줿Τ˹롣 */
				req->status = status;
				/* λ */
				scd_gadget_complete_request_nolock(req, status);
			}
		}
		/* spin_unlock_irqsave */
		SCD_G_SPIN_UNLOCK_IRQ(&scd_ReqDoneQueue.lock, dflags);

#ifdef EP_SKIP_FLUSH2
		state = scd_g_get_devstate(sgadget);
		switch(state){
		/* ResetFLushΤǡֺ︺Τskip롣 */
		 case SCD_GST_DFLT:
		 case SCD_GST_PULLUPON:
		 case SCD_GST_INIT:
		 	break;
		 default:
			if(sgadget->gep[epnum].bIsTx){
				/* FIFOեå */
				scd_g_fifo_flush_tx(sgadget, epnum, SCD_G_FLUSH_NORMAL);
			}else{
				/* FIFOեå */
				scd_g_fifo_flush_rx(sgadget, epnum, SCD_G_FLUSH_NORMAL);
			}
		}
#else
		if(sgadget->gep[epnum].bIsTx){
			/* FIFOեå */
			scd_g_fifo_flush_tx(sgadget, epnum, SCD_G_FLUSH_NORMAL);
		}else{
			/* FIFOեå */
			scd_g_fifo_flush_rx(sgadget, epnum, SCD_G_FLUSH_NORMAL);
		}
#endif
	}
	DBG(2, "==>\n");
}

/*-------------------------------------------------------
	name	: scd_gadget_enable_ep
	argment	: struct usb_ep *ep
			  const struct usb_endpoint_descriptor *desc
	return	: int
	description:
			  to enable endpoint
	date	: 2010/09/10
-------------------------------------------------------*/
static int scd_gadget_enable_ep(struct usb_ep *ep, 
		const struct usb_endpoint_descriptor *desc)
{
	unsigned long flags;
	struct scd_gadget *sgadget = scd_gadget_info;
	struct scd_gep *pEp = (struct scd_gep *)ep;
	const u8 epnum = EP_NUMBER(pEp);

	DBG(2, "<==\n");

	/* EPֹ */
	DBG(2,"epnum %x\n", epnum);

	if(epnum==SCD_G_EP0){
		/* EP0̤ݡ */
		ERR("non-support ep0  \n");
		return -EINVAL;
	}
	
	SCD_G_SPIN_LOCK_IRQ(&pEp->Lock, flags);
	
	/* MAXPоEP */
	ep->maxpacket = desc->wMaxPacketSize;
	pEp->wMaxp = desc->wMaxPacketSize;
	
	DBG(2,"usb_ep maxp:%x scd_gep wMaxp:%x\n", ep->maxpacket, pEp->wMaxp);

	/* ž */
	pEp->bTransferType = desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK;

	DBG(2,"scd_gep TransferType%x\n", pEp->bTransferType);
	
	/* ž */
	pEp->bIsTx 
		= (desc->bEndpointAddress & USB_ENDPOINT_DIR_MASK) ? true : false;

	DBG(2,"scd_gep IsTx %x\n", pEp->bIsTx);
	
	/* EP ͭե饰EPͭ */
	pEp->bEpena = true;

	DBG(2,"scd_gep Epena %x\n", pEp->bEpena);

	SCD_G_SPIN_UNLOCK_IRQ(&pEp->Lock, flags);

	/* ꥹƬ */
	INIT_LIST_HEAD(&pEp->req_list);

	if(pEp->bIsTx){
		/* EP */
		scd_ep_tx_enable(sgadget, ep);
		
	}else{
		/* EP */
		scd_ep_rx_enable(sgadget, ep);
		
	}

	DBG(2, "==>\n");
	return 0;
}

/*-------------------------------------------------------
	name	: scd_gadget_disable_ep
	argment	: struct usb_ep *ep
	return	: int
	description:
			  to disable endpoint
	date	: 2011/01/29
-------------------------------------------------------*/
static int scd_gadget_disable_ep(struct usb_ep *ep)
{
	unsigned long flags;
	unsigned long dflags;
	unsigned long rwflags;
	struct scd_gadget *sgadget	= scd_gadget_info;
	struct scd_device *sdev		= sgadget->g_device;
	struct scd_gep *pEp 		= (struct scd_gep *)ep;
	struct usb_request *pReq, *tmp;
	const u8 epnum = EP_NUMBER(ep);
	u32	reg_val;

	DBG(1, "<== ep:%x \n", epnum);
	
	if(epnum==SCD_G_EP0){
		/* EP0̤ݡ */
		ERR("non-support ep0  \n");
		return -EINVAL;
	}

	SCD_G_SPIN_LOCK_IRQ(&pEp->Lock, flags);
	
	/* EP ͭե饰EP̵ */
	pEp->bEpena = false;
	
	/* 
	 * EPߥޥ 
	 */
	reg_val = SCD__READ(sdev, DAINTMSK);
	
	if(pEp->bIsTx){
		/* Tx */
		reg_val &= ~((1<<epnum)<<SCD_S_DAINTMSK_OUTEPMSK);
	}else{
		/* Rx */
		reg_val &= ~(1<<epnum);
	}
	SCD__WRITE(sdev, DAINTMSK, reg_val);

	SCD_G_SPIN_UNLOCK_IRQ(&pEp->Lock, flags);

	/* spin_lock_irqsave */
	SCD_G_SPIN_LOCK_IRQ(&scd_ReqDoneQueue.lock, dflags);

#ifdef RW_LOCK_ON
	SCD_G_READ_LOCK_IRQ(&pEp->rwLock, rwflags);
#endif	
	/* ߤΥꥯȼ */
	pReq = scd_get_currentreq(pEp);
#ifdef RW_LOCK_ON
	SCD_G_READ_UNLOCK_IRQ(&pEp->rwLock, rwflags);
#endif
	if(pReq&&pReq->status != -ESHUTDOWN){
		list_for_each_entry_safe(pReq, tmp, &pEp->req_list, list) {
			if ((pReq->status == -EINPROGRESS)||(pReq->status == -EALREADY)) {
#ifdef RW_LOCK_ON
				SCD_G_WRITE_LOCK_IRQ(&pEp->rwLock, rwflags);
#endif
				/* ꥹȺ */
				list_del(&pReq->list);
				/* λꥹȺ */
				list_del(&((struct scd_grequest *)pReq)->compl_list);
#ifdef RW_LOCK_ON
				SCD_G_WRITE_UNLOCK_IRQ(&pEp->rwLock, rwflags);
#endif
				pReq->status = -ESHUTDOWN;

				if (pReq->complete) {
					/* λΡʶ */
					pReq->complete(&pEp->end_point, pReq);
				}
			}
		}
	}
	/* spin_unlock_irqsave */
	SCD_G_SPIN_UNLOCK_IRQ(&scd_ReqDoneQueue.lock, dflags);

	if(pEp->bIsTx){
		/* FIFOեå */
		scd_g_fifo_flush_tx(sgadget, epnum, SCD_G_FLUSH_NORMAL);
	}else{
		/* FIFOեå */
		scd_g_fifo_flush_rx(sgadget, epnum, SCD_G_FLUSH_NORMAL);
	}
	SCD_G_SPIN_LOCK_IRQ(&pEp->Lock, flags);

	/* EPžֽ */
	sgadget->gep[epnum].bState 	  = SCD_EP_IDLE;
	sgadget->gep[epnum].bSubState = SCD_EPSUB_IDLE;

	SCD_G_SPIN_UNLOCK_IRQ(&pEp->Lock, flags);
	DBG(1,"==>\n");
	return 0;
}

/*-------------------------------------------------------
	name	: scd_gadget_alloc_request
	argment	: struct usb_ep *ep 
		  	  gfp_t gfp_flags
	return	: usb_request *
	description: 
			  to get memory for a request infomation
	date	: 2010/03/30
-------------------------------------------------------*/
static struct usb_request 
	*scd_gadget_alloc_request(struct usb_ep *ep, gfp_t gfp_flags)
{
	struct scd_grequest *pReq = NULL;
	
	DBG(2, "<==\n");
	
	/* ꥯΰ */
	pReq = kmalloc(sizeof(struct scd_grequest), gfp_flags);
	if (pReq) {
		/* ΰ */
		memset(pReq, 0, sizeof(struct scd_grequest));
		
		/* С */
		INIT_LIST_HEAD(&pReq->request.list);
		INIT_LIST_HEAD(&pReq->compl_list);

		/* 㥹 EPֹ */
		pReq->bEp = EP_NUMBER(ep);
		pReq->bAttribute = SCD_REQ_NONE;
	}
	DBG(2, "==> pReq:%p\n",pReq);
	return (struct usb_request *)pReq;
}

/*-------------------------------------------------------
	name	: scd_gadget_free_request
	argment	: struct usb_ep *ep 
			  struct usb_request *req
	return	: none
	description: to free memory for a request infomation
	date	: 2010/03/30
-------------------------------------------------------*/
static void scd_gadget_free_request(struct usb_ep *ep, struct usb_request *req)
{
	DBG(2, "<== pReq:%p\n", req);
	if(req){
		/* scd_gadget_alloc_requestǼΰβ
		 */
		kfree(req);
	}
	DBG(2, "==>\n");
}

/*-------------------------------------------------------
	name	: scd_gadget_queue
	argment	: struct usb_ep *pEp 
			  struct usb_request *pReq
			  gfp_t gfp_flags
	return	: int
	description: to queue data to the ep list 
	date	: 2011/01/29
-------------------------------------------------------*/
static int scd_gadget_queue(
	struct usb_ep *pEp, struct usb_request *pReq, gfp_t gfp_flags)
{
	struct scd_gadget *sgadget = scd_gadget_info;
	struct scd_gep *gep = (struct scd_gep *)pEp;
	const u8 epnum	= EP_NUMBER(pEp);
	u8	stage;
	int rslt=0;
	unsigned long flags;
	unsigned long rwflags;

	DBG(1, "<== pReq:%p ep:%x \n", pReq,epnum);

	if(epnum >= sgadget->bEps+1 ){
		ERR("error the number of ep is out of range. \n");
		DBG(1,"==>\n");
		return -EINVAL;
	}
	SCD_G_SPIN_LOCK_IRQ(&gep->Lock, flags);
#ifdef RW_LOCK_ON
	SCD_G_WRITE_LOCK_IRQ(&gep->rwLock, rwflags);
#endif
	pReq->actual = 0;
	pReq->status = -EALREADY;	/* λ/status */
	((struct scd_grequest *)pReq)->bEp = epnum;

	DBG(1,"zero:%x short:%x\n", pReq->zero, pReq->short_not_ok);
	if(pReq->length == 0 || pReq->length%gep->wMaxp!=0){
		/* zeroĹե饰 */
		pReq->zero = 0;
		DBG(1,"zero flag:%x \n",pReq->zero);
	}
	/* usb_request list */
	INIT_LIST_HEAD(&pReq->list);
	INIT_LIST_HEAD(&((struct scd_grequest *)pReq)->compl_list);
	
	/* EPꥹȥơ˥ꥯȤɲ */
	list_add_tail(&(pReq->list), &(gep->req_list));
#ifdef RW_LOCK_ON
	SCD_G_WRITE_UNLOCK_IRQ(&gep->rwLock, rwflags);
#endif	
#ifdef RW_LOCK_ON
	SCD_G_READ_LOCK_IRQ(&gep->rwLock, rwflags);
#endif	
	/* ߤΥꥯȤ Ƭݤ*/
	if(pReq != scd_get_currentreq(gep)){
		SCD_G_SPIN_UNLOCK_IRQ(&gep->Lock, flags);
#ifdef RW_LOCK_ON
		SCD_G_READ_UNLOCK_IRQ(&gep->rwLock, rwflags);
#endif	
		DBG(1, "==> queuing only \n");
		return 0;
	}
#ifdef RW_LOCK_ON
	SCD_G_READ_UNLOCK_IRQ(&gep->rwLock, rwflags);
#endif	
	DBG(2,"gep->bEpena:%x pReq:%p\n",gep->bEpena, pReq);
	if (gep->bEpena && pReq){
		/* EPͭ  ꥯȤ¸ߤ롣 */
#ifdef RW_LOCK_ON
		SCD_G_WRITE_LOCK_IRQ(&gep->rwLock, rwflags);
#endif
		pReq->status = -EINPROGRESS;	/* λ/Ϥstatus */

#ifdef RW_LOCK_ON
		SCD_G_WRITE_UNLOCK_IRQ(&gep->rwLock, rwflags);
#endif
		SCD_G_SPIN_UNLOCK_IRQ(&gep->Lock, flags);
		if(epnum == 0){
			/* EP0ơȽ */
			stage = scd_get_ep0stage(sgadget);
			switch(stage){
			case SCD_EP0_STAGE_DATA_OUT:
				/* ǡOUT */
				rslt = scd_g_ep0_data_out(sgadget, gfp_flags);
				break;
			case SCD_EP0_STAGE_DATA_IN:
				/* ǡIN */
				rslt = scd_g_ep0_data_in(sgadget, gfp_flags);
				break;
			case SCD_EP0_STAGE_STATUS_IN:
				/* ơIN */
				rslt = scd_g_ep0_status_in(sgadget, gfp_flags);
				break;
			default:
				rslt = -1;
				ERR("no chenge to %x \n",stage);
			}
			if(rslt<0){
				/* requestʤä-ENOENTȤʤ뤬ǤϤꤨʤ*/
				ERR("no request or error:%x in ep0\n",rslt);
				/* ̤žλ */
				if(pReq){
					pReq->status = -EPERM;	/*  */
					scd_g_complete_ep0_request(sgadget, pReq->status);
				}
			}
		}else if(epnum < sgadget->bEps+SCD_NUM_EP0){
			/* EPɥ쥹ξ¤bEpsʲǤ뤳 */
			/* ž */
			rslt = scd_g_general_trns(sgadget, epnum, gfp_flags);
			if(rslt<0){
				/* requestʤä-ENOENTȤʤ뤬ǤϤꤨʤ*/
				ERR("no request or error:%x in ep %d\n",rslt, epnum);
				/* ̤žλ */
				if(pReq){
					DBG(0,"complete_request\n");
					pReq->status = -EPERM;	/*  */
					scd_gadget_complete_request(pReq, pReq->status);
				}
			}
		}else{
			/* ƬƤƤΤǾĹäݤ */
			ERR("the number of eps over\n");
			rslt = -EINVAL;
		}
	}else{
		if((!gep->bEpena)&&(epnum == SCD_G_EP0)){
			rslt = -EINVAL;
		}
		SCD_G_SPIN_UNLOCK_IRQ(&gep->Lock, flags);
	}
	DBG(1, "==> rslt:%x\n",rslt);
	return rslt;
}
/*-------------------------------------------------------
	name	: scd_gadget_dequeue
	argment	: struct usb_ep *ep
			  struct usb_request *pRequest
	return	: int 
	description: to delete a request queued 
	date	: 2010/04/15
-------------------------------------------------------*/
static int scd_gadget_dequeue(struct usb_ep *ep, struct usb_request *pRequest)
{
	int rslt;

	DBG(2, "<==\n");

	/* ꥯȤδλ */
	rslt = complete_request_force(ep, pRequest, -ECONNRESET);

	DBG(2, "==>\n");
	return rslt;
}

/*-------------------------------------------------------
	name	: complete_request_force
	argment	: struct usb_ep *ep
			  struct usb_request *pRequest
			  int status
	return	: int 
	description: the core function to delete a request queued 
	date	: 2011/02/02
-------------------------------------------------------*/
static int complete_request_force(struct usb_ep *ep, 
			struct usb_request *pRequest, int status)
{
	struct scd_gadget *sgadget = scd_gadget_info;
	struct scd_device *sdev = sgadget->g_device;
	struct scd_gep *pEp = (struct scd_gep *)ep;
	struct usb_request *req, *tmp;
	const u8 epnum	= EP_NUMBER(ep);
	const u8 mode	= sgadget->bTransMode;
	unsigned long flags;
	unsigned long dflags;
	unsigned long rwflags;
	int rslt = 0;
	u32 reg_val;

	DBG(2, "<== epnum:%d\n", epnum);

	if(epnum >= sgadget->bEps+1 ){
		ERR("error the number of ep is out of range. \n");
		DBG(1, "==>\n");
		return -EINVAL;
	}
#ifdef RW_LOCK_ON
	SCD_G_READ_LOCK_IRQ(&pEp->rwLock, rwflags);
#endif
	/* ߤΥꥯȼ */
	req = scd_get_currentreq(pEp);
#ifdef RW_LOCK_ON
	SCD_G_READ_UNLOCK_IRQ(&pEp->rwLock, rwflags);
#endif	
	if(!req){
		/* оݥꥯȤʤ */
		rslt = -ENOENT;
		DBG(1,"no req \n");
		goto skip;
	}
	if(epnum==SCD_G_EP0){	/* EP0 */
		/* EP0ơ */
		req->status = status;

		/* EP0ꥯȴλ */
		/* status-ECONNRESET */
		scd_g_complete_ep0_request(sgadget, status);

		/* FIFO Fluesh */
		/* FIFOեå */
		scd_g_fifo_flush_tx(sgadget, epnum, SCD_G_FLUSH_NORMAL);
		
		/* FIFOեå */
		scd_g_fifo_flush_rx(sgadget, epnum, SCD_G_FLUSH_NORMAL);

		/* ơѹ */
		rslt = scd_chg_ep0stage(sgadget, SCD_EP0_STAGE_SETUP);
		if(rslt){
			/* 顼Ϥꤨʤ */
			ERR("==> error ep0stage not changed to SETUP \n");
			rslt = -EPERM;
			goto skip;
		}
		/* SubStage */
		rslt = scd_set_ctrl_substate(sgadget, SCD_EP_IDLE);
		if(rslt){
			/* 顼Ϥꤨʤ */
			ERR("==> abnormal error  not chaneged to IDLE \n");
			rslt = -EPERM;
			goto skip;
		}
		if(mode == SCD_GADGET_TRANS_DMA){ /* DMAž⡼ */
			/* EP0Хåե */
			scd_g_init_ep0buf(&scd_dma_ep0buf);

			/* Хɥ쥹 */
			scd_dma_ep0buf.dma_buf = (void *)dma_map_single(sdev->dev, scd_dma_ep0buf.buf, sizeof(*(scd_dma_ep0buf.buf))*SCD_DMA_EP0_SZ, DMA_BIDIRECTIONAL);

			/* DMAХåե */
			SCD__WRITE(sdev, DOEPDMA(SCD_G_EP0), scd_dma_ep0buf.dma_buf);

			/* žĹ */
			pEp->dwLen = 0;

			SCD_G_SPIN_LOCK_IRQ(&pEp->Lock, flags);	/* lock */

			/* EPEnaʺsetupѡ */
			reg_val = SCD__READ(sdev, DOEPCTL0);
			reg_val |= SCD_B_DOEPCTL0_EPENA;
			SCD__WRITE(sdev, DOEPCTL0, reg_val);

			SCD_G_SPIN_UNLOCK_IRQ(&pEp->Lock, flags);	/* unlock */

		}
	/*----------------------------------------------------------------*/
	}else{
		/* spin_lock_irqsave */
		SCD_G_SPIN_LOCK_IRQ(&scd_ReqDoneQueue.lock, dflags);
#ifdef RW_LOCK_ON
		SCD_G_READ_LOCK_IRQ(&pEp->rwLock, rwflags);
#endif
	/* ߤΥꥯȼ */
		req = scd_get_currentreq(pEp);
#ifdef RW_LOCK_ON
		SCD_G_READ_UNLOCK_IRQ(&pEp->rwLock, rwflags);
#endif
		/* ¾EPֹ */
		if(req==pRequest){	/* ߤΥꥯȤоݤξ */
			DBG(1,"case 1  \n");

			/* ꥯȤͭξ */
			if(pEp->bIsTx){
				/* FIFOեå */
				scd_g_fifo_flush_tx(sgadget, epnum, SCD_G_FLUSH_NORMAL);
			}else{
				/* FIFOեå */
				scd_g_fifo_flush_rx(sgadget, epnum, SCD_G_FLUSH_NORMAL);
			}
			if(mode == SCD_GADGET_TRANS_DMA){ /* DMAž⡼ */
				/* EPơ֤ */
				rslt = scd_set_epstate(sgadget, epnum, SCD_EP_IDLE);
				if(rslt){
					rslt = -EPERM;
					goto unlock_skip;
				}
				/* žĹ */
				pEp->dwLen = 0;
			}
#ifdef RW_LOCK_ON
			SCD_G_WRITE_LOCK_IRQ(&pEp->rwLock, rwflags);
#endif
			list_del(&req->list);
			list_del(&((struct scd_grequest *)req)->compl_list);
#ifdef RW_LOCK_ON
			SCD_G_WRITE_UNLOCK_IRQ(&pEp->rwLock, rwflags);
#endif
			req->status = status;

			/* lastbufȽ */
			rslt = scd_g_chk_lastbuf(req);
			if(rslt == true){
				/* lastbuf unmap */
				scd_g_free_lastbuf(sgadget, req, epnum);
			}
			/* unmap */
			dma_unmap_req(sdev, req);
			if (req->complete) {
				/* λΡʶ */
				req->complete(&pEp->end_point, req);
				DBG(1,"complete req:%p\n",req);
			}
			goto unlock_skip;
		/*----------------------------------------------------------------*/
		}else{	/* ߤΥꥯȤоݤǤʤ */
			DBG(1,"case 2  \n");
			/* ꥯȤͭξ */
			list_for_each_entry_safe(req, tmp, &pEp->req_list, list) {
				if (req==pRequest){
					if(mode == SCD_GADGET_TRANS_DMA){ /* DMAž⡼ */
						/* EPơ֤ */
						rslt = scd_set_epstate(sgadget, epnum, SCD_EP_IDLE);
						if(rslt){
							rslt = -EPERM;
							goto unlock_skip;
						}
						/* žĹ */
						pEp->dwLen = 0;
					}
#ifdef RW_LOCK_ON
					SCD_G_WRITE_LOCK_IRQ(&pEp->rwLock, rwflags);
#endif
					list_del(&req->list);
					list_del(&((struct scd_grequest *)req)->compl_list);
#ifdef RW_LOCK_ON
					SCD_G_WRITE_UNLOCK_IRQ(&pEp->rwLock, rwflags);
#endif
					req->status = status;

					if (req->complete) {
						/* λΡʶ */
						req->complete(&pEp->end_point, req);
						DBG(1,"complete req:%p\n",req);
					}
				}
				/* 롼ȴ */
				goto unlock_skip;
			}
			/* оݥꥯȤʤ */
			rslt = -ENOENT;
		}
unlock_skip:
		/* spin_unlock_irqsave */
		SCD_G_SPIN_UNLOCK_IRQ(&scd_ReqDoneQueue.lock, dflags);
	}
skip:	
	DBG(2, "==>\n");
	return rslt;
}

/*-------------------------------------------------------
	name	: scd_gadget_set_halt
	argment	: struct usb_ep *ep
			  int value
	return	: int 
	description: to handle STALL function
			  value 0: set stall & fifo flush
				    1: clear stall & toggle reset(except EP0)
	date	: 2010/05/12
-------------------------------------------------------*/
int scd_gadget_set_halt(struct usb_ep *ep, int value)
{
	struct scd_gadget *sgadget = scd_gadget_info;
	struct scd_gep *gep = (struct scd_gep *)ep;
	const u8 epnum = EP_NUMBER(ep);
	unsigned long flags;
	
	DBG(1, "<== val:%d\n",value);

	SCD_G_SPIN_LOCK_IRQ(&gep->Lock, flags);

	/* activeե饰å(1clear_stall) */
	if(!value){
		gep->bEpena = !value;
	}
	SCD_G_SPIN_UNLOCK_IRQ(&gep->Lock, flags);

	/* STALL */
	if(epnum==SCD_G_EP0){
		scd_gadget_set_halt_ep0(sgadget, value);
	}else{
		if(gep->bIsTx){
			scd_gadget_set_halt_in(sgadget, epnum, value);
		}else{
			scd_gadget_set_halt_out(sgadget, epnum, value);
		}
	}
	/* activeե饰ꥢ(0set_stall) */
	if(value){
		gep->bEpena = !value;
	}
	DBG(1, "==>\n");
	return 0;
}
/*-------------------------------------------------------
	name	: scd_gadget_fifo_status
	argment	: struct usb_ep *ep
	return	: int 
	description: non-use
	date	: 2010/02/26
-------------------------------------------------------*/
static int scd_gadget_fifo_status(struct usb_ep *ep)
{
	DBG(2, "<==>\n");
	ERR(" no support func \n");
	return -EINVAL;
}

/*-------------------------------------------------------
	name	: scd_gadget_fifo_flush
	argment	: struct usb_ep *ep
	return	: int 
	description: nonuse
	date	: 2010/02/26
-------------------------------------------------------*/
static void scd_gadget_fifo_flush(struct usb_ep *ep)
{
	struct scd_gadget *sgadget = scd_gadget_info;
	const u8 epnum = EP_NUMBER(ep);
	
	DBG(1, "<==\n");
	
	if(epnum >= sgadget->bEps+1 ){
		ERR("error the number of ep is out of range. \n");
		return;
	}
	/* EPեå׵ */
	scd_gadget_flush_ep_requests((struct scd_gep *)ep, -ECONNRESET);

	DBG(1, "==>\n");
}

/*=======================================================
 *
 * General Operations 
 *
 =======================================================*/

/*-------------------------------------------------------
	name	: scd_gadget_get_frame
	argment	: struct usb_gadget *gadget
	return	: int 
	description: nonuse
	date	: 2010/03/30
-------------------------------------------------------*/
static int scd_gadget_get_frame(struct usb_gadget *gadget)
{
	u32 reg_val;
	struct scd_gadget *sgadget = scd_gadget_info;
	struct scd_device *sdev = sgadget->g_device;

	DBG(2, "<==\n");
	
	/* Get the Frame Number */
	reg_val = SCD__READ(sdev, DSTS);

	DBG(2, "==>\n");
	return (reg_val>>SCD_S_DSTS_SOFFN)&SCD_M_DSTS_SOFFN;
}

/*-------------------------------------------------------
	name	: scd_gadget_wakeup
	argment	: struct usb_gadget *gadget
	return	: int 
	description: nonuse
	date	: 2010/03/30
-------------------------------------------------------*/
static int scd_gadget_wakeup(struct usb_gadget *gadget)
{
	DBG(2, "<==>\n");
	INFO("non-support wakeup\n");
	return 0;
}

/*-------------------------------------------------------
	name	: scd_gadget_set_self_powered
	argment	: struct usb_gadget *gadget, int is_selfpowered
	return	: int
	description: nonuse
	date	: 2010/03/09
-------------------------------------------------------*/
static int scd_gadget_set_self_powered(struct usb_gadget *gadget, int is_selfpowered)
{
	DBG(2, "<==>\n");
	return 0;
}


/*-------------------------------------------------------
	name	: scd_gadget_vbus_session
	argment	: struct usb_gadget *gadget, int is_active
	return	: int
	description: nonuse
	date	: 2010/03/09
-------------------------------------------------------*/
static int scd_gadget_vbus_session(struct usb_gadget *gadget, int is_active)
{
	DBG(2, "<==>\n");
	return -EINVAL;
}

/*-------------------------------------------------------
	name	: scd_gadget_vbus_draw
	argment	: struct usb_gadget *gadget, unsigned mA
	return	: int
	description: nonuse
	date	: 2010/03/09
-------------------------------------------------------*/
static int scd_gadget_vbus_draw(struct usb_gadget *gadget, unsigned mA)
{
	DBG(2, "<==>\n");
	INFO("vbus draw No-support now! \n");
	return 0;
}

/*-------------------------------------------------------
	name	: scd_gadget_pullup
	argment	: struct usb_gadget *gadget, int is_on
	return	: int
	description: control softdiscon for pullup
	date	: 2010/03/09
-------------------------------------------------------*/
static int scd_gadget_pullup(struct usb_gadget *gadget, int is_on)
{
	u32 reg_val;
	struct scd_gadget *sgadget = scd_gadget_info;
	struct scd_device *sdev = sgadget->g_device;
	unsigned long flags;
	int rslt;

	DBG(1, "<== is_on:%x\n", is_on);

	if(is_on){
		/* pullup on */

		/* soft reset */
		scd_device_reset_link(sdev, true);
		/* start_gadgetξ֤ */
		scd_gadget_restart(sgadget);
		/* ڥɥե饰 OFF  */
		sgadget->bSuspended = false;

		/* ǥХ */
		rslt = scd_chg_devstate(SCD_GST_PULLUPON_EN, sgadget);
		if(rslt<0){
			ERR("error! can not change to PULLUP ON_EN state \n");
			DBG(2, "==> err:%d\n",rslt);
			return -EPERM;
		}else if(rslt == SCD_DEVST_SKIP){
			DBG(1,"pullup on skip\n");
			goto skip;
		}
	}else{
		/* pullup off */

		SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);

		/* Softdiscon ON */
		reg_val = SCD__READ(sdev, DCTL);
		reg_val |= SCD_B_DCTL_SFTDISCON;
		SCD__WRITE(sdev, DCTL, reg_val);

		SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);

		/* ǥХ */
		rslt = scd_chg_devstate(SCD_GST_PULLUPOFF_EN, sgadget);
		if(rslt){
			ERR("error! can not change to PULLUP OFF_EN state \n");
			DBG(2, "==> err:%d\n",rslt);
			return -EPERM;
		}
	}
	SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);

	/* 쥸 ڥ/ꥻåȳߵ */
	reg_val = SCD__READ(sdev, GINTMSK);
	reg_val &= ~(SCD_B_GINTMSK_ERLYSUSPMSK|SCD_B_GINTMSK_WKUPINTMSK);
	reg_val |= SCD_B_GINTMSK_USBRSTMSK|SCD_B_GINTMSK_USBSUSPMSK;
	SCD__WRITE(sdev, GINTMSK, reg_val);

	SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);

	/* ǥХ */
	sgadget->bBusState = SCD_BUS_IDLE;
skip:
	DBG(2, "==>\n");

	return 0;
}
/*-------------------------------------------------------
	name	: scd_gadget_ioctl
	argment	: struct usb_gadget *gadget
			  unsigned code, 
			  unsigned long param
	return	: int
	description: ioctl function
	date	: 2010/04/14
-------------------------------------------------------*/
static int scd_gadget_ioctl(struct usb_gadget *gadget, unsigned code, unsigned long param)
{
	int rslt;
	struct scd_gadget *sgadget	= scd_gadget_info;
	
	DBG(2, "<==\n");

	switch (code) {
	case USB_IOCTL_FIFO_RESET:
		/* FIFO reset */
		DBG(2,"ioctl fifo reset \n");
		/* fparam init 
		 * resetcall˰̣ġ
		 */
		scd_fparam_reset(sgadget);

		/* a default state mapping  */
		rslt = scd_fifo_map_reset(sgadget);
		if(rslt != 0){
			ERR("ERROR a mapping in a reset!");
			DBG(2, "==>\n");
			return -EPERM;
		}			
		DBG(2, "==>\n");
		return 0;

	case USB_IOCTL_FIFO_MAP:
		/* FIFO mapping */
		if(sgadget->bFifoCnfg){
			DBG(2,"FIFO MAP start\n");
			rslt = scd_fifo_map_config(sgadget, param);
			if(rslt != 0){
				ERR("ERROR a mapping error in configuration!");
				DBG(2, "==>\n");
				return -EPERM;
			}
		}else{
			ERR("FIFO MAP not applicable now \n");
		}
		DBG(2, "==>\n");
		return 0;

	case USB_IOCTL_FIFO_DUMP:
		/* FIFO info dump */
		INFO("FIFO info dump non-support now! \n");
		DBG(2, "==>\n");
		return 0;
		
	case USB_IOCTL_TESTMODE:
		if (NULL == sgadget){
			DBG(2, "==>\n");
			return -EAGAIN;
		}
		/* ƥȥ⡼ */
		rslt = set_test_mode(sgadget, param);

		DBG(2, "==>\n");
		return rslt;

	case USB_IOCTL_DMA_SET:
		INFO("DMA set non-support! \n");
		DBG(2, "==>\n");
		return 0;
	}
	DBG(2, "==>\n");

	return -EINVAL;
}

/*=======================================================
	Interrupt 
=======================================================*/

/*-------------------------------------------------------
	name	: scd_gadget_irq
	argment	: struct scd_control *control
	return	: irqreturn_t
	description: 
		first interrupt handle on gadget
	date	: 2010/07/15
-------------------------------------------------------*/
irqreturn_t scd_gadget_irq(struct scd_control *control)
{
	struct scd_gadget *sgadget; 
	struct scd_device *sdev;
	irqreturn_t status = IRQ_NONE;
	u32 reg_intval;
	u32 reg_mskval;

#ifdef IRQ_DEBUG
	u32 reg_chk;
#endif
	DBG(2, "<<==\n");
		
	/*  */
	sgadget = container_of((void *)control, struct scd_gadget, g_control);
	sdev = (struct scd_device *)sgadget->g_device;
	
	/*--- GINTSTSװ ---*/
	reg_intval	= SCD__READ(sdev, GINTSTS);
	reg_mskval	= SCD__READ(sdev, GINTMSK);
	reg_intval &= reg_mskval;
	if (reg_intval){
		DBG(2,"#GINTSTS %x\n",reg_intval);
		/* GINTSTSװ */
		scd_g_irq_gintsts(sgadget, reg_intval);
		
		status = IRQ_HANDLED;
	}
#ifdef IRQ_DEBUG
	if(reg_chk&~reg_mskval){
		DBGS(2, "error no care bit! GINTSTS:%x \n",reg_chk);
	}
#endif
	DBG(2,"==>\n");
	return status;
}
/*-------------------------------------------------------
	name	: scd_g_irq_gintsts
	argment	: struct scd_gadget *sgadget, u32 reg_val
	return	: none
	description: 
		for gintsts register
		װ棱Ĥ롣
		װͥ٤ߤ¤ؤƤ롣
	date	: 2010/05/07
-------------------------------------------------------*/
void scd_g_irq_gintsts(struct scd_gadget *sgadget, u32 reg_val)
{
	struct scd_device *sdev = sgadget->g_device;
	u32 reg_intval;
	u32 reg_mskval;

	DBG(2,"<== \n");

	if(reg_val&SCD_B_GINTSTS_USBRST){
		/* USB reset */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_USBRST);
		/* USBХreset */
		scd_g_irq_gintsts_usbrst(sgadget);
	}
	else if(reg_val&SCD_B_GINTSTS_ENUMDONE){
		/* Enumeration Done */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_ENUMDONE);
		/* enumedone߽ */
		scd_g_irq_gintsts_enumedone(sgadget);
	}
	else if(reg_val&SCD_B_GINTSTS_USBSUSP){
		/* USB suspend */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_USBSUSP);
		/* USBڥɽ */
		scd_g_irq_gintsts_usbsusp(sgadget);
	}
	else if(reg_val&SCD_B_GINTSTS_ERLYSUSP){
		/* early suspend */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_ERLYSUSP);
	}
	else if(reg_val&SCD_B_GINTSTS_WKUPINT){
		/* wakeup */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_WKUPINT);
		
		/* resume߽ */
		scd_g_irq_gintsts_resume(sgadget);
	}
	else if(reg_val&SCD_B_GINTSTS_RESETDET){
		/* reset */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_RESETDET);
	}
	else if(reg_val&SCD_B_GINTSTS_INCOMPISOO){
		/* Incompletion ISO OUT */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_INCOMPISOO);

		ERR("GINTSTS non-support Incomliso_out \n");
	}
	else if(reg_val&SCD_B_GINTSTS_INCOMPISOI){
		/* Incompletion ISO IN */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_INCOMPISOI);

		ERR("GINTSTS non-support Incomliso_in \n");
	}
	else if(reg_val&SCD_B_GINTSTS_ISOOUTDROP){
		/* ISO OUT DROP */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_ISOOUTDROP);

		ERR("non-support isoout drop \n");
	}
	else if(reg_val&SCD_B_GINTSTS_GOUTNAKEFF){
		/* out nak eff */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_GOUTNAKEFF);
		/* GlobalNAKEff߽ */
		scd_g_irq_gintsts_goutnakeff(sgadget);
	}
	else if(reg_val&SCD_B_GINTSTS_GINNAKEFF){
		/* in nak eff */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_GINNAKEFF);

		ERR("non-support IN Nak Eff \n");
	}
	else if(reg_val&SCD_B_GINTSTS_RXFLVL){
		/* RxFIFO non-empty ReadOnly */
		/* ߽ */
		scd_g_irq_gintsts_rxflvl(sgadget);
	}
	else if(reg_val&SCD_B_GINTSTS_OEPINT || reg_val&SCD_B_GINTSTS_IEPINT){
		/* OUT or IN ep  */

		/*--- DAINTװ ---*/
		reg_intval	= SCD__READ(sdev, DAINT);
		reg_mskval	= SCD__READ(sdev, DAINTMSK);
		reg_intval &= reg_mskval;
		if (reg_intval){
			/* DAINTװ */
			scd_g_irq_daint(sgadget, reg_intval);
		}
	}
	else if(reg_val&SCD_B_GINTSTS_EOPF){
		/* EOPF */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_EOPF);

		ERR("non-support eopf \n");
	}
	else if(reg_val&SCD_B_GINTSTS_SOF){
		/* SOF */
		/* װꥢ */
		SCD__WRITE(sdev,GINTSTS, SCD_B_GINTSTS_SOF);

		ERR("non-support sof \n");
	}
	else if(reg_val&SCD_B_GINTSTS_OTGINT){
		/* otg interrupt ReadOnly*/
		/* OTG߽ */
		scd_g_irq_gintsts_otgint(sgadget);
	}
	else if(reg_val&SCD_B_GINTSTS_MODEMIS){
		/* mode mismatch */
		/* ⡼ɥߥޥå߽ */
		scd_g_irq_gintsts_modemis(sgadget);
	}
	else{
		ERR("non-support event occured! gintsts:%x\n",reg_val);
	}
	DBG(2,"==> \n");
}
/*-------------------------------------------------------
	name	: scd_g_irq_gintsts_resume
	argment	: struct scd_gadget *sgadget
	return	: none
	description: 
		for resume
	date	: 2010/03/17
-------------------------------------------------------*/
void scd_g_irq_gintsts_resume(struct scd_gadget *sgadget)
{
	scd_gadget_resume(sgadget);
}

/*-------------------------------------------------------
	name	: scd_g_irq_gintsts_enumedone
	argment	: struct scd_gadget *sgadget
	return	: void
	description: 
		for enumeration done
	date	: 2010/03/19
-------------------------------------------------------*/
void scd_g_irq_gintsts_enumedone(struct scd_gadget *sgadget)
{
	scd_gadget_dtct_spd(sgadget);
}

/*-------------------------------------------------------
	name	: scd_g_irq_gintsts_usbrst
	argment	: struct scd_gadget *sgadget
	return	: none
	description: 
		for USB reset
	date	: 2010/03/17
-------------------------------------------------------*/
void scd_g_irq_gintsts_usbrst(struct scd_gadget *sgadget)
{
	scd_gadget_reset(sgadget);
}

/*-------------------------------------------------------
	name	: scd_g_irq_gintsts_usbsusp
	argment	: struct scd_gadget *sgadget
	return	: none
	description: 
		for USB suspend
	date	: 2010/03/17
-------------------------------------------------------*/
void scd_g_irq_gintsts_usbsusp(struct scd_gadget *sgadget)
{
	scd_gadget_suspend(sgadget);
}

/*-------------------------------------------------------
	name	: scd_g_irq_gintsts_rxflvl
	argment	: struct scd_gadget *sgadget
	return	: none
	description: 
		indicate RxFIFO Level
		used at Only CPU mode(non-used now)
	date	: 2010/03/17
-------------------------------------------------------*/
void scd_g_irq_gintsts_rxflvl(struct scd_gadget *sgadget)
{
	struct scd_device *sdev = sgadget->g_device; 
	u32 reg_val;
	int rslt;
	u8	state;

	DBG(2,"<==\n");
	
	if(sgadget->bTransMode == SCD_GADGET_TRANS_CPU){
		/* CPU⡼ RxFlvl߻ */
		reg_val = SCD__READ(sdev, GINTMSK);
		reg_val &= ~SCD_B_GINTMSK_RXFLVLMSK;
		SCD__WRITE(sdev, GINTMSK, reg_val);
	}
	
	/* GRXSTSP쥸ǧ */
	rslt = scd_g_get_grxstsp(sgadget, &grxstsr_info);
	if(rslt){
		ERR("error grxstsp with a invalid return value\n");
		return;
	}
	switch(grxstsr_info.epnum){
	case SCD_G_EP0:
		state = scd_get_ep0stage(sgadget);
		switch(state){
		case SCD_EP0_STAGE_SETUP:
			switch(grxstsr_info.pktsts){
			case SCD_E_GRXSTS_PKTSTS_SETUPTXCMPL:
				rslt = scd_set_ctrl_substate(sgadget, SCD_EP_SETUP_DONE);
				if(rslt) return;	/* ܥ顼 */
				break;
			case SCD_E_GRXSTS_PKTSTS_SETUPDPRX:
				rslt = scd_set_ctrl_substate(sgadget, SCD_EP_GET_SETUP);
				if(rslt) return;	/* ܥ顼 */
				break;
			default:
				ERR("packet status error %x\n", grxstsr_info.pktsts);
			}
			break;
		case SCD_EP0_STAGE_DATA_OUT:
			switch(grxstsr_info.pktsts){
			case SCD_E_GRXSTS_PKTSTS_SETUPTXCMPL:
				rslt = scd_set_ctrl_substate(sgadget, SCD_EP_SETUP_DONE);
				if(rslt) return;	/* ܥ顼 */
				/* SETUPǽ */
				INFO("setup interrupt on the way \n"); 
				/* FIFO꡼ */
				
				break;
			case SCD_E_GRXSTS_PKTSTS_SETUPDPRX:
				rslt = scd_set_ctrl_substate(sgadget, SCD_EP_GET_SETUP);
				if(rslt) return;	/* ܥ顼 */
				break;
			default:
				ERR("packet status error %x\n", grxstsr_info.pktsts);
			}
			break;
		case SCD_EP0_STAGE_STATUS_OUT:
			break;
		default:
			ERR("state error %x\n", state);
		}
		break;
	case SCD_G_EP1:
	case SCD_G_EP2:
	case SCD_G_EP3:
	case SCD_G_EP4:
		break;
	default:
		ERR("non support the number of ep: %x\n",grxstsr_info.epnum);
	}
}

/*-------------------------------------------------------
	name	: scd_g_get_grxstsp
	argment	: struct scd_gadget *sgadget,
			  struct grxstsr_val *rx_sts
	return	: int
	description: 
			  read/pop GRXSTSP	
			  this is only valid for CPU transfer mode.
	date	: 2010/04/16
-------------------------------------------------------*/
static int scd_g_get_grxstsp(struct scd_gadget *sgadget, struct grxstsr_val *rx_sts)
{
	struct scd_device *sdev = sgadget->g_device; 
	u32 reg_val;
	
	reg_val = SCD__READ(sdev, GRXSTSP);
	/* Packet Status */
	rx_sts->pktsts = (reg_val&SCD_M_GRXSTS_PKTSTS)>>SCD_S_GRXSTS_PKTSTS;

	switch(rx_sts->pktsts){
	case SCD_E_GRXSTS_PKTSTS_GOUTNAK:		/* 0x0001 */
		DBG(2, "Rx Global OUT NAK \n");
		break;
	case SCD_E_GRXSTS_PKTSTS_OUTDPRX:		/* 0x0010 */
		DBG(2, "Rx OUT Packet \n");
		break;
	case SCD_E_GRXSTS_PKTSTS_OUTTXCMPL:		/* 0x0011 */
		DBG(2, "Rx OUT completed \n");
		break;
	case SCD_E_GRXSTS_PKTSTS_SETUPTXCMPL:	/* 0x0100 */
		DBG(2, "Rx SETUP completed \n");
		break;
	case SCD_E_GRXSTS_PKTSTS_SETUPDPRX:		/* 0x0110 */
		DBG(2, "Rx SETUP \n");
		break;
	}
	/* Data PID */
	rx_sts->dpid = (reg_val&SCD_M_GRXSTS_DPID)>>SCD_S_GRXSTS_DPID;
	rx_sts->bcnt = (reg_val&SCD_M_GRXSTS_BCNT)>>SCD_S_GRXSTS_BCNT;
	rx_sts->epnum = reg_val&SCD_M_GRXSTS_EPNUM;
	
	return 0;
}

/*-------------------------------------------------------
	name	: scd_g_irq_gintsts_otgint
	argment	: struct scd_gadget *sgadget
	return	: none
	description: 
		for OTG interrupt
	date	: 2010/03/17
-------------------------------------------------------*/
void scd_g_irq_gintsts_otgint(struct scd_gadget *sgadget)
{
	INFO(" non support now \n");
}

/*-------------------------------------------------------
	name	: scd_g_irq_gintsts_modemis
	argment	: struct scd_gadget *sgadget
	return	: none
	description: 
		for Mode mismatch
		Device⡼ɤHostѤΥ쥸
		ȯ롣
		Υ顼ȯ顢modemismatch֥ե饰
		֤롣
	date	: 2010/03/17
-------------------------------------------------------*/
void scd_g_irq_gintsts_modemis(struct scd_gadget *sgadget)
{
	INFO(" non support now \n");
}
/*-------------------------------------------------------
	name	: scd_g_irq_daint
	argment	: struct scd_gadget *sgadget, u32 reg_val
	return	: none
	description: 
		for daint register
	date	: 2010/03/26
-------------------------------------------------------*/
void scd_g_irq_daint(struct scd_gadget *sgadget, u32 daint)
{
	u16 out_int, in_int;
	u8	i;

	DBG(2,"<== int %x \n", daint);
	
	/* װǼ */
	out_int	= (daint>>SCD_S_DAINT_OUTEPINT)&SCD_M_DAINT_OUTEPINT;
	in_int	= daint&SCD_M_DAINT_INEPINT;
	
	/* setup ѤEPenaåȾˤΤ */
	/* IN˽褦ˤ롣 */
	if(in_int){
		if(in_int&SCD_G_DAINT_EP0){
			/* EP0 */
			DBG(2,"inint: %x in ep0 \n",in_int);
			scd_g_ctrl_tx_int(sgadget);
		}else{
			/* EP0ʳ */
			in_int >>= 1;
			i = 1;
			while(i<sgadget->bEps+1){
				if(in_int&1){
					DBG(2,"inint: %x in ep %d \n",in_int, i);
					scd_g_txtrans_int(sgadget, i);
				}
				in_int >>= 1;
				i++;
			}
		}
	}
	if(out_int){
		if(out_int&SCD_G_DAINT_EP0){
			/* EP0 */
			DBG(2,"outint: %x in ep0 \n",out_int);
			scd_g_ctrl_rx_int(sgadget);
		}else{
			/* EP0ʳ */
			out_int >>= 1;
			i = 1;
			while(i<sgadget->bEps+1){
				if(out_int&1){
					DBG(2,"outint: %x in ep %d \n",out_int, i);
					scd_g_rxtrans_int(sgadget, i);
				}
				out_int >>= 1;
				i++;
			}
		}
	}
	DBG(2,"==> \n");
}

/*-------------------------------------------------------
	name	: usb_gadget_register_driver
	argment	: struct usb_gadget_driver *driver
	return	: int
	description: 
		to register to the Upper layer and kernel
	date	: 2010/04/30
-------------------------------------------------------*/
int usb_gadget_register_driver(struct usb_gadget_driver *driver)
{
	struct scd_gadget *sgadget = scd_gadget_info;
	int rslt;
	unsigned long flags;
#ifdef RW_LOCK_ON
	unsigned long rwflags;
#endif	
	DBG(1,"<== registering gadget %s\n", driver->function);

	if (!sgadget) {
		ERR("==>scd_gadget is not exist! \n");
		return -ENODEV;
	}
	
	if(!driver || !driver->bind || driver->speed < USB_SPEED_FULL){
		ERR("==>usb_gadget_driver paramater error! \n");
		return -EINVAL;
	}
	
	if(sgadget->pGadgetDriver){
		ERR("==> pGadgetDriver already exist! \n");
		return -EBUSY;
	}

	SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);

	/*
	 *	Gadget Driver setting
	 */
	/* ɥ饤С̾ */
	sgadget->pGadget->name	= driver->function;	
	/* gadget®HIGH*/
	sgadget->pGadget->speed	= driver->speed;
	/* ɥ饤СХ */
	driver->driver.bus		= NULL;		
#ifdef RW_LOCK_ON
	SCD_G_WRITE_LOCK_IRQ(&sgadget->rwLock, rwflags);
#endif
	/* driver */
	sgadget->pGadgetDriver	= driver;
	/* gadgetǥХɥ饤С */
	sgadget->pGadget->dev.driver = &driver->driver;
#ifdef RW_LOCK_ON
	SCD_G_WRITE_UNLOCK_IRQ(&sgadget->rwLock, rwflags);
#endif	
	SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);

	/* Bindؿ */
	rslt = driver->bind(sgadget->pGadget);
	if(rslt){
#ifdef RW_LOCK_ON
		SCD_G_WRITE_LOCK_IRQ(&sgadget->rwLock, rwflags);
#endif
		/* ǥХɥ饤С */
		sgadget->pGadgetDriver = NULL;
		/* ǥХ gadgetǥХɥ饤С */
		sgadget->pGadget->dev.driver = NULL;
#ifdef RW_LOCK_ON
		SCD_G_WRITE_UNLOCK_IRQ(&sgadget->rwLock, rwflags);
#endif
		DBG(1, "==>\n");
		return rslt;
	}
	DBG(1,"gadget %s registered\n", sgadget->pGadget->name);
	
	return 0;
}

EXPORT_SYMBOL(usb_gadget_register_driver);

/*-------------------------------------------------------
	name	: usb_gadget_unregister_driver
	argment	: struct usb_gadget_driver *driver
	return	: int
	description: 
		to unregister to the Upper layer and kernel
	date	: 2010/03/16
-------------------------------------------------------*/
int usb_gadget_unregister_driver(struct usb_gadget_driver *driver)
{
	struct scd_gadget *sgadget = scd_gadget_info;
	unsigned long flags;
#ifdef RW_LOCK_ON
	unsigned long rwflags;
#endif
	DBG(2, "<==\n");
	DBG(1,"unregistering gadget %s\n", driver->function);

	if(!sgadget){
		DBG(1, "==>\n");
		return -ENODEV;
	}
	if(!driver || !driver->unbind){
		DBG(1, "==>\n");
		return -EINVAL;
	}
	if(sgadget->pGadgetDriver){
		/* ̤ȤηڤΥ */
		driver->unbind(sgadget->pGadget);
	}
	SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);

#ifdef RW_LOCK_ON
	SCD_G_WRITE_LOCK_IRQ(&sgadget->rwLock, rwflags);
#endif
	/* ǥХɥ饤С */
	sgadget->pGadgetDriver = NULL;
	/* ǥХgadgetǥХɥ饤С */
	sgadget->pGadget->dev.driver = NULL;
#ifdef RW_LOCK_ON
	SCD_G_WRITE_UNLOCK_IRQ(&sgadget->rwLock, rwflags);
#endif
	/* INFOɽ */
	DBG(2,"gadget %s unregistered\n", sgadget->pGadget->name);

	SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);

	DBG(1, "==>\n");
	return 0;
}
EXPORT_SYMBOL(usb_gadget_unregister_driver);

/*-------------------------------------------------------
	name	: usb_scd_gadget_init
	argment	: void
	return	: int
	description: 
		the initialization function of a gadget
		called by otg block at first
	date	: 2010/03/16
-------------------------------------------------------*/
int usb_scd_gadget_init(void)
{
	int rslt;

	DBG(3,"<==\n");

	/*	scd_intf setting
		the area of scd_gadget_intf is fix	*/
	scd_gadget_intf.type		= SCD_CONTROL_TYPE_GADGET;
	scd_gadget_intf.probe		= scd_gadget_probe;
	scd_gadget_intf.remove		= scd_gadget_remove;
	scd_gadget_intf.shutdown	= NULL;
	scd_gadget_intf.suspend		= NULL;
	scd_gadget_intf.suspend_late = NULL;
	scd_gadget_intf.resume		= NULL;
	scd_gadget_intf.resume_early = NULL;
	scd_gadget_intf.start		= scd_gadget_start;
	scd_gadget_intf.stop		= scd_gadget_stop;
	scd_gadget_intf.detect_vbus	= scd_gadget_detect_vbus;
	scd_gadget_intf.irq			= scd_gadget_irq;

	/* to register scd_intf */
	rslt = usb_scd_device_set_intf(&scd_gadget_intf);
	
	if(rslt!=0){
		ERR("fail at scd_intf registration \n");
		return rslt;
	}
	return 0;
}

/*-------------------------------------------------------
	name	: usb_scd_gadget_exit
	argment	: void
	return	: none
	description: 
		the exit function of a gadget
		called by otg block at the end
	date	: 2010/03/25
-------------------------------------------------------*/
void usb_scd_gadget_exit(void)
{

	int rslt;
	
	/* to unregister scd_intf */
	rslt = usb_scd_device_unset_intf(&scd_gadget_intf);
	
	if(rslt!=0){
		ERR("fail at scd_intf unregistration \n");
	}
}

/*-------------------------------------------------------
	name	: scd_gadget_probe
	argment	: struct scd_device *device
	return	: struct scd_control *
	description: probe to kernel usb interface
	date	: 2010/09/09
-------------------------------------------------------*/
static struct scd_control *scd_gadget_probe(struct scd_device *device)
{
	struct scd_gadget *sgadget;
	struct scd_gadget_fmapinfo *fmapinfo;
#ifdef RW_LOCK_ON
	unsigned long rwflags;
#endif
	u8	bEp;
#ifdef SCD_DEV_REG
	int ret;
#endif

	DBG(2,"<==\n");
	
	/* scd_gadgetΰ */
	sgadget = (struct scd_gadget *)kmalloc(sizeof(struct scd_gadget), GFP_KERNEL);
	if(sgadget==0){
		ERR( "scd_gadget - no memory\n");
		DBG(2,"==>\n");
		return NULL;
	}
	/* Ep0bufΰ */
	sgadget->pEp0buf = (void *)kmalloc(sizeof(struct scd_ep0_buf), GFP_KERNEL);
	if(sgadget->pEp0buf==0){
		ERR("sgadget->pEp0buf - no memory\n");
		DBG(2,"==>\n");
		goto err1;
	}
	/* ΰ */
	memset(sgadget->pEp0buf, 0, sizeof(struct scd_ep0_buf));

	/* usb_gadget¤Τΰ */
	sgadget->pGadget = (struct usb_gadget *)kmalloc(sizeof(struct usb_gadget), GFP_KERNEL);
	if(!sgadget->pGadget){
		ERR("out of memory\n");
		DBG(2,"==>\n");
		goto err2;
	}
	/* ΰ */
	memset(sgadget->pGadget, 0, sizeof(struct usb_gadget));

	/* RequestDoneQueue  */
	scd_doneque_init();

	/* ꥹƬ */
	INIT_LIST_HEAD(&sgadget->pGadget->ep_list);

	sgadget->bilimit.bWidthSize = SCD__READ(device, GHWCFG3)&SCD_M_GHWCFG3_XFERSIZEWIDTH;
	sgadget->bilimit.bWidthSize = SCD_F_GHWCFG3_XFERSIZEWIDTH(sgadget->bilimit.bWidthSize);
	if(((1<<sgadget->bilimit.bWidthSize)-1) != SCD_MAX_TRNSF_SZ){
		/* ͤȰۤʤϷٹ𤹤롣 */
		ERR("#################### Warning #################### \n");
		ERR("error XferSizeWidth:GHWCFG3 is not expected value %d \n", sgadget->bilimit.bWidthSize);
	}
	DBG(3,"bWidthSize:%d \n",sgadget->bilimit.bWidthSize);

	/* HW󤫤EP */
	sgadget->bEps = SCD__READ(device, GHWCFG2)>>SCD_S_GHWCFG2_NUMDEVEPS;
	sgadget->bEps &= SCD_M_GHWCFG2_NUMDEVEPS;

	if(sgadget->bEps > SCD_NUM_EPS){
		DBG(2, "NumDevEps over maxeps \n");
		sgadget->bEps = SCD_NUM_EPS;
	}
	/* FIFO mappingΰ */
	fmapinfo = 
		(struct scd_gadget_fmapinfo *)kmalloc(sizeof(struct scd_gadget_fmapinfo), GFP_KERNEL);
	if(!fmapinfo){
		ERR("no memory for fmapinfo \n");
		goto err3;
	}
	/* ΰ */
	fmapinfo->max_tx_dfifo_dep 
		= (u32 *)kmalloc(sizeof(u32)*(sgadget->bEps+SCD_NUM_EP0), GFP_KERNEL);
	if(!fmapinfo->max_tx_dfifo_dep){
		ERR("no memory in fifo mapping for depth \n");
		DBG(2,"==>\n");
		goto err4;
	}
	/* link to a global variable */
	scd_gadget_fminfo = fmapinfo;

	/* get areas for FIFO mapping */
	sgadget->fparam.ep_param 
	 = (struct scd_fifo_each *)kmalloc(sizeof(struct scd_fifo_each)*(sgadget->bEps+SCD_NUM_EP0), GFP_KERNEL);
	if(!(sgadget->fparam.ep_param)){
		ERR("no memory in fifo mapping for sgadget->fparam.ep_param \n");
		DBG(2,"==>\n");
		goto err5;
	}
	/*  */
	sgadget->fparam.cnt = 0;

	for(bEp=0; bEp<sgadget->bEps+SCD_NUM_EP0; bEp++){
		
		/* EP¤ν */
		memset(&sgadget->gep[bEp], 0, sizeof(sgadget->gep[bEp]));

		/* gadgetEP¤ EPֹ */
		sgadget->gep[bEp].bEpnum = bEp;
		/* EPlockѿ */
		spin_lock_init(&sgadget->gep[bEp].Lock);
#ifdef RW_LOCK_ON
		/* rqlock  */
		rwlock_init(&sgadget->gep[bEp].rwLock);
#endif
		/* ꥹƬ */
		INIT_LIST_HEAD(&sgadget->gep[bEp].req_list);
		
		/* EP̾ */
		snprintf(aEpNames[bEp], sizeof(aEpNames[bEp]),"ep%d", bEp);
		/* sgadget EP̾ */
		sgadget->gep[bEp].end_point.name = aEpNames[bEp];

		/* sgadget EPopsؿå */
		sgadget->gep[bEp].end_point.ops = &scd_gadget_ep_ops;
		
		/* EPep_listꥹȥإå */
		INIT_LIST_HEAD(&sgadget->gep[bEp].end_point.ep_list);
		
		if(bEp == SCD_G_EP0){
			/* gadget EP0 MAXP */
			sgadget->gep[SCD_G_EP0].end_point.maxpacket = SCD_EP0_MAXPSZ;
			/* gadget EP0ѥåȥ */
			sgadget->gep[SCD_G_EP0].wMaxp = SCD_EP0_MAXPSZ;
			/* ž */
			sgadget->gep[SCD_G_EP0].dwLen = 0;
			/* ֽ */
			sgadget->gep[SCD_G_EP0].bState 	  = SCD_EP_IDLE;
			/* ֽ */
			sgadget->gep[SCD_G_EP0].bSubState = SCD_EPSUB_IDLE;
			/* EP ͭե饰EPͭ */
			sgadget->gep[SCD_G_EP0].bEpena    = true;	/* fix except set_halt */
			/* lastbuf */
			(sgadget->gep[SCD_G_EP0].lbuf).buf  = (u8 *)NULL;
			(sgadget->gep[SCD_G_EP0].lbuf).dma  = (u8 *)NULL;
			(sgadget->gep[SCD_G_EP0].lbuf).size  = 0;
		}else{
			/* gadget EP* MAXP */
			sgadget->gep[bEp].end_point.maxpacket = 0;
			/* gadget EP* ѥåȥ */
			sgadget->gep[bEp].wMaxp = 0;
			/* ž */
			sgadget->gep[bEp].dwLen = 0;
			/* ֽ */
			sgadget->gep[bEp].bState 	= SCD_EP_IDLE;
			/* ֽ */
			sgadget->gep[bEp].bSubState = SCD_EPSUB_IDLE;
			/* lastbuf */
			sgadget->gep[bEp].lbuf.buf  = (u8 *)NULL;
			sgadget->gep[bEp].lbuf.dma  = (u8 *)NULL;
			sgadget->gep[bEp].lbuf.size  = 0;
			/* EPep_list ꥹȥơɲ(EP0ʳ) */
			list_add_tail(&sgadget->gep[bEp].end_point.ep_list, &(sgadget->pGadget->ep_list));
		}
		/* 塼ν */
		init_waitqueue_head(&(sgadget->gep[bEp].wq.ep_stq));	/* ̤ */
		sgadget->gep[bEp].wq.wq_flag = SCD_GQ_IDLE;
#ifdef MUTEX_ON
		init_MUTEX(&(sgadget->gep[bEp].gep_mtx));
#endif
	}
	/*  
	 */
	/* EPlockѿ */
	spin_lock_init(&sgadget->gLock);
#ifdef RW_LOCK_ON
	/* rqlock  */
	rwlock_init(&sgadget->rwLock);
#endif
	/* setupbufferlockѿ */
	spin_lock_init(&scd_dma_ep0buf.lock);
#ifdef MUTEX_ON
	/*  */
	init_MUTEX(&sgadget->g_mtx);
#endif
#ifdef RW_LOCK_ON
	SCD_G_WRITE_LOCK_IRQ(&sgadget->rwLock, rwflags);
#endif
	sgadget->pGadgetDriver = NULL;
#ifdef RW_LOCK_ON
	SCD_G_WRITE_UNLOCK_IRQ(&sgadget->rwLock, rwflags);
#endif
	sgadget->pGadget->ops = &scd_gadget_ops;
	sgadget->pGadget->ep0 = &sgadget->gep[SCD_G_EP0].end_point;

	/* ե饰å */
	sgadget->pGadget->is_dualspeed	= SCD_GADGET_SPEEDDUAL;
	sgadget->pGadget->is_otg		= SCD_GADGET_OTG_SUPPORT;
	sgadget->pGadget->a_hnp_support = SCD_GADGET_A_HNP_SUPPORT;
	
	/* note a name of usb_gadget */
	snprintf(aGadgetName, sizeof(aGadgetName),"not registered yet");
	sgadget->pGadget->name			= aGadgetName;

	/* for fifo mapping */
	memset(&sgadget->fdepth, 0, sizeof(struct scd_gadget_fdepth));

	/* local parameter */
	sgadget->g_device				= device;	/* scd_device */
	scd_gadget_info					= sgadget;  /* Хѿ */

	sgadget->bDevState		= SCD_GST_INIT;			/* Device State */
	sgadget->bEp0Stage		= SCD_EP0_STAGE_INIT;	/* EP0Stage */
	sgadget->bBusState		= SCD_BUS_IDLE;
	sgadget->bSuspended		= false;
	sgadget->bSetAddr		= false;
	sgadget->bFifoCnfg		= false;
	sgadget->bDevAddr		= SCD_DFLT_DEVADDR;		/* device address */
	sgadget->bTestMode		= false;
	sgadget->bTestModeNo	= USB_IOCTL_TESTMODE_NORMAL;
#ifdef CPU_TEST
	sgadget->bTransMode		= SCD_GADGET_TRANS_CPU;	
#else
	sgadget->bTransMode		= SCD_GADGET_TRANS_DMA;	
#endif
	if(sgadget->bTransMode == SCD_GADGET_TRANS_DMA){
		/* DMA setup */
		scd_dma_ep0buf.buf = kmalloc((sizeof(*scd_dma_ep0buf.buf))*SCD_DMA_EP0_SZ, GFP_KERNEL);
		if(!(scd_dma_ep0buf.buf)){
			ERR("no memory in fifo mapping for scd_dma_ep0buf.buf \n");
			DBG(2,"==>\n");
			goto err6;
		}
		for(bEp=0; bEp<sgadget->bEps+SCD_NUM_EP0; bEp++){
			/* last bufferΰ */
			if(bEp == SCD_G_EP0){
				/* lastbuf */
				sgadget->gep[SCD_G_EP0].lbuf.buf = (u8 *)kmalloc(sizeof(u8)*SCD_GRX_MAXFIFOSZ, GFP_KERNEL);
				if(!(sgadget->gep[SCD_G_EP0].lbuf.buf)){
					ERR("no memory in fifo mapping for sgadget->gep[SCD_G_EP0].lbuf.bu \n");
					DBG(2,"==>\n");
					goto err7;
				}
			}else{
				/* lastbuf */
				sgadget->gep[bEp].lbuf.buf  = (u8 *)kmalloc(sizeof(u8)*SCD_GRX_MAXFIFOSZ, GFP_KERNEL);
				if(!(sgadget->gep[bEp].lbuf.buf)){
					ERR("no memory in fifo mapping for "\
						"sgadget->gep[%d].lbuf.buf \n",bEp);
					DBG(2,"==>\n");
					goto err7;
				}
			}
		}
	}
	/* parameter init for FIFO mapping */
	scd_fmapinfo_init(sgadget);

#ifdef SCD_DEV_REG
	/* struct device setting for the sysfs */
	sgadget->pGadget->dev.parent	= device->dev;		/* from scd_device */
	dev_set_name(&sgadget->pGadget->dev, SCD_GADGET_DEV_BUSID);	/* bus_id */
	sgadget->pGadget->dev.release	= scdg_dev_release;
	/* ǥХϿsysfs) */
	ret = device_register(&sgadget->pGadget->dev);
	if(ret){
		 ERR("Unable to register sgadget->pGadget->dev \n");
	}
#endif
	DBG(2,"gadget probed\n");
	DBG(2,"==>\n");

	/* return scd_control */
	return &(sgadget->g_control);

/* memory free code */
err7:
	if(sgadget->bTransMode == SCD_GADGET_TRANS_DMA){
		/* DMA setup */
		kfree(scd_dma_ep0buf.buf);

		for(bEp=0; bEp<sgadget->bEps+SCD_NUM_EP0; bEp++){
			/* last bufferΰ */
			if(bEp == SCD_G_EP0){
				/* lastbuf */
				kfree(sgadget->gep[SCD_G_EP0].lbuf.buf);
			}else{
				/* lastbuf */
				kfree(sgadget->gep[bEp].lbuf.buf);
			}
		}
	}
err6:
	/* FIFO paramΰ */
	kfree(sgadget->fparam.ep_param);
err5:
	/* ΰ */
	kfree(scd_gadget_fminfo->max_tx_dfifo_dep);
err4:
	/* fmapinfo */
	kfree(fmapinfo);
err3:
	/* usb_gadget¤Τΰ */
	kfree(sgadget->pGadget);
err2:
	/* Ep0bufΰ */
	kfree(sgadget->pEp0buf);
err1:
	/* scd_gadgetΰ */
	kfree(sgadget);
	return NULL;
}

/*-------------------------------------------------------
	name	: scd_gadget_remove
	argment	: struct scd_control *control
	return	: int
	description: remove to kernel usb interface
	date	: 2010/09/09
-------------------------------------------------------*/
static int scd_gadget_remove(struct scd_control *control)
{
	struct scd_gadget *sgadget 
		= container_of((void *)control, struct scd_gadget, g_control);
	u8	bEp;

	DBG(1,"<==\n");

	if(sgadget->bTransMode == SCD_GADGET_TRANS_DMA){
		/* DMA setup */
		kfree(scd_dma_ep0buf.buf);

		for(bEp=0; bEp<sgadget->bEps+SCD_NUM_EP0; bEp++){
			/* last bufferΰ */
			if(bEp == SCD_G_EP0){
				/* lastbuf */
				kfree(sgadget->gep[SCD_G_EP0].lbuf.buf);
			}else{
				/* lastbuf */
				kfree(sgadget->gep[bEp].lbuf.buf);
			}
		}
	}
#ifdef SCD_DEV_REG
	/* ǥХϿsysfs) */
	device_unregister(&sgadget->pGadget->dev);
#endif
	/* usb_gadget¤Τΰ */
	kfree(sgadget->pGadget);
	
	/* Ep0bufΰ */
	kfree(sgadget->pEp0buf);
	
	/* FIFO paramΰ */
	kfree(sgadget->fparam.ep_param);
	
	/* scd_gadgetΰ */
	kfree(sgadget);

	/* ΰ */
	kfree(scd_gadget_fminfo->max_tx_dfifo_dep);
	
	/* fmapinfo */
	kfree(scd_gadget_fminfo);

	INFO("gadget removed\n");
	DBG(1,"==>\n");
	return 0;
}
#ifdef SCD_GADGET_PM
/*-------------------------------------------------------
	name	: scd_gadget_pm_suspend
	argment	: struct scd_control *control, pm_message_t state
	return	: int
	description: 
	date	: 2010/04/21
-------------------------------------------------------*/
static int scd_gadget_pm_suspend(struct scd_control *control, pm_message_t state)
{

	INFO(" non support now ");
	return 0;
}

/*-------------------------------------------------------
	name	: scd_gadget_pm_resume
	argment	: struct scd_control *control
	return	: int
	description: nonuse
	date	: 2010/04/21
-------------------------------------------------------*/
static int scd_gadget_pm_resume(struct scd_control *control)
{
	INFO(" non support now ");
	return 0;
}
#endif
/*-------------------------------------------------------
	name	: scd_gadget_start
	argment	: struct scd_control *control
	return	: int
	description: called necessarily before gadget becomes active
	date	: 2010/03/24
-------------------------------------------------------*/
static int scd_gadget_start(struct scd_control *control)
{
	struct scd_gadget *sgadget 
		= container_of((void *)control, struct scd_gadget, g_control);
	struct scd_device *sdev = sgadget->g_device;
	enum usb_device_speed *speed;
	u32 reg_val;
	unsigned long flags;

	DBG(2, ">>\n");
	SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);
	
	/* usb_gadgetǤξ */
	speed = &(sgadget->pGadget->speed);
	
	/* OTGspeed */
	if(test_bit(SCD_DEVICE_ATTR_HS_ENABLE, &sdev->attributes)){
		DBG(3,"set HS speed\n");
		*speed = USB_SPEED_HIGH;
	}else{
		DBG(3,"set FS speed\n");
		*speed = USB_SPEED_FULL;
	}
	/* DCFG쥸꡼ */
	reg_val = SCD__READ(sdev, DCFG);
	/* DescDMA̤ */
	reg_val &= ~SCD_B_DCFG_DESCDMA;
	/* speedӥåȥޥ */
	reg_val &= ~(SCD_M_DCFG_DEVSPD<<SCD_S_DCFG_DEVSPD);
	/* ž®ٳǧ */
	if(*speed == USB_SPEED_HIGH){
		reg_val |= SCD_E_DCFG_DEVSPD_HS<<SCD_S_DCFG_DEVSPD;
	}else{
		reg_val |= SCD_E_DCFG_DEVSPD_FS<<SCD_S_DCFG_DEVSPD;
	}
	/* 0ĹơOUTϥɥ(bit2) */
	reg_val |= SCD_B_DCFG_NZSTSOUTHSHK;

	/* ե졼ֳ(bit12,11) */
	reg_val &= ~(SCD_M_DCFG_PERFRINT<<SCD_S_DCFG_PERFRINT);
	reg_val |= SCD_E_DCFG_PERFRINT_80P<<SCD_S_DCFG_PERFRINT; 	/* HSθ */
	/* DCFG쥸饤 */
	SCD__WRITE(sdev, DCFG, reg_val);
	
	/* (̤) */
	reg_val = SCD__READ(sdev, DTHRCTL);
	reg_val &= ~(SCD_B_DTHRCTL_ARBPRKEN|SCD_B_DTHRCTL_RXTHREN|
		SCD_B_DTHRCTL_ISOTHREN|SCD_B_DTHRCTL_NONISOTHREN);
	SCD__WRITE(sdev, DTHRCTL, reg_val);

	/* Softdiscon */
	reg_val = SCD__READ(sdev, DCTL);
	reg_val |= SCD_B_DCTL_SFTDISCON;
	SCD__WRITE(sdev, DCTL, reg_val);
	
	/* ߥޥ쥸(gadget߶ػ) */
	reg_val = SCD__READ(sdev, GINTMSK);
	reg_val &= SCD_GINTMSK_INIT;
	SCD__WRITE(sdev, GINTMSK, reg_val);

	/* SETUP count  */
	reg_val = SCD_SUPCNT_VAL<<SCD_S_DOEPTSIZ0_SUPCNT;
	SCD__WRITE(sdev, DOEPTSIZ0, reg_val);

	SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);
	DBG(2, "==>\n");

	return 0;
}

/*-------------------------------------------------------
	name	: scd_gadget_stop
	argment	: struct scd_control *control
	return	: int
	description: 
	date	: 2010/03/24
-------------------------------------------------------*/
static int scd_gadget_stop(struct scd_control *control)
{
	struct scd_gadget *sgadget 
		= container_of((void *)control, struct scd_gadget, g_control);
	struct scd_device *sdev;
	u32 reg_val;
	unsigned long flags;

	DBG(2, "<==\n");	
	SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);
	
	/*  */
	sdev = (struct scd_device *)sgadget->g_device;
	/* Softdiscon */
	reg_val = SCD__READ(sdev, DCTL);
	reg_val |= SCD_B_DCTL_SFTDISCON; 
	/* test mode disable */
	reg_val &= ~(SCD_M_DCTL_TSTCTL<<SCD_S_DCTL_TSTCTL);
	SCD__WRITE(sdev, DCTL, reg_val);

	/* ߥޥ쥸 */
	reg_val = SCD__READ(sdev, GINTMSK);
	reg_val &= SCD_GINTMSK_INIT;
	SCD__WRITE(sdev, GINTMSK, reg_val);
	
	SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);

	/* LinkΥեȥꥻåȤԤ */
	/* ƽPullupOFF */
	scd_device_reset_link(sdev, false);

	DBG(2, "==>\n");
	return 0;
}

/*-------------------------------------------------------
	name	: scd_gadget_detect_vbus
	argment	: struct scd_control *control, bool onoff
	return	: none
	description: call scd_gadget_disconnect in vbus off
	date	: 2010/03/24
-------------------------------------------------------*/
static void scd_gadget_detect_vbus(struct scd_control *control, bool onoff)
{
	struct scd_gadget *sgadget 
		= container_of((void *)control, struct scd_gadget, g_control);
	
	DBG(1, "<== %x\n", onoff);

	if(onoff == false){	
#ifdef SCD_POLL_ON
		scd_reset_poll_state(sgadget, SCD_GQ_SHUTDOWN);
#endif
		/* ǻ δؿ */
		scd_gadget_disconnect(sgadget);
	}
	DBG(1, "==>\n");
	return;
}

/*-------------------------------------------------------
	name	: scd_ep_tx_enable
	argment	: struct scd_gadget *sgadget, 
			  struct usb_ep *ep
	return	: none
	description: enable tx endpoint
	date	: 2010/06/03
-------------------------------------------------------*/
static int scd_ep_tx_enable(struct scd_gadget *sgadget, struct usb_ep *ep)
{
	struct scd_device *sdev = sgadget->g_device;
	struct scd_gep *pEp = (struct scd_gep *)ep;
	unsigned long flags;
	u32 reg_val;
	const u8 epnum = EP_NUMBER(ep);

	DBG(2,"<==\n");

	SCD_G_SPIN_LOCK_IRQ(&pEp->Lock, flags);

	/* DIEPCTL쥸꡼ */
	reg_val = SCD__READ(sdev, DIEPCTL(epnum));
	/* MAXP */
	reg_val = (reg_val&~SCD_M_DXEPCTLN_MPS)|(pEp->wMaxp&SCD_M_DXEPCTLN_MPS);
	
	/* EPEna ӥåOFF ž椹롣  */
	if(reg_val&SCD_B_DXEPCTLN_EPENA){
		/* 顼 */
		ERR("Epena bit remains ep:%x \n",epnum);
		SCD_G_SPIN_UNLOCK_IRQ(&pEp->Lock, flags);
		DBG(2,"==>\n");
		return -EPERM;
	}
	
	/* ȥӥå 
	 * ʤ餺ͭˤʤʤSETD1Υꥢʤ
	 */
	reg_val &= ~SCD_B_DXEPCTLN_SETD1PID;	
	reg_val |= SCD_B_DXEPCTLN_SETD0PID;
	
	/* žޥ  */
	reg_val &= ~(SCD_M_DXEPCTLN_EPTYPE<<SCD_S_DXEPCTLN_EPTYPE);
	
	switch(pEp->bTransferType){
	case USB_ENDPOINT_XFER_CONTROL:
		/* ž */
		reg_val |= SCD_E_DXEPCTLN_EPTYPE_CONTROL<<SCD_S_DXEPCTLN_EPTYPE;
		break;
	case USB_ENDPOINT_XFER_ISOC:
		/* ž */
		reg_val |=  SCD_E_DXEPCTLN_EPTYPE_ISO<<SCD_S_DXEPCTLN_EPTYPE;
		break;
	case USB_ENDPOINT_XFER_BULK:
		/* ž */
		reg_val |= SCD_E_DXEPCTLN_EPTYPE_BULK<<SCD_S_DXEPCTLN_EPTYPE;
		break;
	case USB_ENDPOINT_XFER_INT:
		/* ž */
		reg_val |= SCD_E_DXEPCTLN_EPTYPE_INTERRUPT<<SCD_S_DXEPCTLN_EPTYPE;
		break;
	default:
		ERR("non-support Transfer Type %x\n", pEp->bTransferType);
	}
	
	/* TxFIFOֹ */
	reg_val &= ~ (SCD_M_DXEPCTLN_TXFNUM<<SCD_S_DXEPCTLN_TXFNUM);
	reg_val |= (epnum&SCD_M_DXEPCTLN_TXFNUM)<<SCD_S_DXEPCTLN_TXFNUM;

	/* 쥸饤 */
	SCD__WRITE(sdev, DIEPCTL(epnum), reg_val);

#ifdef ENBLE_FLUSH
	SCD_G_SPIN_UNLOCK_IRQ(&pEp->Lock, flags);

	/* FIFOեå */
	scd_g_fifo_flush_tx(sgadget, epnum, SCD_G_FLUSH_NORMAL);

	SCD_G_SPIN_LOCK_IRQ(&pEp->Lock, flags);
#endif	

	/* EPߥޥ */
	reg_val = SCD__READ(sdev, DAINTMSK);

	reg_val |= (1 << epnum)&SCD_M_DAINTMSK_INEPMSK;
	SCD__WRITE(sdev, DAINTMSK, reg_val);

	SCD_G_SPIN_UNLOCK_IRQ(&pEp->Lock, flags);

	DBG(2,"==>\n");
	return 0;
}

/*-------------------------------------------------------
	name	: scd_ep_rx_enable
	argment	: struct scd_gadget *sgadget, 
			  struct usb_ep *ep
	return	: none
	description: enable rx endpoint
	date	: 2010/05/06
-------------------------------------------------------*/
static int scd_ep_rx_enable(struct scd_gadget *sgadget, struct usb_ep *ep)
{
	struct scd_device *sdev = sgadget->g_device;
	struct scd_gep *pEp = (struct scd_gep *)ep;
	unsigned long flags;
	u32 reg_val;
	const u8 epnum = EP_NUMBER(ep);

	DBG(2,"<==\n");	

	SCD_G_SPIN_LOCK_IRQ(&pEp->Lock, flags);

	/* DOEPCTL쥸꡼ */
	reg_val = SCD__READ(sdev, DOEPCTL(epnum));
	/* MAXP */
	reg_val = (reg_val&~SCD_M_DXEPCTLN_MPS)|(pEp->wMaxp&SCD_M_DXEPCTLN_MPS);
	
	/* EPEna ӥåOFF ž椹롣  */
	if(reg_val&SCD_B_DXEPCTLN_EPENA){
		/* 顼 */
		ERR("Epena bit remains \n");
		SCD_G_SPIN_UNLOCK_IRQ(&pEp->Lock, flags);
		DBG(2,"==>\n");
		return -EPERM;
	}	
	/* ȥӥå 
	 * ʤ餺ͭˤʤʤSETD1Υꥢʤ
	 */
	reg_val &= ~SCD_B_DXEPCTLN_SETD1PID;	
	reg_val |= SCD_B_DXEPCTLN_SETD0PID;
	
	/* žޥ  */
	reg_val &= ~(SCD_M_DXEPCTLN_EPTYPE<<SCD_S_DXEPCTLN_EPTYPE);
	
	switch(pEp->bTransferType){
	case USB_ENDPOINT_XFER_CONTROL:
		/* ž */
		reg_val |= SCD_E_DXEPCTLN_EPTYPE_CONTROL<<SCD_S_DXEPCTLN_EPTYPE;
		break;
	case USB_ENDPOINT_XFER_ISOC:
		/* ž */
		reg_val |=  SCD_E_DXEPCTLN_EPTYPE_ISO<<SCD_S_DXEPCTLN_EPTYPE;
		break;
	case USB_ENDPOINT_XFER_BULK:
		/* ž */
		reg_val |= SCD_E_DXEPCTLN_EPTYPE_BULK<<SCD_S_DXEPCTLN_EPTYPE;
		break;
	case USB_ENDPOINT_XFER_INT:
		/* ž */
		reg_val |= SCD_E_DXEPCTLN_EPTYPE_INTERRUPT<<SCD_S_DXEPCTLN_EPTYPE;
		break;
	default:
		ERR("non-support Transfer Type %x\n", pEp->bTransferType);
	}
	
	/* TxFIFOֹ */
	reg_val &= ~ (SCD_M_DXEPCTLN_TXFNUM<<SCD_S_DXEPCTLN_TXFNUM);
	reg_val |= (epnum&SCD_M_DXEPCTLN_TXFNUM)<<SCD_S_DXEPCTLN_TXFNUM;

	/* 쥸饤 */
	SCD__WRITE(sdev, DOEPCTL(epnum), reg_val);

#ifdef ENBLE_FLUSH
	SCD_G_SPIN_UNLOCK_IRQ(&pEp->Lock, flags);

	/* FIFOեå */
	scd_g_fifo_flush_rx(sgadget, epnum, SCD_G_FLUSH_NORMAL);

	SCD_G_SPIN_LOCK_IRQ(&pEp->Lock, flags);
#endif	
	
	/* EPߥޥ */
	reg_val = SCD__READ(sdev, DAINTMSK);
	reg_val |= ((1 << epnum)&SCD_M_DAINTMSK_OUTEPMSK)<<SCD_S_DAINTMSK_OUTEPMSK;
	SCD__WRITE(sdev, DAINTMSK, reg_val);

	SCD_G_SPIN_UNLOCK_IRQ(&pEp->Lock, flags);

	DBG(2,"==>\n");	
	return 0;
}

/*-------------------------------------------------------
	name	: set_test_mode
	argment	: struct scd_gadget *sgadget, 
			  int mode
	return	: int
	description: set testmode
	date	: 2010/05/31
-------------------------------------------------------*/
static int set_test_mode(struct scd_gadget *sgadget, unsigned long param)
{
	int rslt=0;

	DBG(1, "<==\n");

	/* test mode enable */
	sgadget->bTestMode = true;
	
	switch (param) {
	case USB_IOCTL_TESTMODE_NORMAL:

		DBG(3, "TESTMODE_NORMAL\n");
		sgadget->bTestModeNo = SCD_E_DCTL_TSTCTL_DISABLE;
		break;
	case USB_IOCTL_TESTMODE_TEST_J:
		/* TEST_J */
		DBG(3, "TEST_J\n");
		sgadget->bTestModeNo = SCD_E_DCTL_TSTCTL_J;
		break;
	case USB_IOCTL_TESTMODE_TEST_K:
		/* TEST_K */
		DBG(3, "TEST_K\n");
		sgadget->bTestModeNo = SCD_E_DCTL_TSTCTL_K;
		break;
	case USB_IOCTL_TESTMODE_SE0_NAK:
		/* TEST_SE0_NAK */
		DBG(3, "TEST_SE0_NAK\n");
		sgadget->bTestModeNo = SCD_E_DCTL_TSTCTL_SE0NAK;
		break;
	case USB_IOCTL_TESTMODE_TEST_PACKET:
		/* TEST_PACKET */
		DBG(3, "TEST_PACKET\n");
		sgadget->bTestModeNo = SCD_E_DCTL_TSTCTL_PACKET;
		break;
	case USB_IOCTL_TESTMODE_FORCE_ENABLE:
		/* FORCE_ENABLE */
		DBG(3, "FORCE_ENABLE\n");
		sgadget->bTestModeNo = SCD_E_DCTL_TSTCTL_FORCE;
		break;
	default:
		/* non-support */
		/* test mode disable */
		sgadget->bTestMode = false;

		DBG(1, "==> a non-support test mode! \n");
		rslt = -EINVAL;
	}	
	DBG(1, "==>\n");
	return rslt;
}

/*-------------------------------------------------------
	name	: scd_exec_test_mode
	argment	: struct scd_gadget *sgadget, 
	return	: int
	description: set testmode
	date	: 2010/05/31
-------------------------------------------------------*/
int scd_exec_test_mode(struct scd_gadget *sgadget)
{
	struct scd_device *sdev = sgadget->g_device;
	int rslt=0;
	u32 reg_val;

	DBG(1, "<==\n");

	switch (sgadget->bTestModeNo) {
	case SCD_E_DCTL_TSTCTL_DISABLE:

		DBG(3, "TESTMODE_NORMAL\n");

		reg_val = SCD__READ(sdev, DCTL);
		/* test mode disable */
		reg_val &= ~(SCD_M_DCTL_TSTCTL<<SCD_S_DCTL_TSTCTL);
		SCD__WRITE(sdev, DCTL, reg_val);
		break;
	case SCD_E_DCTL_TSTCTL_J:
		/* TEST_J */
		DBG(3, "TEST_J\n");

		reg_val = SCD__READ(sdev, DCTL);
		/* test J mode */
		reg_val |= SCD_E_DCTL_TSTCTL_J<<SCD_S_DCTL_TSTCTL;
		SCD__WRITE(sdev, DCTL, reg_val);
		break;
	case SCD_E_DCTL_TSTCTL_K:
		/* TEST_K */
		DBG(3, "TEST_K\n");

		reg_val = SCD__READ(sdev, DCTL);
		/* test K mode */
		reg_val |= SCD_E_DCTL_TSTCTL_K<<SCD_S_DCTL_TSTCTL;
		SCD__WRITE(sdev, DCTL, reg_val);
		break;
	case SCD_E_DCTL_TSTCTL_SE0NAK:
		/* TEST_SE0_NAK */
		DBG(3, "TEST_SE0_NAK\n");

		reg_val = SCD__READ(sdev, DCTL);
		/* test SE0NAKmode */
		reg_val |= SCD_E_DCTL_TSTCTL_SE0NAK<<SCD_S_DCTL_TSTCTL;
		SCD__WRITE(sdev, DCTL, reg_val);
		break;
	case SCD_E_DCTL_TSTCTL_PACKET:
		/* TEST_PACKET */
		DBG(3, "TEST_PACKET\n");

		reg_val = SCD__READ(sdev, DCTL);
		/* test Packet mode */
		reg_val |= SCD_E_DCTL_TSTCTL_PACKET<<SCD_S_DCTL_TSTCTL;
		SCD__WRITE(sdev, DCTL, reg_val);
		break;
	case SCD_E_DCTL_TSTCTL_FORCE:
		/* FORCE_ENABLE */
		DBG(3, "FORCE_ENABLE\n");

		reg_val = SCD__READ(sdev, DCTL);
		/* Force Enable mode */
		reg_val |= SCD_E_DCTL_TSTCTL_FORCE<<SCD_S_DCTL_TSTCTL;
		SCD__WRITE(sdev, DCTL, reg_val);
		break;
	default:
		/* non-support */
		DBG(1, "==> a non-support test mode! \n");
		rslt = -EINVAL;
	}	
	/* test mode disable */
	sgadget->bTestMode = false;

	DBG(1,"==>\n");
	return rslt;
}
#ifdef SCD_DEV_REG
/*-------------------------------------------------------
	name	: scdg_dev_release
	argment	: struct device *dev, 
	return	: none
	description: the release function in struct device
	date	: 2010/09/09
-------------------------------------------------------*/
static void scdg_dev_release(struct device *dev)
{
	/* nothing  */
}
#endif

/*-------------------------------------------------------
	name	: scd_gadget_restart
	argment	: struct scd_gadget *sgadget
	return	: int
	description: reset to the status of the gadget_start at pullup on
	date	: 2010/11/08
-------------------------------------------------------*/
static int scd_gadget_restart(struct scd_gadget *sgadget)
{
	struct scd_device *sdev = sgadget->g_device;
	enum usb_device_speed *speed;
	u32 reg_val;
	unsigned long flags;

	DBG(1, "<==\n");
	SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);

	/* usb_gadgetǤξ */
	speed = &(sgadget->pGadget->speed);
	
	/* OTGspeed */
	if(test_bit(SCD_DEVICE_ATTR_HS_ENABLE, &sdev->attributes)){
		DBG(3,"set HS speed\n");
		*speed = USB_SPEED_HIGH;
	}else{
		DBG(3,"set FS speed\n");
		*speed = USB_SPEED_FULL;
	}
	/* DCFG쥸꡼ */
	reg_val = SCD__READ(sdev, DCFG);
	/* DescDMA̤ */
	reg_val &= ~SCD_B_DCFG_DESCDMA;

	/* 0ĹơOUTϥɥ(bit2) */
	reg_val |= SCD_B_DCFG_NZSTSOUTHSHK;

	/* ե졼ֳ(bit12,11) */
	reg_val &= ~(SCD_M_DCFG_PERFRINT<<SCD_S_DCFG_PERFRINT);
	reg_val |= SCD_E_DCFG_PERFRINT_80P<<SCD_S_DCFG_PERFRINT; 	/* HSθ */
	/* DCFG쥸饤 */
	SCD__WRITE(sdev, DCFG, reg_val);
	
	/* (̤) */
	reg_val = SCD__READ(sdev, DTHRCTL);
	reg_val &= ~(SCD_B_DTHRCTL_ARBPRKEN|SCD_B_DTHRCTL_RXTHREN|
		SCD_B_DTHRCTL_ISOTHREN|SCD_B_DTHRCTL_NONISOTHREN);
	SCD__WRITE(sdev, DTHRCTL, reg_val);

	/* ߥޥ쥸(gadget߶ػ) */
	reg_val = SCD__READ(sdev, GINTMSK);
	reg_val &= SCD_GINTMSK_INIT;
	SCD__WRITE(sdev, GINTMSK, reg_val);

	/* SETUP count  */
	reg_val = SCD_SUPCNT_VAL<<SCD_S_DOEPTSIZ0_SUPCNT;
	SCD__WRITE(sdev, DOEPTSIZ0, reg_val);

	SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);
	DBG(2, "==>\n");

	return 0;
}
