/*--------------------------------------------------------------------------*
 * drivers/usb/scd/scd_debug.h
 * 
 * Debug definitions
 * 
 * Copyright 2011 Sony Corporation
 *
 * This file is part of the HS-OTG Controller Driver SCD.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *---------------------------------------------------------------------------*/

#ifndef _USB_SCD_DEBUG_H_
#define _USB_SCD_DEBUG_H_

/*
 *	In-place debug options
 *	You can use debug function to valid follow defines.
 */

/* __printk function, outputs to the ring buffer and display outputs from
procfs. The ring buffer is defined scd_info.c */
//#define SCD_STR_LOG_ON
/* Logging spinlock and rwlock. It must be the define SCD_STR_LOG_ON
at same time. */
//#define _SCD_LOCK_DEBUG

/*
 *	Traditional debug print
 */

extern int scd_debug_level;

#if SCD_DEBUG > 0

#define _if_dbg_level(level) \
	if ( (level>=-1 && scd_debug_level>=level) || scd_debug_level==level )
#define xprintk(level, facility, format, args...) \
        do { _if_dbg_level(level) { \
                printk(facility "%s %d: " format , \
                       __FUNCTION__, __LINE__ , ## args); \
        } } while (0)
#define xprintks(level, format, args...) \
        do { _if_dbg_level(level) { printk(format,  ## args); } } while (0)

#define DBG(level,fmt,args...) xprintk(level,KERN_INFO,fmt, ## args)
#define DBGS(level,fmt,args...) xprintks(level, fmt, ## args)
#define ERR(fmt,args...) xprintk(-1,KERN_INFO,fmt, ## args)
#define INFO(fmt,args...) xprintk(-1,KERN_INFO,fmt, ## args)

#else

#define DBG(fmt,args...)	do {}  while (0)
#define DBGS(fmt,args...)	do {}  while (0)
#define ERR(fmt,args...)  	do {}  while (0)
#define INFO(fmt,args...) 	do {}  while (0)

#endif	/* SCD_DEBUG */

/*
 *	Ring buffer print
 */

#ifdef SCD_STR_LOG_ON
extern int __printk(const char *fmt, ...);
#endif	/* SCD_STR_LOG_ON */

/*
 *	Lock debug (need to set ring buffer print option)
 */

#ifdef _SCD_LOCK_DEBUG

#define RT_MUTEX_OWNER_PENDING	1UL
#define RT_MUTEX_HAS_WAITERS	2UL
#define RT_MUTEX_OWNER_MASKALL	3UL

static inline struct task_struct *rt_mutex_owner(struct rt_mutex *lock)
{
	return (struct task_struct *)
		((unsigned long)lock->owner & ~RT_MUTEX_OWNER_MASKALL);
}

static inline struct task_struct *rt_mutex_real_owner(struct rt_mutex *lock)
{
	return (struct task_struct *)
		((unsigned long)lock->owner & ~RT_MUTEX_HAS_WAITERS);
}

#undef read_lock
#define read_lock(LOCK) \
	__printk("%s rls(%p?=%p,%d)\n", __FUNCTION__, \
		 current, rt_mutex_real_owner(&(LOCK)->lock), \
		 (LOCK)->read_depth); \
	rt_read_lock((LOCK)); \
	__printk("%s rle\n", __FUNCTION__);
#undef read_unlock
#define read_unlock(LOCK) \
	__printk("%s rus\n", __FUNCTION__); \
	rt_read_unlock((LOCK)); \
	__printk("%s rue\n", __FUNCTION__);

#undef read_lock_irqsave
#define read_lock_irqsave(LOCK, FLAGS) \
	__printk("%s rlis(%p?=%p,%d)\n", __FUNCTION__, \
		 current, rt_mutex_real_owner(&(LOCK)->lock), \
		 (LOCK)->read_depth); \
	rt_read_lock_irqsave((LOCK)); \
	__printk("%s rlie\n", __FUNCTION__);
#undef read_unlock_irqrestore
#define read_unlock_irqrestore(LOCK, FLAGS) \
	__printk("%s ruis\n", __FUNCTION__); \
	rt_read_unlock((LOCK)); \
	__printk("%s ruie\n", __FUNCTION__);

#undef write_lock
#define write_lock(LOCK) \
	__printk("%s wls(%p?=%p,%d)\n", __FUNCTION__, \
		 current, rt_mutex_real_owner(&(LOCK)->lock), \
		 (LOCK)->read_depth); \
	rt_write_lock((LOCK)); \
	__printk("%s wle\n", __FUNCTION__);
#undef write_unlock
#define write_unlock(LOCK) \
	__printk("%s wus\n", __FUNCTION__); \
	rt_write_unlock((LOCK)); \
	__printk("%s wue\n", __FUNCTION__);

#undef write_lock_irqsave
#define write_lock_irqsave(LOCK, FLAGS) \
	__printk("%s wlis(%p?=%p,%d)\n", __FUNCTION__, \
		 current, rt_mutex_real_owner(&(LOCK)->lock), \
		 (LOCK)->read_depth); \
	rt_write_lock_irqsave((LOCK)); \
	__printk("%s wlie\n", __FUNCTION__);
#undef write_unlock_irqrestore
#define write_unlock_irqrestore(LOCK, FLAGS) \
	__printk("%s wuis\n", __FUNCTION__); \
	rt_write_unlock((LOCK)); \
	__printk("%s wuie\n", __FUNCTION__);

#undef spin_lock
#define spin_lock(LOCK) \
	__printk("%s sls(%p?=%p)\n", __FUNCTION__, \
		 current, rt_mutex_real_owner(&(LOCK)->lock)); \
	rt_spin_lock((LOCK)); \
	__printk("%s sle\n", __FUNCTION__);
#undef spin_unlock
#define spin_unlock(LOCK) \
	__printk("%s sus\n", __FUNCTION__); \
	rt_spin_unlock((LOCK)); \
	__printk("%s sue\n", __FUNCTION__);

#undef spin_lock_irqsave
#define spin_lock_irqsave(LOCK, FLAGS) \
	__printk("%s slis(%p?=%p)\n", __FUNCTION__, \
		 current, rt_mutex_real_owner(&(LOCK)->lock)); \
	rt_spin_lock((LOCK)); \
	__printk("%s slie\n", __FUNCTION__);
#undef spin_unlock_irqrestore
#define spin_unlock_irqrestore(LOCK, FLAGS) \
	__printk("%s suis\n", __FUNCTION__); \
	rt_spin_unlock((LOCK)); \
	__printk("%s suie\n", __FUNCTION__);

#endif	/* _SCD_LOCK_DEBUG */

#endif	/* _USB_SCD_DEBUG_H_ */
