/*--------------------------------------------------------------------------*
 * drivers/usb/scd/scd_device.c
 * 
 * OTG and, Platform bus driver operations
 * 
 * Copyright 2011, 2012 Sony Corporation
 *
 * This file is part of the HS-OTG Controller Driver SCD.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/io.h>
#include <linux/platform_device.h>

#include <linux/usb.h>
#include <linux/usb/gadget.h>
#include <linux/version.h>
#include <linux/usb/scd/usb_otg_control.h>
#include "scd.h"
#include "scd_debug.h"
#include "scd_otg.h"
#include "scd_register.h"
#include "scd_version.h"

#define A_DEVICE	0
#define B_DEVICE	1

#define NUM_CONTROL_TYPE	4	/* gadget, hcd, sysdep, info */

/* port speed */
#define SCD_PORT_HS 0x0
#define SCD_PORT_FS 0x1
#define SCD_PORT_LS 0x2

/* ޥꥹ */
struct scd_device_timer {
	struct scd_timer t;
	struct scd_device *sdev;
	struct list_head timer_list;
};

/* OTGơ */
#define _S_B(LVL)	(4 * (LVL)) /* state bit field is 4bit * level shift */
#define _S_G(LVL, NUM)	((NUM) << _S_B(LVL)) /* state group */
#define _S_M(LVL)	(~0 << _S_B(LVL)) /* state level mask */
#define _S(NAME)	(SCD_DEVICE_S_##NAME) /* device name */
enum scd_device_state {
	/* level 2 */
	SCD_DEVICE_S_POWERED_ON	= _S_G(2, 0),
	SCD_DEVICE_S_PROBED	= _S_G(2, 1),
	SCD_DEVICE_S_REMOVED	= _S_G(2, 2),
	SCD_DEVICE_S_POWERED_OFF = _S_G(2, 3),
	/* level 1 */
	SCD_DEVICE_S_PROBED_INIT =	_S(PROBED)|_S_G(1, 0),
	SCD_DEVICE_S_STOP	=	_S(PROBED)|_S_G(1, 1),
	SCD_DEVICE_S_IDLE	=	_S(PROBED)|_S_G(1, 2),
	SCD_DEVICE_S_B_PERIPH	=	_S(PROBED)|_S_G(1, 4),
	SCD_DEVICE_S_A_HOST	=	_S(PROBED)|_S_G(1, 5),
	/* level 0 */
	SCD_DEVICE_S_IDLE_INIT	=		_S(IDLE)|_S_G(0, 0),
	SCD_DEVICE_S_AB_VBUSED	=		_S(IDLE)|_S_G(0, 1),
	SCD_DEVICE_S_AB_SILENT	=		_S(IDLE)|_S_G(0, 2),
	SCD_DEVICE_S_A_H_DISCON	=		_S(A_HOST)|_S_G(0, 1),
	SCD_DEVICE_S_A_H_DEBOUNCE =		_S(A_HOST)|_S_G(0, 2),
	SCD_DEVICE_S_A_H_CONN	=		_S(A_HOST)|_S_G(0, 3),
};

enum scd_device_trigger {
	SCD_DEVICE_T_NONE,
	SCD_DEVICE_T_PROBE,
	SCD_DEVICE_T_REMOVE,
	SCD_DEVICE_T_SHUTDOWN,
	SCD_DEVICE_T_STOP,
	SCD_DEVICE_T_START_CONTROL,
	SCD_DEVICE_T_TEST_BUS,
	SCD_DEVICE_T_STOP_GADGET,
	SCD_DEVICE_T_STOP_HOST,
	SCD_DEVICE_T_START_GADGET,
	SCD_DEVICE_T_START_HOST,
	SCD_DEVICE_T_STOP_CONTROL,
	SCD_DEVICE_T_CHANGE_VBUS,
	SCD_DEVICE_T_CONNECT,
	SCD_DEVICE_T_DEBOUNCE,
	SCD_DEVICE_T_DISCONNECT,
};

/* OTG¤ */
struct scd_device_control {
	struct scd_device sdev;
	struct usb_otg_control otgc;
	struct platform_driver pdrv;
	struct scd_control *control[NUM_CONTROL_TYPE];

	scd_otg_notify notify_hook;

	spinlock_t lock;
	struct mutex state_mutex;
	struct tasklet_struct event_tasklet;
	struct workqueue_struct *trigger_wq;
	atomic_t delayed_check_bus_count;

	struct list_head timer_anchor; /* timer listϥॢȽ¤٤ */
	struct list_head event_anchor; /* eventFIFO */

	enum scd_device_state state;
	int irqn;
	int cid;		/* last cid */
	enum usb_otg_vbus_stat vbus; /* last vbus */

	struct list_head scd_list;
};

static inline
struct scd_device_control *otgc_to_scd(struct usb_otg_control* otgc)
{
	return container_of(otgc, struct scd_device_control, otgc);
}

static inline
struct scd_device_control *sdev_to_scd(struct scd_device* sdev)
{
	return container_of(sdev, struct scd_device_control, sdev);
}

/* scdɥ饤дꥹ */
/* ʣLinkбǽOTG APIλ;塢¼1ĤΤߤб */
static LIST_HEAD(scd_anchor);	/* scd listprobe */

/* ¾ */
DEFINE_RWLOCK(scd_lock);        /* - unset_intf-ߴintfs¾ */
                                /* - probe/remove sdev¾ */
                                /* - probe/remove control¾ */

/* 󥿡եؿѥƥå */
static struct scd_intf *intfs[NUM_CONTROL_TYPE];

/* 󥿡եؿ󥢥ޥ */
#define get_gadget_intf() intfs[SCD_CONTROL_TYPE_GADGET]
#define get_hcd_intf() intfs[SCD_CONTROL_TYPE_HCD]
#define get_sysdep_intf() container_of(intfs[SCD_CONTROL_TYPE_SYSDEP], \
				       struct scd_sysdep_intf, base_intf)
#define get_info_intf() intfs[SCD_CONTROL_TYPE_INFO]
#define gadget_control(SCD) ((SCD)->control[SCD_CONTROL_TYPE_GADGET])
#define hcd_control(SCD) ((SCD)->control[SCD_CONTROL_TYPE_HCD])
#define sysdep_control(SCD) ((SCD)->control[SCD_CONTROL_TYPE_SYSDEP])
#define info_control(SCD) ((SCD)->control[SCD_CONTROL_TYPE_INFO])

/* ܴؿ */
static void exec_begin(struct scd_device_control *scd);
static void exec_terminate(struct scd_device_control *scd);
static void exec_inactivate(struct scd_device_control *scd);
static int try_start_control(struct scd_device_control *scd);
static int try_stop_control(struct scd_device_control *scd);
static void exec_activate(struct scd_device_control *scd);
static int test_bus(struct scd_device_control *scd);
static void exec_suspend(struct scd_device_control *scd);
static void exec_resume(struct scd_device_control *scd);
static int try_start_gadget(struct scd_device_control *scd);
static int try_start_host(struct scd_device_control *scd);
static void exec_stop_gadget(struct scd_device_control *scd);
static void exec_stop_host(struct scd_device_control *scd);
static void exec_connect(struct scd_device_control *scd);
static void exec_disconnect(struct scd_device_control *scd);

/*
 *	⡼ɽ
 */

/* ե륿줿OTGơȼؿ */
inline static
enum scd_device_state scd_device_get_state(struct scd_device_control *scd,
					   u8 level)
{
	return scd->state & _S_M(level);
}

/* is_host : HW a-deviceư */
inline bool is_host(struct scd_device *sdev)
{
	u32 gintsts;

	gintsts = SCD__READ(sdev, GINTSTS);

	return (gintsts & SCD_B_GINTSTS_CURMOD);
}

/* is_device : HW b-deviceư */
inline bool is_device(struct scd_device *sdev)
{
	return !is_host(sdev);
}

/* is_gadget : SW gadgetư */
inline bool is_gadget(struct scd_device_control *scd)
{
	return (scd_device_get_state(scd, 1) == SCD_DEVICE_S_B_PERIPH);
}

/* is_hcd : SW hcdư */
inline bool is_hcd(struct scd_device_control *scd)
{
	return (scd_device_get_state(scd, 1) == SCD_DEVICE_S_A_HOST);
}

/*
 *	쥸
 */

/* D+/-饤Pullup, gadget startPullup OFF */
/* gadget⥢ǽ뤿¾ */
/* A[?S] */
static void control_pullup(struct scd_device *sdev, bool onoff)
{
	unsigned long flags;
	u32 dctl;

	/*** Gadget Pullup OFF ***/
	if (is_device(sdev)) { /* Device ModeΤ */
		spin_lock_irqsave(&sdev->lock, flags);
		dctl = SCD__READ(sdev, DCTL);
		/* D Pullup OFF */
		dctl = (onoff ? dctl & ~SCD_B_DCTL_SFTDISCON :
			dctl | SCD_B_DCTL_SFTDISCON);
		SCD__WRITE(sdev, DCTL, dctl);
		DBG(2, "dctl=0x%x\n", dctl);
		spin_unlock_irqrestore(&sdev->lock, flags);
	}
}

/* ߤ */
/* A[iS] */
static void stop_regs(struct scd_device *sdev)
{
        /* GINTMSK, GINTSTS, GOTGINTꥢ */
        SCD__WRITE(sdev, GINTMSK, 0);
        SCD__WRITE(sdev, GINTSTS, ~0);
        SCD__WRITE(sdev, GOTGINT, ~0);
}

/* ߤγ */
/* A[iS] */
static void start_regs(struct scd_device *sdev)
{
	u32 gintmsk;

        /* GINTMSK set */
        gintmsk = (SCD_B_GINTMSK_SESSREQINTMSK |
		   SCD_B_GINTMSK_CONIDSTSCHNGMSK |
		   SCD_B_GINTMSK_OTGINTMSK |
		   SCD_B_GINTMSK_MODEMISMSK);
        SCD__WRITE(sdev, GINTMSK, gintmsk);
	BUG_ON(SCD__READ(sdev, GINTMSK) != gintmsk);
}

/* 쥸ν */
/* A[iS] */
static void init_regs(struct scd_device *sdev, bool pullup)
{
	u32 gahbcfg;
	u32 gusbcfg;
#define _HBSTLEN_INC16		0x7

	DBG(2, ">>\n");

	/* GOTGINT */
	SCD__READ(sdev, GOTGINT);

        /*
          
         */
        /*** GAHBCFG ***/
        gahbcfg = SCD__READ(sdev, GAHBCFG);
        /* DMA mode */
        gahbcfg |= SCD_B_GAHBCFG_DMAEN;
        /* AHB burst length(INC16) */
        gahbcfg |= SCD_SFT__FIELD(_HBSTLEN_INC16, GAHBCFG, HBSTLEN);
        /* Global interrupt unmask */
        gahbcfg |= SCD_B_GAHBCFG_GLBLINTRMSK;
        SCD__WRITE(sdev, GAHBCFG, gahbcfg);

        /*** GUSBCFG ***/
        gusbcfg = SCD__READ(sdev, GUSBCFG);
        gusbcfg &= ~(SCD_B_GUSBCFG_HNPCAP |
		     SCD_B_GUSBCFG_SRPCAP |
		     SCD_B_GUSBCFG_DDRSEL |
		     SCD_B_GUSBCFG_PHYSEL |
		     SCD_B_GUSBCFG_ULPIUTMISEL |
		     SCD_B_GUSBCFG_PHYIF);
        SCD__WRITE(sdev, GUSBCFG, gusbcfg);

	/* Link */
	stop_regs(sdev);

	/* Device ModePullup OFF */
	control_pullup(sdev, pullup);
}

/* SRPCap쥸 - Host⡼ɻ */
/* A[?S] */
static void set_srpcap_regs(struct scd_device *sdev)
{
	u32 gusbcfg;

	gusbcfg = SCD__READ(sdev, GUSBCFG);
	gusbcfg |= SCD_B_GUSBCFG_SRPCAP; /* a-device - set SRPCAP */
	SCD__WRITE(sdev, GUSBCFG, gusbcfg);
}

/* SRPCap쥸β - Device⡼ɻ */
/* A[?S] */
static void clear_srpcap_regs(struct scd_device *sdev)
{
	u32 gusbcfg;

	gusbcfg = SCD__READ(sdev, GUSBCFG);
	/* b-deviceSRPCAP򥻥åȤƤȡD-饤Ű */
	/* 1.3[V]ȤʤäƤޤ᥯ꥢ */
	gusbcfg &= ~SCD_B_GUSBCFG_SRPCAP; /* b-device - clear SRPCAP */
	SCD__WRITE(sdev, GUSBCFG, gusbcfg);
}

/* Ͼ */
/* A[iS] */
static void suspend_regs(struct scd_device *sdev)
{
	u32 pcgcctl;

	pcgcctl = SCD__READ(sdev, PCGCCTL);
	pcgcctl |= (SCD_B_PCGCCTL_GATEHCLK |
		    SCD_B_PCGCCTL_STOPPCLK);
	SCD__WRITE(sdev, PCGCCTL, pcgcctl);
}

/* ư */
/* A[iS] */
static void resume_regs(struct scd_device *sdev)
{
	u32 pcgcctl;

	pcgcctl = SCD__READ(sdev, PCGCCTL);
	pcgcctl &= ~(SCD_B_PCGCCTL_GATEHCLK |
		     SCD_B_PCGCCTL_STOPPCLK);
	SCD__WRITE(sdev, PCGCCTL, pcgcctl);
}

/* 
 * OTGơȥޥ
 */
/* ܴؿ deivce_trans_state */
/* A[is] */
static int device_trans_state(struct scd_device_control *scd,
			      enum scd_device_trigger trigger)
{
	int status;

	BUG_ON(in_atomic());
	BUG_ON(in_interrupt());

	mutex_lock(&scd->state_mutex);

	status = 0;

device_trans_state_again:
	DBG(3, "device_trans_state s=%x,t=%d\n", scd->state, trigger);

	/* exit state - level 2 : probed */
	switch (scd_device_get_state(scd, 2)) {
	case SCD_DEVICE_S_PROBED:
		switch (trigger) {
		case SCD_DEVICE_T_REMOVE:
		case SCD_DEVICE_T_SHUTDOWN:
			exec_terminate(scd);
			goto device_trans_state_trans;
		default:
			break;
		}
		break;
	default:
		break;
	}
	/* exit state - level 1 : probed_init, stop, idle, b_periph, a_host */
	switch (scd_device_get_state(scd, 1)) {
	case SCD_DEVICE_S_IDLE:
		switch (trigger) {
		case SCD_DEVICE_T_STOP_CONTROL:
		case SCD_DEVICE_T_CHANGE_VBUS:
			exec_inactivate(scd);
			goto device_trans_state_trans;
		default:
			break;
		}
		break;
	case SCD_DEVICE_S_B_PERIPH:
		switch (trigger) {
		case SCD_DEVICE_T_STOP_GADGET:
			exec_stop_gadget(scd);
			goto device_trans_state_trans;
		default:
			break;
		}
		break;
	case SCD_DEVICE_S_A_HOST:
		switch (trigger) {
		case SCD_DEVICE_T_STOP_HOST:
			disable_irq(scd->irqn);
			exec_stop_host(scd);
			enable_irq(scd->irqn);
			goto device_trans_state_trans;
		default:
			break;
		}
		break;
	default:
		break;
	}
	/* exit state - level 0 : ab_vbused, ab_silent,
				a_h_discon, a_h_debounce, a_h_conn */
	switch (scd_device_get_state(scd, 0)) {
	case SCD_DEVICE_S_A_H_CONN:
		switch (trigger) {
		case SCD_DEVICE_T_DISCONNECT:
			exec_disconnect(scd);
			goto device_trans_state_trans;
		default:
			break;
		}
		break;
	default:
		break;
	}

device_trans_state_trans:
	/* transition state - level 2 : probed */
	switch (scd_device_get_state(scd, 2)) {
	case SCD_DEVICE_S_POWERED_ON:
		if (trigger == SCD_DEVICE_T_PROBE) {
			scd->state = SCD_DEVICE_S_PROBED;
			goto device_trans_state_entry;
		}
		break;
	case SCD_DEVICE_S_PROBED:
		if (trigger == SCD_DEVICE_T_REMOVE) {
			scd->state = SCD_DEVICE_S_REMOVED;
			goto device_trans_state_entry;
		}
		else if (trigger == SCD_DEVICE_T_SHUTDOWN) {
			scd->state= SCD_DEVICE_S_POWERED_OFF;
			goto device_trans_state_entry;
		}
		break;
	default:
		break;
	}
	/* transition state - level 1 : stop, b_periph, a_host */
	/* SCD_DEVICE_S_PROBED == SCD_DEVICE_S_PROBED_INIT */
	switch (scd_device_get_state(scd, 1)) {
	case SCD_DEVICE_S_PROBED_INIT:
		if (trigger == SCD_DEVICE_T_STOP) {
			scd->state = SCD_DEVICE_S_STOP;
			goto device_trans_state_entry;
		}
		break;
	case SCD_DEVICE_S_STOP:
		if (trigger == SCD_DEVICE_T_START_CONTROL) {
			status = try_start_control(scd);
			if (status == 0) {
				scd->state = SCD_DEVICE_S_IDLE_INIT;
			}
			goto device_trans_state_entry;
		}
		else if (trigger == SCD_DEVICE_T_REMOVE) {
			scd->state = SCD_DEVICE_S_REMOVED;
			goto device_trans_state_entry;
		}
		break;
	case SCD_DEVICE_S_B_PERIPH:
		if (trigger == SCD_DEVICE_T_STOP_GADGET) {
			scd->state = SCD_DEVICE_S_IDLE_INIT;
			goto device_trans_state_entry;
		}
		else if (trigger == SCD_DEVICE_T_CHANGE_VBUS) {
			goto device_trans_state_entry;
		}
		break;
	case SCD_DEVICE_S_A_HOST:
		if (trigger == SCD_DEVICE_T_STOP_HOST) {
			scd->state = SCD_DEVICE_S_IDLE_INIT;
			goto device_trans_state_entry;
		}
		else if (trigger == SCD_DEVICE_T_CHANGE_VBUS) {
			goto device_trans_state_entry;
		}
		break;
	case SCD_DEVICE_S_IDLE:
		if (trigger == SCD_DEVICE_T_STOP_CONTROL) {
			status = try_stop_control(scd);
			scd->state = SCD_DEVICE_S_STOP;
			goto device_trans_state_entry;
		}
		else if (trigger == SCD_DEVICE_T_CHANGE_VBUS) {
			scd->state = SCD_DEVICE_S_IDLE;
			goto device_trans_state_entry;
		}
		else if (trigger == SCD_DEVICE_T_START_GADGET) {
			/* scd_irqγ߿ʬθ֤ܤ*/
			/* ޤdisable_irqˤƳߤʤ褦ˤ */
			disable_irq(scd->irqn);
			status = try_start_gadget(scd);
			if (status == 0) {
				scd->state = SCD_DEVICE_S_B_PERIPH;
			}
			else {
				scd->state = SCD_DEVICE_S_IDLE;
			}
			enable_irq(scd->irqn);
			goto device_trans_state_entry;
		}
		else if (trigger == SCD_DEVICE_T_START_HOST) {
			/* scd_irqγ߿ʬθ֤ܤ*/
			/* ޤdisable_irqˤƳߤʤ褦ˤ */
			disable_irq(scd->irqn);
			status = try_start_host(scd);
			if (status == 0) {
				scd->state = SCD_DEVICE_S_A_H_DISCON;
			}
			else {
				scd->state = SCD_DEVICE_S_IDLE;
			}
			enable_irq(scd->irqn);
			goto device_trans_state_entry;
		}
		break;
	default:
		break;
	}
	/* transition state - level 0 : idle_init, AB_VBUSED, ab_silent */
	/* SCD_DEVICE_S_IDLE == SCD_DEVICE_S_IDLE_INIT */
	switch (scd_device_get_state(scd, 0)) {
	case SCD_DEVICE_S_IDLE_INIT:
		if (trigger == SCD_DEVICE_T_TEST_BUS) {
			if (test_bus(scd)) {
				scd->state = SCD_DEVICE_S_AB_VBUSED;
			}
			else {
				scd->state = SCD_DEVICE_S_AB_SILENT;
			}
		}
		goto device_trans_state_entry;
	case SCD_DEVICE_S_A_H_DISCON:
		if (trigger ==SCD_DEVICE_T_CONNECT) {
			scd->state = SCD_DEVICE_S_A_H_DEBOUNCE;
		}
		goto device_trans_state_entry;
	case SCD_DEVICE_S_A_H_DEBOUNCE:
		if (trigger ==SCD_DEVICE_T_DEBOUNCE) {
			scd->state = SCD_DEVICE_S_A_H_CONN;
		}
		else if (trigger ==SCD_DEVICE_T_DISCONNECT) {
			scd->state = SCD_DEVICE_S_A_H_DISCON;
		}
		goto device_trans_state_entry;
	case SCD_DEVICE_S_A_H_CONN:
		if (trigger ==SCD_DEVICE_T_DISCONNECT) {
			scd->state = SCD_DEVICE_S_A_H_DISCON;
		}
		goto device_trans_state_entry;
	default:
		break;
	}
	status = -EBUSY;
	goto device_trans_state_end;

device_trans_state_entry:
	/* clear trigger for trans again */
	/* entry state - level 1 : stop, idle */
	/* entry state - level 0 : ab_silent, AB_VBUSED */
	/* (goto ..._again)ˤϡstatusѲ̵ */
	switch (scd->state) {
	case SCD_DEVICE_S_PROBED:
		exec_begin(scd);
		trigger = SCD_DEVICE_T_STOP;
		goto device_trans_state_again;
	case SCD_DEVICE_S_IDLE:
		exec_activate(scd);
		trigger = SCD_DEVICE_T_TEST_BUS;
		goto device_trans_state_again;
	case SCD_DEVICE_S_AB_VBUSED:
		exec_resume(scd);
		break;
	case SCD_DEVICE_S_AB_SILENT:
		exec_suspend(scd);
		break;
	case SCD_DEVICE_S_A_H_CONN:
		exec_connect(scd);
		break;
	default:
		break;
	}

device_trans_state_end:
	mutex_unlock(&scd->state_mutex);

	WARN(status < 0, "usb_scd: Unexpected trigger %d at state=%x(%d)\n",
	     trigger, scd->state, status);
	DBG(3, "device_trans_state_end %d at state=%x(%d)\n",
	    trigger, scd->state, status);

	return status;
}

/*
 *	ơȥޥκƵŪȥꥬ
 */
struct scd_trigger_work {
	struct scd_device_control *scd;
	enum scd_device_trigger trigger;
	struct work_struct work;
};

static void signal_trigger(struct work_struct *work)
{
	struct scd_trigger_work *stw;

	stw = container_of(work, struct scd_trigger_work, work);
	BUG_ON(!stw);

	device_trans_state(stw->scd, stw->trigger);
	kfree(stw);
}

static int device_trans_state_interrupt(struct scd_device_control *scd,
					 enum scd_device_trigger trigger)
{
	int status;
	struct scd_trigger_work *stw;

	stw = (struct scd_trigger_work *)kmalloc(sizeof(*stw), GFP_ATOMIC);
	WARN(!stw, "usb_scd cannot alloc memory for scd_trigger_work %x,%d\n",
	     scd->state, trigger);
	if (!stw) {
		scd_warn(&scd->sdev, "work alloc failed\n");
		return -ENOMEM;
	}

	stw->scd = scd;
	stw->trigger = trigger;
	INIT_WORK(&stw->work, signal_trigger);
	status = queue_work(scd->trigger_wq, &stw->work);
	if (status < 0) {
		scd_error(&scd->sdev, "queue_work failed %d\n", status);
		kfree(stw);
		BUG();
	}

	return status;
}

/*
 *	̤ؤΥ٥
 */

/* event container - ߤtaskletؤΥ٥ȽƥȾ费 */
struct scd_otg_event_container {
	union usb_otg_event event;
	struct list_head event_list;
};

/*
 * ؿsignal_event
 * signal_event_interruptϿ줿OTG٥ȤtaskletΥƥȤ
 * ̤إ٥ȤΤ
 */
/* A[is] */
static void signal_event(unsigned long data)
{
	unsigned long flags;
	struct scd_device_control *scd;

	struct list_head temp_anchor;
	struct scd_otg_event_container *ec;
	struct scd_otg_event_container *temp_ec;

	DBG(2, ">>\n");

	scd = (struct scd_device_control *)data;

	/* ¾θ¸٥Ȥqueueذư
	   scdΥ٥queueڤΥƼ¹Ԥ */

	INIT_LIST_HEAD(&temp_anchor);

	/* ¾泫 */
	spin_lock_irqsave(&scd->lock, flags);

	/* ϿƤ뤹٤ƤΥ٥Ȥ򲾥ꥹȤذư */
	list_for_each_entry_safe(ec, temp_ec, &scd->event_anchor, event_list) {
		/* ٥ȤqueueǼ¹ԤΤǰqueuetailذư */
		list_move_tail(&ec->event_list, &temp_anchor);
	}

	/* ¾潪λ */
	spin_unlock_irqrestore(&scd->lock, flags);

	/* ꥹȤΥ٥ȤˤĤƽ */
	list_for_each_entry_safe(ec, temp_ec, &temp_anchor, event_list) {
		list_del(&ec->event_list);
		/* HookͥǾ̤ - notifyHook⥸塼뤫Ƥ */
		if (scd->otgc.otg_core == NULL) {
			goto skip_notify;
		}
		if (scd->notify_hook) {
			scd->notify_hook(&ec->event);
		}
		else {
			usb_otg_core_notify(scd->otgc.otg_core, &ec->event);
		}
	skip_notify:
		/* ƥʤ˴ */
		kfree(ec);
	}
}

/*
 * ؿsignal_event_interrupt
 * ߥƥȤOTG٥Ȥtasklet𤷤OTGơȥޥ
 * Ϥ */
/* ߥƥ */
/* ƤӽФؿˤscd->lockspin_lockɬ */
/* A[IS] */
static void signal_event_interrupt(struct scd_device *sdev,
				   union usb_otg_event *event)
{
	struct scd_device_control *scd;
	struct scd_otg_event_container *event_container;

	scd = sdev_to_scd(sdev);

	DBG(2, "<<=>>\n");

	/* ٥ȥƥʤγ */
	event_container = (struct scd_otg_event_container *)
	    kmalloc(sizeof(struct scd_otg_event_container), GFP_ATOMIC);
	if (!event_container) {
		/* kmallocǤʤä硢ߤμꤳܤˤʤ */
		return;
	}

	/* ٥ȤΥԡ */
	memcpy(&event_container->event, event, sizeof(*event));

	INIT_LIST_HEAD(&event_container->event_list);

	list_add_tail(&event_container->event_list, &scd->event_anchor);

	DBG(3, "t=%d v=%d\n", event->type, event->cid.value);

	/* taskletΥ塼 */
	tasklet_schedule(&scd->event_tasklet);
}

/* CIDͤͤȰۤʤС̤ΤƺǿͤȤƵ */
/* ߥƥ */
/* ƤӽФؿˤscd->lockspin_lockɬ */
/* A[IS]  */
static void notify_cid(struct scd_device_control *scd, int current_cid)
{
	union usb_otg_event otg_event;

	otg_event.type = USB_OTG_EVENT_TYPE_CID;
	otg_event.cid.port = 0;
	if (scd->cid != current_cid) { /* Ѳ̵ */
		otg_event.cid.value = current_cid;
		signal_event_interrupt(&scd->sdev, &otg_event);
		scd->cid = current_cid;
	}
}

/* VbusͤͤȰۤʤС̤ΤƺǿͤȤƵ */
/*   b-deviceξ OFF<->LOW<->SESS */
/*   a-deviceξ OFF<->LOW<->SESS<->VALID */
/*   䴰 */
/* ߥƥ */
/* ƤӽФؿˤscd->lockspin_lockɬ */
/* A[IS]  */
static void notify_vbus(struct scd_device_control *scd,
			enum usb_otg_vbus_stat current_vbus)
{			
	union usb_otg_event otg_event;

	otg_event.type = USB_OTG_EVENT_TYPE_VBUS;
	otg_event.vbus.port = 0;
	while (scd->vbus != current_vbus) { /* Ѳ̵ */
		if (scd->vbus < current_vbus) {
			scd->vbus++;
		}
		else {
			scd->vbus--;
		}
		otg_event.vbus.vbus_stat = scd->vbus;
		signal_event_interrupt(&scd->sdev, &otg_event);
		device_trans_state_interrupt(scd, SCD_DEVICE_T_CHANGE_VBUS);
	}
}

/* Bus֤θͤ¨Τ */
/* ֤: State machineAB_VBUSEDܾ˰פtrueʳfalse */
/* A[IS] */
static bool check_and_notify_bus(struct scd_device_control *scd)
{
	int cid;
	enum usb_otg_vbus_stat vbus;
	struct scd_device *sdev;
	u32 gotgctl;

	sdev = &scd->sdev;

	gotgctl = SCD__READ(sdev, GOTGCTL);
	cid = gotgctl & SCD_B_GOTGCTL_CONIDSTS ? B_DEVICE : A_DEVICE;

	if (cid == B_DEVICE) {
		clear_srpcap_regs(sdev); /* b-device */
	}
	else {
		set_srpcap_regs(sdev); /* a-device */
	}

	vbus = (is_device(sdev) ?
		(gotgctl & SCD_B_GOTGCTL_BSESVLD ? USB_OTG_VBUS_STAT_SESS :
		 USB_OTG_VBUS_STAT_OFF) : /* DeviceModeξ Session or OFF */
		(gotgctl & SCD_B_GOTGCTL_ASESVLD ? USB_OTG_VBUS_STAT_VALID :
		 USB_OTG_VBUS_STAT_OFF)); /* HostModeξ Valid or OFF */

	/* Х֤ν - Ѳʤ */
	notify_cid(scd, cid);
	notify_vbus(scd, vbus);

	return (vbus != USB_OTG_VBUS_STAT_OFF);
}

#define DELAYED_CHECK_BUS_TIME 10000 /* 10[ms] */
#define DELAYED_CHECK_BUS_MAX  1000 /* 1000*_TIME = 10[s] */

static int delayed_check_bus(struct scd_timer *timer)
{
	bool active;
	unsigned long flags;
	struct scd_device_control *scd;

	scd = (struct scd_device_control *)timer->id;

	/* ¾泫 */
	spin_lock_irqsave(&scd->lock, flags);

	active = check_and_notify_bus(scd);

	/* ¾潪λ */
	spin_unlock_irqrestore(&scd->lock, flags);

	/* stop_hostVbus OFFԤƸХߤ */
	/* Host DISCONN100ms夫10[s]֥ư */
	if (active && !atomic_dec_and_test(&scd->delayed_check_bus_count)) {
		/* Vbus OFF10møBUS֤ƥå */
		scd_device_add_timer(&scd->sdev, DELAYED_CHECK_BUS_TIME,
				     delayed_check_bus, scd);
	}
	else {
		atomic_set(&scd->delayed_check_bus_count, 0);
	}

	return 0;
}

/*
 *	scd irq - ߥϥɥ
 */
/* A[Is] */
static irqreturn_t scd_irq(int irq, void *data)
{
	struct scd_intf *intf;
	struct scd_device_control *scd;
	struct scd_device *sdev;
	struct scd_control *control;

	irqreturn_t ret;
	u32 gotgctl;
	u32 gintsts;
	u32 gotgint;

	DBG(2, ">>\n");

	intf = NULL;
	control = NULL;
	ret = IRQ_NONE;

	scd = (struct scd_device_control *)data;
	sdev = &scd->sdev;

	spin_lock(&scd->lock);

	gotgctl = SCD__READ(sdev, GOTGCTL);
	gintsts = SCD__READ(sdev, GINTSTS);
	gotgint = SCD__READ(sdev, GOTGINT);
	DBG(3, "gintsts=0x%08x gotgint=0x%08x\n", gintsts, gotgint);
	if (gintsts & SCD_B_GINTSTS_MODEMIS) {
		scd_error(sdev, "otg mode missmatch\n");
		SCD__WRITE(sdev, GINTSTS, SCD_B_GINTSTS_MODEMIS);

		ret = IRQ_HANDLED;
		goto scd_irq_unlock;
	}
	else if (gintsts & SCD_B_GINTSTS_CONIDSTSCHNG) {
		int cid;

		cid = gotgctl & SCD_B_GOTGCTL_CONIDSTS ? B_DEVICE : A_DEVICE;

		if (cid == B_DEVICE) {
			clear_srpcap_regs(sdev); /* b-device */
		}
		else {
			set_srpcap_regs(sdev); /* a-device */
		}

		/* CID */
		notify_cid(scd, cid);

		/* DeviceModePullup off */
		control_pullup(sdev, false);

		SCD__WRITE(sdev, GINTSTS, SCD_B_GINTSTS_CONIDSTSCHNG);

		ret = IRQ_HANDLED;
		goto scd_irq_unlock;
	}
	else if (gintsts & SCD_B_GINTSTS_SESSREQINT) {
		/* Vbus(ON) */
		/* b-deviceΤ */
		notify_vbus(scd, USB_OTG_VBUS_STAT_SESS);
		intf = get_gadget_intf();
		control = gadget_control(scd);
		if (intf && control && is_gadget(scd)) {
			intf->detect_vbus(control, true);
		}

		SCD__WRITE(sdev, GINTSTS, SCD_B_GINTSTS_SESSREQINT);

		ret = IRQ_HANDLED;
		goto scd_irq_unlock;
	}

	if (gotgint & SCD_B_GOTGINT_SESENDDET) {
		/* Vbus(OFF) */
		/* b-deviceΤ */
		notify_vbus(scd, USB_OTG_VBUS_STAT_OFF);
		intf = get_gadget_intf();
		control = gadget_control(scd);
		if (intf && control && is_gadget(scd)) {
			intf->detect_vbus(control, false);
		}

		SCD__WRITE(sdev, GOTGINT, SCD_B_GOTGINT_SESENDDET);

		ret = IRQ_HANDLED;
		goto scd_irq_unlock;
	}
	else if (gotgint & SCD_B_GOTGINT_DBNCEDONE) {
		device_trans_state_interrupt(scd, SCD_DEVICE_T_DEBOUNCE);

		SCD__WRITE(sdev, GOTGINT, SCD_B_GOTGINT_DBNCEDONE);

		ret = IRQ_HANDLED;
		goto scd_irq_unlock;
	}

scd_irq_unlock:
	spin_unlock(&scd->lock);

	/* װotgξϽλ */
	if (ret == IRQ_HANDLED) {
		goto scd_irq_end;
	}

	/* װotgǤ̵ */
	if (is_gadget(scd)) {
		intf = get_gadget_intf();
		control = gadget_control(scd);
	}
	else if (is_hcd(scd)) {
		intf = get_hcd_intf();
		control = hcd_control(scd);
	}

	if (intf && control) {
		ret = intf->irq(control);
	}

	/* ïʤä硢ɽƽλ */
	if (ret != IRQ_HANDLED) {
		err("IRQ_NONE %x %08x, %08x/%08x %08x\n",
		    scd->state, SCD__READ(sdev, GOTGCTL),
		    gintsts, SCD__READ(sdev, GINTMSK), gotgint);
	}

scd_irq_end:

	DBG(2, "<<\n");

	return ret;
}

/*
 *	scd_device
 */


/*
  timerϢ
 */

/* ޥꥹȤ˴ޤޤ륿ޤ椫Ǥ㤤ޤλ */
/* ƥ¸HWޤꤹ */
/* A[?S] */
static void update_sysdep_timeout(struct scd_device_control *scd)
{
	struct scd_sysdep_intf *sysdep_intf;
	struct scd_control* control;
	struct scd_device_timer *dtimer;
	scd_usec expire;

	sysdep_intf = get_sysdep_intf();
	control = sysdep_control(scd);

	if (!sysdep_intf || !control) {
		return;
	}

	/* ösysdep timeoutϥ󥻥뤹 */
	sysdep_intf->cancel_timeout(control);
	/* timeoutԤdtimerкǤ㤤ޤꤹ */
	if (!list_empty(&scd->timer_anchor)) {
		dtimer = list_first_entry(&scd->timer_anchor,
					  struct scd_device_timer, timer_list);
		expire = dtimer->t.expire;
		if (scd_cmp_usec(expire, sysdep_intf->get_time(control)) < 0) {
			expire = sysdep_intf->get_time(control);
		}
		sysdep_intf->set_timeout(control, expire);
		DBG(2, "%d:set_timeout(%d) -> expire=%d\n",
		    (s32)sysdep_intf->get_time(control), (s32)expire,
		    (s32)dtimer->t.expire);
	}
}

/* ٤ƤtimerŪߤ */
/* A[iS] */
static void stop_all_timer(struct scd_device_control *scd)
{
	struct scd_sysdep_intf *sysdep_intf;
	struct scd_control *control;

	sysdep_intf = get_sysdep_intf();
	control = sysdep_control(scd);
	if (sysdep_intf && control) {
		sysdep_intf->cancel_timeout(control);
	}
}

/* stop_all_timerޥꥹȤƻư */
/* A[iS] */
static void restart_all_timer(struct scd_device_control *scd)
{
	update_sysdep_timeout(scd);
}

/* ٤Ƥtimer򥭥󥻥뤹롧struct platform_driver:remove, shutdown */
/* scd_device_unset_intfƤФ */
/* A[iS] */
static void cancel_all_timer(struct scd_device_control *scd)
{
	unsigned long flags;
	struct scd_sysdep_intf *sysdep_intf;
	struct scd_control *control;

	struct list_head temp_anchor;
	struct scd_device_timer *dtimer;
	struct scd_device_timer *temp_dtimer;

	sysdep_intf = get_sysdep_intf();
	control = sysdep_control(scd);
	if (sysdep_intf && control) {
		sysdep_intf->cancel_timeout(control);
	}

	/* ¾θ¸ޤqueueذư
	   scdΥqueueڤΥƼ¹Ԥ */

	INIT_LIST_HEAD(&temp_anchor);

	/* ¾泫 */
	spin_lock_irqsave(&scd->lock, flags);

	/* ϿƤ뤹٤ƤtimerˤĤƽ */
	list_for_each_entry_safe(dtimer, temp_dtimer, &scd->timer_anchor,
				 timer_list) {
		/* ХåʤΤä˥塼󥰤ɬפʤ
		   Ȥꤢtailդ*/
		list_move_tail(&dtimer->timer_list, &temp_anchor);
	}

	/* ¾潪λ */
	spin_unlock_irqrestore(&scd->lock, flags);

	/* ϿƤ뤹٤ƤtimerˤĤƽ */
	list_for_each_entry_safe(dtimer, temp_dtimer, &temp_anchor,
				 timer_list) {
		list_del(&dtimer->timer_list);
		kfree(dtimer);
	}
}

/* A[??] */
static struct scd_timer *_add_timer(
	struct scd_device *sdev, scd_usec timeout,
	scd_timer_callback cb, void *id)
{
	unsigned long flags;
	struct scd_device_control *scd;
	struct scd_sysdep_intf *sysdep_intf;
	struct scd_control *control;
	struct scd_timer *timer;
	struct scd_device_timer *dtimer;
	struct scd_device_timer *near_dtimer;
	scd_usec expire;

	scd = sdev_to_scd(sdev);

	sysdep_intf = get_sysdep_intf();
	control = sysdep_control(scd);
	timer = NULL;

	if (!sysdep_intf || !control) {
		goto add_timer_end;
	}
	else {
		expire = timeout + sysdep_intf->get_time(control);
	}

	/* 椫ƤФ硢GFPե饰ATOMIC */
	dtimer = (struct scd_device_timer*)kmalloc(sizeof(*dtimer),
						   (in_interrupt() ? GFP_ATOMIC :
						    GFP_KERNEL));
	if (dtimer == NULL) {
		scd_error(sdev, "timer allocation error %lu,%x,%x\n",
			  timeout, (u32)cb, (u32)id);
		goto add_timer_end;
	}
	dtimer->t.expire = expire;
	dtimer->t.cb = cb;
	dtimer->t.id = id;
	INIT_LIST_HEAD(&dtimer->timer_list);

	/* scd timer list¾泫 */
	/* ɤΥƥȤƤФƤɤ褦irqsave¾ */
	spin_lock_irqsave(&scd->lock, flags);

	DBG(3, "compare expire [%d]<=>", (s32)expire);
	list_for_each_entry(near_dtimer, &scd->timer_anchor, timer_list) {
		DBGS(3, "[%d]", (s32)near_dtimer->t.expire);
		if (scd_cmp_usec(near_dtimer->t.expire, expire) > 0) {
			break;
		}
	}
	DBGS(3, "\n");
	/* timer listtimeoutʤΤǡӤnear_timer */
	/* ᤯timeoutϤľadd */
	list_add_tail(&dtimer->timer_list, &near_dtimer->timer_list);

	/* ּ㤤ޤsysdeptimeoutꤹ */
	update_sysdep_timeout(scd);

	timer = &dtimer->t;

	spin_unlock_irqrestore(&scd->lock, flags);

add_timer_end:

	return timer;
}

/* A[??] */
static int cancel_timer(struct scd_device *sdev, struct scd_timer *timer)
{
	unsigned long flags;
	struct scd_device_control *scd;
	struct scd_device_timer *dtimer;

	scd = sdev_to_scd(sdev);

	/* ¾泫 */
	/* ɤΥƥȤƤФƤɤ褦irqsave¾ */
	spin_lock_irqsave(&scd->lock, flags);

	dtimer = container_of(timer, struct scd_device_timer, t);

	/* ꥿ޤ˼㤤ޤsysdeptimeoutꤹ */
	list_del(&dtimer->timer_list);
	update_sysdep_timeout(scd);

	/* ¾潪λ */
	spin_unlock_irqrestore(&scd->lock, flags);

	kfree(dtimer);

	return 0;
}

/* A[??] */
static scd_usec get_time(struct scd_device *sdev)
{
	scd_usec time;

	struct scd_device_control *scd;
	struct scd_sysdep_intf *sysdep_intf;
	struct scd_control *control;

	scd = sdev_to_scd(sdev);

	sysdep_intf = get_sysdep_intf();
	control = sysdep_control(scd);

	if (!sysdep_intf || !control) {
		time = 0;
	}
	else {
		time = sysdep_intf->get_time(control);
	}

	return time;
}

/* hcdǥХ³ǤΤ */
/* A[Is] */
static int connect(struct scd_device *sdev, bool onoff, int port)
{
	struct scd_device_control *scd;

	scd = sdev_to_scd(sdev);

	/* spinlockϳǤν */

	if (!onoff) {	/* DEVICEǻ˥եȥꥻåȤ򤫤ʤƤϤʤʤ */
		spin_lock(&scd->lock);

		/* OFF */
		resume_regs(sdev);

		/* եȥꥻå */
		SCD__WRITE(sdev, GRSTCTL, SCD_B_GRSTCTL_CSFTRST);
		udelay(1); /* at least 3PHY clock */

		/* 쥸 */
		init_regs(sdev, false);
		start_regs(sdev);

		spin_unlock(&scd->lock);

		/* CID/VbusѲ - ǻ
		   եȥꥻåȤ쥸ޤǤȯCID/VbusѲ
		   Τ롣start_regsʹߤѲϳߤȯ */
		/* եȥꥻåľϥ쥸ͤHW֤ˤʤ뤿ᡢ1[s]
                   ¬ΡCID35[ms]Υե륿뤿ᡣ */
		if (atomic_xchg(&scd->delayed_check_bus_count,
				DELAYED_CHECK_BUS_MAX) == 0) {
			scd_device_add_timer(&scd->sdev,
					     DELAYED_CHECK_BUS_TIME * 100,
					     delayed_check_bus, scd);
		}

		device_trans_state_interrupt(scd, SCD_DEVICE_T_DISCONNECT);
	}
	else {
		spin_lock(&scd->lock);

		/* CID/VbusѲ - ³
		   PortPower ONˤVbus ONΤ*/
		check_and_notify_bus(scd);

		spin_unlock(&scd->lock);

		device_trans_state_interrupt(scd, SCD_DEVICE_T_CONNECT);
	}

	return 0;
}

/* hcdŸONΤ */
/* A[?s] */
static int power(struct scd_device *sdev, bool onoff, int port)
{
	/* Power ONVbusErrorե饰򥯥ꥢ */
	if (onoff) {
		clear_bit(SCD_DEVICE_ATTR_VBUSERRORED, &sdev->attributes);
	}

	return 0;
}

/* hcdVbusErrorߤcallback */
/* A[Is] */
static int vbuserror(struct scd_device *sdev, int port)
{
	struct scd_device_control *scd;
	union usb_otg_event otg_event;

	scd = sdev_to_scd(sdev);

	/* VbusErrorե饰ΩäƤʤСե饰ΩƤƾ̤Τ */
	if (!test_and_set_bit(SCD_DEVICE_ATTR_VBUSERRORED, &sdev->attributes)) {
		otg_event.type = USB_OTG_EVENT_TYPE_VBUSERROR;
		spin_lock(&scd->lock);
		signal_event_interrupt(sdev, &otg_event);
		spin_unlock(&scd->lock);
	}

	return 0;
}

/* sysdepOvercurrentߤcallback */
/* A[Is] */
static int overcurrent(struct scd_device *sdev, int port)
{
	struct scd_device_control *scd;
	struct scd_intf *intf;
	struct scd_control *control;
	union usb_otg_event otg_event;

	scd = sdev_to_scd(sdev);

	/* VbusErrorե饰ΩäƤʤСե饰ΩƤƾ̤Τ */
	if (!test_and_set_bit(SCD_DEVICE_ATTR_VBUSERRORED, &sdev->attributes)) {
		otg_event.type = USB_OTG_EVENT_TYPE_VBUSERROR;
		spin_lock(&scd->lock);
		signal_event_interrupt(sdev, &otg_event);
		spin_unlock(&scd->lock);
	}

	intf = get_hcd_intf();
	control = hcd_control(scd);

	if (intf && control) {
		intf->overcurrent(control, port);
	}

	return 0;
}

/* gadgetLinkΥեȥꥻåȤ׵ᤵ */
/* PulluptrueˤSoftDisconn򥯥ꥢ */
/* A[?s] */
static int reset_link(struct scd_device *sdev, bool pullup)
{
	struct scd_device_control *scd;
	unsigned long flags;
	unsigned long allflags;

	scd = sdev_to_scd(sdev);

	/* OTG쥸ȶ˥֥å֥쥸Υ¾ */
	spin_lock_irqsave(&scd->lock, flags);

	/* OFF */
	resume_regs(sdev);

	/* ֥å֥쥸Υ¾ */
	spin_lock_irqsave(&sdev->lock, allflags);

	/* եȥꥻå */
	SCD__WRITE(sdev, GRSTCTL, SCD_B_GRSTCTL_CSFTRST);

	udelay(1); /* at least 3PHY clock */

	spin_unlock_irqrestore(&sdev->lock, allflags);

	/* 쥸 */
	init_regs(sdev, pullup);
	start_regs(sdev);

	spin_unlock_irqrestore(&scd->lock, flags);

	return 0;
}

/* sysdepTimer timeoutcallback */
/* ƥȤsysdepμ˰¸ */
/* A[?s] */
static int timeout(struct scd_device *sdev)
{
	unsigned long flags;
	struct scd_device_control *scd;
	struct scd_device_timer *dtimer;

	scd = sdev_to_scd(sdev);

	/* ɤΥƥȤƤФƤɤ褦irqsave¾ */
	spin_lock_irqsave(&scd->lock, flags);

	/* timeoutԤdtimer̵ϰ۾ */
	if (list_empty(&scd->timer_anchor)) {
		ERR("There are no waiting timer\n");
		spin_unlock_irqrestore(&scd->lock, flags);
		return -ENOENT;
	}
	dtimer = list_first_entry(&scd->timer_anchor, struct scd_device_timer,
				  timer_list);

	/* ॢȤޤ
	   ˼㤤ޤsysdeptimeoutꤹ */
	list_del(&dtimer->timer_list);
	update_sysdep_timeout(scd);

	/* ¾ */
	spin_unlock_irqrestore(&scd->lock, flags);

	/* timerХåθƤӽФ */
	dtimer->t.cb(&dtimer->t);

	/* Хåtimer֥Ȥޤdtimer֥ȤϺ */
	kfree(dtimer);

	return 0;
}

/* A[??] */
static int show_info(struct scd_device_control *scd,
		     enum scd_message_type type, const char *fmt, va_list ap)
{
	unsigned long flags;
	struct scd_device_timer *dtimer;
	struct scd_otg_event_container *ec;

	u8 i;

	spin_lock_irqsave(&scd->lock, flags);

	scd_info(&scd->sdev, "dev_name{%s}\n", dev_name(scd->sdev.dev));
	scd_info(&scd->sdev, "otg_state{0x%x}\n", scd->state);
	scd_info(&scd->sdev, "cid{%d}\n", scd->cid);
	scd_info(&scd->sdev, "vbus{%d}\n", scd->vbus);
	scd_info(&scd->sdev, "irqn{%d}\n", scd->irqn);
	scd_info(&scd->sdev, "attributes{0x%x}\n", scd->sdev.attributes);
	scd_info(&scd->sdev, "lock{%d}\n", (bool)spin_is_locked(&scd->lock));
	scd_info(&scd->sdev, "control");
	for (i = 0; i < NUM_CONTROL_TYPE; i++) {
		scd_print(&scd->sdev, "[%d]{0x%x} ", i, (u32)scd->control[i]);
	}
	scd_print(&scd->sdev, "\n");

	/* ϿƤ뤹٤ƤtimerϤ */
	scd_info(&scd->sdev, "current_time{%d}\n", (s32)get_time(&scd->sdev));
	i = 0;
	scd_info(&scd->sdev, "timer");
	list_for_each_entry(dtimer, &scd->timer_anchor, timer_list) {
		scd_print(&scd->sdev, "[%d]{%d} ", i, (s32)dtimer->t.expire);
		i++;
	}
	scd_print(&scd->sdev, "\n");

	/* ϿƤ뤹٤ƤeventϤ */
	i = 0;
	scd_info(&scd->sdev, "event");

	read_lock(&scd_lock);

	list_for_each_entry(ec, &scd->event_anchor, event_list) {
		scd_print(&scd->sdev, "[%d]{%d}", i, ec->event.type);
		i++;
	}

	read_unlock(&scd_lock);

	scd_print(&scd->sdev,  "\n");
	scd_info(&scd->sdev, "=== end of scd device info ===\n");

	spin_unlock_irqrestore(&scd->lock, flags);

	return 0;
}

/* A[??] */
static int vmessage(struct scd_device *sdev, enum scd_control_type recipient,
		    enum scd_message_type type, const char *fmt, va_list ap)
{
	int status;
	struct scd_device_control *scd;
	struct scd_control *control;
	struct scd_intf *intf;

	scd = sdev_to_scd(sdev);
	status = 0;

	if (recipient == SCD_CONTROL_TYPE_DEVICE) {
		DBG(0, "device vmessage %s\n", fmt);
#define CMPCMD(FMT, CMD) strncmp((FMT), (CMD), sizeof(CMD) - 1)
		if (!CMPCMD(fmt, "pullup on")) {
			control_pullup(sdev, true);
		}
		else if (!CMPCMD(fmt, "pullup off")) {
			control_pullup(sdev, false);
		}
		else {
			status = show_info(scd, type, fmt, ap);
		}
		goto vmessage_end;
	}

	control = scd->control[recipient];
	intf = intfs[recipient];

	if (intf && control && intf->vmessage) {
		status = intf->vmessage(control, type, fmt, ap);
	}
	else {
		status = -ENOENT;
	}

vmessage_end:

	return status;
}

/* A[??] */
static int message(struct scd_device *sdev, enum scd_control_type recipient,
		   enum scd_message_type type, const char *fmt, ...)
{
	va_list ap;
	int status;

	va_start(ap, fmt);

	status = vmessage(sdev, recipient, type, fmt, ap);

	va_end(ap);

	return status;
}

static struct scd_device_ops device_ops = {
	.add_timer = _add_timer,
	.cancel_timer = cancel_timer,
	.get_time = get_time,
	.connect = connect,
	.power = power,
	.vbuserror = vbuserror,
	.overcurrent = overcurrent,
	.reset_link = reset_link,
	.timeout = timeout,
	.vmessage = vmessage,
	.message = message,
};

/*
 *  ¾
 */
/* scd */
static void stop_scd(struct scd_device_control *scd)
{
	/* Link */
	disable_irq(scd->irqn);

	/* ٤Ƥtimer, tasklet, workqueueߤ */
	stop_all_timer(scd);
	tasklet_kill(&scd->event_tasklet);
	flush_workqueue(scd->trigger_wq);
}

/* scdƳ */
static void restart_scd(struct scd_device_control *scd)
{
	/* ٤ƤtimerƳ */
	restart_all_timer(scd);

	/* LinkߤƳ */
	enable_irq(scd->irqn);
}


/*
 * ܴؿ
 */
/* entry to PROBED */
static void exec_begin(struct scd_device_control *scd)
{
	int status;
	unsigned long flags;
	struct scd_sysdep_intf *sysdep_intf;

	/* sysdep start */
	sysdep_intf = get_sysdep_intf();
	BUG_ON(sysdep_intf == NULL); /* STOP֤sysdep probeѤ */
	status = sysdep_intf->base_intf.start(sysdep_control(scd));
	if (status < 0) {
		ERR("fail to start sysdep\n");
		BUG();
		return;
	}

	spin_lock_irqsave(&scd->lock, flags);

	/* Link쥸߳ */
	init_regs(&scd->sdev, false);

	/* ON */
	suspend_regs(&scd->sdev);
	set_bit(SCD_DEVICE_ATTR_SUSPENDED, &scd->sdev.attributes);

	spin_unlock_irqrestore(&scd->lock, flags);
}

/* exit from PROBED */
static void exec_terminate(struct scd_device_control *scd)
{
	int status;
	unsigned long flags;
	struct scd_sysdep_intf *sysdep_intf;

	spin_lock_irqsave(&scd->lock, flags);

	/* Link쥸 */
	stop_regs(&scd->sdev);

	/* ON */
	suspend_regs(&scd->sdev);
	set_bit(SCD_DEVICE_ATTR_SUSPENDED, &scd->sdev.attributes);

	spin_unlock_irqrestore(&scd->lock, flags);

	/* sysdep stop */
	sysdep_intf = get_sysdep_intf();
	BUG_ON(sysdep_intf == NULL);
	status = sysdep_intf->base_intf.stop(sysdep_control(scd));
	if (status < 0) {
		ERR("fail to stop sysdep\n");
		BUG();
		return;
	}
}

/* exit from IDLE */
static void exec_inactivate(struct scd_device_control *scd)
{
	unsigned long flags;

	spin_lock_irqsave(&scd->lock, flags);

	/* OFF */
	resume_regs(&scd->sdev);

	/* Link쥸 */
	stop_regs(&scd->sdev);

	/* ON */
	suspend_regs(&scd->sdev);
	set_bit(SCD_DEVICE_ATTR_SUSPENDED, &scd->sdev.attributes);

	spin_unlock_irqrestore(&scd->lock, flags);
}

/* before a transition to IDLE */
static int try_start_control(struct scd_device_control *scd)
{
	int status;

	/* ߥϥɥ */
	status = request_irq(scd->irqn, scd_irq, IRQF_DISABLED,
			     dev_name(scd->sdev.dev), (void *)scd);

	if (status == 0) {
		set_bit(SCD_DEVICE_ATTR_IRQ_REQUESTED, &scd->sdev.attributes);
	}

	atomic_set(&scd->delayed_check_bus_count, 0);

	return status;
}

static int try_stop_control(struct scd_device_control *scd)
{
	atomic_set(&scd->delayed_check_bus_count, 0);

	/* ߥե꡼ */
	if (test_and_clear_bit(SCD_DEVICE_ATTR_IRQ_REQUESTED,
			       &scd->sdev.attributes)) {
		free_irq(scd->irqn, (void *)scd);
	}

	return 0;
}

/* entry to IDLE */
static void exec_activate(struct scd_device_control *scd)
{
	unsigned long flags;

	spin_lock_irqsave(&scd->lock, flags);

	/* OFF */
	resume_regs(&scd->sdev);
	clear_bit(SCD_DEVICE_ATTR_SUSPENDED, &scd->sdev.attributes);

	/* Link쥸 */
	start_regs(&scd->sdev);

	spin_unlock_irqrestore(&scd->lock, flags);
}

/* evaluate to transit from IDLE_INIT */
static int test_bus(struct scd_device_control *scd)
{
	int status;
	unsigned long flags;

	spin_lock_irqsave(&scd->lock, flags);

	status = check_and_notify_bus(scd);

	spin_unlock_irqrestore(&scd->lock, flags);

	return status;
}

/* entry to AB_SILENT */
static void exec_suspend(struct scd_device_control *scd)
{
	unsigned long flags;

	spin_lock_irqsave(&scd->lock, flags);

	/* ON */
	suspend_regs(&scd->sdev);
	set_bit(SCD_DEVICE_ATTR_SUSPENDED, &scd->sdev.attributes);

	spin_unlock_irqrestore(&scd->lock, flags);
}

/* entry to AB_VBUSED */
static void exec_resume(struct scd_device_control *scd)
{
	unsigned long flags;

	spin_lock_irqsave(&scd->lock, flags);

	/* OFF */
	resume_regs(&scd->sdev);
	clear_bit(SCD_DEVICE_ATTR_SUSPENDED, &scd->sdev.attributes);

	spin_unlock_irqrestore(&scd->lock, flags);
}

/* try to start gadget from AB_VBUSED */
static int try_start_gadget(struct scd_device_control *scd)
{
	unsigned long flags;
	struct scd_intf *intf;
	struct scd_control *control;

	intf = get_gadget_intf();
	control = gadget_control(scd);

	if (!intf || !control) {
		return -EPERM;
	}

	spin_lock_irqsave(&scd->lock, flags);

	/* OFF */
	resume_regs(&scd->sdev);
	clear_bit(SCD_DEVICE_ATTR_SUSPENDED, &scd->sdev.attributes);

	spin_unlock_irqrestore(&scd->lock, flags);

	/* start gadget intf */
	return intf->start(control);
}

/* exit from B_PERIPH */
static void exec_stop_gadget(struct scd_device_control *scd)
{
	struct scd_intf *intf;
	struct scd_control *control;

	intf = get_gadget_intf();
	control = gadget_control(scd);

	if (!intf || !control) {
		return;
	}

	/* gadget stop */
	intf->stop(control);
}

/* try to start host from AB_SILENT */
static int try_start_host(struct scd_device_control *scd)
{
	int status;
	unsigned long flags;
	struct scd_intf *intf;
	struct scd_sysdep_intf *sysdep_intf;
	struct scd_control *control;

	intf = get_hcd_intf();
	control = hcd_control(scd);

	if (!intf || !control) {
		return -EPERM;
	}

	spin_lock_irqsave(&scd->lock, flags);

	/* OFF */
	resume_regs(&scd->sdev);
	clear_bit(SCD_DEVICE_ATTR_SUSPENDED, &scd->sdev.attributes);

	spin_unlock_irqrestore(&scd->lock, flags);

	/* host start */
	status = intf->start(control);

	if (status == 0) {
		sysdep_intf = get_sysdep_intf();
		control = sysdep_control(scd);

		if (sysdep_intf && control) {
			sysdep_intf->power(control, true, 0);
		}
	}

	return status;
}

/* exit from A_HOST */
static void exec_stop_host(struct scd_device_control *scd)
{
	struct scd_intf *intf;
	struct scd_sysdep_intf *sysdep_intf;
	struct scd_control *control;

	sysdep_intf = get_sysdep_intf();
	control = sysdep_control(scd);

	if (sysdep_intf && control) {
		sysdep_intf->power(control, false, 0);
	}

	intf = get_hcd_intf();
	control = hcd_control(scd);

	if (!intf || !control) {
		return;
	}

	/* host stop */
	intf->stop(control);

	/* stop host10mðʹߤ˰Vbus OFFΤԤ */
	if (atomic_xchg(&scd->delayed_check_bus_count,
			DELAYED_CHECK_BUS_MAX) == 0) {
		scd_device_add_timer(&scd->sdev, DELAYED_CHECK_BUS_TIME,
				     delayed_check_bus, scd);
	}
}

/* entry to A_H_CONN */
static void exec_connect(struct scd_device_control *scd)
{
	union usb_otg_event otg_event;

	otg_event.type = USB_OTG_EVENT_TYPE_CON;
	otg_event.con.value = 1; /* as connect */
	signal_event_interrupt(&scd->sdev, &otg_event);
}

/* exit from A_H_CONN */
static void exec_disconnect(struct scd_device_control *scd)
{
	union usb_otg_event otg_event;

	otg_event.type = USB_OTG_EVENT_TYPE_CON;
	otg_event.con.value = 0; /* as disconnect */
	signal_event_interrupt(&scd->sdev, &otg_event);
}

/* force stop scd - STOP֤ؤζ */
static void force_stopping_scd(struct scd_device_control *scd)
{
	enum scd_device_state state;
	
	state = scd_device_get_state(scd, 1);
	if (state == SCD_DEVICE_S_A_HOST) {
		device_trans_state(scd, SCD_DEVICE_T_STOP_HOST);
	}
	else if (state == SCD_DEVICE_S_B_PERIPH) {
		device_trans_state(scd, SCD_DEVICE_T_STOP_GADGET);
	}

	state = scd_device_get_state(scd, 1);
	if (state == SCD_DEVICE_S_IDLE) {
		device_trans_state(scd, SCD_DEVICE_T_STOP_CONTROL);
	}
}

/* A[iS] */
static struct scd_control *probe_control(struct scd_device_control *scd,
					 struct scd_intf *intf)
{
	int status;
	struct scd_control *control;

	control = intf->probe(&scd->sdev);

	if (control && intf->type == SCD_CONTROL_TYPE_SYSDEP) {
		status = device_trans_state(scd, SCD_DEVICE_T_PROBE);
		if (status < 0) {
			ERR("cannot probe sysdep control\n");
			status = intf->remove(control);
			control = NULL;
			if (status < 0) {
				/* scd_error is not use here. Because sysdep */
				ERR("sysdep remove error\n");
				/* is not working, so scd_info too */
			}
			BUG_ON(status < 0);
		}
	}

	return control;
}

/* A[iS] */
static int remove_control(struct scd_device_control *scd, struct scd_intf *intf)
{
	int status;

	if (intf->type == SCD_CONTROL_TYPE_SYSDEP) {
		/* sysdep unsetξ硢٤Ƥtimer򥭥󥻥뤹 */
		cancel_all_timer(scd);
		status = device_trans_state(scd, SCD_DEVICE_T_REMOVE);
		if (status < 0) {
			ERR("sysdep remove error\n");
			return status;
		}
	}

	return intf->remove(scd->control[intf->type]);
}

/* A[is] */
int usb_scd_device_set_intf(struct scd_intf *intf)
{
	struct scd_device_control *scd;
	struct scd_device_control *temp_scd;
	struct scd_control *control;

	if (!intf || intf->type >= NUM_CONTROL_TYPE) {
		return -EINVAL;
	}

	read_lock(&scd_lock);

	list_for_each_entry_safe(scd, temp_scd, &scd_anchor, scd_list) {
		/* sysdepscd_deviceʬΥʤΤǡplatform probeκݤ */
		/* ɬsysdepprobe롣 */
		/* info, gadget, hcdprobesysdep probe˸ƤФ */
		/* ΤHCLKON֤Խ */
		if (scd->control[intf->type] == NULL) {
			stop_scd(scd);
			control = probe_control(scd, intf);
			if (control == NULL) {
				ERR("probe error(%d,%d)\n", __LINE__);
				intfs[intf->type] = intf;
				restart_scd(scd);
				read_unlock(&scd_lock);
				usb_scd_device_unset_intf(intf);
				return -EPERM;
			}
			scd->control[intf->type] = control;
			restart_scd(scd);
		}
	}

	intfs[intf->type] = intf;

	read_unlock(&scd_lock);

	return 0;
}

EXPORT_SYMBOL(usb_scd_device_set_intf);

/* A[is] */
int usb_scd_device_unset_intf(struct scd_intf *intf)
{
	int status;
	struct scd_device_control *scd;
	struct scd_device_control *temp_scd;

	if (!intf || intf->type >= NUM_CONTROL_TYPE) {
		return -EINVAL;
	}

	if (intfs[intf->type] != intf) {
		return -EINVAL;
	}

	read_lock(&scd_lock);

	list_for_each_entry_safe(scd, temp_scd, &scd_anchor, scd_list) {
		if (scd->control[intf->type]) {
			stop_scd(scd);
			status = remove_control(scd, intf);
			if (status < 0) {
				read_unlock(&scd_lock);
				ERR("remove error(%d,%d)\n", __LINE__, status);
				return status;
			}
			scd->control[intf->type] = NULL;
			restart_scd(scd);
		}
	}

	intfs[intf->type] = NULL;

	read_unlock(&scd_lock);

	return 0;
}

EXPORT_SYMBOL(usb_scd_device_unset_intf);

/*
 *	usb_otg_control
 */

/*
 * ؿget_port_info
 * USBݡȾ롣
 */
/* A[?s] */
static int
get_port_info(struct usb_otg_control *otg_control,
	      struct usb_otg_control_port_info *info)
{
	info->nr = 1;		/* ݡȿ1 */
	info->current_port = 0;	/* ȥݡ0 */

	return 0;
}

/*
 * ؿselect_port
 * USBݡȤ򤹤롣
 */
/* A[?s] */
static int select_port(struct usb_otg_control *otg_control, u32 pn)
{
	if (pn != 0) {
		return -EIO;
	}

	return 0;
}

/*
 * ؿstart_control
 * Controller򳫻Ϥ롣
 */
/* A[?s] */
static int start_control(struct usb_otg_control *otg_control)
{
	int status;
	unsigned long flags;
	struct scd_device_control *scd;

	scd = otgc_to_scd(otg_control);

	spin_lock_irqsave(&scd->lock, flags);

	/* Хֽ - start controlΤΤ */
	scd->cid = -1;		/* unknown value */
	scd->vbus = USB_OTG_VBUS_STAT_UNKNOWN;

	spin_unlock_irqrestore(&scd->lock, flags);

	status = device_trans_state(scd, SCD_DEVICE_T_START_CONTROL);

	return status;
}

/*
 * ؿstop_control
 * Controllerߤ롣
 */
/* A[?s] */
static int stop_control(struct usb_otg_control *otg_control)
{
	int status;
	struct scd_device_control *scd;

	scd = otgc_to_scd(otg_control);

	status = device_trans_state(scd, SCD_DEVICE_T_STOP_CONTROL);

	return status;
}

/*
 * ؿstart_gadget
 * ControllerGADGET⡼ɤˤ롣
 */
/* A[?s] */
static int start_gadget(struct usb_otg_control *otg_control)
{
	int status;
	struct scd_device_control *scd;

	scd = otgc_to_scd(otg_control);

	status = device_trans_state(scd, SCD_DEVICE_T_START_GADGET);

	return status;
}

/*
 * ؿstop_gadget
 * ControllerIDLE⡼ɤˤ롣
 */
/* A[?s] */
static int stop_gadget(struct usb_otg_control *otg_control)
{
	int status;
	struct scd_device_control *scd;

	scd = otgc_to_scd(otg_control);

	status = device_trans_state(scd, SCD_DEVICE_T_STOP_GADGET);

	return status;
}

/*
 * ؿenable_rchost
 * ControllerRCHOST⡼ɤˤʤ뤳ȤĤ롣
 */
/* A[?s] */
static int enable_rchost(struct usb_otg_control *otg_control)
{
	return -EPROTONOSUPPORT; /* ̤ݡAPI */
}

/*
 * ؿdisable_rchost
 * ControllerRCHOST⡼ɤˤʤ뤳ȤԵĤȤ롣
 */
/* A[?s] */
static int disable_rchost(struct usb_otg_control *otg_control)
{
	return -EPROTONOSUPPORT; /* ̤ݡAPI */
}

/*
 * ؿstop_rchost
 * ControllerIDLE⡼ɤˤ롣
 */
/* A[?s] */
static int stop_rchost(struct usb_otg_control *otg_control)
{
	return -EPROTONOSUPPORT; /* ̤ݡAPI */
}

/*
 * ؿstart_host
 * ControllerHOST⡼ɤˤ롣
 */
/* A[?s] */
static int start_host(struct usb_otg_control *otg_control)
{
	int status;
	struct scd_device_control *scd;

	status = 0;

	scd = otgc_to_scd(otg_control);

	status = device_trans_state(scd, SCD_DEVICE_T_START_HOST);

	return status;
}

/*
 * ؿstop_host
 * ControllerIDLE⡼ɤˤ롣
 */
/* A[?s] */
static int stop_host(struct usb_otg_control *otg_control)
{
	int status;
	struct scd_device_control *scd;

	scd = otgc_to_scd(otg_control);

	status = device_trans_state(scd, SCD_DEVICE_T_STOP_HOST);

	return status;
}

/*
 * ؿstart_rcgadget
 * ControllerRCGADGET⡼ɤˤ롣
 */
/* A[?s] */
static int
start_rcgadget(struct usb_otg_control *otg_control, unsigned int port)
{
	return -EPROTONOSUPPORT; /* ̤ݡAPI */
}

/*
 * ؿstop_rcgadget
 * ControllerIDLE⡼ɤˤ롣
 */
/* A[?s] */
static int stop_rcgadget(struct usb_otg_control *otg_control, unsigned int port)
{
	return -EPROTONOSUPPORT; /* ̤ݡAPI */
}

/*
 * ؿget_mode
 * OTG⡼ɤ롣
 */
/* A[?s] */
static int get_mode(struct usb_otg_control *otg_control)
{
	int mode;
	struct scd_device_control *scd;

	scd = otgc_to_scd(otg_control);

	/* @probed - level 1 : probed_init, stop, idle, b_periph, a_host */
	switch (scd_device_get_state(scd, 1)) {
	case SCD_DEVICE_S_STOP:
		mode = USB_OTG_CONTROL_STOP;
		break;
	case SCD_DEVICE_S_IDLE:
		mode = USB_OTG_CONTROL_IDLE;
		break;
	case SCD_DEVICE_S_B_PERIPH:
		mode = USB_OTG_CONTROL_GADGET;
		break;
	case SCD_DEVICE_S_A_HOST:
		mode = USB_OTG_CONTROL_HOST;
		break;
	default:
		/* error */
		mode = 0;
		break;
	}

	return mode;
}

/*
 * ؿrequest_session
 * ControllerSRPѥ륷󥰤׵᤹롣
 */
/* A[?s] */
static int request_session(struct usb_otg_control *otg_control)
{
	return -EPROTONOSUPPORT; /* ̤ݡAPI */
}

/*
 * ؿset_speed
 * ̿ԡɤꤹ롣
 */
/* A[?s] */
static int set_speed(struct usb_otg_control *otg_control, unsigned int speed)
{
	struct scd_device_control *scd;

	if (speed != USB_OTG_HS_ENABLE && speed != USB_OTG_HS_DISABLE) {
		return -EINVAL;
	}

	scd = otgc_to_scd(otg_control);

	/*
	  attributes
	 */
	if (speed == USB_OTG_HS_ENABLE) {
		set_bit(SCD_DEVICE_ATTR_HS_ENABLE, &scd->sdev.attributes);
	}
	else {
		clear_bit(SCD_DEVICE_ATTR_HS_ENABLE, &scd->sdev.attributes);
	}

	return 0;
}

/*
 * ؿget_speed
 * ̿ԡɤ롣
 */
/* A[?s] */
static enum usb_otg_control_speed get_speed(struct usb_otg_control *otg_control)
{
	enum usb_otg_control_speed speed;
	u8 num;
	u32 hprt;
	struct scd_device_control *scd;

	scd = otgc_to_scd(otg_control);

	/* Υԡɤ - PHYκ祹ԡɤǤ̵ */
	hprt = SCD__READ(&scd->sdev, HPRT);
	/* speed: USB_OTG_SPEED_HS/FS/LS */
	num = (hprt >> SCD_S_HPRT_PRTSPD) & SCD_M_HPRT_PRTSPD;

	if (num == SCD_PORT_HS) {
		speed = USB_SPEED_HIGH;
	}
	else if (num == SCD_PORT_FS) {
		speed = USB_SPEED_FULL;
	}
	else if (num == SCD_PORT_LS) {
		speed = USB_SPEED_LOW;
	}
	else {
		speed = USB_SPEED_UNKNOWN;
	}
	
	return speed;
}

/*
 * ؿset_test_mode
 * ƥȥ⡼ɤꤹ롣
 */
/* A[?s] */
static int
set_test_mode(struct usb_otg_control *otg_control,
	      enum usb_otg_control_test_mode test_mode)
{
	int retval;

	/* ǿOTG APIǤϥƥȥ⡼ɤ򥵥ݡȤʤ */
	retval = -EPROTONOSUPPORT; /* ̤ݡAPI */

	return retval;
}

/*
 * ؿget_test_mode
 * ƥȥ⡼ɤ롣
 */
/* A[?s] */
static enum usb_otg_control_test_mode
get_test_mode(struct usb_otg_control *otg_control)
{
	enum usb_otg_control_test_mode test_mode;

	/* ǿOTG APIǤϥƥȥ⡼ɤ򥵥ݡȤʤ */
	test_mode = USB_OTG_TEST_MODE_NORMAL; /* ̤ݡAPI */

	return test_mode;
}

int ioctl(struct usb_otg_control *otg_control, unsigned code, unsigned long data)
{
	struct usb_otg_line_state *lstat;
	struct scd_device_control *scd;
	u32 hprt;

	if (!data) {
		return -EINVAL;
	}

	switch (code) {
	case USB_IOCTL_GET_LINE_STATE:
		scd = otgc_to_scd(otg_control);
		hprt = SCD__READ(&scd->sdev, HPRT);

		lstat = (struct usb_otg_line_state *)data;
		lstat->dp = (hprt & SCD_B_HPRT_PRTLNSTS_DP
			     ? USB_OTG_LINE_STATE_HIGH : USB_OTG_LINE_STATE_LOW);
		lstat->dm = (hprt & SCD_B_HPRT_PRTLNSTS_DM
			     ? USB_OTG_LINE_STATE_HIGH : USB_OTG_LINE_STATE_LOW);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static struct usb_otg_control_ops otg_control_ops = {
	.get_port_info = get_port_info,
	.select_port = select_port,
	.start_control = start_control,
	.stop_control = stop_control,
	.start_gadget = start_gadget,
	.stop_gadget = stop_gadget,
	.enable_rchost = enable_rchost,
	.disable_rchost = disable_rchost,
	.stop_rchost = stop_rchost,
	.start_host = start_host,
	.stop_host = stop_host,
	.start_rcgadget = start_rcgadget,
	.stop_rcgadget = stop_rcgadget,
	.get_mode = get_mode,
	.request_session = request_session,
	.set_speed = set_speed,
	.get_speed = get_speed,
	.set_test_mode = set_test_mode,
	.get_test_mode = get_test_mode,
	.ioctl = ioctl,
};

/* scd_listκǽscd */
static struct scd_device_control *get_first_scd(void)
{
	struct scd_device_control *scd;

	if (list_empty(&scd_anchor)) {
		scd = NULL;
		goto get_first_scd_end;
	}

	scd = container_of(scd_anchor.next,
			   struct scd_device_control, scd_list);

get_first_scd_end:
	return scd;
}

/*
 * ؿusb_otg_control_register_core
 * OTG Core򥳥ȥϿ롣
 */
/* A[is] */
int usb_otg_control_register_core(struct usb_otg_core *otg_core)
{
	int status;
	struct scd_device_control *scd;

	/*
	  ˸OTG APIǤϰĤOTG(Link IP)Ǥʤ
	 */

	read_lock(&scd_lock);

	scd = get_first_scd();
	if (scd == NULL) {
		status = -ENOENT;
		goto usb_otg_control_register_end;
	}

	scd->otgc.otg_core = otg_core;
	status = usb_otg_core_bind(otg_core, &scd->otgc);
	if (status < 0) {
		scd->otgc.otg_core = NULL;
	}

usb_otg_control_register_end:
	read_unlock(&scd_lock);

	return status;

}

EXPORT_SYMBOL(usb_otg_control_register_core);

/*
 * ؿusb_otg_control_unregister_core
 * OTG Core򥳥ȥϿ롣
 */
/* A[is] */
int usb_otg_control_unregister_core(struct usb_otg_core *otg_core)
{
	int status;
	struct scd_device_control *scd;

	/*
	  ˸OTG APIǤϰĤOTG(Link IP)Ǥʤ
	 */
	read_lock(&scd_lock);

	scd = get_first_scd();
	if (scd == NULL) {
		status = -ENOENT;
		goto usb_otg_control_unregister_end;
	}

	/* ֤STOP֤ض */
	force_stopping_scd(scd);
	/* scd¾Τ */
	stop_scd(scd);
	spin_lock(&scd->lock);
	stop_regs(&scd->sdev);
	spin_unlock(&scd->lock);

	if (scd->otgc.otg_core != otg_core) {
		status = -EINVAL;
		goto usb_otg_control_unregister_end;
	}
	scd->otgc.otg_core = NULL;
	status = usb_otg_core_unbind(otg_core, &scd->otgc);

usb_otg_control_unregister_end:
	read_unlock(&scd_lock);

	return status;
}

EXPORT_SYMBOL(usb_otg_control_unregister_core);

/*
 *	platform_driver
 */

static int remove_intfs(struct scd_device_control *scd)
{
	s8 i;
	int status;

	for (i = 0; i < NUM_CONTROL_TYPE; i++) {
		DBG(3, "remove control intfs[%d]=0x%x\n", i, (u32)intfs[i]);
		if (intfs[i]) {
			status = remove_control(scd, intfs[i]);
			if (status < 0) {
				/* controlremove顼⥨顼̵뤷 */
				/* ʹߤcontrolremove */
				ERR("control remove failed (%d,%d)\n", i, status);
			}
			scd->control[intfs[i]->type] = NULL;
		}
	}

	return status;
}

/* A[is] */
static int probe(struct platform_device *pdev)
{
	s8 i;
	int status;
	unsigned long flags;
	void* regs;
	int irqn;
	unsigned int start;
	unsigned int len;
	struct resource *rsc;
	struct scd_device_control *scd;
	struct scd_control *control;

	status = 0;

	/* ֹ */
	irqn = platform_get_irq(pdev, 0);
	if (irqn < 0) {
		return irqn;
	}

	/* pdevIO꥽ - IORESOURCE_IOԲĤ */
	rsc = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (rsc == NULL) {
		ERR("get resource failed\n");
		status = -ENXIO;
		goto probe_err0;
	}
	start = rsc->start;
	len = rsc->end - rsc->start;

	regs = ioremap_nocache(start, len);
	if (regs == NULL) {
		ERR("ioremap failed\n");
		status = -ENOMEM;
		goto probe_err0;
	}

	scd = kzalloc(sizeof(*scd), GFP_KERNEL);
	if (!scd) {
		ERR("scd alloc failed\n");
		status = -ENOMEM;
		goto probe_err1;
	}

	scd->trigger_wq = create_singlethread_workqueue(dev_name(&pdev->dev));
	if (!scd->trigger_wq) {
		DBG(0, "scd->trigger_wq create failed\n");
		status = -ENOMEM;
		goto probe_err2;
	}

	/* ǥХ֥Ƚ */
	scd->irqn = irqn;
	scd->state = SCD_DEVICE_S_POWERED_ON;
	scd->cid = -1;		/* CIDͤ-1 */
	scd->vbus = USB_OTG_VBUS_STAT_UNKNOWN;
	INIT_LIST_HEAD(&scd->timer_anchor);
	INIT_LIST_HEAD(&scd->event_anchor);
	spin_lock_init(&scd->lock);
	mutex_init(&scd->state_mutex);
	tasklet_init(&scd->event_tasklet, signal_event, (unsigned long)scd);

	/*  󥿡ե scd_devicedev */
	scd->sdev.dev = &pdev->dev;
	scd->sdev.regs = regs;
	scd->sdev.ops = &device_ops;
	set_bit(SCD_DEVICE_ATTR_HS_ENABLE, &scd->sdev.attributes);
	spin_lock_init(&scd->sdev.lock);
	scd->otgc.ops = &otg_control_ops;
	scd->otgc.name = dev_name(&pdev->dev);

	INIT_LIST_HEAD(&scd->scd_list);
	/* list¾ */
	write_lock_irqsave(&scd_lock, flags);
	list_add(&scd->scd_list, &scd_anchor);
	write_unlock_irqrestore(&scd_lock, flags);

	/* controller probe */
	for (i = NUM_CONTROL_TYPE - 1; i >= 0; i--) {
		if (intfs[i]) {
			DBG(2, "probe control intfs[%d]=0x%x\n",
			    i, (u32)intfs[i]);
			control = probe_control(scd, intfs[i]);
			scd->control[intfs[i]->type] = control;
			if (scd->control[intfs[i]->type] == NULL) {
				ERR("control probe failed(%d,%d)\n", i, status);
				status = -EPERM;
				goto probe_err3;
			}
		}
	}

	platform_set_drvdata(pdev, scd);

	goto probe_end;

	/* 顼 */
probe_err3:
	remove_intfs(scd);
	destroy_workqueue(scd->trigger_wq);
probe_err2:
	kfree(scd);
probe_err1:
	iounmap(regs);
probe_err0:
probe_end:
	return status;
}

/* A[is] */
static int remove(struct platform_device *pdev)
{
	int status;
	unsigned long flags;
	struct scd_device_control *scd;

	status = 0;

	scd = platform_get_drvdata(pdev);
	if (scd == NULL) {
		/* removeƤ */
		WARN(0, "usb_scd already removed\n");
		goto remove_end;
	}
	platform_set_drvdata(pdev, NULL);

	/* ¾Τscdưߤ */
	stop_scd(scd);
	stop_regs(&scd->sdev);

	write_lock_irqsave(&scd_lock, flags);
	list_del(&scd->scd_list);
	write_unlock_irqrestore(&scd_lock, flags);

	/*
	  controllers and interfaces remove
	 */
	remove_intfs(scd);
	iounmap(scd->sdev.regs);
	destroy_workqueue(scd->trigger_wq);
	kfree(scd);

remove_end:
	return status;
}

/* controllers and interfaces resume (sysdepʳresume) */
static void resume_intfs(struct scd_device_control *scd)
{
	s8 i;

	for (i = NUM_CONTROL_TYPE - 2; i >= 0; i--) {
		if (intfs[i] && intfs[i]->resume) {
			DBG(2, "suspend control intfs[%d]=0x%x\n",
			    i, (u32)intfs[i]);
			intfs[i]->resume(scd->control[intfs[i]->type]);
		}
	}
}

/* A[is] */
static int suspend(struct platform_device *pdev, pm_message_t message)
{
	s8 i;
	int status;
	unsigned long flags;
	struct scd_device_control *scd;
	struct scd_sysdep_intf *sysdep_intf;

	status = 0;

	scd = platform_get_drvdata(pdev);
	if (scd == NULL) {
		/* removeƤ */
		WARN(0, "usb_scd already removed\n");
		status = -ENODEV;
		goto suspend_err0;
	}

	/* scd */
	stop_scd(scd);

	sysdep_intf = get_sysdep_intf();
	BUG_ON(sysdep_intf == NULL); /* STOP֤sysdep probeѤ */

	/* PRETHAWsysdep suspend */
	if (message.event == PM_EVENT_PRETHAW) {
		status = sysdep_intf->base_intf.suspend(sysdep_control(scd),
							message);
		if (status < 0) {
			err("usb_scd fail to suspend sysdep\n");
			goto suspend_err1;
		}
	}

	/* timerϤΤޤݻ */

	spin_lock_irqsave(&scd->lock, flags);

	/* ƥȥsuspendLink쥸ǽȤ */
	resume_regs(&scd->sdev);

	if (message.event == PM_EVENT_PRETHAW) {
		/* Link쥸 */
		init_regs(&scd->sdev, false);
	}

	spin_unlock_irqrestore(&scd->lock, flags);

	/*
	  controller suspend (sysdepʳ)
	 */
	for (i = 0; i < NUM_CONTROL_TYPE - 1; i++) {
		if (intfs[i] && intfs[i]->suspend) {
			DBG(2, "suspend control intfs[%d]=0x%x\n",
			    i, (u32)intfs[i]);
			status = intfs[i]->
				suspend(scd->control[intfs[i]->type], message);
			if (status < 0) {
				ERR("control probe failed(%d,%d)\n", i, status);
				goto suspend_err2;
			}
		}
	}

	/* PRETHAWʳUSB HWť⡼ɤ졢Ǹsysdep suspend */
	if (message.event != PM_EVENT_PRETHAW) {
		spin_lock_irqsave(&scd->lock, flags);

		suspend_regs(&scd->sdev);

		spin_unlock_irqrestore(&scd->lock, flags);

		status = sysdep_intf->base_intf.suspend(sysdep_control(scd),
							message);
		if (status < 0) {
			err("usb_scd fail to suspend sysdep\n");
			goto suspend_end;
		}
	}

suspend_err2:
	resume_intfs(scd);
suspend_err1:
	restart_scd(scd);
suspend_err0:
suspend_end:
	return status;
}

/* A[is] */
static int resume(struct platform_device *pdev)
{
	int status;
	unsigned long flags;
	struct scd_device_control *scd;
	struct scd_sysdep_intf *sysdep_intf;

	status = 0;

	scd = platform_get_drvdata(pdev);
	if (scd == NULL) {
		/* removeƤ */
		WARN(0, "usb_scd already removed\n");
		status = -ENODEV;
		goto resume_end;
	}

	sysdep_intf = get_sysdep_intf();
	BUG_ON(sysdep_intf == NULL); /* STOP֤sysdep probeѤ */

	status = sysdep_intf->base_intf.resume(sysdep_control(scd));
	if (status < 0) {
		err("usb_scd fail to resume sysdep\n");
		goto resume_err0;
	}

	/* timerϤΤޤݻ */

	spin_lock_irqsave(&scd->lock, flags);

	/* ƥȥsuspendLink쥸ǽȤ */
	resume_regs(&scd->sdev);

#if defined(SCD_POWER_OFF_SUSPEND)
	/* PM_EVENT_SUSPENDŸOFF륷ƥǤresume˥쥸 */
	init_regs(&scd->sdev, false);
#endif	/* defined(SCD_POWER_OFF_SUSPEND) */

	/* ư⡼ɻunmask */
	/*   unmaskꤷƤ¾controlORǽȤ */
	switch (scd_device_get_state(scd, 1)) {
	case SCD_DEVICE_S_IDLE:
	case SCD_DEVICE_S_B_PERIPH:
	case SCD_DEVICE_S_A_HOST:
		start_regs(&scd->sdev);
		break;
	default:
		break;
	}

	spin_unlock_irqrestore(&scd->lock, flags);

	/* controllers and interfaces resume (sysdepʳresume) */
	resume_intfs(scd);

	spin_lock_irqsave(&scd->lock, flags);

	/* 3ơȰʳSuspendUSB HWť⡼ɤ */
	switch (scd->state) {
	case SCD_DEVICE_S_AB_VBUSED:
	case SCD_DEVICE_S_B_PERIPH:
	case SCD_DEVICE_S_A_HOST:
		break;
	default:
		suspend_regs(&scd->sdev);
		break;
	}

	spin_unlock_irqrestore(&scd->lock, flags);

	/* ư⡼ɻ߳ */
	switch (scd_device_get_state(scd, 1)) {
	case SCD_DEVICE_S_IDLE:
	case SCD_DEVICE_S_B_PERIPH:
	case SCD_DEVICE_S_A_HOST:
		restart_scd(scd);
		break;
	default:
		break;
	}

resume_err0:
resume_end:
	return status;
}

/* A[is] */
static void shutdown(struct platform_device *pdev)
{
	s8 i;
	unsigned long flags;
	struct scd_device_control *scd;

	scd = platform_get_drvdata(pdev);
	if (scd == NULL) {
		/* removeƤ */
		WARN(0, "usb_scd already removed\n");
		return;
	}
	platform_set_drvdata(pdev, NULL);

	spin_lock_irqsave(&scd->lock, flags);

	/* Link - free_irqϤƤ뤬ǰΤ */
	stop_scd(scd);
	stop_regs(&scd->sdev);

	spin_unlock_irqrestore(&scd->lock, flags);

	/* ٤Ƥtimer򥭥󥻥뤹 */
	cancel_all_timer(scd);

	/*
	  controller shutdown
	 */
	for (i = 0; i < NUM_CONTROL_TYPE; i++) {
		if (intfs[i] && intfs[i]->shutdown) {
			DBG(2, "shutdown control intfs[%d]=0x%x\n",
			    i, (u32)intfs[i]);
			intfs[i]->shutdown(scd->control[intfs[i]->type]);
		}
	}
}

static struct platform_driver scd_platform_driver = {
	.probe = probe,
	.remove = remove,
	.suspend = suspend,
	.resume = resume,
	.shutdown = shutdown,
	.driver = {
		.name = "usb_scd",
		.owner = THIS_MODULE,
	},
};

/*
 *	OTG API Hook
 */
/* ǥХ̾OTGȥ */
struct usb_otg_control *scd_otg_lookup_control_by_name(const char* name)
{
	struct scd_device_control *scd;
	struct usb_otg_control *otgc;

#if defined(BUS_ID_SIZE)
#define MAX_DEV_NAME_LEN BUS_ID_SIZE
#elif defined(KOBJ_NAME_LEN)
#define MAX_DEV_NAME_LEN KOBJ_NAME_LEN
#else
/* ȼкȤ̾׽Ϳ٤ǥХ̾Ĺʤ */
#error "usb_scd : not defined maximum device name length"
#endif
	otgc = NULL;
	read_lock(&scd_lock);
	list_for_each_entry(scd, &scd_anchor, scd_list) {
		if (strncmp(name, dev_name(scd->sdev.dev),
			    MAX_DEV_NAME_LEN) == 0) {
			otgc = &scd->otgc;
			break;
		}
	}
	read_unlock(&scd_lock);

	return otgc;
}
EXPORT_SYMBOL(scd_otg_lookup_control_by_name);

/* OTG API NotifyؿHook */
int scd_otg_hook_notify(struct usb_otg_control *control, scd_otg_notify new)
{
	struct scd_device_control *scd;

	scd = otgc_to_scd(control);
	scd->notify_hook = new;

	return 0;
}
EXPORT_SYMBOL(scd_otg_hook_notify);

/* OTG API ؿHook */
struct usb_otg_control_ops *scd_otg_hook_ops(struct usb_otg_control *control,
					     struct usb_otg_control_ops *new)
{
	control->ops = new ? new : &otg_control_ops;

	return &otg_control_ops;
}
EXPORT_SYMBOL(scd_otg_hook_ops);


/*
 *	init/exit
 */

int __init usb_scd_module_init(void)
{
	memset(intfs, 0, sizeof(*intfs) * NUM_CONTROL_TYPE);

	usb_scd_sysdep_init();
	usb_scd_info_init();
#if defined(CONFIG_USB_SCD_GADGET)
	usb_scd_gadget_init();
#endif
#if defined(CONFIG_USB_SCD_HOST)
	usb_scd_hcd_init();
#endif

	return platform_driver_register(&scd_platform_driver);
}
/*
  StaticLinkmodule_init̵ˤ
 */
#if !defined(SCD_STATIC_LINK)
module_init(usb_scd_module_init);
#endif

void __exit usb_scd_module_exit(void)
{
	platform_driver_unregister(&scd_platform_driver);

#if defined(CONFIG_USB_SCD_HOST)
	usb_scd_hcd_exit();
#endif
#if defined(CONFIG_USB_SCD_GADGET)
	usb_scd_gadget_exit();
#endif
	usb_scd_info_exit();
	usb_scd_sysdep_exit();
}
module_exit(usb_scd_module_exit);

MODULE_DESCRIPTION(DRIVER_GADGET_INFO);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_LICENSE("GPL");
