/*-----------------------------------------------------------------------------
 * drivers/usb/scd/scd_hep.c
 * 
 * Host endpoint(channel) operations
 * 
 * Copyright 2011, 2012 Sony Corporation
 *
 * This file is part of the HS-OTG Controller Driver SCD.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/module.h>
#include <linux/device.h>
#include <linux/memory.h>
#include <linux/interrupt.h>
#include <linux/dma-mapping.h>
#include <asm/io.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb/gadget.h>
#include <linux/usb/scd/usb_otg_control.h>
#include <linux/usb/scd/scd_hub.h>
#include "../core/hub.h"
#include "../core/hcd.h"

#include "scd.h"
#include "scd_debug.h"
#include "scd_host.h"
#include "scd_register.h"
#include "scd_hfifo.h"

/* hep states and triggers */
enum {
	SCD_HEP_S_CNTL_IDLE = 0,
	SCD_HEP_S_CNTL_WAIT_TX_SETUP_CPU,
	SCD_HEP_S_CNTL_WAIT_TX_DATA_CPU,
	SCD_HEP_S_CNTL_WAIT_RX_DATA_CPU,
	SCD_HEP_S_CNTL_WAIT_RX_DATA_CPU_HALT,
	SCD_HEP_S_CNTL_WAIT_RX_STATUS_CPU,
	SCD_HEP_S_CNTL_WAIT_RX_STATUS_CPU_HALT,
	SCD_HEP_S_CNTL_WAIT_TX_STATUS_CPU,

	SCD_HEP_S_CNTL_WAIT_TX_SETUP_DMA,
	SCD_HEP_S_CNTL_WAIT_TX_DATA_DMA_MULTI,
	SCD_HEP_S_CNTL_WAIT_TX_DATA_DMA_ZEROP,
	SCD_HEP_S_CNTL_WAIT_RX_DATA_DMA_MULTI,
	SCD_HEP_S_CNTL_WAIT_RX_DATA_DMA_SHORTP,
	SCD_HEP_S_CNTL_WAIT_RX_STATUS_DMA,
	SCD_HEP_S_CNTL_WAIT_TX_STATUS_DMA,
};

enum {
	SCD_HEP_S_TX_IDLE = 0,
	SCD_HEP_S_TX_WAIT_DMA_MULTI,
	SCD_HEP_S_TX_WAIT_DMA_ZEROP,
	SCD_HEP_S_TX_WAIT_DMA,
	SCD_HEP_S_TX_WAIT_CPU,
};

enum {
	SCD_HEP_S_RX_IDLE = 0,
	SCD_HEP_S_RX_WAIT_DMA_MULTI,
	SCD_HEP_S_RX_WAIT_DMA_SHORTP,
	SCD_HEP_S_RX_WAIT_DMA,
	SCD_HEP_S_RX_WAIT_CPU,
	SCD_HEP_S_RX_WAIT_CPU_HALT,
};

enum {
	SCD_HEP_T_START,
	SCD_HEP_T_RESET,
	SCD_HEP_T_LINK_XFERCOMPL_INTR,
	SCD_HEP_T_LINK_RXFLVL_INTR,
	SCD_HEP_T_LINK_CHHLT_INTR,
	SCD_HEP_T_LINK_ERROR_INTR,
};

enum {
	SCD_HEP_CNTL_IN,
	SCD_HEP_CNTL_OUT,
	SCD_HEP_CNTL_NODATA,
};

/* PID */
#define PID_DATA0 SCD_B_HCTSIZN_PID_DATA0
#define PID_DATA2 SCD_B_HCTSIZN_PID_DATA2
#define PID_DATA1 SCD_B_HCTSIZN_PID_DATA1
#define PID_MDATA SCD_B_HCTSIZN_PID_MDATA
#define PID_SETUP SCD_B_HCTSIZN_PID_SETUP
#define PID_AUTO 0xf0

/* error */
#define SCD_HEP_E_NOBUFFER	(1 << 0)
#define SCD_HEP_E_MAXPOVER	(1 << 1)
#define SCD_HEP_E_NORESP	(1 << 2)
#define SCD_HEP_E_RXSTALL	(1 << 3)
#define SCD_HEP_E_OVERFLOW	(1 << 4)
#define SCD_HEP_E_SHORT 	(1 << 5)
#define SCD_HEP_E_XACTERR 	(1 << 6)
#define SCD_HEP_E_AHBERR 	(1 << 7)

/* FIFO flush timeout */
#define SCD_HS_FLUSH_TIMEOUT 	25
#define SCD_FS_FLUSH_TIMEOUT 	800

/* Transaction Error retry count */
#define SCD_RETRY_COUNT		3

/* transfer_buffer_length divide actual maximum length */
#define SCD_MAX_PKTCNT		SCD_M_HCTSIZN_PKTCNT
#define SCD_MAX_XFERSIZE	SCD_M_HCTSIZN_XFERSIZE

/* Proto define of internal functions */

static int hep_cntl_trans_state(struct usb_hcd *hcd, struct scd_hep *ep,
				u32 trigger);
static int hep_tx_trans_state(struct usb_hcd *hcd, struct scd_hep *ep,
			      u32 trigger);
static int hep_rx_trans_state(struct usb_hcd *hcd, struct scd_hep *ep,
			      u32 trigger);

/* internal functions */

/* common internal functions  */
inline static u8 hep_hash(struct usb_host_endpoint *hep, int devnum, u8 epnum)
{
	return epnum - 1;
}

inline static int hep_hep2epnum(struct usb_host_endpoint *hep)
{
	return hep->desc.bEndpointAddress & USB_ENDPOINT_NUMBER_MASK;
}

inline int hep_check_ep_type(struct usb_endpoint_descriptor *epd)
{
	return usb_endpoint_type(epd);
}

inline static u8 hep_get_tx_index(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_host *shost = hcd_to_shost(hcd);

	return ep - &shost->ep_out[0] + 1;
}

inline static u8 hep_get_rx_index(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_host *shost = hcd_to_shost(hcd);

	return ep - &shost->ep_in[0] + 1;
}

static void hep_unmask_tx_intr(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 ep_type;
	u8 epnum;
	u32 gintmsk, haintmsk, hcintmsk;

	DBG(2, "<==\n");

	ep_type = hep_check_ep_type(&ep->hep->desc);
	if(ep_type == USB_ENDPOINT_XFER_CONTROL){
		epnum = 0;
	}
	else {
		epnum = hep_hep2epnum(ep->hep);
	}

	/* Unmask interrupt of Non-Periodec TxFIFO or Periodec TxFIFO */
	gintmsk = SCD__READ(sdev, GINTMSK);
#ifdef SLAVE_MODE_DEBUG
	if (!sdev->dmac) {
		switch(ep_type){
		case USB_ENDPOINT_XFER_CONTROL:
		case USB_ENDPOINT_XFER_BULK:
		    gintmsk |= SCD_B_GINTMSK_NPTXFEMPMSK;
		    break;
		case USB_ENDPOINT_XFER_INT:
		case USB_ENDPOINT_XFER_ISOC:
		    gintmsk |= SCD_B_GINTMSK_PTXFEMPMSK;
		    break;
		}
		SCD__WRITE(sdev, GINTMSK, gintmsk);
	}
#endif /* SLAVE_MODE_DEBUG */

	/* Unmask interrupt of channel */
	gintmsk |= SCD_B_GINTMSK_HCHINTMSK |
		SCD_B_GINTMSK_PRTINTMSK |
		SCD_B_GINTMSK_INCOMPLPMSK;
#ifdef SLAVE_MODE_DEBUG
	if (!sdev->dmac) {
		gintmsk |= SCD_B_GINTMSK_RXFLVLMSK;
	}
#endif /* SLAVE_MODE_DEBUG */
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	/* Unmask HAINTMSK for channel interrupt */
	haintmsk = SCD__READ(sdev, HAINTMSK);
	haintmsk |= SCD_F_HAINTMSK_CH(epnum);
	SCD__WRITE(sdev, HAINTMSK, haintmsk);

	/* Unmask HCINTMSK for channel interrupt */
	hcintmsk = SCD__READ(sdev, HCINTMSK(epnum));
	hcintmsk |= SCD_B_HCINTMSKN_DATATGLERR |
		SCD_B_HCINTMSKN_FRMOVRUN |
		SCD_B_HCINTMSKN_BBLERR |
		SCD_B_HCINTMSKN_XACTERR |
		SCD_B_HCINTMSKN_STALL |
		SCD_B_HCINTMSKN_AHBERR |
		SCD_B_HCINTMSKN_CHHLTD;
#ifdef SLAVE_MODE_DEBUG
	if (!sdev->dmac) {
		hcintmsk |= SCD_B_HCINTMSKN_NYET |
			SCD_B_HCINTN_ACK |
			SCD_B_HCINTMSKN_NAK |
			SCD_B_HCINTMSKN_XFERCOMPL;
	}
#endif /* SLAVE_MODE_DEBUG */
	SCD__WRITE(sdev, HCINTMSK(epnum), hcintmsk);

	DBG(2,"==>\n");
}

static void hep_mask_tx_intr(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 ep_type;
	u8 epnum;
	u32 haintmsk, gintmsk, hcint = 0x0;

	DBG(2, "<==\n");

	ep_type = hep_check_ep_type(&ep->hep->desc);
	if(ep_type == USB_ENDPOINT_XFER_CONTROL){
		epnum = 0;
	}
	else {
		epnum = hep_hep2epnum(ep->hep);
	}

	/* Mask HCINTMSK about host transfer */
	hcint &= 0x00000000;
	SCD__WRITE(sdev, HCINT(epnum), hcint);

	/* Mask HAINTMSK for channel interrupt */
	haintmsk = SCD__READ(sdev, HAINTMSK);
	haintmsk &= ~(SCD_F_HAINTMSK_CH(epnum));
	SCD__WRITE(sdev, HAINTMSK, haintmsk);

	/* Mask except GINTMSK.PRTINTMSK */
	gintmsk = SCD__READ(sdev, GINTMSK);
	gintmsk &= ~(SCD_B_GINTMSK_HCHINTMSK |
		 SCD_B_GINTMSK_INCOMPLPMSK);
#ifdef SLAVE_MODE_DEBUG
	if (!sdev->dmac) {
		gintmsk &= ~SCD_B_GINTMSK_RXFLVLMSK;
	}
#endif /* SLAVE_MODE_DEBUG */
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	/* Mask GINTMSK.NPTxFEmpMsk */
#ifdef SLAVE_MODE_DEBUG
	if (!sdev->dmac) {
		switch(ep_type){
		case USB_ENDPOINT_XFER_CONTROL:
		case USB_ENDPOINT_XFER_BULK:
		    gintmsk &= ~SCD_B_GINTMSK_NPTXFEMPMSK;
		    break;
		case USB_ENDPOINT_XFER_INT:
		case USB_ENDPOINT_XFER_ISOC:
		    gintmsk &= ~SCD_B_GINTMSK_PTXFEMPMSK;
		    break;
		}
		SCD__WRITE(sdev, GINTMSK, gintmsk);
	}
#endif /* SLAVE_MODE_DEBUG */


	DBG(2,"==>\n");
}

static void hep_unmask_rx_intr(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 ep_type;
	u8 epnum;
	u32 gintmsk, haintmsk, hcintmsk;

	DBG(2, "<==\n");

	ep_type = hep_check_ep_type(&ep->hep->desc);
	if(ep_type == USB_ENDPOINT_XFER_CONTROL){
		epnum = 0;
	}
	else {
		epnum = hep_hep2epnum(ep->hep);
	}

	/* Unmask interrupt of Non-Periodec TxFIFO or Periodec TxFIFO */
	gintmsk = SCD__READ(sdev, GINTMSK);
#ifdef SLAVE_MODE_DEBUG
	if (!sdev->dmac) {
		switch(ep_type){
		case USB_ENDPOINT_XFER_CONTROL:
		case USB_ENDPOINT_XFER_BULK:
		    gintmsk |= SCD_B_GINTMSK_NPTXFEMPMSK;
		    break;
		case USB_ENDPOINT_XFER_INT:
		case USB_ENDPOINT_XFER_ISOC:
		    gintmsk |= SCD_B_GINTMSK_PTXFEMPMSK;
		    break;
		}
		SCD__WRITE(sdev, GINTMSK, gintmsk);
	}
#endif /* SLAVE_MODE_DEBUG */

	/* Unmask interrupt of channel */
	gintmsk |= SCD_B_GINTMSK_HCHINTMSK |
		SCD_B_GINTMSK_PRTINTMSK |
		SCD_B_GINTMSK_INCOMPLPMSK;
#ifdef SLAVE_MODE_DEBUG
	if (!sdev->dmac) {
		gintmsk |= SCD_B_GINTMSK_RXFLVLMSK;
	}
#endif /* SLAVE_MODE_DEBUG */
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	/* Unmask HAINTMSK for channel interrupt */
	haintmsk = SCD__READ(sdev, HAINTMSK);
	haintmsk |= SCD_F_HAINTMSK_CH(epnum);
	SCD__WRITE(sdev, HAINTMSK, haintmsk);

	/* Unmask HCINTMSK for channel interrupt */
	hcintmsk = SCD__READ(sdev, HCINTMSK(epnum));
	hcintmsk |= SCD_B_HCINTMSKN_DATATGLERR |
		SCD_B_HCINTMSKN_FRMOVRUN |
		SCD_B_HCINTMSKN_BBLERR |
		SCD_B_HCINTMSKN_XACTERR |
		SCD_B_HCINTMSKN_STALL |
		SCD_B_HCINTMSKN_AHBERR |
		SCD_B_HCINTMSKN_CHHLTD;
#ifdef SLAVE_MODE_DEBUG
	if (!sdev->dmac) {
		hcintmsk |= SCD_B_HCINTMSKN_NYET |
			SCD_B_HCINTN_ACK |
			SCD_B_HCINTMSKN_NAK |
			SCD_B_HCINTMSKN_XFERCOMPL;
	}
#endif /* SLAVE_MODE_DEBUG */
	SCD__WRITE(sdev, HCINTMSK(epnum), hcintmsk);

	DBG(2,"==>\n");
}

static void hep_mask_rx_intr(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 ep_type;
	u8 epnum;
	u32 haintmsk, gintmsk, hcint = 0x0;

	DBG(2, "<==\n");

	ep_type = hep_check_ep_type(&ep->hep->desc);
	if(ep_type == USB_ENDPOINT_XFER_CONTROL){
		epnum = 0;
	}
	else {
		epnum = hep_hep2epnum(ep->hep);
	}

	/* Mask HCINTMSK about host transfer */
	hcint &= 0x00000000;
	SCD__WRITE(sdev, HCINT(epnum), hcint);

	/* Mask HAINTMSK for channel interrupt */
	haintmsk = SCD__READ(sdev, HAINTMSK);
	haintmsk &= ~(SCD_F_HAINTMSK_CH(epnum));
	SCD__WRITE(sdev, HAINTMSK, haintmsk);

	/* Mask except GINTMSK.PRTINTMSK */
	gintmsk = SCD__READ(sdev, GINTMSK);
	gintmsk &= ~(SCD_B_GINTMSK_HCHINTMSK |
		 SCD_B_GINTMSK_INCOMPLPMSK);
#ifdef SLAVE_MODE_DEBUG
	if (!sdev->dmac) {
		gintmsk &= ~SCD_B_GINTMSK_RXFLVLMSK;
	}
#endif /* SLAVE_MODE_DEBUG */
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	/* Mask GINTMSK.NPTxFEmpMsk */
#ifdef SLAVE_MODE_DEBUG
	if (!sdev->dmac) {
		switch(ep_type){
		case USB_ENDPOINT_XFER_CONTROL:
		case USB_ENDPOINT_XFER_BULK:
		    gintmsk &= ~SCD_B_GINTMSK_NPTXFEMPMSK;
		    break;
		case USB_ENDPOINT_XFER_INT:
		case USB_ENDPOINT_XFER_ISOC:
		    gintmsk &= ~SCD_B_GINTMSK_PTXFEMPMSK;
		    break;
		}
		SCD__WRITE(sdev, GINTMSK, gintmsk);
	}
#endif /* SLAVE_MODE_DEBUG */

	DBG(2,"==>\n");
}


static void hep_disable_ch(struct usb_hcd *hcd, u8 epnum)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u32 hcchar;

	/* Disable channel */
	hcchar = SCD_READ(sdev, SCD_R_HCCHAR(epnum));
	if(hcchar & SCD_B_HCCHARN_CHENA){
		hcchar |= SCD_B_HCCHARN_CHDIS;
		SCD__WRITE(sdev, HCCHAR(epnum), hcchar);
	}
}

static void hep_disable_tx(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	hep_disable_ch(hcd, epnum);

	DBG(2,"==>\n");
}

static void hep_disable_rx(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	hep_disable_ch(hcd, epnum);

	DBG(2,"==>\n");
}

static void hep_enable_ch(struct usb_hcd *hcd, u8 epnum)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u32 hcchar;

	/* Enable channel */
	hcchar = SCD_READ(sdev, SCD_R_HCCHAR(epnum));
	if(!(hcchar & SCD_B_HCCHARN_CHENA)){
		hcchar &= ~(SCD_B_HCCHARN_CHDIS);
		hcchar |= SCD_B_HCCHARN_CHENA;
		SCD__WRITE(sdev, HCCHAR(epnum), hcchar);
	}
}

static void hep_enable_tx(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	hep_enable_ch(hcd, epnum);

	DBG(2,"==>\n");
}

static void hep_enable_rx(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	hep_enable_ch(hcd, epnum);

	DBG(2,"==>\n");
}

inline u32 hep_get_remain(struct urb *urb)
{
	return urb->transfer_buffer_length - urb->actual_length;
}


inline static u8 hep_get_fifo_type(struct usb_endpoint_descriptor *epd)
{
	bool is_bulk = usb_endpoint_xfer_bulk(epd);
	int is_in = usb_endpoint_dir_in(epd);

	return is_bulk && is_in ? 2 : 1;
}

static struct scd_hep *hep_inquiry_ep(struct usb_hcd *hcd,
				      struct usb_device *udev,
				      struct usb_host_endpoint *hep,
				      struct scd_hep *ep_table, u8 *nep)
{
	struct scd_hep *ep;
	u8 epnum = hep_hep2epnum(hep);
	u8 cur_epnum;
	u8 hidx;
	u8 i;

	DBG(2, "<==\n");

	hidx = hep_hash(hep, udev->devnum, epnum);

	i = 0;
	do {
		ep = &ep_table[hidx];
		if (!ep->dev && !ep->hep) {
			(*nep)++;
			ep->dev = udev;
			ep->hep = hep;
			break;
		}

		cur_epnum = hep_hep2epnum(ep->hep);
		if (cur_epnum == epnum && ep->dev &&
		    ep->dev->devnum == udev->devnum) {
			/* match! */
			break;
		}

		i++;
		hidx++;
		if (i == SCD_NUM_HEPS) {
			/* no match and no space for new ep */
			ep = NULL;
			break;
		}
		if (hidx == SCD_NUM_HEPS) {
			/* reset hidx to search from top of ep table */
			hidx = 0;
		}
	} while (1);

	DBG(2,"==>\n");
	return ep;
}

static void hep_release_ep(struct usb_hcd *hcd,
			   struct usb_host_endpoint *hep,
			   struct scd_hep *ep_table, u8 *nep, u8 other_nep)
{
	struct scd_hep *ep;
	u8 hidx;

	DBG(2, "<==\n");

	hidx = 0;

	do {
		ep = &ep_table[hidx];
		if (ep->hep == hep) {
			/* match! */
			ep->dev = NULL;
			ep->hep = NULL;
			(*nep)--;
			if (*nep + other_nep == 0) {
				/* If a rest appears, 
				   only one EP should clear. (not flush) */
				//scd_host_fifo_reset(hcd);
			}
			break;
		}

		hidx++;

	} while (hidx != SCD_NUM_HEPS);

	DBG(2,"==>\n");
}

static struct scd_hep *hep_search_ep(struct scd_port *port, u8 epadr)
{
	struct scd_host *shost = hub_to_shost(port_to_hub(port));
	struct scd_hep *ep;
	struct scd_hep *last_ep;
	bool is_in =
		(epadr & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN ? true : false;

	DBG(2, "<==\n");

	ep = is_in ? &shost->ep_in[0] : &shost->ep_out[0];
	last_ep = &ep[SCD_NUM_HEPS - 1];
	do {
		if (ep->hep && ep->hep->desc.bEndpointAddress == epadr)
			return ep;
		ep++;
	} while (ep != last_ep);

	DBG(2,"==>\n");
	return NULL;
}

#ifdef SLAVE_MODE_DEBUG
static void hep_set_interval_reg(struct usb_hcd *hcd, u32 intv_reg, u8 index,
			     struct usb_endpoint_descriptor *epd)
{
	/* nothing */
}
#endif /* SLAVE_MODE_DEBUG */

static void hep_get_buf(struct urb *urb, struct usb_endpoint_descriptor *epd,
			struct scd_dma_channel *dmach, char **buf, u32 *len)
{
	u8 ep_type = hep_check_ep_type(epd);
	u16 maxp = epd->wMaxPacketSize;
	u32 remain;

	DBG(2, "<==\n");

#ifdef SLAVE_MODE_DEBUG
	*buf = (dmach ?
		(char *)urb->transfer_dma + urb->actual_length :
		(char *)urb->transfer_buffer + urb->actual_length);
#else
	*buf = (char *)urb->transfer_buffer + urb->actual_length;
#endif /* SLAVE_MODE_DEBUG */

	
	remain = hep_get_remain(urb);
#ifdef SLAVE_MODE_DEBUG
	*len = dmach && (ep_type==USB_ENDPOINT_XFER_CONTROL||
	    ep_type==USB_ENDPOINT_XFER_BULK) ? remain : min((u32)maxp, remain);
#else
	*len = (ep_type==USB_ENDPOINT_XFER_CONTROL||
	    ep_type==USB_ENDPOINT_XFER_BULK) ? remain : min((u32)maxp, remain);
#endif /* SLAVE_MODE_DEBUG */

	DBG(2,"==>\n");
}

inline static void hep_add_actual(struct urb *urb, u32 len)
{
	urb->actual_length += len;
}

static long hep_check_bandwidth(struct usb_device *udev, struct scd_hep *ep)
{
	int old_alloc;
	int new_alloc;
	long bustime;
	bool is_in = usb_endpoint_dir_in(&ep->hep->desc);
	u16 maxp = ep->hep->desc.wMaxPacketSize;

	DBG(2, "<==\n");

	old_alloc = udev->bus->bandwidth_allocated;
	bustime = NS_TO_US(usb_calc_bus_time(udev->speed, is_in, false, maxp));
	new_alloc = old_alloc + (int)bustime;

	if (new_alloc > FRAME_TIME_MAX_USECS_ALLOC) {
		bustime = -ENOSPC;
	}

	DBG(2,"==>\n");
	return bustime;
}

static void hep_claim_bandwidth(struct usb_device *udev, struct scd_hep *ep,
				long bustime)
{
	udev->bus->bandwidth_allocated += bustime;
	udev->bus->bandwidth_int_reqs++;
}

static void hep_release_bandwidth(struct usb_device *udev, struct scd_hep *ep)
{
	udev->bus->bandwidth_allocated -= hep_check_bandwidth(udev, ep);
	udev->bus->bandwidth_int_reqs--;
}

static int hep_assign_bandwidth(struct usb_device *udev, struct scd_hep *ep)
{
	long bustime;

	DBG(2, "<==\n");

	bustime = hep_check_bandwidth(udev, ep);
	if (bustime < 0) {
		return bustime;
	}

	hep_claim_bandwidth(udev, ep, bustime);

	DBG(2,"==>\n");
	return 0;
}

static void hep_complete(struct usb_hcd *hcd, struct scd_hep *ep,
			 struct urb *urb, int status)
{
	unsigned int pipe = urb->pipe;
	int type = usb_pipetype(pipe);

	DBG(2, "<==\n");

	if (!((type == PIPE_CONTROL)||(type == PIPE_BULK))) {
		hep_release_bandwidth(urb->dev, ep);
	}
  	urb->hcpriv = NULL;
	urb->status = status;

	if (status || urb->actual_length != urb->transfer_buffer_length) {
		DBG(1, "hep_complete: error complete "
			"(devnum:%d,epadr:%x,status:%d,act:%d,expect:%d,"
			"flags:%x)\n",
			ep->dev->devnum, ep->hep->desc.bEndpointAddress,
			status, urb->actual_length, urb->transfer_buffer_length,
			urb->transfer_flags);
	}
	if (type == PIPE_CONTROL) {
		dma_unmap_single(&ep->dev->dev, (dma_addr_t)ep->status_packet, 
		    STATUS_PKT_MAXP, DMA_BIDIRECTIONAL);
	}
	usb_hcd_unlink_urb_from_ep(hcd, urb);

	spin_unlock(&(hcd_to_shost(hcd))->lock);
	usb_hcd_giveback_urb(hcd, urb, status);
	spin_lock(&(hcd_to_shost(hcd))->lock);
	if (ep->urb == urb)
		ep->urb = NULL;

	DBG(2,"==>\n");
}

static void hep_error(struct usb_hcd *hcd, struct scd_hep *ep, u32 error)
{
	int status;

	DBG(2, "<==\n");

	if (error & SCD_HEP_E_NOBUFFER) {
		DBG(3, "SCD_HEP_E_NOBUFFER\n");
		status = -ENOMEM;
	}
	else if (error & SCD_HEP_E_MAXPOVER) {
		DBG(3, "SCD_HEP_E_MAXPOVER\n");
		status = -ETIME;
	}
	else if (error & SCD_HEP_E_NORESP) {
		DBG(3, "SCD_HEP_E_NORESP\n");
		status = -ETIME;
	}
	else if (error & SCD_HEP_E_RXSTALL) {
		DBG(3, "SCD_HEP_E_RXSTALL\n");
		status = -EPIPE;
	}
	else if (error & SCD_HEP_E_OVERFLOW) {
		DBG(3, "SCD_HEP_E_OVERFLOW\n");
		status = -EOVERFLOW;
	}
	else if (error & SCD_HEP_E_SHORT) {
		DBG(3, "SCD_HEP_E_SHORT\n");
		status = -EREMOTEIO;
	}
	else if (error & SCD_HEP_E_XACTERR ) {
		DBG(3, "SCD_HEP_E_XACTERR\n");
		status = -ETIME;
	}
	else if (error & SCD_HEP_E_AHBERR ) {
		DBG(3, "SCD_HEP_E_AHBERR\n");
		if(ep->urb) {
			status = usb_pipein(ep->urb->pipe)
				? -ENOSR  /* hc couldn't read */
				: -ECOMM; /* hc couldn't write */
		}
		else {
			status = -ERANGE; /* unknown */
		}
	}
	else {
		status = -ERANGE; /* unknown */
	}

	hep_complete(hcd, ep, ep->urb, status);

	DBG(2,"==>\n");
}

static u32 hep_check_rx_len_error(u16 maxp, u32 expected, u16 count,
				  bool short_not_ok)
{
	u32 error = 0;

	DBG(2, "<==\n");
#ifdef TRANSFER_FLAGS_DEBUG
	if (short_not_ok && count != 0 && count < maxp) {
		error |= SCD_HEP_E_SHORT;
	}
#endif
	if (count > maxp) {
		error |= SCD_HEP_E_MAXPOVER;
	}
	if (count > expected) {
		error |= SCD_HEP_E_OVERFLOW;
	}

	DBG(2,"==>\n");
	return error;
}

inline bool hep_check_bulk(struct scd_hep *ep)
{
	return usb_endpoint_xfer_bulk(&ep->hep->desc);
}

inline bool hep_check_short(struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	u16 maxp = ep->hep->desc.wMaxPacketSize;

	return hep_get_remain(urb) < maxp ? true : false;
}

inline static void hep_set_pid(struct scd_hep *ep, u32 pid)
{
	if (ep) {
		ep->pid = pid;
	}
}

inline static u8 hep_cntl_get_type(struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	struct usb_ctrlrequest *request =
		(struct usb_ctrlrequest *)urb->setup_packet;
	u8 type;

	DBG(2, "<==\n");

	if (request->bRequestType & USB_DIR_IN) {
		type = SCD_HEP_CNTL_IN;
	}
	else if (request->wLength == 0) {
		type = SCD_HEP_CNTL_NODATA;
	}
	else {
		type = SCD_HEP_CNTL_OUT;
	}

	DBG(2,"==>\n");
	return type;
}

/* tx transaction */

inline static void hep_tx_unmask_intr(struct usb_hcd *hcd, struct scd_hep *ep)
{
	hep_unmask_tx_intr(hcd, ep);
}

inline static void hep_tx_mask_intr(struct usb_hcd *hcd, struct scd_hep *ep)
{
	hep_mask_tx_intr(hcd, ep);
}

inline static void hep_tx_disable(struct usb_hcd *hcd, struct scd_hep *ep)
{
	hep_disable_tx(hcd, ep);
}

inline static void hep_tx_enable(struct usb_hcd *hcd, struct scd_hep *ep)
{
	hep_enable_tx(hcd, ep);
}

inline static void hep_tx_reset(struct usb_hcd *hcd, struct scd_hep *ep)
{
	scd_hep_tx_flush_fifo(hcd, ep);
}

static void hep_tx_self_start(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct urb *urb;
	struct usb_host_endpoint *hep = ep->hep;

	DBG(2, "<==\n");

	if (ep->urb || !hep || list_empty(&hep->urb_list)) {
		DBG(2,"==>\n");
		return;
	}
	urb = list_entry(hep->urb_list.next, struct urb, urb_list);
	ep->urb = urb;

	hep_tx_trans_state(hcd, ep, SCD_HEP_T_START);

	DBG(2,"==>\n");
}

inline static void hep_tx_set_pid(struct scd_hep *ep, u32 pid)
{
	hep_set_pid(ep, pid);
}

#ifdef SLAVE_MODE_DEBUG
static void hep_tx_unmask_hcint(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u32 hcintmsk;

	DBG(2, "<==\n");

	/* Unmask HCINTMSK for channel interrupt */
	hcintmsk = SCD__READ(sdev, HCINTMSK(epnum));
	hcintmsk |= SCD_B_HCINTMSKN_DATATGLERR |
		SCD_B_HCINTMSKN_FRMOVRUN |
		SCD_B_HCINTMSKN_BBLERR |
		SCD_B_HCINTMSKN_XACTERR |
		SCD_B_HCINTMSKN_STALL |
		SCD_B_HCINTMSKN_AHBERR |
		SCD_B_HCINTMSKN_CHHLTD;
#ifdef SLAVE_MODE_DEBUG
	if (!sdev->dmac) {
		hcintmsk |= SCD_B_HCINTMSKN_NYET |
			SCD_B_HCINTN_ACK |
			SCD_B_HCINTMSKN_NAK |
			SCD_B_HCINTMSKN_XFERCOMPL;
	}
#endif /* SLAVE_MODE_DEBUG */
	SCD__WRITE(sdev, HCINTMSK(epnum), hcintmsk);

	DBG(2,"==>\n");
}
#endif /* SLAVE_MODE_DEBUG */

static void hep_tx_set_ch_size_multi(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum,u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u16 maxp;
	u32 pid;
	u32 remain, pktcnt;
	u32 hctsiz;

	DBG(2, "<==\n");

	maxp = ep->hep->desc.wMaxPacketSize;
	pid = ep->pid;

	/* DOPNG, PID, PKTCNT, XFERSIZE */
	/* DOPNG is set BULK/CONTROL Transfer */
	/* PID = DATA0, (30:29)=2b00 */
	remain = hep_get_remain(ep->urb);
	pktcnt = (remain/maxp) + (((remain % maxp) || (0 == remain))? 1 : 0 );
	/* Care only PKTCNT, because PKTCNT*512 < XFERSIZE. */
	if(SCD_MAX_PKTCNT < pktcnt) {
		pktcnt =  SCD_MAX_PKTCNT;
		remain = pktcnt * maxp;
	}
	hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
	if (pid != PID_AUTO){
		/* Clear bit to set */
		hctsiz = pid;
		hep_tx_set_pid(ep, PID_AUTO);
	}
	else {
		/* Clear bit to set, Unclear pid bit */
		hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
		hctsiz &= (SCD_M_HCTSIZN_PID << SCD_S_HCTSIZN_PID);
	}
	hctsiz |= ((((ep_type==USB_ENDPOINT_XFER_CONTROL)||
		(ep_type==USB_ENDPOINT_XFER_BULK)) && !ep->urb->actual_length)? 
		SCD_B_HCTSIZN_DOPNG : 0x0 ) | 
		(pktcnt<<SCD_S_HCTSIZN_PKTCNT) | 
		remain;
	SCD__WRITE(sdev, HCTSIZ(epnum), hctsiz );

	DBG(2,"==>\n");
}

static void hep_tx_set_ch_size(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum,u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u16 maxp;
	u32 pid;
	u32 remain, pktcnt, xfersize;
	u32 hctsiz;

	DBG(2, "<==\n");

	maxp = ep->hep->desc.wMaxPacketSize;
	pid = ep->pid;

	/* DOPNG, PID, PKTCNT, XFERSIZE */
	/* DOPNG is set BULK/CONTROL Transfer */
	/* PID = DATA0, (30:29)=2b00 */
	remain = hep_get_remain(ep->urb);
	pktcnt = (0 <= remain ? 1 : 0 );
	xfersize = (maxp < remain ? maxp : remain);
	hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
	if (pid != PID_AUTO){
		/* Clear bit to set */
		hctsiz = pid;
		hep_tx_set_pid(ep, PID_AUTO);
	}
	else {
		/* Clear bit to set, Unclear pid bit */
		hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
		hctsiz &= (SCD_M_HCTSIZN_PID << SCD_S_HCTSIZN_PID);
	}
	hctsiz |= ((((ep_type==USB_ENDPOINT_XFER_CONTROL)||
		(ep_type==USB_ENDPOINT_XFER_BULK)) && !ep->urb->actual_length)? 
		SCD_B_HCTSIZN_DOPNG : 0x0 ) | 
		(pktcnt<<SCD_S_HCTSIZN_PKTCNT) | 
		xfersize;
	SCD__WRITE(sdev, HCTSIZ(epnum), hctsiz );

	DBG(2,"==>\n");
}

static void hep_cntl_tx_setup_ch_size(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum,u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u16 maxp;
	u32 remain, pktcnt, xfersize;
	u32 hctsiz;

	DBG(2, "<==\n");

	maxp = ep->hep->desc.wMaxPacketSize;

	/* DOPNG, PID, PKTCNT, XFERSIZE */
	/* PID = SETUP */
	remain = 8; /* SETUP Packet */
	pktcnt = (0 <= remain ? 1 : 0 );
	xfersize = (maxp < remain ? maxp : remain);
	/* Clear bit to set */
	hctsiz = (SCD_B_HCTSIZN_PID_SETUP |
		(pktcnt << SCD_S_HCTSIZN_PKTCNT) |
		xfersize);
	SCD__WRITE(sdev, HCTSIZ(epnum), hctsiz );

	DBG(2,"==>\n");
}

static void hep_cntl_tx_status_ch_size(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum,u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u16 maxp;
	u32 remain, pktcnt, xfersize;
	u32 hctsiz;

	DBG(2, "<==\n");

	maxp = ep->hep->desc.wMaxPacketSize;

	/* DOPNG, PID, PKTCNT, XFERSIZE */
	/* PID = SETUP */
	remain = 0; /* STATUS Packet */
	pktcnt = (0 <= remain ? 1 : 0 );
	xfersize = (maxp < remain ? maxp : remain);
	/* Clear bit to set */
	hctsiz = (SCD_B_HCTSIZN_PID_DATA1 |
		(pktcnt << SCD_S_HCTSIZN_PKTCNT) |
		xfersize);
	SCD__WRITE(sdev, HCTSIZ(epnum), hctsiz );

	DBG(2,"==>\n");
}

static void hep_tx_set_dma_addr(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	char* buf;

	DBG(2, "<==\n");
	buf = (char *)(ep->urb->transfer_dma + ep->urb->actual_length);
	SCD__WRITE(sdev, HCDMA(epnum), (u32)buf & (u32)SCD_M_HCDMAN_DMAADDR);
	DBG(2,"==>\n");
}

static void hep_tx_reset_dma_addr(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	char* buf;
	u8 ep_type;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u32 remain, pktcnt, actual = 0;
	u32 hctsiz;

	DBG(2, "<==\n");

	ep_type = hep_check_ep_type(&ep->hep->desc);

	switch(ep_type){
	case USB_ENDPOINT_XFER_CONTROL:
	case USB_ENDPOINT_XFER_BULK:
	    remain = hep_get_remain(ep->urb);
	    pktcnt = (remain/maxp) + (((remain % maxp) || (0 == remain))? 1 : 0 );
	    hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
	    actual = remain ? (pktcnt - ((hctsiz >> SCD_S_HCTSIZN_PKTCNT) & 
	        SCD_M_HCTSIZN_PKTCNT)) * maxp : 0;
	    break;
	case USB_ENDPOINT_XFER_INT:
	case USB_ENDPOINT_XFER_ISOC:
	    actual = 0;
	    break;
	}

	buf = (char *)(ep->urb->transfer_dma + ep->urb->actual_length + actual);
	SCD__WRITE(sdev, HCDMA(epnum), (u32)buf & (u32)SCD_M_HCDMAN_DMAADDR );

	DBG(2,"==>\n");
}

static void hep_cntl_set_setup_dma_addr(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	char* buf;

	DBG(2, "<==\n");
	buf = (char *)ep->urb->setup_dma;
	SCD__WRITE(sdev, HCDMA(epnum), (u32)buf & (u32)SCD_M_HCDMAN_DMAADDR );
	DBG(2,"==>\n");
}

static void hep_cntl_set_status_dma_addr(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	char* buf;

	DBG(2, "<==\n");
	buf = (char *)ep->status_dma;
	SCD__WRITE(sdev, HCDMA(epnum), (u32)buf & (u32)SCD_M_HCDMAN_DMAADDR );
	DBG(2,"==>\n");
}

static void hep_tx_set_ch_char(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum, u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 is_tx;
	u16 speed;
	u32 lspddev, epdir;
	u32 hccharn, hfnum;

	DBG(2, "<==\n");

	speed = port_get_speed(&(hcd_to_shost(hcd))->hub.port[0]);
	lspddev = ((speed==USB_OTG_SPEED_LS)? SCD_B_HCCHARN_LSPDDEV : 0x0);
	if (!epnum){
		is_tx = (SCD_HEP_CNTL_OUT == hep_cntl_get_type(ep));
	}
	else {
		is_tx = usb_endpoint_dir_out(&ep->hep->desc);
	}
	epdir = (is_tx? 0x0 : SCD_B_HCCHARN_EPDIR);

	hccharn = SCD__READ(sdev, HCCHAR(epnum));
	/* Clear bit to set */
	hccharn &= (~(SCD_B_HCCHARN_CHDIS) & 0xfff00000); /* HCCHARN init */
	/* tx packet next frame at interrupt transfer or isoc transfer */
	if((ep_type==USB_ENDPOINT_XFER_INT)||
	    (ep_type==USB_ENDPOINT_XFER_ISOC)){
		hfnum = (SCD__READ(sdev, HFNUM)) & SCD_M_HFNUM_FRNUM;
		if (hfnum % 2) {
			/* Set next odd frame */
			hccharn |= SCD_B_HCCHARN_ODDFRM;
		}
		else {
			/* Set next even frame */
			hccharn &= ~(SCD_B_HCCHARN_ODDFRM);
		}
	}

	hccharn &= ~(SCD_M_HCCHARN_DEVADDR << SCD_S_HCCHARN_DEVADDR);
	hccharn |= ((ep->urb->dev->state > USB_STATE_DEFAULT ? ep->urb->dev->devnum : 0)
		    << SCD_S_HCCHARN_DEVADDR) |
							/* device address */
		(0x01 << SCD_S_HCCHARN_MCEC) |		/* MC set 1 */
		(ep_type << SCD_S_HCCHARN_EPTYPE) |	/* ep_type same as USB_ENDPOINT_XFER_? */
		lspddev |				/* LS device? */
		epdir |					/* ep dir */
		(epnum << SCD_S_HCCHARN_EPNUM) |	/* ep number */
		ep->hep->desc.wMaxPacketSize ;		/* max packet size */
	SCD__WRITE(sdev, HCCHAR(epnum), hccharn);

	DBG(2,"==>\n");
}

static void hep_cntl_tx_ch_char(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum, u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	bool is_tx;
	u16 speed;
	u32 lspddev, epdir;
	u32 hccharn, hfnum;

	DBG(2, "<==\n");

	speed = port_get_speed(&(hcd_to_shost(hcd))->hub.port[0]);
	lspddev = ((speed==USB_OTG_SPEED_LS)? SCD_B_HCCHARN_LSPDDEV : 0x0);
	is_tx = true;
	epdir = (is_tx? 0x0 : SCD_B_HCCHARN_EPDIR);

	hccharn = SCD__READ(sdev, HCCHAR(epnum));
	/* Clear bit to set */
	hccharn &= (~SCD_B_HCCHARN_CHDIS & 0xfff00000); /* HCCHARN init */
	/* tx packet next frame at interrupt transfer or isoc transfer */
	if((ep_type==USB_ENDPOINT_XFER_INT)||
	    (ep_type==USB_ENDPOINT_XFER_ISOC)){
		hfnum = (SCD__READ(sdev, HFNUM)) & SCD_M_HFNUM_FRNUM;
		if (hfnum % 2) {
			/* Set next odd frame */
			hccharn |= SCD_B_HCCHARN_ODDFRM;
		}
		else {
			/* Set next even frame */
			hccharn &= ~(SCD_B_HCCHARN_ODDFRM);
		}
	}

	hccharn &= ~(SCD_M_HCCHARN_DEVADDR << SCD_S_HCCHARN_DEVADDR);
	hccharn |= ((ep->urb->dev->state > USB_STATE_DEFAULT ? ep->urb->dev->devnum : 0)
		    << SCD_S_HCCHARN_DEVADDR) |
							/* device address */
		(0x01 << SCD_S_HCCHARN_MCEC) |		/* MC set 1 */
		(ep_type << SCD_S_HCCHARN_EPTYPE) |	/* ep_type same as USB_ENDPOINT_XFER_? */
		lspddev |				/* LS device? */
		epdir |					/* ep dir */
		(epnum << SCD_S_HCCHARN_EPNUM) |	/* ep number */
		ep->hep->desc.wMaxPacketSize ;		/* max packet size */
	SCD__WRITE(sdev, HCCHAR(epnum), hccharn);

	DBG(2,"==>\n");
}

static int hep_tx_data_dma_multi(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_tx_set_ch_size_multi(hcd, ep, epnum, ep_type);
	hep_tx_set_dma_addr(hcd, ep, epnum);
	hep_tx_set_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}

static int hep_tx_data_dma(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_tx_set_ch_size(hcd, ep, epnum, ep_type);
	hep_tx_set_dma_addr(hcd, ep, epnum);
	hep_tx_set_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}

static int hep_tx_data_cpu(struct usb_hcd *hcd, struct scd_hep *ep)
{
	char* buf;
	u8 epnum;
	u8 ep_type;
	u32 len;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_tx_set_ch_size(hcd, ep, epnum, ep_type);
	hep_get_buf(ep->urb, &ep->hep->desc, NULL, &buf, &len);
	scd_device_host_load_fifo(hcd, ep_type, epnum, len, (u32 *)buf);
	hep_tx_set_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}

static u32 hep_tx_check_error(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 epnum;
	u32 error = 0;
	u32 hcint;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);

	hcint = SCD__READ(sdev, HCINT(epnum));

	error = 0;
	if (hcint  & SCD_B_HCINTN_FRMOVRUN) {
		error |= SCD_HEP_E_OVERFLOW;
	}
	if (hcint  & SCD_B_HCINTN_XACTERR) {
		error |= SCD_HEP_E_XACTERR;
	}
	if (hcint  & SCD_B_HCINTN_STALL) {
		error |= SCD_HEP_E_RXSTALL;
	}
	if (hcint  & SCD_B_HCINTN_AHBERR) {
		error |= SCD_HEP_E_AHBERR;
	}

	DBG(2,"==>\n");
	return error;
}

static bool hep_tx_update_actual_multi(struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u32 last_remain, pktcnt;
	bool zerop;

	DBG(2, "<==\n");

	last_remain = hep_get_remain(ep->urb);
	pktcnt = (last_remain/maxp) + (((last_remain % maxp) || (0 == last_remain))? 1 : 0 );
	/* Care only PKTCNT, because PKTCNT*512 < XFERSIZE. */
	if(SCD_MAX_PKTCNT < pktcnt) {
		pktcnt =  SCD_MAX_PKTCNT;
		last_remain = pktcnt * maxp;
	}
	/* If HCTSIZn was readable, using it is better. */
	hep_add_actual(urb, last_remain);

	zerop = last_remain && !(last_remain%maxp) && 
		urb->transfer_flags & URB_ZERO_PACKET ? true : false;

	DBG(2,"==>\n");
	return zerop;
}

inline static void hep_tx_update_actual(struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u32 remain, add_len;

	DBG(2, "<==\n");

	remain = hep_get_remain(urb);
	add_len = remain < maxp ? remain : maxp;
	/* If HCTSIZn was readable, using it is better. */
	hep_add_actual(urb, add_len);

	DBG(2,"==>\n");
}

#ifdef SLAVE_MODE_DEBUG
static u32 hep_tx_get_nextsz(struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	char* buf;
	u32 len;

	DBG(2, "<==\n");
	hep_get_buf(urb, &ep->hep->desc, ep->dmach, &buf, &len);

	DBG(2,"==>\n");
	return urb->transfer_buffer_length + len - urb->actual_length;
}
#endif /* SLAVE_MODE_DEBUG */

static bool hep_tx_check_remain(struct scd_hep *ep, u32 *nextsz)
{
	struct urb *urb = ep->urb;
	char* buf;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u32 len;
	u32 lastsz;

	DBG(2, "<==\n");

	hep_get_buf(ep->urb, &ep->hep->desc, ep->dmach, &buf, &len);
	lastsz = hep_get_remain(urb);
	hep_add_actual(urb, len);
	*nextsz = hep_get_remain(urb);

	DBG(2,"==>\n");
	return (lastsz && !(lastsz%maxp) && urb->transfer_flags & URB_ZERO_PACKET) ||
		*nextsz ? true : false;
}

static void hep_tx_set_last_toggle(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 epnum;
	u32 hctsiz;
	bool toggle = false;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
	hctsiz &= (SCD_M_HCTSIZN_PID << SCD_S_HCTSIZN_PID);
	if(hctsiz == PID_DATA1){
		toggle = true;
	}
	else if(toggle == PID_DATA0){
		toggle = false;
	}
	else {
		/* error */
	}
	usb_settoggle(ep->dev, epnum, true, toggle);

	DBG(2,"==>\n");
}

static void hep_tx_set_new_toggle(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 epnum;
	u32 hctsiz;
	bool toggle;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	toggle = usb_gettoggle(ep->dev, epnum, true);

	hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
	if (toggle) {
		hctsiz &= ~(SCD_M_HCTSIZN_PID << SCD_S_HCTSIZN_PID);
		hctsiz |= (PID_DATA1);
	}
	else {
		hctsiz &= ~(SCD_M_HCTSIZN_PID << SCD_S_HCTSIZN_PID);
		hctsiz |= (PID_DATA0);
	}
	SCD__WRITE(sdev, HCTSIZ(epnum), hctsiz);

	DBG(2,"==>\n");
}

static void hep_tx_dma_sync_start(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	struct scd_hcd *shcd = (hcd_to_shost(hcd))->shcd;

	DBG(2, "<==\n");

	if (!urb->transfer_buffer) {
		urb->transfer_buffer = ioremap(urb->transfer_dma,
					       urb->transfer_buffer_length);
		if (!urb->transfer_buffer) {
			return;
		}
		ep->attributes |= SCD_HEP_ATR_IOREMAP;
	}

	if (urb->transfer_dma) {
		dma_sync_single_for_cpu(shcd->dev, urb->transfer_dma,
					urb->transfer_buffer_length,
					DMA_FROM_DEVICE);
	}

	DBG(2,"==>\n");
}

static void hep_tx_dma_sync_end(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	struct scd_hcd *shcd = (hcd_to_shost(hcd))->shcd;

	DBG(2, "<==\n");

	if (urb->transfer_dma) {
		dma_sync_single_for_device(shcd->dev, urb->transfer_dma,
					   urb->transfer_buffer_length,
					   DMA_FROM_DEVICE);
	}

	if (ep->attributes & SCD_HEP_ATR_IOREMAP) {
		iounmap(urb->transfer_buffer);
		urb->transfer_buffer = NULL;
		ep->attributes &= ~(SCD_HEP_ATR_IOREMAP);
	}

	DBG(2,"==>\n");
}

static int hep_tx_trans_state(struct usb_hcd *hcd, struct scd_hep *ep,
			      u32 trigger)
{
	u8 epnum;
	bool reset = trigger == SCD_HEP_T_RESET ? true : false;
	bool remain = false;
	bool zerop = false; 
	u8 ep_type = 0;
	u32 nextsz = 0;
	u32 error = 0;

	DBG(2, "<==\n");
	DBG(3, "ep->state (exit) = %lu\n", (unsigned long)ep->state);

	epnum = hep_hep2epnum(ep->hep);

	/* exit state */
	switch (ep->state) {
	case SCD_HEP_S_TX_IDLE:
		if ((trigger == SCD_HEP_T_LINK_CHHLT_INTR) ||
		    (trigger == SCD_HEP_T_LINK_ERROR_INTR)) {
			/* Nothing to do */
			return 0;
		}
		hep_tx_set_new_toggle(hcd, ep);
		ep->retrycount = 0;
		if (!reset) {
#ifdef SLAVE_MODE_DEBUG
			if (!sdev->dmac) {
				nextsz = hep_tx_get_nextsz(ep);
				hep_tx_dma_sync_start(hcd, ep);
			}
#endif /* SLAVE_MODE_DEBUG */
			ep_type = hep_check_ep_type(&ep->hep->desc);
		}
		break;
	case SCD_HEP_S_TX_WAIT_DMA_MULTI:
		if (!reset) {
			error = hep_tx_check_error(hcd, ep);
			if (!error) {
				zerop = hep_tx_update_actual_multi(ep);
				remain = hep_get_remain(ep->urb);
			}
		}
		break;
	case SCD_HEP_S_TX_WAIT_DMA_ZEROP:
		if (!reset) {
			error = hep_tx_check_error(hcd, ep);
			if (!error) {
				hep_tx_update_actual(ep);
			}
		}
		break;
	case SCD_HEP_S_TX_WAIT_DMA:
		if (!reset) {
			error = hep_tx_check_error(hcd, ep);
			if (!error) {
				remain = hep_tx_check_remain(ep, &nextsz);
			}
		}
		break;
	case SCD_HEP_S_TX_WAIT_CPU:
		if (!reset) {
			error = hep_tx_check_error(hcd, ep);
			if (!error) {
				remain = hep_tx_check_remain(ep, &nextsz);
			}
		}
		if (reset || error || !remain) {
			hep_tx_dma_sync_end(hcd, ep);
		}
		break;
	}

	/* transition state */
	if (trigger == SCD_HEP_T_RESET ||
	    trigger == SCD_HEP_T_LINK_ERROR_INTR ||
	    ((trigger == SCD_HEP_T_LINK_CHHLT_INTR ||
	      trigger == SCD_HEP_T_LINK_XFERCOMPL_INTR )&& error)) {
		ep->state = SCD_HEP_S_TX_IDLE;
		goto enter_state;
	}
	switch (ep->state) {
	case SCD_HEP_S_TX_IDLE:
		if (trigger == SCD_HEP_T_START) {
#ifdef SLAVE_MODE_DEBUG
			if (sdev->dmac) {
#endif /* SLAVE_MODE_DEBUG */
				if(ep_type == USB_ENDPOINT_XFER_BULK) {
					ep->state = SCD_HEP_S_TX_WAIT_DMA_MULTI;
					goto enter_state;
				}
				else if (ep_type == USB_ENDPOINT_XFER_INT) {
					ep->state = SCD_HEP_S_TX_WAIT_DMA;
					goto enter_state;
				}
#ifdef SLAVE_MODE_DEBUG
			}
			else {
				ep->state = SCD_HEP_S_TX_WAIT_CPU;
				goto enter_state;
			}
#endif /* SLAVE_MODE_DEBUG */
		}
		break;
	case SCD_HEP_S_TX_WAIT_DMA_MULTI:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				if (remain) {
					/* Do not change state */
					goto enter_state;
				}
				else {
					if (!zerop) {
						ep->state = SCD_HEP_S_TX_IDLE;
						goto enter_state;
					}
					else {
						ep->state = SCD_HEP_S_TX_WAIT_DMA_ZEROP;
						goto enter_state;
					}
				}
			}
		}
		break;
	case SCD_HEP_S_TX_WAIT_DMA_ZEROP:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				ep->state = SCD_HEP_S_TX_IDLE;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_TX_WAIT_DMA:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				if (remain) {
					/* Do not change state */
					goto enter_state;
				}
				else {
					ep->state = SCD_HEP_S_TX_IDLE;
					goto enter_state;
				}
			}

		}
		break;
	case SCD_HEP_S_TX_WAIT_CPU:
		if (trigger == SCD_HEP_T_LINK_XFERCOMPL_INTR) {
			if (!error) {
				if (remain) {
					ep->state = SCD_HEP_S_TX_WAIT_CPU;
					goto enter_state;
				}
				else {
					ep->state = SCD_HEP_S_TX_IDLE;
					goto enter_state;
				}
			}
		}
		break;
	}
	goto error_end;

	/* enter state */
enter_state:
	switch (ep->state) {
	case SCD_HEP_S_TX_IDLE:
		hep_tx_disable(hcd, ep);
		hep_tx_set_last_toggle(hcd, ep);
		if (!reset) {
			if (!error) {
				hep_complete(hcd, ep, ep->urb, 0);
				hep_tx_self_start(hcd, ep);
			}
			else {
				hep_error(hcd, ep, error);
			}
		}
		break;
	case SCD_HEP_S_TX_WAIT_DMA_MULTI:
		hep_tx_unmask_intr(hcd, ep);
		hep_tx_data_dma_multi(hcd, ep);
		hep_tx_enable(hcd, ep);
		break;
	case SCD_HEP_S_TX_WAIT_DMA_ZEROP:
		hep_tx_unmask_intr(hcd, ep);
		hep_tx_data_dma(hcd, ep);
		hep_tx_enable(hcd, ep);
		break;
	case SCD_HEP_S_TX_WAIT_DMA:
		hep_tx_unmask_intr(hcd, ep);
		hep_tx_data_dma(hcd, ep);
		hep_tx_enable(hcd, ep);
		break;
	case SCD_HEP_S_TX_WAIT_CPU:
		hep_tx_unmask_intr(hcd, ep);
		hep_tx_data_cpu(hcd, ep);
		hep_tx_enable(hcd, ep);
		break;
	}

	DBG(3, "ep->state (enter) = %lu\n", (unsigned long)ep->state);
	DBG(2,"==>\n");
	return 0;

error_end:
	DBG(3, "ep->state (enter) = %lu\n", (unsigned long)ep->state);
	DBG(2,"==>\n");
	return -EINVAL;
}

/* rx transaction */

inline static void hep_rx_unmask_intr(struct usb_hcd *hcd, struct scd_hep *ep)
{
	hep_unmask_rx_intr(hcd, ep);
}

inline static void hep_rx_mask_intr(struct usb_hcd *hcd, struct scd_hep *ep)
{
	hep_mask_rx_intr(hcd, ep);
}

inline static void hep_rx_disable(struct usb_hcd *hcd, struct scd_hep *ep)
{
	hep_disable_rx(hcd, ep);
}

inline static void hep_rx_enable(struct usb_hcd *hcd, struct scd_hep *ep)
{
	hep_enable_rx(hcd, ep);
}

inline static void hep_rx_reset(struct usb_hcd *hcd, struct scd_hep *ep)
{
	scd_hep_rx_flush_fifo(hcd, ep);
}

static void hep_rx_self_start(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct urb *urb;
	struct usb_host_endpoint *hep = ep->hep;
	
	DBG(2, "<==\n");

	if (ep->urb || !hep || list_empty(&hep->urb_list)) {
		DBG(2,"==>\n");
		return;
	}
	urb = list_entry(hep->urb_list.next, struct urb, urb_list);
	ep->urb = urb;

	hep_rx_trans_state(hcd, ep, SCD_HEP_T_START);

	DBG(2,"==>\n");
}

inline static void hep_rx_set_pid(struct scd_hep *ep, u32 pid)
{
	hep_set_pid(ep, pid);
}

#ifdef SLAVE_MODE_DEBUG
static void hep_rx_unmask_hcint(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u32 hcintmsk;

	DBG(2, "<==\n");

	/* Unmask HCINTMSK for channel interrupt */
	hcintmsk = SCD__READ(sdev, HCINTMSK(epnum));
	hcintmsk |= SCD_B_HCINTMSKN_DATATGLERR |
		SCD_B_HCINTMSKN_FRMOVRUN |
		SCD_B_HCINTMSKN_BBLERR |
		SCD_B_HCINTMSKN_XACTERR |
		SCD_B_HCINTMSKN_STALL |
		SCD_B_HCINTMSKN_AHBERR |
		SCD_B_HCINTMSKN_CHHLTD;
#ifdef SLAVE_MODE_DEBUG
	if (!sdev->dmac) {
		hcintmsk |= SCD_B_HCINTMSKN_NYET |
			SCD_B_HCINTN_ACK |
			SCD_B_HCINTMSKN_NAK |
			SCD_B_HCINTMSKN_XFERCOMPL;
	}
#endif /* SLAVE_MODE_DEBUG */
	SCD__WRITE(sdev, HCINTMSK(epnum), hcintmsk );

	DBG(2,"==>\n");
}
#endif /* SLAVE_MODE_DEBUG */

static void hep_rx_set_ch_size_multi(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum,u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u16 maxp;
	u32 pid;
	u32 remain, pktcnt, xfersize;
	u32 hctsiz;

	DBG(2, "<==\n");

	maxp = ep->hep->desc.wMaxPacketSize;
	pid = ep->pid;

	/* DOPNG, PID, PKTCNT, XFERSIZE */
	/* DOPNG is set BULK/CONTROL Transfer */
	/* PID = DATA0, (30:29)=2b00 */
	remain = hep_get_remain(ep->urb);
	if (remain == 0) {
		pktcnt = 1;
	}
	else {
		/* except short packet */
		pktcnt = remain / maxp;
		/* Care only PKTCNT, because PKTCNT*512 < XFERSIZE. */
		if(SCD_MAX_PKTCNT < pktcnt) {
			pktcnt =  SCD_MAX_PKTCNT;
		}	
	}
	/* Setting Xfersize of IN Transfer in maxp*n, not bytes. */
	xfersize = pktcnt * maxp; 
	if (pid != PID_AUTO){
		/* Clear bit to set */
		hctsiz = pid;
		hep_rx_set_pid(ep, PID_AUTO);
	}
	else {
		/* Clear bit to set, Unclear pid bit */
		hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
		hctsiz &= (SCD_M_HCTSIZN_PID << SCD_S_HCTSIZN_PID);
	}
	hctsiz |= (pktcnt<<SCD_S_HCTSIZN_PKTCNT) |
		xfersize;
	SCD__WRITE(sdev, HCTSIZ(epnum), hctsiz );

	DBG(2,"==>\n");
}

static void hep_rx_set_ch_size(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum,u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u16 maxp;
	u32 pid;
	u32 remain, pktcnt, xfersize;
	u32 hctsiz;

	DBG(2, "<==\n");

	maxp = ep->hep->desc.wMaxPacketSize;
	pid = ep->pid;

	/* DOPNG, PID, PKTCNT, XFERSIZE */
	/* DOPNG is set BULK/CONTROL Transfer */
	/* PID = DATA0, (30:29)=2b00 */
	remain = hep_get_remain(ep->urb);
	pktcnt = (0 <= remain ? 1 : 0 );
	/* Setting Xfersize of IN Transfer in maxp*n, not bytes. */
	xfersize = pktcnt * maxp; 
	if (pid != PID_AUTO){
		/* Clear bit to set */
		hctsiz = pid;
		hep_rx_set_pid(ep, PID_AUTO);
	}
	else {
		/* Clear bit to set, Unclear pid bit */
		hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
		hctsiz &= (SCD_M_HCTSIZN_PID << SCD_S_HCTSIZN_PID);
	}
	hctsiz |= (pktcnt<<SCD_S_HCTSIZN_PKTCNT) |
		xfersize;
	SCD__WRITE(sdev, HCTSIZ(epnum), hctsiz );

	DBG(2,"==>\n");
}

static void hep_cntl_rx_status_ch_size(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum,u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u16 maxp;
	u32 pid;
	u32 remain, pktcnt, xfersize;
	u32 hctsiz;

	DBG(2, "<==\n");

	maxp = ep->hep->desc.wMaxPacketSize;
	pid = SCD_B_HCTSIZN_PID_DATA1;

	/* DOPNG, PID, PKTCNT, XFERSIZE */
	/* DOPNG is set BULK/CONTROL Transfer */
	/* PID = DATA0, (30:29)=2b00 */
	remain = 0;
	pktcnt = (0 <= remain ? 1 : 0 );
	/* Setting Xfersize of IN Transfer in maxp*n, not bytes. */
	xfersize = remain; 
	hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
	hctsiz = pid |
		(pktcnt<<SCD_S_HCTSIZN_PKTCNT) |
		xfersize;
	SCD__WRITE(sdev, HCTSIZ(epnum), hctsiz );

	DBG(2,"==>\n");
}

static void hep_rx_set_dma_addr(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	char* buf;

	DBG(2, "<==\n");
	buf = (char *)(ep->urb->transfer_dma + ep->urb->actual_length);
	SCD__WRITE(sdev, HCDMA(epnum), (u32)buf&(u32)SCD_M_HCDMAN_DMAADDR );
	DBG(2,"==>\n");
}

static void hep_rx_set_dma_addr_short(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	char* buf;

	DBG(2, "<==\n");
	buf = (char *)ep->packet_dma;
	SCD__WRITE(sdev, HCDMA(epnum), (u32)buf&(u32)SCD_M_HCDMAN_DMAADDR );
	DBG(2,"==>\n");
}

static void hep_rx_set_ch_char_multi(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum, u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	bool is_tx;
	u16 speed;
	u32 lspddev, epdir;
	u32 hccharn, hfnum;

	DBG(2, "<==\n");

	speed = port_get_speed(&(hcd_to_shost(hcd))->hub.port[0]);
	lspddev = ((speed==USB_OTG_SPEED_LS)? SCD_B_HCCHARN_LSPDDEV : 0x0);
	if (!epnum){
		is_tx = (SCD_HEP_CNTL_OUT == hep_cntl_get_type(ep));
	}
	else {
		is_tx = usb_endpoint_dir_out(&ep->hep->desc);
	}
	epdir = (is_tx? 0x0 : SCD_B_HCCHARN_EPDIR);

	hccharn = SCD__READ(sdev, HCCHAR(epnum));
	/* Clear bit to set */
	hccharn &= (~SCD_B_HCCHARN_CHDIS & 0xfff00000); /* HCCHARN init */
	/* rx packet next frame at interrupt transfer or isoc transfer */
	if((ep_type==USB_ENDPOINT_XFER_INT)||
	    (ep_type==USB_ENDPOINT_XFER_ISOC)){
		hfnum = (SCD__READ(sdev, HFNUM)) & SCD_M_HFNUM_FRNUM;
		if (hfnum % 2) {
			/* Set next odd frame */
			hccharn |= SCD_B_HCCHARN_ODDFRM;
		}
		else {
			/* Set next even frame */
			hccharn &= ~SCD_B_HCCHARN_ODDFRM;
		}
	}

	hccharn &= ~(SCD_M_HCCHARN_DEVADDR << SCD_S_HCCHARN_DEVADDR);
	hccharn |= ((ep->urb->dev->state > USB_STATE_DEFAULT ? ep->urb->dev->devnum : 0)
		    << SCD_S_HCCHARN_DEVADDR) |
							/* device address */
		(0x01 << SCD_S_HCCHARN_MCEC) |		/* MC set 1 */
		(ep_type << SCD_S_HCCHARN_EPTYPE) |	/* ep_type same as USB_ENDPOINT_XFER_? */
		lspddev |				/* LS device? */
		epdir |					/* ep dir */
		(epnum << SCD_S_HCCHARN_EPNUM) |	/* ep number */
		ep->hep->desc.wMaxPacketSize ;		/* max packet size */
	SCD__WRITE(sdev, HCCHAR(epnum), hccharn);

	DBG(2,"==>\n");
}

static void hep_rx_set_ch_char(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum, u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	bool is_tx;
	u16 speed;
	u32 lspddev, epdir;
	u32 hccharn, hfnum;

	DBG(2, "<==\n");

	speed = port_get_speed(&(hcd_to_shost(hcd))->hub.port[0]);
	lspddev = ((speed==USB_OTG_SPEED_LS)? SCD_B_HCCHARN_LSPDDEV : 0x0);
	if (!epnum){
		is_tx = (SCD_HEP_CNTL_OUT == hep_cntl_get_type(ep));
	}
	else {
		is_tx = usb_endpoint_dir_out(&ep->hep->desc);
	}
	epdir = (is_tx? 0x0 : SCD_B_HCCHARN_EPDIR);

	hccharn = SCD__READ(sdev, HCCHAR(epnum));
	/* Clear bit to set */
	hccharn &= (~SCD_B_HCCHARN_CHDIS & 0xfff00000); /* HCCHARN init */
	/* rx packet next frame at interrupt transfer or isoc transfer */
	if((ep_type==USB_ENDPOINT_XFER_INT)||
	    (ep_type==USB_ENDPOINT_XFER_ISOC)){
		hfnum = (SCD__READ(sdev, HFNUM)) & SCD_M_HFNUM_FRNUM;
		if (hfnum % 2) {
			/* Set next odd frame */
			hccharn |= SCD_B_HCCHARN_ODDFRM;
		}
		else {
			/* Set next even frame */
			hccharn &= ~(SCD_B_HCCHARN_ODDFRM);
		}
	}

	hccharn &= ~(SCD_M_HCCHARN_DEVADDR << SCD_S_HCCHARN_DEVADDR);
	hccharn |= ((ep->urb->dev->state > USB_STATE_DEFAULT ? ep->urb->dev->devnum : 0)
		    << SCD_S_HCCHARN_DEVADDR) |
							/* device address */
		(0x01 << SCD_S_HCCHARN_MCEC) |		/* MC set 1 */
		(ep_type << SCD_S_HCCHARN_EPTYPE) |	/* ep_type same as USB_ENDPOINT_XFER_? */
		lspddev |				/* LS device? */
		epdir |					/* ep dir */
		(epnum << SCD_S_HCCHARN_EPNUM) |	/* ep number */
		ep->hep->desc.wMaxPacketSize ;		/* max packet size */
	SCD__WRITE(sdev, HCCHAR(epnum), hccharn);

	DBG(2,"==>\n");
}

#ifdef SLAVE_MODE_DEBUG
static void hep_rx_set_ch_char_short(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum, u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	bool is_tx;
	u16 speed;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u32 lspddev, epdir;
	u32 hccharn, hfnum;
	u32 remain;

	DBG(2, "<==\n");

	speed = port_get_speed(&(hcd_to_shost(hcd))->hub.port[0]);
	lspddev = ((speed==USB_OTG_SPEED_LS)? SCD_B_HCCHARN_LSPDDEV : 0x0);
	if (!epnum){
		is_tx = (SCD_HEP_CNTL_OUT == hep_cntl_get_type(ep));
	}
	else {
		is_tx = usb_endpoint_dir_out(&ep->hep->desc);
	}
	epdir = (is_tx? 0x0 : SCD_B_HCCHARN_EPDIR);
	/* Setting remain length for maxp over packet */
	remain = hep_get_remain(ep->urb);
	maxp = (maxp < remain ? maxp : remain);

	hccharn = SCD__READ(sdev, HCCHAR(epnum));
	/* Clear bit to set */
	hccharn &= (~SCD_B_HCCHARN_CHDIS & 0xfff00000); /* HCCHARN init */
	/* rx packet next frame at interrupt transfer or isoc transfer */
	if((ep_type==USB_ENDPOINT_XFER_INT)||
	    (ep_type==USB_ENDPOINT_XFER_ISOC)){
		hfnum = (SCD__READ(sdev, HFNUM)) & SCD_M_HFNUM_FRNUM;
		if (hfnum % 2) {
			/* Set next odd frame */
			hccharn |= SCD_B_HCCHARN_ODDFRM;
		}
		else {
			/* Set next even frame */
			hccharn &= ~SCD_B_HCCHARN_ODDFRM;
		}
	}

	hccharn &= ~(SCD_M_HCCHARN_DEVADDR << SCD_S_HCCHARN_DEVADDR);
	hccharn |= ((ep->urb->dev->state > USB_STATE_DEFAULT ? ep->urb->dev->devnum : 0)
		    << SCD_S_HCCHARN_DEVADDR) |
							/* device address */
		(0x01 << SCD_S_HCCHARN_MCEC) |		/* MC set 1 */
		(ep_type << SCD_S_HCCHARN_EPTYPE) |	/* ep_type same as USB_ENDPOINT_XFER_? */
		lspddev |				/* LS device? */
		epdir |					/* ep dir */
		(epnum << SCD_S_HCCHARN_EPNUM) |	/* ep number */
		maxp ;					/* max packet size */

	SCD__WRITE(sdev, HCCHAR(epnum), hccharn);

	DBG(2,"==>\n");
}
#endif /* SLAVE_MODE_DEBUG */

static void hep_cntl_rx_ch_char(struct usb_hcd *hcd, struct scd_hep *ep, u8 epnum, u8 ep_type)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	bool is_tx;
	u16 speed;
	u32 lspddev, epdir;
	u32 hccharn, hfnum;

	DBG(2, "<==\n");

	speed = port_get_speed(&(hcd_to_shost(hcd))->hub.port[0]);
	lspddev = ((speed==USB_OTG_SPEED_LS)? SCD_B_HCCHARN_LSPDDEV : 0x0);
	is_tx = false;
	epdir = (is_tx? 0x0 : SCD_B_HCCHARN_EPDIR);

	hccharn = SCD__READ(sdev, HCCHAR(epnum));
	/* Clear bit to set */
	hccharn &= (~(SCD_B_HCCHARN_CHDIS) & 0xfff00000); /* HCCHARN init */
	/* rx packet next frame at interrupt transfer or isoc transfer */
	if((ep_type==USB_ENDPOINT_XFER_INT)||
	    (ep_type==USB_ENDPOINT_XFER_ISOC)){
		hfnum = (SCD__READ(sdev, HFNUM)) & SCD_M_HFNUM_FRNUM;
		if (hfnum % 2) {
			/* Set next odd frame */
			hccharn |= SCD_B_HCCHARN_ODDFRM;
		}
		else {
			/* Set next even frame */
			hccharn &= ~(SCD_B_HCCHARN_ODDFRM);
		}
	}

	hccharn &= ~(SCD_M_HCCHARN_DEVADDR << SCD_S_HCCHARN_DEVADDR);
	hccharn |= ((ep->urb->dev->state > USB_STATE_DEFAULT ? ep->urb->dev->devnum : 0)
		    << SCD_S_HCCHARN_DEVADDR) |
							/* device address */
		(0x01 << SCD_S_HCCHARN_MCEC) |		/* MC set 1 */
		(ep_type << SCD_S_HCCHARN_EPTYPE) |	/* ep_type same as USB_ENDPOINT_XFER_? */
		lspddev |				/* LS device? */
		epdir |					/* ep dir */
		(epnum << SCD_S_HCCHARN_EPNUM) |	/* ep number */
		ep->hep->desc.wMaxPacketSize ;		/* max packet size */
	SCD__WRITE(sdev, HCCHAR(epnum), hccharn);

	DBG(2,"==>\n");
}

static int hep_rx_data_dma_multi(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_rx_set_ch_size_multi(hcd, ep, epnum, ep_type);
	hep_rx_set_dma_addr(hcd, ep, epnum);
	hep_rx_set_ch_char_multi(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}


static int hep_rx_data_dma(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_rx_set_ch_size(hcd, ep, epnum, ep_type);
	hep_rx_set_dma_addr(hcd, ep, epnum);
	hep_rx_set_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}

static int hep_rx_data_dma_short(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_rx_set_ch_size(hcd, ep, epnum, ep_type);
	hep_rx_set_dma_addr_short(hcd, ep, epnum);
	hep_rx_set_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}

static int hep_rx_data_cpu(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_rx_set_ch_size(hcd, ep, epnum, ep_type);
	hep_rx_set_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}

static u32 hep_rx_check_error(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 epnum;
	u32 error = 0;
	u32 hcint;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);

	hcint = SCD__READ(sdev, HCINT(epnum));

	error = 0;
	if (hcint  & SCD_B_HCINTN_FRMOVRUN) {
		error |= SCD_HEP_E_OVERFLOW;
	}
	if (hcint  & SCD_B_HCINTN_BBLERR) {
		error |= SCD_HEP_E_MAXPOVER;
	}
	if (hcint  & SCD_B_HCINTN_XACTERR) {
		error |= SCD_HEP_E_XACTERR;
	}
	if (hcint  & SCD_B_HCINTN_STALL) {
		error |= SCD_HEP_E_RXSTALL;
	}
	if (hcint  & SCD_B_HCINTN_AHBERR) {
		error |= SCD_HEP_E_AHBERR;
	}

	DBG(2,"==>\n");
	return error;
}

static bool hep_rx_update_actual_multi(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 epnum;
	u32 remain, pktcnt, xfersize, actual;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u32 hctsiz;

	DBG(2, "<==\n");

	/* If HCTSIZn was readable, using it is better. */
	epnum = hep_hep2epnum(ep->hep);
	remain = hep_get_remain(urb);
	/* except short packet */
	pktcnt = remain / maxp;
	/* Care only PKTCNT, because PKTCNT*512 < XFERSIZE. */
	if(SCD_MAX_PKTCNT < pktcnt) {
		pktcnt =  SCD_MAX_PKTCNT;
	}	
	xfersize = pktcnt * maxp; 
	hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
	actual = xfersize - (hctsiz & SCD_M_HCTSIZN_XFERSIZE);
	hep_add_actual(urb, actual);
	remain = hep_get_remain(urb);

	DBG(2,"==>\n");

	return (remain && !(urb->actual_length % maxp)) ? true : false;
}

static bool hep_rx_update_actual(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 epnum;
	u32 remain, pktcnt, xfersize;
	u32 actual_last_packet;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u32 hctsiz;

	DBG(2, "<==\n");

	/* If HCTSIZn was readable, using it is better. */
	epnum = hep_hep2epnum(ep->hep);
	remain = hep_get_remain(urb);
	pktcnt = (0 <= remain ? 1 : 0 );
	/* Setting Xfersize of IN Transfer in maxp*n, not bytes. */
	xfersize = pktcnt * maxp; 
	hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
	actual_last_packet = xfersize - (hctsiz & SCD_M_HCTSIZN_XFERSIZE);
	hep_add_actual(urb, actual_last_packet);
	remain = hep_get_remain(urb);

	DBG(2,"==>\n");

	return (remain && !(urb->actual_length % maxp)) ? true : false;
}

static int hep_rx_check_len(struct usb_hcd *hcd, struct scd_hep *ep,
			    u32 *error)
{
	struct urb *urb = ep->urb;
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	char* buf;
	u32 expected;
	u32 remain, pktcnt, xfersize;
	u32 hctsiz;
	u16 rxcount;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u8 epnum;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	hep_get_buf(urb, &ep->hep->desc, ep->dmach, &buf, &expected);
	remain = hep_get_remain(urb);
	pktcnt = (0 <= remain ? 1 : 0 );
	/* Setting Xfersize of IN Transfer in maxp*n, not bytes. */
	xfersize = pktcnt * maxp; 
	hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
	rxcount = xfersize - (hctsiz & SCD_M_HCTSIZN_XFERSIZE);
	*error |= hep_check_rx_len_error(maxp, expected, rxcount,
					 urb->transfer_flags & URB_SHORT_NOT_OK);

	DBG(2,"==>\n");
	return rxcount;
}

#ifdef SLAVE_MODE_DEBUG
/* takeing overdata out from FIFO */
inline static void hep_rx_clear_fifo(struct usb_hcd *hcd, u8 index)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	void *regs = mdev_link_reg(sdev);
	u16 rxcount;
	u8 offset = SCD_FIFO_OFFSET(index);

	rxcount = SCD_READ_CSR16(regs, SCD_O_HDRC_RXCOUNT, index);
	while (rxcount >= 4) {
		SCD_READ32(regs, offset);
		rxcount -= 4;
	}
	while (rxcount > 0) {
		SCD_READ8(regs, offset);
		rxcount--;
	}
}
#endif /* SLAVE_MODE_DEBUG */

static void hep_rx_dma_sync_start(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	struct scd_hcd *shcd = (hcd_to_shost(hcd))->shcd;

	DBG(2, "<==\n");

	if (!urb->transfer_buffer) {
		urb->transfer_buffer = ioremap(urb->transfer_dma,
					       urb->transfer_buffer_length);
		if (!urb->transfer_buffer) {
			return;
		}
		ep->attributes |= SCD_HEP_ATR_IOREMAP;
	}

	if (urb->transfer_dma) {
		dma_sync_single_for_cpu(shcd->dev, urb->transfer_dma,
					urb->transfer_buffer_length,
					DMA_TO_DEVICE);
	}

	DBG(2,"==>\n");
}

static void hep_rx_dma_sync_end(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	struct scd_hcd *shcd = (hcd_to_shost(hcd))->shcd;

	DBG(2, "<==\n");

	if (urb->transfer_dma) {
		dma_sync_single_for_device(shcd->dev, urb->transfer_dma,
					   urb->transfer_buffer_length,
					   DMA_TO_DEVICE);
	}

	if (ep->attributes & SCD_HEP_ATR_IOREMAP) {
		iounmap(urb->transfer_buffer);
		urb->transfer_buffer = NULL;
		ep->attributes &= ~SCD_HEP_ATR_IOREMAP;
	}

	DBG(2,"==>\n");
}

static bool hep_rx_get_data(struct usb_hcd *hcd, struct scd_hep *ep,
			    u32 *error)
{
	struct urb *urb= ep->urb;
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	char* buf;
	u8 epnum;
	u8 ep_type;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u16 rxcount;
	u32 expected;
	u32 len;
	u32 gintmsk, grxstsp, pktsts;
	bool remain = false;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	/* Masking RxFlvl */
	gintmsk = SCD__READ(sdev, GINTMSK);
	gintmsk &= ~(SCD_B_GINTMSK_RXFLVLMSK);
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	hep_get_buf(urb, &ep->hep->desc, NULL, &buf, &expected);
	if (!urb->transfer_buffer) {
		*error |= SCD_HEP_E_NOBUFFER;
		return false;
	}
	grxstsp = SCD__READ(sdev, GRXSTSP);
	pktsts = (grxstsp >> SCD_S_GRXSTS_PKTSTS) & SCD_M_GRXSTS_PKTSTS;
	if (pktsts==0x0010) {  /* IN data packet received */
		rxcount = (grxstsp >> SCD_B_GRXSTS_BCNT) & SCD_M_GRXSTS_BCNT;
		if(rxcount > 0){
			len = min(expected, (u32)rxcount);
			hep_add_actual(urb, rxcount);
			scd_device_host_unload_fifo(hcd, ep_type, epnum, len, (u32 *)buf);
			/* If a rest appears, 
			   only one EP should clear. (not flush) */
			//scd_hep_rx_flush_fifo(hcd, ep);
			remain = (urb->transfer_buffer_length > urb->actual_length) ? true : false;
			*error |= hep_check_rx_len_error(maxp, expected, rxcount,
				urb->transfer_flags & URB_SHORT_NOT_OK);
		}
	}
	/* UnMasking RxFlvl */
	gintmsk |= SCD_B_GINTMSK_RXFLVLMSK;
	SCD__WRITE(sdev, GINTMSK, gintmsk);


	DBG(2,"==>\n");
	return remain;
}

static bool hep_rx_get_data_from_buffer(struct usb_hcd *hcd, struct scd_hep *ep,
			    u32 *error)
{
	struct urb *urb= ep->urb;
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	char* buf;
	u8 epnum;
	u8 ep_type;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u16 rxcount;
	u32 expected;
	u32 len;
	u32 hctsiz, xfersize;
	bool remain;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_get_buf(urb, &ep->hep->desc, NULL, &buf, &expected);
	if (!urb->transfer_buffer) {
		*error |= SCD_HEP_E_NOBUFFER;
		return false;
	}

	xfersize = maxp; 
	hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
	rxcount = xfersize - (hctsiz & SCD_M_HCTSIZN_XFERSIZE);
	if(rxcount > 0){
			len = min(expected, (u32)rxcount);
		if( !(urb->transfer_flags & URB_NO_TRANSFER_DMA_MAP) ){
			hep_rx_dma_sync_start(hcd, ep);
		}
		memcpy(buf, ep->packet_buffer, len);
		if( !(urb->transfer_flags & URB_NO_TRANSFER_DMA_MAP) ){
			hep_rx_dma_sync_end(hcd, ep);
		}
		/* If a rest appears, 
		   only one EP should clear. (not flush) */
		//scd_hep_rx_flush_fifo(hcd, ep);
	}
	remain = (urb->transfer_buffer_length > urb->actual_length) ? true : false;
	*error |= hep_check_rx_len_error(maxp, expected, rxcount,
		urb->transfer_flags & URB_SHORT_NOT_OK);

	DBG(2,"==>\n");
	return remain;
}

#ifdef SLAVE_MODE_DEBUG
static void hep_rx_alloc_buffer(struct scd_hep *ep)
{
	u16 maxp = ep->hep->desc.wMaxPacketSize;

	DBG(2, "<==\n");

	ep->packet_buffer = kmalloc(maxp, GFP_KERNEL);

	DBG(2,"==>\n");
}

static void hep_rx_free_buffer(struct scd_hep *ep)
{
	DBG(2, "<==\n");

	kfree(ep->packet_buffer);
	ep->packet_buffer = NULL;

	DBG(2,"==>\n");
}

static void hep_rx_dma_map_single(struct scd_hep *ep)
{
	u16 maxp = ep->hep->desc.wMaxPacketSize;

	ep->packet_dma = (dma_addr_t)dma_map_single(&ep->dev->dev, 
	    (void *)ep->packet_buffer, maxp, DMA_BIDIRECTIONAL);
}

static void hep_rx_dma_unmap_single(struct scd_hep *ep)
{
	u16 maxp = ep->hep->desc.wMaxPacketSize;

	dma_unmap_single(&ep->dev->dev, (dma_addr_t)ep->packet_dma, 
	    maxp, DMA_BIDIRECTIONAL);
}
#endif /* SLAVE_MODE_DEBUG */

static bool hep_rx_get_status(struct usb_hcd *hcd, struct scd_hep *ep,
			    u32 *error)
{
	return hep_rx_get_data(hcd, ep, error);
}

static bool hep_rx_check_remain(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u32 nextsz;

	DBG(2, "<==\n");

	hep_rx_update_actual(hcd,ep);
	nextsz = hep_get_remain(urb);

	DBG(2,"==>\n");
	return (nextsz && !(urb->actual_length % maxp)) ? true : false;

}

static bool hep_rx_check_short(struct scd_hep *ep)
{
	struct urb *urb = ep->urb;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u32 nextsz;

	DBG(2, "<==\n");

	nextsz = hep_get_remain(urb);

	DBG(2,"==>\n");
	return ( nextsz && (nextsz < maxp) && !(urb->actual_length % maxp)) ? 
		true : false;
}

static void hep_rx_set_last_toggle(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 epnum;
	u32 hctsiz;
	bool toggle = false;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
	hctsiz &= (SCD_M_HCTSIZN_PID << SCD_S_HCTSIZN_PID);
	if(hctsiz == PID_DATA1){
		toggle = true;
	}
	else if(toggle == PID_DATA0){
		toggle = false;
	}
	else {
		/* error */
	}
	usb_settoggle(ep->dev, epnum, false, toggle);

	DBG(2,"==>\n");
}

static void hep_rx_set_new_toggle(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 epnum;
	u32 hctsiz;
	bool toggle;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	toggle = usb_gettoggle(ep->dev, epnum, false);

	hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
	if (toggle) {
		hctsiz &= ~(SCD_M_HCTSIZN_PID << SCD_S_HCTSIZN_PID);
		hctsiz |= (PID_DATA1);
	}
	else {
		hctsiz &= ~(SCD_M_HCTSIZN_PID << SCD_S_HCTSIZN_PID);
		hctsiz |= (PID_DATA0);
	}
	SCD__WRITE(sdev, HCTSIZ(epnum), hctsiz);

	DBG(2,"==>\n");
}

static int hep_rx_trans_state(struct usb_hcd *hcd, struct scd_hep *ep,
			      u32 trigger)
{
	u8 epnum;
	bool reset = trigger == SCD_HEP_T_RESET ? true : false;
	bool remain = false;
	bool shortp = false;
	u8 ep_type = 0;
	u32 rxcount = 0;
	u32 error = 0;

	DBG(2, "<==\n");
	DBG(3, "ep->state (exit) = %lu\n", (unsigned long)ep->state);

	epnum = hep_hep2epnum(ep->hep);

	/* exit state */
	switch (ep->state) {
	case SCD_HEP_S_RX_IDLE:
		if ((trigger == SCD_HEP_T_LINK_CHHLT_INTR) ||
		    (trigger == SCD_HEP_T_LINK_ERROR_INTR)) {
			/* Nothing to do */
			return 0;
		}
		hep_rx_set_new_toggle(hcd, ep);
		ep->retrycount = 0;
		if (!reset) {
#ifdef SLAVE_MODE_DEBUG
			if (!sdev->dmac) {
				hep_rx_dma_sync_start(hcd, ep);
			}
#endif /* SLAVE_MODE_DEBUG */
			shortp = hep_rx_check_short(ep);
			ep_type = hep_check_ep_type(&ep->hep->desc);
		}
		break;
	case SCD_HEP_S_RX_WAIT_DMA_MULTI:
		if (!reset) {
			error = hep_rx_check_error(hcd, ep);
			if (!error) {
				remain = hep_rx_update_actual_multi(hcd,ep);
				shortp = hep_rx_check_short(ep);
			}
		}
		break;
	case SCD_HEP_S_RX_WAIT_DMA:
		if (!reset) {
			error = hep_rx_check_error(hcd, ep);
			if (!error) {
				remain = hep_rx_check_remain(hcd, ep);
				shortp = hep_rx_check_short(ep);
			}
		}
		break;
	case SCD_HEP_S_RX_WAIT_DMA_SHORTP:
		if (!reset) {
			error = hep_rx_check_error(hcd, ep);
			if (!error) {
				rxcount = hep_rx_check_len(hcd, ep, &error);
				remain = hep_rx_get_data_from_buffer(hcd, ep, &error);
				hep_rx_update_actual(hcd,ep);
			}
		}
		break;
	case SCD_HEP_S_RX_WAIT_CPU:
		if (!reset) {
			error = hep_rx_check_error(hcd, ep);
			if (!error) {
				rxcount = hep_rx_check_len(hcd, ep, &error);
				remain = hep_rx_get_data(hcd, ep, &error);
			}
		}
		if (reset || error || !remain) {
			hep_rx_dma_sync_end(hcd, ep);
		}
		break;
	case SCD_HEP_S_RX_WAIT_CPU_HALT:
		if (!reset) {
			error = hep_rx_check_error(hcd, ep);
			if (!error) {
				if (trigger == SCD_HEP_T_LINK_RXFLVL_INTR) {
					rxcount = hep_rx_check_len(hcd, ep, &error);
					remain = hep_rx_get_data(hcd, ep, &error);
				}
				else if (trigger == SCD_HEP_T_LINK_XFERCOMPL_INTR) {
					hep_rx_disable(hcd, ep);
				}
			}
		}
		break;
	}

	/* transition state */
	if (trigger == SCD_HEP_T_RESET ||
	    trigger == SCD_HEP_T_LINK_ERROR_INTR ||
	    ((trigger == SCD_HEP_T_LINK_CHHLT_INTR ||
	      trigger == SCD_HEP_T_LINK_RXFLVL_INTR)&& error)) {
		ep->state = SCD_HEP_S_RX_IDLE;
		goto enter_state;
	}
	switch (ep->state) {
	case SCD_HEP_S_RX_IDLE:
		if (trigger == SCD_HEP_T_START) {
#ifdef SLAVE_MODE_DEBUG
			if (sdev->dmac) {
#endif /* SLAVE_MODE_DEBUG */
				if(ep_type == USB_ENDPOINT_XFER_BULK) {
					if ((!shortp) && ep->urb->transfer_buffer_length) {
						ep->state = SCD_HEP_S_RX_WAIT_DMA_MULTI;
						goto enter_state;
					}
					else {
						ep->state = SCD_HEP_S_RX_WAIT_DMA_SHORTP;
						goto enter_state;
					}
				}
				else if(ep_type == USB_ENDPOINT_XFER_INT) {
					if ((!shortp) && ep->urb->transfer_buffer_length) {
						ep->state = SCD_HEP_S_RX_WAIT_DMA;
						goto enter_state;
					}
					else {
						ep->state = SCD_HEP_S_RX_WAIT_DMA_SHORTP;
						goto enter_state;
					}
				}
#ifdef SLAVE_MODE_DEBUG
			}
			else {
				ep->state = SCD_HEP_S_RX_WAIT_CPU;
				goto enter_state;
			}
#endif /* SLAVE_MODE_DEBUG */
		}
		break;
	case SCD_HEP_S_RX_WAIT_DMA_MULTI:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				if(remain) {
					if(!shortp) {
						/* Reenter same state */
						goto enter_state;
					}
					else {
						ep->state = 
						  SCD_HEP_S_RX_WAIT_DMA_SHORTP;
						goto enter_state;
					}
				}
				else {
					ep->state = SCD_HEP_S_RX_IDLE;
					goto enter_state;
				}
			}
		}
		break;
	case SCD_HEP_S_RX_WAIT_DMA:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				if (remain) {
					if (!shortp) {
						/* Reenter same state */
						goto enter_state;
					}
					else {
						ep->state = 
						  SCD_HEP_S_RX_WAIT_DMA_SHORTP;
						goto enter_state;
					}
				}
				else {
					ep->state = SCD_HEP_S_RX_IDLE;
					goto enter_state;
				}
			}
		}
		break;
	case SCD_HEP_S_RX_WAIT_DMA_SHORTP:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				ep->state = SCD_HEP_S_RX_IDLE;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_RX_WAIT_CPU:
		if (trigger == SCD_HEP_T_LINK_RXFLVL_INTR) {
			if (!error) {
				if (remain) {
					/* Do not change state */
					goto enter_state;
				}
				else {
					ep->state = SCD_HEP_S_RX_WAIT_CPU_HALT;
					goto enter_state;
				}
			}
		}
		break;
	case SCD_HEP_S_RX_WAIT_CPU_HALT:
		if (trigger == SCD_HEP_T_LINK_RXFLVL_INTR) {
			if (!error) {
				/* Do not change state */
			}
		}
		else if (trigger == SCD_HEP_T_LINK_XFERCOMPL_INTR) {
			if (!error) {
				ep->state = SCD_HEP_S_RX_WAIT_CPU_HALT;
			}
		}
		else if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				ep->state = SCD_HEP_S_RX_IDLE;
			}
		}
		break;
	}
	goto error_end;


	/* enter state */
enter_state:
	switch (ep->state) {
	case SCD_HEP_S_RX_IDLE:
		hep_rx_disable(hcd, ep);
		hep_rx_set_last_toggle(hcd, ep);
		if (!reset) {
			if (!error) {
				hep_complete(hcd, ep, ep->urb, 0);
				hep_rx_self_start(hcd, ep);
			}
			else {
				hep_error(hcd, ep, error);
			}
		}
		break;
	case SCD_HEP_S_RX_WAIT_DMA_MULTI:
		hep_rx_unmask_intr(hcd, ep);
		hep_rx_data_dma_multi(hcd, ep);
		hep_rx_enable(hcd, ep);
		break;
	case SCD_HEP_S_RX_WAIT_DMA:
		hep_rx_unmask_intr(hcd, ep);
		hep_rx_data_dma(hcd, ep);
		hep_rx_enable(hcd, ep);
		break;
	case SCD_HEP_S_RX_WAIT_DMA_SHORTP:
		hep_rx_unmask_intr(hcd, ep);
		hep_rx_data_dma_short(hcd, ep);
		hep_rx_enable(hcd, ep);
		break;
	case SCD_HEP_S_RX_WAIT_CPU:
		hep_rx_unmask_intr(hcd, ep);
		hep_rx_data_cpu(hcd, ep);
		hep_rx_enable(hcd, ep);
		break;
	case SCD_HEP_S_RX_WAIT_CPU_HALT:
		/* Noting to do */
		break;
	}


	DBG(3, "ep->state (enter) = %lu\n", (unsigned long)ep->state);
	DBG(2,"==>\n");
	return 0;

error_end:
	DBG(3, "ep->state (enter) = %lu\n", (unsigned long)ep->state);
	DBG(2,"==>\n");
	return -EINVAL;
}

/* global functions */

void scd_hep_clear_atr(struct scd_port *port, u8 epadr, u32 attribute)
{
	struct scd_hep *ep;

	DBG(2, "<==\n");

	ep = hep_search_ep(port, epadr);
	if (ep) {
		ep->attributes &= ~attribute;
	}

	DBG(2,"==>\n");
}

void scd_hep_set_atr(struct scd_port *port, u8 epadr, u32 attribute)
{
	struct scd_hep *ep;

	DBG(2, "<==\n");

	ep = hep_search_ep(port, epadr);
	if (ep) {
		ep->attributes |= attribute;
	}

	DBG(2,"==>\n");
}

void scd_hep_cntl_init(struct usb_hcd *hcd)
{
	/* Nothing to do */
}
EXPORT_SYMBOL(scd_hep_cntl_init);

void scd_hep_tx_init(struct usb_hcd *hcd)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	int i;
	scd_usec timein, timeout, now;
	u32 hcchar;

	DBG(2, "<==\n");

	/* Make all channels to disable, before flush  */
	for( i = 0; i < SCD_NUM_HEPS; i++ ){
		hep_disable_ch(hcd, i);
		/* Checking CHHLTD to complete disable */
		do {
			hcchar = SCD__READ(sdev, HCCHAR(i));  
		} while ( hcchar & SCD_B_HCCHARN_CHENA );
	}

	/* Flush FIFO */
	hcchar = SCD__READ(sdev, GRSTCTL );
	/* TXFNUM(10:6) = */
	/* 5h10: Flush all the transmit FIFOs in device or host mode */
	/* 5h0:  Non-periodic TxFIFO flush in Host mode */
	/* 5h1:  Periodic TxFIFO flush in Host mode */
	hcchar |= SCD_B_GRSTCTL_TXFFLSH | (0x10 << SCD_S_GRSTCTL_TXFNUM);
	SCD__WRITE(sdev, GRSTCTL, hcchar );
	timeout = SCD_FS_FLUSH_TIMEOUT;
	timein = scd_device_get_time(sdev);
	do {
		hcchar = SCD__READ(sdev, GRSTCTL );
	        now = scd_device_get_time(sdev);
	        if((now - timein) < 0){
			timein = now;  /* It takes double time at overflow */
	        }
	} while ((hcchar & SCD_B_GRSTCTL_TXFFLSH) && 
		 ((now - timein) < timeout) );
	udelay(1); /* Wait 8clk of 30MHz */

	DBG(2,"==>\n");
}
EXPORT_SYMBOL(scd_hep_tx_init);

void scd_hep_rx_init(struct usb_hcd *hcd)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	int i;
	scd_usec timein, timeout, now;
	u32 hcchar;

	DBG(2, "<==\n");

	/* Make all channels to disable, before flush  */
	for( i = 0; i < SCD_NUM_HEPS; i++ ){
		hep_disable_ch(hcd, i);
		/* Checking CHHLTD to complete disable */
		do {
			hcchar = SCD__READ(sdev, HCCHAR(i));  
		} while ( hcchar & SCD_B_HCCHARN_CHENA );
	}

	/* Flush FIFO */
	hcchar = SCD__READ(sdev, GRSTCTL );
	hcchar |= SCD_B_GRSTCTL_RXFFLSH ;
	SCD__WRITE(sdev, GRSTCTL, hcchar );
	timeout = SCD_FS_FLUSH_TIMEOUT;
	timein = scd_device_get_time(sdev);
	do {
		hcchar = SCD__READ(sdev, GRSTCTL );
	        now = scd_device_get_time(sdev);
	        if((now - timein) < 0){
			timein = now;  /* It takes double time at overflow */
	        }
	} while ((hcchar & SCD_B_GRSTCTL_RXFFLSH) && 
		 ((now - timein) < timeout) );
	udelay(1); /* Wait 8clk of 30MHz */

	DBG(2,"==>\n");
}
EXPORT_SYMBOL(scd_hep_rx_init);

void scd_hep_ch0_init(struct usb_hcd *hcd)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	int epnum = -1;
	u8 ep_type = USB_ENDPOINT_XFER_INT;
	u32 lspddev = 0x0;
	u32 epdir = 0x0;
	u32 hccharn;

	hccharn = SCD__READ(sdev, HCCHAR(epnum));
	/* Clear bit to set */
	hccharn &= (~SCD_B_HCCHARN_CHDIS & 0xfff00000); /* HCCHARN init */

	hccharn |= 
		(0x01 << SCD_S_HCCHARN_MCEC) |		/* MC set 1 */
		(ep_type << SCD_S_HCCHARN_EPTYPE) |	/* ep_type same as USB_ENDPOINT_XFER_? */
		lspddev |				/* LS device? */
		epdir |					/* ep dir */
		(0 << SCD_S_HCCHARN_EPNUM) |		/* ep number */
		1024 ;					/* max packet size */
	SCD__WRITE(sdev, HCCHAR(epnum), hccharn);
}
EXPORT_SYMBOL(scd_hep_ch0_init);

void scd_hep_tx_flush_fifo(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	scd_usec timein, timeout, now;
	u8 epnum;
	u16 speed;
	u32 hcchar;

	DBG(2, "<==\n");

	/* Clean tx FIFO of epnum */

	/* Checking CHHLTD to complete disable */
	epnum = hep_hep2epnum(ep->hep);
	hep_disable_ch(hcd, epnum);
	do {
		hcchar = SCD__READ(sdev, HCCHAR(epnum));  
	} while ( hcchar & SCD_B_HCCHARN_CHENA );

	/* Flush FIFO */
	hcchar = SCD__READ(sdev, GRSTCTL );

	switch(hep_check_ep_type(&ep->hep->desc)){
	case USB_ENDPOINT_XFER_CONTROL:
	case USB_ENDPOINT_XFER_BULK:
	    /* TXFNUM(10:6) = */
	    /* 5h10: Flush all the transmit FIFOs in device or host mode */
	    /* 5h0:  Non-periodic TxFIFO flush in Host mode */
	    /* 5h1:  Periodic TxFIFO flush in Host mode */
	    hcchar |= SCD_B_GRSTCTL_TXFFLSH | (0x0 << SCD_S_GRSTCTL_TXFNUM);
	    SCD__WRITE(sdev, GRSTCTL, hcchar );
	    break;
	case USB_ENDPOINT_XFER_INT:
	case USB_ENDPOINT_XFER_ISOC:
	    /* 5h1:  Periodic TxFIFO flush in Host mode */
	    hcchar |= SCD_B_GRSTCTL_TXFFLSH | (0x1 << SCD_S_GRSTCTL_TXFNUM);
	    SCD__WRITE(sdev, GRSTCTL, hcchar );
	    break;
	}
	speed = port_get_speed(&(hcd_to_shost(hcd))->hub.port[0]);
	timeout = ((speed==USB_OTG_SPEED_HS)? 
		SCD_HS_FLUSH_TIMEOUT : SCD_FS_FLUSH_TIMEOUT);
	timein = scd_device_get_time(sdev);
	do {
		hcchar = SCD__READ(sdev, GRSTCTL );
	        now = scd_device_get_time(sdev);
	        if((now - timein) < 0){
			timein = now;  /* It takes double time at overflow */
	        }
	} while ((hcchar & SCD_B_GRSTCTL_TXFFLSH) && 
		 ((now - timein) < timeout) );
	udelay(1); /* Wait 8clk of 30MHz */

	DBG(2,"==>\n");
}

void scd_hep_rx_flush_fifo(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	scd_usec timein, timeout, now;
	u8 epnum;
	u16 speed;
	u32 hcchar;

	DBG(2, "<==\n");

	/* Clean rx FIFO of epnum */

	/* Checking CHHLTD to complete disable */
	epnum = hep_hep2epnum(ep->hep);
	hep_disable_ch(hcd, epnum);
	do {
		hcchar = SCD__READ(sdev, HCCHAR(epnum));  
	} while ( hcchar & SCD_B_HCCHARN_CHENA );

	/* Flush FIFO */
	hcchar = SCD__READ(sdev, GRSTCTL );
	hcchar |= SCD_B_GRSTCTL_RXFFLSH ;
	SCD__WRITE(sdev, GRSTCTL, hcchar );
	speed = port_get_speed(&(hcd_to_shost(hcd))->hub.port[0]);
	timeout = ((speed==USB_OTG_SPEED_HS)? 
		SCD_HS_FLUSH_TIMEOUT : SCD_FS_FLUSH_TIMEOUT);
	timein = scd_device_get_time(sdev);
	do {
		hcchar = SCD__READ(sdev, GRSTCTL );
	        now = scd_device_get_time(sdev);
	        if((now - timein) < 0){
			timein = now;  /* It takes double time at overflow */
	        }
	} while ((hcchar & SCD_B_GRSTCTL_RXFFLSH) && 
		 ((now - timein) < timeout) );
	udelay(1); /* Wait 8clk of 30MHz */

	DBG(2,"==>\n");
}

static void scd_hep_cntl_irq(struct usb_hcd *hcd, u8 epnum)
{
	struct scd_host *shost = hcd_to_shost(hcd);
	struct scd_device *sdev = shost->sdev;
	struct scd_hep *ep = &shost->ep0;
	u32 hcint, hctsiz;

	hcint = SCD__READ(sdev, HCINT(epnum));
	if(hcint & (SCD_B_HCINTN_FRMOVRUN |
	    SCD_B_HCINTN_AHBERR)) {
		hep_cntl_trans_state(hcd, ep, SCD_HEP_T_LINK_ERROR_INTR);
	}
	else if(hcint & SCD_B_HCINTN_STALL) {
		SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_CHHLTD);
		hep_cntl_trans_state(hcd, ep, SCD_HEP_T_LINK_ERROR_INTR);
	}
	else if((hcint & SCD_B_HCINTN_XACTERR) || 
	        (hcint & SCD_B_HCINTN_BBLERR))
	{
		SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_CHHLTD);
		hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
		SCD__WRITE(sdev, HCTSIZ(epnum), hctsiz & ~SCD_B_HCTSIZN_DOPNG);
		if(ep->retrycount < (SCD_RETRY_COUNT - 1))
		{
			SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_XACTERR);
			SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_BBLERR);
			if ( ep->state == SCD_HEP_S_CNTL_WAIT_TX_SETUP_DMA) {
				hep_cntl_set_setup_dma_addr(hcd, ep, epnum);
			}
			else if(
			  ep->state == SCD_HEP_S_CNTL_WAIT_TX_DATA_DMA_MULTI ||
			  ep->state == SCD_HEP_S_CNTL_WAIT_TX_DATA_DMA_ZEROP) {
				hep_tx_reset_dma_addr(hcd, ep, epnum);
			}
			hep_enable_ch(hcd, epnum);
			ep->retrycount++;
		}
		else
		{
			hep_cntl_trans_state(hcd, ep, SCD_HEP_T_LINK_ERROR_INTR);
			SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_XACTERR);
			SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_BBLERR);
			ep->retrycount = 0;
		}
	}
	else {
		hep_cntl_trans_state(hcd, ep, SCD_HEP_T_LINK_CHHLT_INTR);
	}
}

static void scd_hep_tx_irq(struct usb_hcd *hcd, u8 epnum)
{
	struct scd_host *shost = hcd_to_shost(hcd);
	struct scd_device *sdev = shost->sdev;
	struct scd_hep *ep = &shost->ep_out[epnum - 1];
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u32 hcint, hctsiz;
	u32 pktcnt, pktcnt_org, buflen_org;

	hcint = SCD__READ(sdev, HCINT(epnum));
	if(hcint & (SCD_B_HCINTN_FRMOVRUN |
	    SCD_B_HCINTN_AHBERR)) {
		hep_tx_trans_state(hcd, ep, SCD_HEP_T_LINK_ERROR_INTR);
	}
	else if(hcint & SCD_B_HCINTN_STALL) {
		SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_CHHLTD);
		hep_tx_trans_state(hcd, ep, SCD_HEP_T_LINK_ERROR_INTR);
	}
	else if((hcint & SCD_B_HCINTN_XACTERR) || 
	        (hcint & SCD_B_HCINTN_BBLERR))
	{
		SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_CHHLTD);
		hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
		if(!(hctsiz & SCD_B_HCTSIZN_DOPNG)) {
			pktcnt = (hctsiz >> SCD_S_HCTSIZN_PKTCNT) & SCD_M_HCTSIZN_PKTCNT;
			if(2 < pktcnt){
				hctsiz += maxp*2;
			}
			else if(1 < pktcnt){
				hctsiz += maxp;
			}
		}
		SCD__WRITE(sdev, HCTSIZ(epnum), hctsiz & ~SCD_B_HCTSIZN_DOPNG);
		/* Get original packet count */
		buflen_org = ep->urb->transfer_buffer_length;
		pktcnt_org = (buflen_org/maxp) + (((buflen_org % maxp) || (0 == buflen_org))? 1 : 0 );
		/* Care only PKTCNT, because PKTCNT*512 < XFERSIZE. */
		if(SCD_MAX_PKTCNT < pktcnt_org) {
			pktcnt_org =  SCD_MAX_PKTCNT;
		}
#ifdef RETRY_COUNT_DEBUG
		if(ep->retrycount < (pktcnt_org * (SCD_RETRY_COUNT - 1)))
#else
		if(1) /* Retry infinite time */
#endif /* RETRY_COUNT_DEBUG */
		{
			SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_XACTERR);
			SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_BBLERR);
			hep_tx_reset_dma_addr(hcd, ep, epnum);
			hep_enable_ch(hcd, epnum);
			ep->retrycount++;
		}
		else
		{
			DBG(1, "usb_scd XACTERR over RETRY_COUNT\n");
			hep_tx_trans_state(hcd, ep, SCD_HEP_T_LINK_ERROR_INTR);
			SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_XACTERR);
			SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_BBLERR);
			ep->retrycount = 0;
		}
	}
	else {
		hep_tx_trans_state(hcd, ep, SCD_HEP_T_LINK_CHHLT_INTR);
	}
}

static void scd_hep_rx_irq(struct usb_hcd *hcd, u8 epnum)
{
	struct scd_host *shost = hcd_to_shost(hcd);
	struct scd_device *sdev = shost->sdev;
	struct scd_hep *ep = &shost->ep_in[epnum - 1];
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u32 hcint, hctsiz;
	u32 pktcnt_org, buflen_org;

	hcint = SCD__READ(sdev, HCINT(epnum));
	if(hcint & (SCD_B_HCINTN_FRMOVRUN |
	    SCD_B_HCINTN_AHBERR)) {
		hep_rx_trans_state(hcd, ep, SCD_HEP_T_LINK_ERROR_INTR);
	}
	else if(hcint & SCD_B_HCINTN_STALL) {
		SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_CHHLTD);
		hep_rx_trans_state(hcd, ep, SCD_HEP_T_LINK_ERROR_INTR);
	}
	else if((hcint & SCD_B_HCINTN_XACTERR) || 
	        (hcint & SCD_B_HCINTN_BBLERR))
	{
		SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_CHHLTD);
		hctsiz = SCD__READ(sdev, HCTSIZ(epnum));
		SCD__WRITE(sdev, HCTSIZ(epnum), hctsiz & ~SCD_B_HCTSIZN_DOPNG);
		/* Get original packet count */
		buflen_org = ep->urb->transfer_buffer_length;
		pktcnt_org = (buflen_org/maxp) + (((buflen_org % maxp) || (0 == buflen_org))? 1 : 0 );
		/* Care only PKTCNT, because PKTCNT*512 < XFERSIZE. */
		if(SCD_MAX_PKTCNT < pktcnt_org) {
			pktcnt_org =  SCD_MAX_PKTCNT;
		}
#ifdef RETRY_COUNT_DEBUG
		if(ep->retrycount < (pktcnt_org * (SCD_RETRY_COUNT - 1)))
#else
		if(1) /* Retry infinite time */
#endif /* RETRY_COUNT_DEBUG */
		{
			SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_XACTERR);
			SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_BBLERR);
			hep_enable_ch(hcd, epnum);
			ep->retrycount++;
		}
		else
		{
			DBG(1, "usb_scd XACTERR over RETRY_COUNT\n");
			hep_rx_trans_state(hcd, ep, SCD_HEP_T_LINK_ERROR_INTR);
			SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_XACTERR);
			SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_BBLERR);
			ep->retrycount = 0;
		}
	}
	else {
		hep_rx_trans_state(hcd, ep, SCD_HEP_T_LINK_CHHLT_INTR);
	}
}

static void scd_hep_epnum_irq(struct usb_hcd *hcd, u8 epnum)
{
	struct scd_host *shost = hcd_to_shost(hcd);
	struct scd_device *sdev;
	bool is_out;
	u32 hccharn, epdir;

	DBG(2, "<==\n");

	sdev = shost->sdev;

	if(epnum == 0){
		scd_hep_cntl_irq(hcd, epnum);
	}
	else {
		hccharn = SCD__READ(sdev, HCCHAR(epnum));
		epdir = SCD_B_HCCHARN_EPDIR;
		is_out = ((hccharn & epdir) == 0);

		if (is_out) {
			scd_hep_tx_irq(hcd, epnum);
		}
		else {
			scd_hep_rx_irq(hcd, epnum);
		}
	}

	DBG(2,"==>\n");
}

irqreturn_t scd_hep_irq(struct usb_hcd *hcd)
{
	struct scd_host *shost = hcd_to_shost(hcd);
	struct scd_device *sdev;
	u8 epnum;
	u32 haint, hcint;
	irqreturn_t ret;
	
	DBG(2, "<==\n");

	ret = IRQ_NONE;
	sdev = shost->sdev;

	/* Call side did spin_lock() */

	haint = SCD__READ(sdev, HAINT);
	for (epnum = 0; epnum < SCD_NUM_HEPS; epnum++){
		if((haint & SCD_M_HAINT_HAINT) & (SCD_F_HAINT_CH(epnum))){
			DBG(3, "haint = 0x%lx\n", (long)haint);
			hcint = SCD__READ(sdev, HCINT(epnum));
			if (hcint & SCD_B_HCINTN_DATATGLERR) {
				DBG(3, "HCINTN_DATATGLERR\n");
				SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_CHHLTD);
				SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_DATATGLERR);
				/* Nothing to do */

				ret = IRQ_HANDLED;
				goto irq_end;
			}
			else if (hcint & SCD_B_HCINTN_FRMOVRUN) {
				DBG(3, "HCINTN_FRMOVRUN\n");
				scd_hep_epnum_irq(hcd, epnum);
				SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_FRMOVRUN);

				ret = IRQ_HANDLED;
				goto irq_end;
			}
			else if (hcint & SCD_B_HCINTN_BBLERR) {
				DBG(3, "HCINTN_BBLERR\n");
				scd_hep_epnum_irq(hcd, epnum);
				/* Clear bit at func of upper line */

				ret = IRQ_HANDLED;
				goto irq_end;
			}
			else if (hcint & SCD_B_HCINTN_XACTERR) {
				DBG(3, "HCINTN_XACTERR\n");
				scd_hep_epnum_irq(hcd, epnum);
				/* Clear bit at func of upper line */

				ret = IRQ_HANDLED;
				goto irq_end;
			}
			else if (hcint & SCD_B_HCINTN_STALL) {
				DBG(3, "HCINTN_STALL\n");
				scd_hep_epnum_irq(hcd, epnum);
				SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_STALL);

				ret = IRQ_HANDLED;
				goto irq_end;
			}
			else if (hcint & SCD_B_HCINTN_AHBERR) {
				DBG(3, "HCINTN_AHBERR\n");
				scd_hep_epnum_irq(hcd, epnum);
				SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_AHBERR);

				ret = IRQ_HANDLED;
				goto irq_end;
			}
			else if (hcint & SCD_B_HCINTN_CHHLTD) {
				DBG(3, "HCINTN_CHHLTD\n");
				/* Clear bit before func */
				SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_CHHLTD);
				scd_hep_epnum_irq(hcd, epnum);

				ret = IRQ_HANDLED;
				goto irq_end;
			}
			else if (hcint & SCD_B_HCINTN_XFERCOMPL) {
				DBG(3, "HCINTN_XFERCOMPL\n");
				SCD__WRITE(sdev, HCINT(epnum), SCD_B_HCINTN_XFERCOMPL);
				/* Nothing to do */

				ret = IRQ_HANDLED;
				goto irq_end;
			}
		}
	}
irq_end:

	DBG(2,"==>\n");
	return ret;
}

/* operation functions */

int scd_hep_urb_enqueue(struct usb_hcd *hcd, 
			struct urb *urb, gfp_t mem_flags)
{
	struct scd_host *shost;
	struct scd_hcd *shcd;
	struct usb_device *udev;
	struct scd_hep *ep_table;
	struct scd_hep *ep;
	u8 *nep;
	u8 epnum = hep_hep2epnum(urb->ep);
	unsigned long flags;
	unsigned int pipe = urb->pipe;
	int status = 0;
	int is_out = !usb_pipein(pipe);
	int type = usb_pipetype(pipe);

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	shcd = (hcd_to_shost(hcd))->shcd;
	udev = urb->dev;

	spin_lock_irqsave(&shost->lock, flags);

	if (unlikely(SCD_NUM_HEPS_HCD <= epnum)){
		status = -ENOENT;
		goto done;
	}
	status = usb_hcd_link_urb_to_ep(hcd, urb);
	if (unlikely(status)) {
		goto done;
	}

	if (type == PIPE_CONTROL) {
		ep = &shost->ep0;

		ep->status_dma = (dma_addr_t)dma_map_single(&ep->dev->dev, 
		  (void *)ep->status_packet, STATUS_PKT_MAXP, DMA_BIDIRECTIONAL);
	}
	else {
		ep_table = is_out ? &shost->ep_out[0] : &shost->ep_in[0];
		nep = is_out ? &shost->nep_out : &shost->nep_in;
		ep = hep_inquiry_ep(hcd, udev, urb->ep, ep_table, nep);
		if (!ep) {
			dev_warn(shcd->dev, "no pipe resource"
				 "(devnum%d,epadr:%.2x)", udev->devnum,
				 urb->ep->desc.bEndpointAddress);
			status = -ENOMEM;
			goto error_end;
		}
		if (type != PIPE_BULK) {
			status = hep_assign_bandwidth(udev, ep);
			if (status < 0) {
				dev_warn(shcd->dev,
					 "no bandwdith resource"
					 "(devnum:%d,epadr:%.2x,sts=%d)",
					 udev->devnum,
					 urb->ep->desc.bEndpointAddress, status);
				goto error_end;
			}
		}
	}

	if (ep->urb != NULL) {
		status = 0;
		urb->hcpriv = (void *)ep;
		goto done;
	}

	ep->urb = urb;
	ep->hep = urb->ep;
	ep->dev = udev;
	urb->hcpriv = (void *)ep;

	/* hep_enqueue */
	switch (type) {
	case PIPE_CONTROL:
		status = hep_cntl_trans_state(hcd, ep, SCD_HEP_T_START);
		goto done;
	case PIPE_BULK:
	case PIPE_INTERRUPT:
		if (is_out) {
			status = hep_tx_trans_state(hcd, ep, SCD_HEP_T_START);
		}
		else {
			status = hep_rx_trans_state(hcd, ep, SCD_HEP_T_START);
		}
		goto done;
	default:
		status = -ENOSPC;
		goto error_end;
	}

error_end:
	usb_hcd_unlink_urb_from_ep(hcd, urb);
done:
	spin_unlock_irqrestore(&shost->lock, flags);

	DBG(2,"==>\n");
	return status;
}

int scd_hep_urb_dequeue(struct usb_hcd *hcd, struct urb *urb, int status)
{
	struct scd_host *shost;
	struct scd_device *sdev;
	struct usb_device *udev;
	struct scd_hep *ep;
	struct urb *ep_urb;
	unsigned long flags;
	unsigned int pipe = urb->pipe;
	int scd_status = 0;
	int is_out = !usb_pipein(pipe);
	int type = usb_pipetype(pipe);

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;
	udev = urb->dev;

	ep = (struct scd_hep *)urb->hcpriv;
	ep_urb = ep->urb;

	spin_lock_irqsave(&shost->lock, flags);
	scd_status = usb_hcd_check_unlink_urb(hcd, urb, status);
	if (scd_status) {
		goto done_not_comp;
	}

	if (ep_urb != urb) {
		goto done;
	}

	/* hep_enqueue */
	switch (type) {
	case PIPE_CONTROL:
		scd_status = hep_cntl_trans_state(hcd, &shost->ep0, SCD_HEP_T_RESET);
		break;
	case PIPE_BULK:
	case PIPE_INTERRUPT:
		if (is_out) {
			scd_status = hep_tx_trans_state(hcd, ep, SCD_HEP_T_RESET);
		}
		else {
			scd_status = hep_rx_trans_state(hcd, ep, SCD_HEP_T_RESET);
		}
		break;
	default:
		scd_status = -ENOSPC;
		break;
	}

done:
	hep_complete(hcd, ep, urb, status);
done_not_comp:
	spin_unlock_irqrestore(&shost->lock, flags);

	DBG(2,"==>\n");
	return scd_status;
}

/* control transaction */

inline static void hep_cntl_disable(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum = 0;

	hep_disable_ch(hcd, epnum);
}

inline static void hep_cntl_enable(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum = 0;

	hep_disable_ch(hcd, epnum);
}

inline static void hep_cntl_reset(struct usb_hcd *hcd)
{
	scd_hep_cntl_init(hcd);
}

inline static void hep_cntl_set_pid(struct scd_hep *ep, u32 pid)
{
	hep_set_pid(ep, pid);
}

static void hep_cntl_self_start(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct urb *urb;
	struct usb_host_endpoint *hep = ep->hep;
	
	DBG(2, "<==\n");

	if (ep->urb || !hep || list_empty(&hep->urb_list)) {
		DBG(2,"==>\n");
		return;
	}
	urb = list_entry(hep->urb_list.next, struct urb, urb_list);
	ep->urb = urb;

	hep_cntl_trans_state(hcd, ep, SCD_HEP_T_START);

	DBG(2,"==>\n");
}

static u32 hep_cntl_check_error(struct usb_hcd *hcd)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	u8 epnum = 0;
	u32 error = 0;
	u32 hcint;

	DBG(2, "<==\n");

	hcint = SCD__READ(sdev, HCINT(epnum));

	error = 0;
	if (hcint  & SCD_B_HCINTN_FRMOVRUN) {
		error |= SCD_HEP_E_OVERFLOW;
	}
	if (hcint  & SCD_B_HCINTN_BBLERR) {
		error |= SCD_HEP_E_MAXPOVER;
	}
	if (hcint  & SCD_B_HCINTN_XACTERR) {
		error |= SCD_HEP_E_XACTERR;
	}
	if (hcint  & SCD_B_HCINTN_STALL) {
		error |= SCD_HEP_E_RXSTALL;
	}
	if (hcint  & SCD_B_HCINTN_AHBERR) {
		error |= SCD_HEP_E_AHBERR;
	}

	DBG(2,"==>\n");
	return error;
}

inline static void hep_cntl_unmask_intr(struct usb_hcd *hcd, struct scd_hep *ep)
{
	hep_unmask_rx_intr(hcd, ep);
}

inline static void hep_cntl_mask_intr(struct usb_hcd *hcd, struct scd_hep *ep)
{
	hep_mask_rx_intr(hcd, ep);
}

static int hep_cntl_tx_setup_dma(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = 0;
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_cntl_tx_setup_ch_size(hcd, ep, epnum, ep_type);
	hep_cntl_set_setup_dma_addr(hcd, ep, epnum);
	hep_cntl_tx_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}

static int hep_cntl_tx_setup_cpu(struct usb_hcd *hcd, struct scd_hep *ep)
{
	unsigned char* buf;
	u8 epnum;
	u8 ep_type;
	u32 len;

	DBG(2, "<==\n");

	epnum = 0;
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_cntl_tx_setup_ch_size(hcd, ep, epnum, ep_type);
	buf = ep->urb->setup_packet;
	len = sizeof(ep->urb->setup_packet);
	scd_device_host_load_fifo(hcd, ep_type, epnum, len, (u32 *)buf);
	hep_cntl_tx_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;

}

static int hep_cntl_tx_data_dma_multi(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = 0;
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_tx_set_ch_size_multi(hcd, ep, epnum, ep_type);
	hep_tx_set_dma_addr(hcd, ep, epnum);
	hep_tx_set_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}

static int hep_cntl_tx_data_dma(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = 0;
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_tx_set_ch_size(hcd, ep, epnum, ep_type);
	hep_tx_set_dma_addr(hcd, ep, epnum);
	hep_tx_set_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}

static int hep_cntl_tx_data_cpu(struct usb_hcd *hcd, struct scd_hep *ep)
{
	char* buf;
	u8 epnum;
	u8 ep_type;
	u32 len;

	DBG(2, "<==\n");

	epnum = 0;
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_tx_set_ch_size(hcd, ep, epnum, ep_type);
	hep_get_buf(ep->urb, &ep->hep->desc, NULL, &buf, &len);
	scd_device_host_load_fifo(hcd, ep_type, epnum, len, (u32 *)buf);
	hep_tx_set_ch_char(hcd, ep, epnum, ep_type);

	return 0;
}

static int hep_cntl_tx_status_dma(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = 0;
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_cntl_tx_status_ch_size(hcd, ep, epnum, ep_type);
	hep_cntl_set_status_dma_addr(hcd, ep, epnum);
	hep_cntl_tx_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}

static int hep_cntl_tx_status_cpu(struct usb_hcd *hcd, struct scd_hep *ep)
{
	char buf[4];  /* Minimum of 32bit alignment */
	u8 epnum;
	u8 ep_type;
	u32 len;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_cntl_tx_status_ch_size(hcd, ep, epnum, ep_type);
	len = 0;
	scd_device_host_load_fifo(hcd, ep_type, epnum, len, (u32 *)&buf);
	hep_cntl_tx_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}

#ifdef SLAVE_MODE_DEBUG
static bool hep_cntl_check_tx_remain(struct scd_hep *ep)
{
	bool remain;
	char* buf;
	u32 len;
	
	DBG(2, "<==\n");

	hep_get_buf(ep->urb, &ep->hep->desc, NULL, &buf, &len);
	hep_add_actual(ep->urb, len);
	remain = hep_get_remain(ep->urb);

	DBG(2,"==>\n");
	return remain;
}
#endif /* SLAVE_MODE_DEBUG */

static void hep_cntl_rx_data_dma_multi(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_rx_set_ch_size_multi(hcd, ep, epnum, ep_type);
	hep_rx_set_dma_addr(hcd, ep, epnum);
	hep_rx_set_ch_char_multi(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return;
}

#ifdef SLAVE_MODE_DEBUG
static void hep_cntl_rx_data_dma(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_rx_set_ch_size(hcd, ep, epnum, ep_type);
	hep_rx_set_dma_addr(hcd, ep, epnum);
	hep_rx_set_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return;
}
#endif /* SLAVE_MODE_DEBUG */

static void hep_cntl_rx_data_dma_short(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_rx_set_ch_size(hcd, ep, epnum, ep_type);
	hep_rx_set_dma_addr_short(hcd, ep, epnum);
	hep_rx_set_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return;
}

static void hep_cntl_rx_data_cpu(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_rx_set_ch_size(hcd, ep, epnum, ep_type);
	hep_rx_set_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return;
}

#ifdef SLAVE_MODE_DEBUG
static bool hep_cntl_rx_get_data(struct usb_hcd *hcd, struct scd_hep *ep,
				u32 *error)
{
	struct urb *urb= ep->urb;
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	char* buf;
	u8 epnum;
	u8 ep_type;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u16 rxcount;
	u32 expected;
	u32 len;
	u32 gintmsk, grxstsp, pktsts;
	bool remain;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	/* Masking RxFlvl */
	gintmsk = SCD__READ(sdev, GINTMSK);
	gintmsk &= ~(SCD_B_GINTMSK_RXFLVLMSK);
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	hep_get_buf(urb, &ep->hep->desc, NULL, &buf, &expected);
	if (!urb->transfer_buffer) {
		*error |= SCD_HEP_E_NOBUFFER;
		return false;
	}
	grxstsp = SCD__READ(sdev, GRXSTSP);
	pktsts = (grxstsp >> SCD_S_GRXSTS_PKTSTS) & SCD_M_GRXSTS_PKTSTS;
	if (pktsts==0x0010) {  /* IN data packet received */
		rxcount = (grxstsp >> SCD_B_GRXSTS_BCNT) & SCD_M_GRXSTS_BCNT;
		if(rxcount > 0){
			len = min(expected, (u32)rxcount);
			hep_add_actual(urb, rxcount);
			scd_device_host_unload_fifo(hcd, ep_type, epnum, len, (u32 *)buf);
			remain = (urb->transfer_buffer_length > urb->actual_length) ? true : false;
			*error |= hep_check_rx_len_error(maxp, expected, rxcount,
				urb->transfer_flags & URB_SHORT_NOT_OK);
		}
	}
	/* UnMasking RxFlvl */
	gintmsk |= SCD_B_GINTMSK_RXFLVLMSK;
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	DBG(2,"==>\n");
	return remain;
}

static bool hep_cntl_rx_get_status(struct usb_hcd *hcd, struct scd_hep *ep,
				u32 *error)
{
	struct urb *urb= ep->urb;
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	char buf[4];  /* Minimum of 32bit alignment */
	u8 epnum;
	u8 ep_type;
	u16 maxp = ep->hep->desc.wMaxPacketSize;
	u16 rxcount;
	u32 expected;
	u32 len;
	u32 gintmsk, grxstsp, pktsts;
	bool remain;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	/* Masking RxFlvl */
	gintmsk = SCD__READ(sdev, GINTMSK);
	gintmsk &= ~(SCD_B_GINTMSK_RXFLVLMSK);
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	expected = sizeof(buf);
	grxstsp = SCD__READ(sdev, GRXSTSP);
	pktsts = (grxstsp >> SCD_S_GRXSTS_PKTSTS) & SCD_M_GRXSTS_PKTSTS;
	if (pktsts==0x0010) {  /* IN data packet received */
		rxcount = (grxstsp >> SCD_B_GRXSTS_BCNT) & SCD_M_GRXSTS_BCNT;
		if(rxcount > 0){
			len = min(expected, (u32)rxcount);
			hep_add_actual(urb, rxcount);
			scd_device_host_unload_fifo(hcd, ep_type, epnum, len, (u32 *)buf);
			remain = (urb->transfer_buffer_length > urb->actual_length) ? true : false;
			*error |= hep_check_rx_len_error(maxp, expected, rxcount,
				urb->transfer_flags & URB_SHORT_NOT_OK);
		}
	}
	/* UnMasking RxFlvl */
	gintmsk |= SCD_B_GINTMSK_RXFLVLMSK;
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	DBG(2,"==>\n");
	return remain;
}
#endif /* SLAVE_MODE_DEBUG */

static int hep_cntl_rx_status_dma(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_cntl_rx_status_ch_size(hcd, ep, epnum, ep_type);
	hep_cntl_set_status_dma_addr(hcd, ep, epnum);
	hep_cntl_rx_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;
}

static int hep_cntl_rx_status_cpu(struct usb_hcd *hcd, struct scd_hep *ep)
{
	u8 epnum;
	u8 ep_type;

	DBG(2, "<==\n");

	epnum = hep_hep2epnum(ep->hep);
	ep_type = hep_check_ep_type(&ep->hep->desc);

	hep_rx_set_ch_size(hcd, ep, epnum, ep_type);
	hep_rx_set_ch_char(hcd, ep, epnum, ep_type);

	DBG(2,"==>\n");
	return 0;

}

#ifdef SLAVE_MODE_DEBUG
static void hep_cntl_dma_sync_start(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_hcd *shcd = (hcd_to_shost(hcd))->shcd;
	struct urb *urb = ep->urb;
	u8 type;
	enum dma_data_direction dir;

	DBG(2, "<==\n");

	type = hep_cntl_get_type(ep);
	if (type == SCD_HEP_CNTL_OUT) {
		dir = DMA_FROM_DEVICE;
	}
	else if (type == SCD_HEP_CNTL_IN) {
		dir = DMA_TO_DEVICE;
	}
	else {
		goto no_data;
	}

	if (!urb->transfer_buffer) {
		urb->transfer_buffer = ioremap(urb->transfer_dma,
					       urb->transfer_buffer_length);
		if (!urb->transfer_buffer) {
			goto no_data;
		}
		ep->attributes |= SCD_HEP_ATR_IOREMAP;
	}

	if (urb->transfer_dma) {
		dma_sync_single_for_cpu(shcd->dev, urb->transfer_dma,
					urb->transfer_buffer_length, dir);
	}

no_data:
	if (urb->setup_dma) {
		dma_sync_single_for_cpu(shcd->dev, urb->setup_dma, 8,
					DMA_FROM_DEVICE);
	}

	DBG(2,"==>\n");
}

static void hep_cntl_dma_sync_end(struct usb_hcd *hcd, struct scd_hep *ep)
{
	struct scd_device *sdev = (hcd_to_shost(hcd))->sdev;
	struct urb *urb = ep->urb;
	u8 type;
	enum dma_data_direction dir;

	DBG(2, "<==\n");

	type = hep_cntl_get_type(ep);
	if (type == SCD_HEP_CNTL_OUT) {
		dir = DMA_FROM_DEVICE;
	}
	else if (type == SCD_HEP_CNTL_IN) {
		dir = DMA_TO_DEVICE;
	}
	else {
		goto no_data;
	}

	if (urb->transfer_dma) {
		dma_sync_single_for_device(shcd->dev, urb->transfer_dma,
					   urb->transfer_buffer_length, dir);
	}

	if (ep->attributes & SCD_HEP_ATR_IOREMAP) {
		iounmap(urb->transfer_buffer);
		urb->transfer_buffer = NULL;
		ep->attributes &= ~(SCD_HEP_ATR_IOREMAP);
	}

no_data:
	if (urb->setup_dma) {
		dma_sync_single_for_device(shcd->dev, urb->setup_dma, 8,
					   DMA_FROM_DEVICE);
	}

	DBG(2,"==>\n");
}
#endif /* SLAVE_MODE_DEBUG */

static int hep_cntl_trans_state(struct usb_hcd *hcd, struct scd_hep *ep,
				u32 trigger)
{
	u32 error = 0;
	u32 nextsz;
	bool reset = trigger == SCD_HEP_T_RESET ? true : false;
	bool remain = false;
	bool zerop = false;
	bool shortp = false;
	u8 type = SCD_HEP_CNTL_NODATA;

	DBG(2, "<==\n");
	DBG(3, "ep->state (exit) = %lu\n", (unsigned long)ep->state);

	/* exit state */
	switch (ep->state) {
	case SCD_HEP_S_CNTL_IDLE:
		if ((trigger == SCD_HEP_T_LINK_CHHLT_INTR) ||
		    (trigger == SCD_HEP_T_LINK_ERROR_INTR)) {
			/* Nothing to do */
			return 0;
		}
		if (!reset) {
			hep_cntl_set_pid( ep, PID_SETUP );
			ep->retrycount = 0;
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_SETUP_DMA:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
			if (!error) {
				shortp = hep_rx_check_short(ep);
				type=hep_cntl_get_type(ep);
				hep_cntl_set_pid(ep, PID_DATA1);
				ep->retrycount = 0;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_SETUP_CPU:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
			if (!error) {
				type=hep_cntl_get_type(ep);
				remain = hep_tx_check_remain(ep, &nextsz); 
				hep_tx_dma_sync_start(hcd, ep);
				hep_cntl_set_pid(ep, PID_DATA1);
				ep->retrycount = 0;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_DATA_DMA_MULTI:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
			if (!error) {
				zerop = hep_tx_update_actual_multi(ep);
				remain = hep_get_remain(ep->urb);
				if (!zerop) {
					hep_cntl_set_pid(ep, PID_DATA1);
					ep->retrycount = 0;
				}
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_DATA_DMA_ZEROP:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
			if (!error) {
				hep_tx_update_actual(ep);
				remain = hep_tx_check_remain(ep, &nextsz);
				if (!remain) {
					hep_cntl_set_pid(ep, PID_DATA1);
					ep->retrycount = 0;
				}
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_DATA_CPU:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
			if (!error) {
				remain = hep_tx_check_remain(ep, &nextsz);
				hep_tx_dma_sync_end(hcd, ep);
				hep_cntl_set_pid(ep, PID_DATA1);
				ep->retrycount = 0;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_DATA_DMA_MULTI:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
			if (!error) {
				remain = hep_rx_update_actual_multi(hcd, ep);
				shortp = hep_rx_check_short(ep);
				if (!shortp) {
					hep_cntl_set_pid(ep, PID_DATA1);
					ep->retrycount = 0;
				}
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_DATA_DMA_SHORTP:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
			if (!error) {
				hep_rx_check_len(hcd, ep, &error);
				remain = hep_rx_get_data_from_buffer(hcd, ep, &error);
				hep_rx_update_actual(hcd,ep);
				hep_cntl_set_pid(ep, PID_DATA1);
				ep->retrycount = 0;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_DATA_CPU:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
			if (!error) {
				hep_rx_check_len(hcd, ep, &error);
				remain = hep_rx_get_data(hcd, ep, &error);
			}
		}
		if (reset || error || !remain) {
			hep_cntl_set_pid(ep, PID_DATA1);
			ep->retrycount = 0;
			hep_rx_dma_sync_end(hcd, ep);
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_STATUS_DMA:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_STATUS_CPU:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
			remain = hep_tx_check_remain(ep, &nextsz);
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_STATUS_DMA:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_STATUS_CPU:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
			if (!error) {
				hep_rx_check_len(hcd, ep, &error);
				remain=hep_rx_get_status(hcd, ep, &error);
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_STATUS_CPU_HALT:
		if (!reset) {
			error = hep_cntl_check_error(hcd);
			if (!error) {
				if (trigger == SCD_HEP_T_LINK_RXFLVL_INTR) {
					hep_rx_check_len(hcd, ep, &error);
					remain=hep_rx_get_status(hcd, ep, &error);
				}
				else if (trigger == SCD_HEP_T_LINK_XFERCOMPL_INTR) {
				}
			}
		}
		break;
	}

	/* transition state */
	if (trigger == SCD_HEP_T_RESET ||
	    trigger == SCD_HEP_T_LINK_ERROR_INTR ||
	    ((trigger == SCD_HEP_T_LINK_RXFLVL_INTR || 
	      trigger == SCD_HEP_T_LINK_XFERCOMPL_INTR || 
	      trigger == SCD_HEP_T_LINK_CHHLT_INTR) && error)) {
		ep->state = SCD_HEP_S_CNTL_IDLE;
		goto enter_state;
	}
	switch (ep->state) {
	case SCD_HEP_S_CNTL_IDLE:
		if (trigger == SCD_HEP_T_START) {
#ifdef SLAVE_MODE_DEBUG
			if (sdev->dmac) {
#endif /* SLAVE_MODE_DEBUG */
				ep->state = SCD_HEP_S_CNTL_WAIT_TX_SETUP_DMA;
#ifdef SLAVE_MODE_DEBUG
			}
			else {
				ep->state = SCD_HEP_S_CNTL_WAIT_TX_SETUP_CPU;
			}
#endif /* SLAVE_MODE_DEBUG */
			goto enter_state;
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_SETUP_DMA:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (type == SCD_HEP_CNTL_OUT && !error) {
				ep->state = 
				    SCD_HEP_S_CNTL_WAIT_TX_DATA_DMA_MULTI;
				goto enter_state;
			}
			else if (type == SCD_HEP_CNTL_IN && !error) {
				if ((!shortp) && ep->urb->transfer_buffer_length) {
					ep->state = 
					 SCD_HEP_S_CNTL_WAIT_RX_DATA_DMA_MULTI;
				}
				else {
					ep->state = 
					 SCD_HEP_S_CNTL_WAIT_RX_DATA_DMA_SHORTP;
				}
				goto enter_state;
			}
			else if (type == SCD_HEP_CNTL_NODATA && !error) {
				ep->state = SCD_HEP_S_CNTL_WAIT_RX_STATUS_DMA;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_SETUP_CPU:
		if (trigger == SCD_HEP_T_LINK_XFERCOMPL_INTR) {
			if (type == SCD_HEP_CNTL_OUT && !error) {
				ep->state = SCD_HEP_S_CNTL_WAIT_TX_DATA_CPU;
				goto enter_state;
			}
			else if (type == SCD_HEP_CNTL_IN && !error) {
				ep->state = SCD_HEP_S_CNTL_WAIT_RX_DATA_CPU;
				goto enter_state;
			}
			else if (type == SCD_HEP_CNTL_NODATA && !error) {
				ep->state = SCD_HEP_S_CNTL_WAIT_RX_STATUS_CPU;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_DATA_DMA_MULTI:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				if (remain) {
					/* Do not change state */
					goto enter_state;
				}
				else {
					if (zerop) {
						ep->state = SCD_HEP_S_CNTL_WAIT_TX_DATA_DMA_ZEROP;
						goto enter_state;
					}
					else if (!zerop && !error) {
						ep->state = SCD_HEP_S_CNTL_WAIT_RX_STATUS_DMA;
						goto enter_state;
					}
				}
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_DATA_DMA_ZEROP:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				ep->state = SCD_HEP_S_CNTL_WAIT_RX_STATUS_DMA;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_DATA_CPU:
		if (trigger == SCD_HEP_T_LINK_XFERCOMPL_INTR) {
			if (remain && !error) {
				goto enter_state;
			}
			if (!remain && !error) {
				ep->state = SCD_HEP_S_CNTL_WAIT_RX_STATUS_CPU;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_DATA_DMA_MULTI:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				if (remain) {
					if (!shortp) {
						/* Reenter same state */
						goto enter_state;
					}
					else {
						ep->state = 
						 SCD_HEP_S_CNTL_WAIT_RX_DATA_DMA_SHORTP;
						goto enter_state;
					}
				}
				else {
						ep->state = 
						  SCD_HEP_S_CNTL_WAIT_TX_STATUS_DMA;
						goto enter_state;
				}
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_DATA_DMA_SHORTP:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				ep->state = SCD_HEP_S_CNTL_WAIT_TX_STATUS_DMA;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_DATA_CPU:
		if (trigger == SCD_HEP_T_LINK_RXFLVL_INTR) {
			if (remain && !error) {
				goto enter_state;
			}
			else if (!remain && !error) {
				ep->state = SCD_HEP_S_CNTL_WAIT_RX_DATA_CPU_HALT;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_DATA_CPU_HALT:
		if (trigger == SCD_HEP_T_LINK_RXFLVL_INTR) {
			if (!error) {
				goto enter_state;
			}
		}
		if (trigger == SCD_HEP_T_LINK_XFERCOMPL_INTR) {
			if (!error) {
				goto enter_state;
			}
		}
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				ep->state = SCD_HEP_S_CNTL_WAIT_TX_STATUS_CPU;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_STATUS_DMA:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				ep->state = SCD_HEP_S_CNTL_IDLE;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_STATUS_CPU:
		if (trigger == SCD_HEP_T_LINK_RXFLVL_INTR) {
			if (!error) {
				ep->state = SCD_HEP_S_CNTL_WAIT_RX_STATUS_CPU_HALT;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_STATUS_CPU_HALT:
		if (trigger == SCD_HEP_T_LINK_RXFLVL_INTR) {
			if (!error) {
				goto enter_state;
			}
		}
		if (trigger == SCD_HEP_T_LINK_XFERCOMPL_INTR) {
			if (!error) {
				goto enter_state;
			}
		}
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				ep->state = SCD_HEP_S_CNTL_IDLE;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_STATUS_DMA:
		if (trigger == SCD_HEP_T_LINK_CHHLT_INTR) {
			if (!error) {
				ep->state = SCD_HEP_S_CNTL_IDLE;
				goto enter_state;
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_STATUS_CPU:
		if (trigger == SCD_HEP_T_LINK_XFERCOMPL_INTR) {
			if (!error) {
				ep->state = SCD_HEP_S_CNTL_IDLE;
				goto enter_state;
			}
		}
		break;
	}
	goto error_end;

	/* enter state */
enter_state:
	switch (ep->state) {
	case SCD_HEP_S_CNTL_IDLE:
		hep_cntl_disable(hcd, ep);
		if (!reset) {
			if (!error) {
				hep_complete(hcd, ep, ep->urb, 0);
				hep_cntl_self_start(hcd, ep);
			}
			else {
				hep_error(hcd, ep, error);
			}
		}
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_SETUP_DMA:
		hep_cntl_unmask_intr(hcd, ep);
		hep_cntl_tx_setup_dma(hcd, ep);
		hep_tx_enable(hcd, ep);
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_SETUP_CPU:
		hep_cntl_unmask_intr(hcd, ep);
		hep_cntl_tx_setup_cpu(hcd, ep);
		hep_tx_enable(hcd, ep);
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_DATA_DMA_MULTI:
		hep_cntl_tx_data_dma_multi(hcd, ep);
		hep_tx_enable(hcd, ep);
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_DATA_DMA_ZEROP:
		hep_cntl_tx_data_dma(hcd, ep);
		hep_tx_enable(hcd, ep);
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_DATA_CPU:
		hep_cntl_tx_data_cpu(hcd, ep);
		hep_tx_enable(hcd, ep);
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_DATA_DMA_MULTI:
		hep_cntl_rx_data_dma_multi(hcd, ep);
		hep_rx_enable(hcd, ep);
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_DATA_DMA_SHORTP:
		hep_cntl_rx_data_dma_short(hcd, ep);
		hep_rx_enable(hcd, ep);
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_DATA_CPU:
		hep_cntl_rx_data_cpu(hcd, ep);
		hep_rx_enable(hcd, ep);
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_DATA_CPU_HALT:
		/* Noting to do */
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_STATUS_DMA:
		hep_cntl_tx_status_dma(hcd, ep);
		hep_tx_enable(hcd, ep);
		break;
	case SCD_HEP_S_CNTL_WAIT_TX_STATUS_CPU:
		hep_cntl_tx_status_cpu(hcd, ep);
		hep_tx_enable(hcd, ep);
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_STATUS_DMA:
		hep_cntl_rx_status_dma(hcd, ep);
		hep_rx_enable(hcd, ep);
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_STATUS_CPU:
		hep_cntl_rx_status_cpu(hcd, ep);
		hep_rx_enable(hcd, ep);
		break;
	case SCD_HEP_S_CNTL_WAIT_RX_STATUS_CPU_HALT:
		/* Noting to do */
		break;
	}

	DBG(3, "ep->state (enter) = %lu\n", (unsigned long)ep->state);
	DBG(2,"==>\n");
	return 0;

error_end:
	DBG(3, "ep->state (enter) = %lu\n", (unsigned long)ep->state);
	DBG(2,"==>\n");
	return -EINVAL;
}

void scd_hep_endpoint_disable(struct usb_hcd *hcd, struct usb_host_endpoint *hep)
{
	struct scd_host *shost;
	struct scd_device *sdev;
	struct scd_hep *ep_table;
	struct scd_hep *ep;
	struct usb_endpoint_descriptor *epd = &hep->desc;
	int is_out = usb_endpoint_dir_out(epd);
	int epnum = hep_hep2epnum(hep);
	unsigned long flags;
	u8 *nep;
	u8 other_nep;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;

	spin_lock_irqsave(&shost->lock, flags);

	if (epnum == 0) {
		ep = &shost->ep0;
		ep->status_dma = 0x0;
		ep->dev = NULL;
		ep->hep = NULL;
	}
	else {
		ep_table = is_out ? &shost->ep_out[0] : &shost->ep_in[0];
		nep = is_out ? &shost->nep_out : &shost->nep_in;
		other_nep = !is_out ? shost->nep_out : shost->nep_in;
		hep_release_ep(hcd, hep, ep_table, nep, other_nep);
	}
	spin_unlock_irqrestore(&shost->lock, flags);

	DBG(2,"==>\n");
}
