/*-----------------------------------------------------------------------------
 * drivers/usb/scd/scd_hfifo.c
 * 
 * Host FIFO operations
 * 
 * Copyright 2011 Sony Corporation
 *
 * This file is part of the HS-OTG Controller Driver SCD.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/kernel.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb/gadget.h>
#include <linux/usb/scd/usb_otg_control.h>
#include <linux/usb/scd/scd_ioctl.h>
#include "../core/hub.h"
#include "../core/hcd.h"

#include "scd.h"
#include "scd_debug.h"
#include "scd_host.h"
#include "scd_register.h"
#include "scd_hfifo.h"


/* FIFO func */

/*
 * Load an FIFO
 *
 * @param hcd controller driver instance pointer
 * @param ep_type endpoint type
 * @param index local endpoint
 * @param len how many bytes to load
 * @param src data source buffer
 */
void scd_device_host_load_fifo(struct usb_hcd *hcd, 
	u8 ep_type, u8 index, u32 len, u32* src)
{
	struct scd_host *shost;
	u16 fidx;
	u16 htx_fifo_addr;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);

	switch(ep_type){
	case USB_ENDPOINT_XFER_CONTROL:
	case USB_ENDPOINT_XFER_BULK:
	    htx_fifo_addr = SCD_R_DFIFO + shost->tx_fifo_addr[0];
	    break;
	case USB_ENDPOINT_XFER_INT:
	case USB_ENDPOINT_XFER_ISOC:
	    htx_fifo_addr = SCD_R_DFIFO + shost->tx_fifo_addr[1];
	    break;
	default:
	    ERR("ep_type error \n");
	    return;
	}

	/* Writing doublewords */
	for (fidx = 0; (fidx*4) < len; fidx++) {
		SCD_WRITE(shost->sdev, htx_fifo_addr, src[fidx]);
	}

	DBG(2,"==>\n");
}
EXPORT_SYMBOL(scd_device_host_load_fifo);

/*
 * Unload an FIFO
 *
 * @param hcd controller driver instance pointer
 * @param ep_type endpoint type
 * @param index local endpoint
 * @param len how many bytes to unload
 * @param src data buffer
 */
void scd_device_host_unload_fifo(struct usb_hcd *hcd, u8 ep_type,
	u8 index, u32 len, u32* src)
{
	struct scd_host *shost;
	u16 fidx;
	u16 hrx_fifo_addr;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
        hrx_fifo_addr = SCD_R_DFIFO;

	/* Writing doublewords */
	for (fidx = 0; (fidx*4) < len; fidx++) {
		src[fidx] = SCD_READ(shost->sdev, hrx_fifo_addr);
	}

	DBG(2,"==>\n");
}
EXPORT_SYMBOL(scd_device_host_unload_fifo);

/**
 * Reset host dynamic FIFO size
 *
 * @param hcd controller driver instance pointer
 */
void scd_host_fifo_reset(struct usb_hcd *hcd)
{
	DBG(2, "<==\n");
	scd_host_fifo_map(hcd);
}
EXPORT_SYMBOL(scd_host_fifo_reset);

/**
 * Map host dynamic FIFO for specified ep
 *
 * @param hcd controller driver instance pointer
 */
#define LARGEST_MAXP_RXFIFO 1024
#define LARGEST_MAXP_NP_TXFIFO 512
#define LARGEST_MAXP_P_TXFIFO 1024
#define STATUS_INFOMATION 1
#define TRANSFER_COMPLETE 1
#define LOCATION_EACH_NP_FOR_NAK SCD_NUM_HEPS
#define CONSTANT_AHB_LATENCY 1

void scd_host_fifo_map(struct usb_hcd *hcd)
{
	struct scd_host *shost;
	u16 maxp_rxfifo;
	u16 maxp_np_txfifo, maxp_p_txfifo;
	u16 hrx_fifo_siz;
	u16 htx_npfifo_siz, htx_pfifo_siz;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);

	/* Sheared FIFO is set in about lagest MAXP */
	maxp_rxfifo    = LARGEST_MAXP_RXFIFO;
	maxp_np_txfifo = LARGEST_MAXP_NP_TXFIFO;
	maxp_p_txfifo  = LARGEST_MAXP_P_TXFIFO;

	/* Calculation for sheared FIFO */
	if ( 1 /* sdev->dmac */) {
		/* DMA mode */
		hrx_fifo_siz   = (CONSTANT_AHB_LATENCY + 1) * 
			((maxp_rxfifo / 4) + STATUS_INFOMATION +
			TRANSFER_COMPLETE) + LOCATION_EACH_NP_FOR_NAK;
	}
	else {
		/* Slave mode */
		hrx_fifo_siz   = (CONSTANT_AHB_LATENCY + 1) * 
			((maxp_rxfifo / 4) + STATUS_INFOMATION + 
			TRANSFER_COMPLETE);
	}
	htx_npfifo_siz = (CONSTANT_AHB_LATENCY + 1) * (maxp_np_txfifo / 4);
	/* Not over maximum interrupt/iso transfer ratio in case of HS */
	htx_pfifo_siz  = 
		min((CONSTANT_AHB_LATENCY + 1) * (maxp_np_txfifo / 4),
		    (CONSTANT_AHB_LATENCY + 1) * (1500 / 4));

	/* Set register for sheared FIFO */
	shost->rx_fifo_siz = hrx_fifo_siz;
	shost->tx_fifo_size[0] = htx_npfifo_siz;
	shost->tx_fifo_addr[0] = hrx_fifo_siz;
	shost->tx_fifo_size[1] = htx_pfifo_siz;
	shost->tx_fifo_addr[1] = shost->tx_fifo_addr[0] + shost->tx_fifo_size[0];
	SCD__WRITE(shost->sdev, GRXFSIZ, (shost->rx_fifo_siz & SCD_M_GRXFSIZ_RXFDEP));
	SCD__WRITE(shost->sdev, GNPTXFSIZ, 
		(shost->tx_fifo_size[0] << SCD_S_GNPTXFSIZ_NPTXFDEP)|
		(shost->tx_fifo_addr[0] & SCD_M_GNPTXFSIZ_NPTXFSTADDR));
	SCD__WRITE(shost->sdev, HPTXFSIZ,  
		(shost->tx_fifo_size[1] << SCD_S_HPTXFSIZ_PTXFSIZE)|
		(shost->tx_fifo_addr[1] & SCD_M_HPTXFSIZ_PTXFSTADDR));

	/* Flush FIFO */
	scd_hep_cntl_init(hcd);
	scd_hep_tx_init(hcd);
	scd_hep_rx_init(hcd);
	scd_hep_ch0_init(hcd);

	DBG(2,"==>\n");
}
EXPORT_SYMBOL(scd_host_fifo_map);

void scd_fifo_dump(struct usb_hcd *hcd)
{
	/* It is nonimplement */
}
