/*-----------------------------------------------------------------------------
 * drivers/usb/scd/scd_host.c
 * 
 * Host mode operations
 * 
 * Copyright 2011, 2012 Sony Corporation
 *
 * This file is part of the HS-OTG Controller Driver SCD.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *---------------------------------------------------------------------------*/


#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/io.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/dma-mapping.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>

#include "../core/hub.h"
#include "../core/hcd.h"
#include "scd.h"
#include "scd_version.h"
#include "scd_register.h"
#include "scd_host.h"
#include "scd_hfifo.h"
#include "scd_debug.h"

#define HOST_DEVICE_NAME "host"
#define LARGEST_MAXP_RXFIFO_BUF 1024

static inline
struct scd_hcd *control_to_shcd(struct scd_control* h_control)
{
	return container_of(h_control, struct scd_hcd, h_control);
}

static inline
struct usb_hcd *shcd_to_hcd(struct scd_hcd *shcd)
{
	return (struct usb_hcd *)(shcd->hcd);
}

static inline
struct usb_hcd *control_to_hcd(struct scd_control* h_control)
{
	return shcd_to_hcd(control_to_shcd(h_control));
}

/*
 *	host init functions
 */

static void host_init_hub(struct usb_hcd *hcd)
{
	struct scd_host *shost;

	DBG(2, "<==\n");
	shost = hcd_to_shost(hcd);
	scd_port_init(&shost->hub.port[0]);
	DBG(2,"==>\n");
}

static void host_init_hep(struct usb_hcd *hcd)
{
	struct scd_host *shost;
	u8 i;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);

	scd_hep_tx_init(hcd);
	scd_hep_rx_init(hcd);
	scd_hep_ch0_init(hcd);
	for (i = 0; i < SCD_NUM_HEPS; i++) {
		spin_lock_init(&shost->ep_in[i].lock);
		spin_lock_init(&shost->ep_out[i].lock);
	}

	DBG(2,"==>\n");
}

static inline void host_init_regs(struct usb_hcd *hcd)
{
	struct scd_host *shost;
	struct scd_device *sdev;
	int i;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;

	/* Initialize Host disconn registers */
	SCD__WRITE(sdev, GINTSTS, SCD_B_GINTSTS_DISCONNINT);

	/* Initialize all Host registers */
	SCD__WRITE(sdev, HCFG,    0x00000200);  /* RESVALID(15:8)=8d2 */
	SCD__WRITE(sdev, HFIR,    0x0000ea60);  /* FRINT(15:0)=16d60000 */
	SCD__WRITE(sdev, HAINTMSK,0x00000000);  
	SCD__WRITE(sdev, HPRT,    0x00000000);  
	for( i = 0; i < SCD_NUM_HEPS; i++ ){
		SCD__WRITE(sdev, HCCHAR(i),    0x00000000);  
		SCD__WRITE(sdev, HCSPLT(i),    0x00000000);  
		SCD__WRITE(sdev, HCINTMSK(i),  0x00000000); 
		SCD__WRITE(sdev, HCTSIZ(i),    0x00000000); 
		SCD__WRITE(sdev, HCDMA(i),     0x00000000); 
		SCD__WRITE(sdev, HCDMA(i),     0x00000000); 
		SCD__WRITE(sdev, HCDMA(i),     0x00000000); 
	}

	DBG(2,"==>\n");
}

static inline void host_exit_regs(struct usb_hcd *hcd)
{
	struct scd_host *shost;
	struct scd_device *sdev;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;

	/* Exit all Host registers */
	SCD__WRITE(sdev, HPRT,    0x00000000);  /* Session end */

	DBG(2,"==>\n");
}

static inline void host_run_regs(struct usb_hcd *hcd)
{
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 gintmsk;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;

        /* GINTMSK set */
	gintmsk = SCD__READ(sdev, GINTMSK);
        gintmsk |= (SCD_B_GINTMSK_WKUPINTMSK	|
		   SCD_B_GINTMSK_DISCONNINTMSK |
		   SCD_B_GINTMSK_HCHINTMSK |
		   SCD_B_GINTMSK_PRTINTMSK |
		   SCD_B_GINTMSK_MODEMISMSK);
        SCD__WRITE(sdev, GINTMSK, gintmsk);

	DBG(2,"==>\n");
}

static inline void host_stop_regs(struct usb_hcd *hcd)
{
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 gintmsk;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;

        /* GINTMSK set */
	gintmsk = SCD__READ(sdev, GINTMSK);
        gintmsk &= ~(SCD_B_GINTMSK_WKUPINTMSK	|
		   SCD_B_GINTMSK_DISCONNINTMSK |
        	   SCD_B_GINTMSK_HCHINTMSK |
		   SCD_B_GINTMSK_PRTINTMSK);
        SCD__WRITE(sdev, GINTMSK, gintmsk);

	DBG(2,"==>\n");
}

/*
 *	hcd interface ops
 */

static int scd_host_reset(struct usb_hcd *hcd)
{
	struct scd_host *shost;
	struct scd_device *sdev;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;

	host_init_regs(hcd);

	/* Flush all FIFO */
	scd_host_fifo_reset(hcd);
	host_init_hub(hcd);
	host_init_hep(hcd);

	DBG(2,"==>\n");
	return 0;
}

static int scd_host_start(struct usb_hcd *hcd)
{
	struct scd_host *shost;
	struct scd_device *sdev;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;

	hcd->state = HC_STATE_RUNNING;

	host_run_regs(hcd);

	DBG(2,"==>\n");
	return 0;
}

#ifdef SUSRES
static int scd_host_suspend(struct usb_hcd *hcd, pm_message_t message)
{
	struct scd_host *shost;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);

	if (hcd->state != HC_STATE_SUSPENDED) {
		return -EINVAL;
	}

	if (message.event == PM_EVENT_FREEZE) {
		host_init_hub(hcd);
	}

	clear_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags);

	DBG(2,"==>\n");
	return 0;
}

static int scd_host_resume(struct usb_hcd *hcd)
{
	DBG(2, "<==\n");

	set_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags);

	host_init_regs(hcd);
	host_run_regs(hcd);

	/* 顼꥿ܴؿƤӽФ¦usb_hc_diedԤΤ˾ޤ */

	DBG(2,"==>\n");
	return 0;
}
#endif /* SUSRES */

static void scd_host_stop(struct usb_hcd *hcd)
{
	struct scd_host *shost;
	struct scd_device *sdev;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;

	/* Init register, Clear 0 */
	host_stop_regs(hcd);
	host_exit_regs(hcd);

	hcd->state = HC_STATE_HALT;

	DBG(2,"==>\n");
}

static void scd_host_shutdown(struct usb_hcd *hcd)
{
	/* nothing */
}

static int scd_host_get_frame_number(struct usb_hcd *hcd)
{
	struct scd_host *shost;
	struct scd_device *sdev;
	int frame;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;

	frame = (int)(SCD__READ(sdev, HFNUM)) & SCD_M_HFNUM_FRNUM;

	DBG(2,"==>\n");
	return frame;
}

static void scd_relinquish_port(struct usb_hcd *hcd, int portnum)
{
	DBG(2, "<==\n");

	/* It is nonimplement */
	if (hcd->has_tt)
		return;

	DBG(2,"==>\n");
}

static int scd_port_handed_over(struct usb_hcd *hcd, int portnum)
{
	DBG(2, "<==\n");

	/* It is nonimplement */
	if (hcd->has_tt)
		return 0;

	DBG(2,"==>\n");
	return 0;
}

static struct hc_driver scd_hc_driver = {
	.description = "scd_hcd",
	.product_desc = NULL,
	.hcd_priv_size = sizeof(struct scd_host),
	.irq = NULL,
	.flags = HCD_USB2,
	/* host operations */
	.reset = scd_host_reset,
	.start = scd_host_start,
	/* pci_suspendpci_resume(scd_intfsusres) */
	.stop = scd_host_stop,
	.shutdown = scd_host_shutdown,
	.get_frame_number = scd_host_get_frame_number,
	/* hep operations */
	.urb_enqueue = scd_hep_urb_enqueue,
	.urb_dequeue = scd_hep_urb_dequeue,
	.endpoint_disable = scd_hep_endpoint_disable,
	/* hub operations */
	.hub_status_data = scd_hub_status_data,
	.hub_control = scd_hub_control,
	.bus_suspend = scd_hub_bus_suspend,
	.bus_resume = scd_hub_bus_resume,
	.relinquish_port = scd_relinquish_port,
	.port_handed_over = scd_port_handed_over,
};

/*
 *	scd intf host ops
 */

static void hcd_release (struct device *dev)
{
	/* Nothing to do */
}

static struct scd_hcd *scd_host_create_shcd(struct scd_device *sdev)
{
	struct scd_hcd *shcd;
	struct usb_hcd *hcd;
	struct scd_host *shost;
	struct scd_hep *ep;
	int i, j;
	int retval;

	DBG(2, "<==\n");

	/* Allocate shcd */
	shcd = (struct scd_hcd *)kzalloc(sizeof(struct scd_hcd), GFP_KERNEL);
	if(shcd == 0){
		return NULL;
	}
	/* Allocate device */
	shcd->dev = kzalloc (sizeof(struct device), GFP_KERNEL);
	if(shcd->dev == 0){
		kfree(shcd);
		return NULL;
	}
	/* Allocate driver */
	shcd->dev->driver = kzalloc (sizeof(struct device_driver), GFP_KERNEL);
	if(shcd->dev->driver == 0){
		kfree(shcd->dev);
		kfree(shcd);
		return NULL;
	}

	/* Register device */
	dev_set_name(shcd->dev, HOST_DEVICE_NAME);
	strncpy(shcd->dev->bus_id, HOST_DEVICE_NAME, sizeof(shcd->dev->bus_id));
	shcd->dev->parent = sdev->dev;
	shcd->dev->dma_mask = sdev->dev->dma_mask;
	shcd->dev->coherent_dma_mask = sdev->dev->coherent_dma_mask;
	shcd->dev->release = hcd_release;
	/* register_driver not necessary */
	shcd->dev->driver->name = shcd->dev->bus_id;

	retval = device_register (shcd->dev);

	if(retval){
		kfree(shcd->dev->driver);
		kfree(shcd->dev);
		kfree(shcd);
		return NULL;
	}

	/* Allocate hcd */
	hcd = usb_create_hcd(&scd_hc_driver, shcd->dev, "SCDXXX");
	if(hcd == 0){
		device_unregister (shcd->dev);
		kfree(shcd->dev->driver);
		kfree(shcd->dev);
		kfree(shcd);
		return NULL;
	}

	/* Allocate packet_dma */
	shost = hcd_to_shost(hcd);
	ep = &shost->ep0;
	ep->packet_buffer =
	    dma_alloc_coherent(shcd->dev, LARGEST_MAXP_RXFIFO_BUF, 
	    &ep->packet_dma, GFP_KERNEL);
	if(ep->packet_buffer == 0){
		kfree(shcd->dev->driver);
		kfree(shcd->dev);
		kfree(shcd);
		return NULL;
	}
	for( i = 1; i < SCD_NUM_HEPS_HCD; i++ ){
		ep = &(shost->ep_in[i - 1]);
		ep->packet_buffer =
		    dma_alloc_coherent(shcd->dev, (size_t)LARGEST_MAXP_RXFIFO_BUF, 
		    &ep->packet_dma, GFP_KERNEL);
		if(ep->packet_buffer == 0){
			for( j = (i - 1); 0 < j; j-- ){
				ep = &(shost->ep_in[j - 1]);
				dma_free_coherent(shcd->dev, (size_t)LARGEST_MAXP_RXFIFO_BUF, 
				    ep->packet_buffer, ep->packet_dma);
			}
			ep = &shost->ep0;
			dma_free_coherent(shcd->dev, (size_t)LARGEST_MAXP_RXFIFO_BUF, 
			    ep->packet_buffer, ep->packet_dma);
			if (hcd) {
				if (hcd->state != HC_STATE_HALT) {
					/* Host⡼ɻrmmod */
					usb_remove_hcd(hcd);
				}
				usb_put_hcd(hcd);
				hcd = NULL;
			}
			device_unregister (shcd->dev);
			kfree(shcd->dev->driver);
			kfree(shcd->dev);
			kfree(shcd);
			return NULL;
		}
	}

	hcd_to_shost(hcd)->sdev = sdev;
	shcd->hcd = hcd;
	spin_lock_init(&(hcd_to_shost(hcd)->lock));

	DBG(2,"==>\n");
	return shcd;
}

static struct scd_intf scd_host_intf;

static struct scd_control *scd_host_intf_probe(struct scd_device *sdev)
{
	struct scd_hcd *shcd;

	DBG(2, "<==\n");

	/* Allocate usb_hcd, scd_host */
	shcd = scd_host_create_shcd(sdev);
	if(shcd < 0){
		ERR( "scd_hcd - no memory\n");
		return NULL;
	}

	shcd->h_control.intf = (struct scd_intf *)&scd_host_intf;

	/* return scd_control */

	DBG(2,"==>\n");
	return &shcd->h_control;
}

static int scd_host_intf_remove(struct scd_control *h_control)
{
	struct scd_hcd *shcd;
	struct usb_hcd *hcd;
	struct scd_host *shost;
	struct scd_hep *ep;
	int i;

	DBG(2, "<==\n");

	shcd = control_to_shcd(h_control);
	hcd = control_to_hcd(h_control);

	/* free packet_dma */
	shost = hcd_to_shost(hcd);
	ep = &shost->ep0;
	dma_free_coherent(shcd->dev, (size_t)LARGEST_MAXP_RXFIFO_BUF, 
	    ep->packet_buffer, ep->packet_dma);
	for( i = 1; i < SCD_NUM_HEPS_HCD; i++ ){
		ep = &(shost->ep_in[i - 1]);
		dma_free_coherent(shcd->dev, (size_t)LARGEST_MAXP_RXFIFO_BUF, 
		    ep->packet_buffer, ep->packet_dma);
	}

	if (hcd) {
		if (hcd->state != HC_STATE_HALT) {
			/* Host⡼ɻrmmod */
			usb_remove_hcd(hcd);
		}
		usb_put_hcd(hcd);
		hcd = NULL;
	}
	device_unregister (shcd->dev);
	kfree(shcd->dev->driver);
	kfree(shcd->dev);
	kfree(shcd);

	DBG(2,"==>\n");
	return 0;
}

static int scd_host_intf_start(struct scd_control *h_control)
{
	struct scd_hcd *shcd;
	struct usb_hcd *hcd;
	int status;

	DBG(2, "<==\n");

	shcd = control_to_shcd(h_control);
	hcd = shcd_to_hcd(shcd);

	if (hcd) {
		hcd->driver->reset(hcd);
	}

	if (hcd) {
		status = usb_add_hcd(hcd, 0, 0);
		if (status < 0) {
			return status;
		}
	}

	DBG(2,"==>\n");
	return 0;
}

static int scd_host_intf_stop(struct scd_control *h_control)
{
	struct scd_hcd *shcd;
	struct usb_hcd *hcd;

	DBG(2, "<==\n");

	shcd = control_to_shcd(h_control);
	hcd = shcd_to_hcd(shcd);

	if (hcd && hcd->state != HC_STATE_HALT) {
		/* rmmodƤʤ */
		usb_remove_hcd(hcd);
	}

	DBG(2,"==>\n");
	return 0;
}

irqreturn_t scd_host_intf_irq(struct scd_control *h_control)
{
	struct scd_hcd *shcd;
	struct usb_hcd *hcd;
	struct scd_host *shost;
	struct scd_device *sdev;
	irqreturn_t ret;
	u32 gintsts;
	u32 gotgint;
	u32 haint;
	u32 hprt;

	DBG(2, "<==\n");

	ret = IRQ_NONE;

	shcd = control_to_shcd(h_control);
	hcd = shcd_to_hcd(shcd);
	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;

	spin_lock(&shost->lock);
	gintsts = SCD__READ(sdev, GINTSTS);
	if (gintsts & SCD_B_GINTSTS_WKUPINT) {
		DBG(3, "GINTSTS_WKUPINT\n");
		ret = scd_hub_irq(hcd, 0);

		goto irq_end;
	}
	/* OTG driver handle SCD_B_GINTSTS_SESSREQINT, Host do not care. */
	else if (gintsts & SCD_B_GINTSTS_DISCONNINT) {
		DBG(3, "GINTSTS_DISCONNINT\n");
		ret = scd_hub_irq(hcd, 0);

		goto irq_end;
	}
	/* OTG driver handle SCD_B_GINTSTS_CONIDSTSCHNG, Host do not care. */
	else if (gintsts & SCD_B_GINTSTS_PTXFEMP) {
		/* SLAVE MODE ONLY */
	}
	else if (gintsts & SCD_B_GINTSTS_HCHINT) {
		goto irq_haint;
	}
	else if (gintsts & SCD_B_GINTSTS_PRTINT) {
		goto irq_hprt;

	}
	else if (gintsts & SCD_B_GINTSTS_INCOMPLP) {
		DBG(3, "GINTSTS_INCOMPLP\n");
		/* During construction */
		SCD__WRITE(sdev, GINTSTS, SCD_B_GINTSTS_INCOMPLP);

		ret = IRQ_HANDLED;
		goto irq_end;
	}
	else if (gintsts & SCD_B_GINTSTS_NPTXFEMP) {
		/* SLAVE MODE ONLY */
	}
	else if (gintsts & SCD_B_GINTSTS_RXFLVL) {
		DBG(3, "GINTSTS_RXFLVL\n");
		/* SLAVE MODE ONLY */
		SCD__WRITE(sdev, GINTSTS, SCD_B_GINTSTS_RXFLVL);

		ret = IRQ_HANDLED;
		goto irq_end;
	}
	else if (gintsts & SCD_B_GINTSTS_SOF) {
		DBG(3, "GINTSTS_SOF\n");
	}
	else if (gintsts & SCD_B_GINTSTS_OTGINT) {
		goto irq_gotgint;
	}
	/* OTG driver handle SCD_B_GINTSTS_MODEMIS, Host do not care. */

irq_gotgint:
	gotgint = SCD__READ(sdev, GOTGINT);
	if (gotgint & SCD_B_GOTGINT_DBNCEDONE) {
		DBG(3, "GINTSTS_DBNCEDONE\n");
		/* During construction */
		SCD__WRITE(sdev, GOTGINT, SCD_B_GOTGINT_DBNCEDONE);

		ret = IRQ_HANDLED;
		goto irq_end;
	}
	else if (gotgint & SCD_B_GOTGINT_ADEVTOUTCHG) {
		DBG(3, "GINTSTS_ADEVTOUTCHG\n");
		/* During construction */
		SCD__WRITE(sdev, GOTGINT, SCD_B_GOTGINT_ADEVTOUTCHG);

		ret = IRQ_HANDLED;
		goto irq_end;
	}

irq_hprt:
	hprt = SCD__READ(sdev, HPRT);
	if (hprt & SCD_B_HPRT_PRTOVRCURRCHNG) {
		DBG(3, "HPRT_PRTOVRCURRCHNG\n");
		/* During construction */
		ret = scd_hub_irq(hcd, 0);
		goto irq_end;
	}
	else if (hprt & SCD_B_HPRT_PRTENCHNG) {
		DBG(3, "HPRT_PRTENCHNG\n");
		/* During construction */
		ret = scd_hub_irq(hcd, 0);

		goto irq_end;
	}
	else if (hprt & SCD_B_HPRT_PRTCONNDET) {
		DBG(3, "HPRT_PRTCONNDET\n");
		/* During construction */
		ret = scd_hub_irq(hcd, 0);

		goto irq_end;
	}

irq_haint:
	haint = SCD__READ(sdev, HAINT);
	if (haint & SCD_M_HAINT_HAINT) {
		DBG(3, "HAINT_HAINT\n");
		ret = scd_hep_irq(hcd);

		goto irq_end;
	}

	DBG(1,"IRQ_NONE scd gintsts=0x%x gotgint=0x%x\n",
		 gintsts, gotgint);

irq_end:
	spin_unlock(&shost->lock);

	DBG(2,"==>\n");
	return ret;
}

static void scd_host_intf_overcurrent(struct scd_control *h_control, int port)
{
	struct scd_hcd *shcd;
	struct usb_hcd *hcd;

	DBG(2, "<==\n");

	shcd = control_to_shcd(h_control);
	hcd = shcd_to_hcd(shcd);
	scd_hub_overcurrent(hcd, port);

	DBG(2,"==>\n");
}

static struct scd_intf scd_host_intf = {
	.type = SCD_CONTROL_TYPE_HCD,
	.probe = scd_host_intf_probe,
	.remove = scd_host_intf_remove,
	.start = scd_host_intf_start,
	.stop = scd_host_intf_stop,
	.irq = scd_host_intf_irq,
	.overcurrent = scd_host_intf_overcurrent,
};


/* ؿ - kernel module managerƤФ */
int __init usb_scd_hcd_init(void)
{
	usb_scd_device_set_intf(&scd_host_intf);

	return 0;
}
module_init(usb_scd_hcd_init);


/* λؿ */
void __exit usb_scd_hcd_exit(void)
{
	usb_scd_device_unset_intf(&scd_host_intf);
}
module_exit(usb_scd_hcd_exit);

MODULE_DESCRIPTION(DRIVER_HOST_INFO);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_LICENSE("GPL");
