/*-----------------------------------------------------------------------------
 * drivers/usb/scd/scd_host.h
 * 
 * Host common definitions
 * 
 * Copyright 2011, 2012 Sony Corporation
 *
 * This file is part of the HS-OTG Controller Driver SCD.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *---------------------------------------------------------------------------*/

#ifndef _USB_SCD_HOST_H_
#define _USB_SCD_HOST_H_

#define SCD_NUM_HEPS		5 /* Number of Hardware Host endpoints */
#define SCD_NUM_HEPS_HCD	4 /* Number of hcd driver endpoints */
#define SCD_HEP_EP0_IDX		0
#define STATUS_PKT_MAXP		64
#define SCD_RESET_TIME		50000 /* us */
#define SCD_RESUME_TIME		20000 /* us */

struct scd_hcd {
	struct scd_control h_control;

	struct device *dev;
	struct usb_hcd *hcd;

	spinlock_t lock;
};

struct scd_hep {
	u32 state;
	spinlock_t lock;
	struct scd_dma_channel *dmach;
	struct urb *urb;
	struct usb_device *dev;
	struct usb_host_endpoint *hep;
	u32 attributes;
	u32 pid;
	u32 retrycount;
	u8 *packet_buffer;
	dma_addr_t packet_dma;
	u8 status_packet[STATUS_PKT_MAXP];
	dma_addr_t status_dma;
};

#define SCD_HEP_ATR_IOREMAP	(1 << 0)
#define SCD_HEP_ATR_NO_DMA	(1 << 1)

#define SCD_HEP_ATR_IOREMAP	(1 << 0)

struct scd_port {
	u32 state;
	spinlock_t lock;
	struct usb_port_status status;
	struct scd_timer* timer;
};

#define SCD_NUM_HUB_PORT	1

struct scd_hub {
	struct scd_port port[SCD_NUM_HUB_PORT];
};

struct scd_host_ops {
	struct usb_hcd *(*create_hcd)(struct scd_device *);
	int (*add_hcd)(struct usb_hcd *, unsigned int, unsigned long);
	void (*remove_hcd)(struct usb_hcd *);
	void (*put_hcd)(struct usb_hcd *);
	void (*notify_over_current)(struct scd_port *);
	irqreturn_t (*link_irq)(int, void *);
};

struct scd_host {
	struct scd_device *sdev;
	struct scd_hcd *shcd;

	u8 nep_in;
	u8 nep_out;

	struct scd_hub hub;
	struct scd_hep ep0;
	struct scd_hep ep_in[SCD_NUM_HEPS];
	struct scd_hep ep_out[SCD_NUM_HEPS];

	spinlock_t lock;
	u16 rx_fifo_siz;
	u16 tx_fifo_size[2];  /* htx_npfifo_siz  and htx_pfifo_siz */
	u16 tx_fifo_addr[2];  /* htx_npfifo_addr and htx_pfifo_addr */
};

/* functions */

static inline struct scd_host *hcd_to_shost(struct usb_hcd *hcd)
{
	return (struct scd_host *)(hcd->hcd_priv);
}

static inline struct usb_hcd *shost_to_hcd(struct scd_host *shost)
{
	return container_of((void *)shost, struct usb_hcd, hcd_priv);
}

static inline struct scd_hub *port_to_hub(struct scd_port *port)
{
	return container_of((void *)port, struct scd_hub, port);
}

static inline struct scd_host *hub_to_shost(struct scd_hub *hub)
{
	return container_of((void *)hub, struct scd_host, hub);
}

/* scd_host */
extern irqreturn_t scd_host_link_irq(int irq, void *data);

/* scd_hub */
extern int scd_hub_status_data (struct usb_hcd *hcd, char *buf);
extern int scd_hub_control(struct usb_hcd *hcd, u16 typeReq, u16 wValue,
			   u16 wIndex, char *buf, u16 wLength);
extern int scd_hub_bus_suspend(struct usb_hcd *);
extern int scd_hub_bus_resume(struct usb_hcd *);
extern irqreturn_t scd_hub_irq(struct usb_hcd*, u8);
extern void scd_hub_overcurrent(struct usb_hcd *, int port_num);

/* scd_port */
extern void scd_port_init(struct scd_port *port);
extern void scd_port_notify_over_current(struct scd_port* port);

/* scd_hep */
extern void scd_hep_set_atr(struct scd_port *port, u8 epadr, u32 attribute);
extern void scd_hep_clear_atr(struct scd_port *port, u8 epadr, u32 attribute);
extern void scd_hep_cntl_init(struct usb_hcd *hcd);
extern void scd_hep_tx_init(struct usb_hcd *hcd);
extern void scd_hep_rx_init(struct usb_hcd *hcd);
extern void scd_hep_ch0_init(struct usb_hcd *hcd);
extern void scd_hep_tx_flush_fifo(struct usb_hcd *hcd, struct scd_hep *ep);
extern void scd_hep_rx_flush_fifo(struct usb_hcd *hcd, struct scd_hep *ep);
extern irqreturn_t scd_hep_irq(struct usb_hcd *hcd);
extern int scd_hep_urb_enqueue(struct usb_hcd *hcd,
			       struct urb *urb, gfp_t mem_flags);
extern int scd_hep_urb_dequeue(struct usb_hcd *hcd, struct urb *urb, int status);
extern void scd_hep_endpoint_disable(struct usb_hcd *hcd,
				     struct usb_host_endpoint *ep);
enum usb_device_speed port_get_speed(struct scd_port *port);

#endif /* _USB_SCD_HOST_H_ */
