/*-----------------------------------------------------------------------------
 * drivers/usb/scd/scd_hub.c
 * 
 * Host hub functions
 * 
 * Copyright 2011 Sony Corporation
 *
 * This file is part of the HS-OTG Controller Driver SCD.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/module.h>
#include <linux/device.h>
#include <linux/memory.h>
#include <linux/interrupt.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb/gadget.h>
#include <linux/usb/scd/usb_otg_control.h>
#include <linux/usb/scd/scd_hub.h>
#include "../core/hub.h"
#include "../core/hcd.h"

#include "scd.h"
#include "scd_debug.h"
#include "scd_host.h"
#include "scd_hfifo.h"
#include "scd_register.h"

/* proto type of internal functions */
static int port_trans_state(struct scd_port *port, u32 trigger);

/* hub states and triggers */
enum {
	SCD_PORT_S_POWER_OFF = 0,
	SCD_PORT_S_DISCONNECTED,
	SCD_PORT_S_DISABLED,
	SCD_PORT_S_RESETTING,
	SCD_PORT_S_ENABLED,
	SCD_PORT_S_SUSPENDED,
	SCD_PORT_S_RESUMING,
};

enum {
	SCD_PORT_T_CLEAR_PORT_ENABLE,
	SCD_PORT_T_CLEAR_PORT_SUSPEND,
	SCD_PORT_T_CLEAR_PORT_POWER,
	SCD_PORT_T_SET_PORT_SUSPEND,
	SCD_PORT_T_SET_PORT_POWER,
	SCD_PORT_T_SET_PORT_RESET,
	SCD_PORT_T_RESET_END,
	SCD_PORT_T_RESUME_END,
	SCD_PORT_T_DETECT_REMOTEWAKEUP,
	SCD_PORT_T_DETECT_CONNECT,
	SCD_PORT_T_DETECT_DISCONNECT,
};

/* descriptor for root hub */
static const struct usb_hub_descriptor hub_desc = {
	.bDescLength = 9,
	.bDescriptorType = 0x29,
	.bNbrPorts = 1,
	.wHubCharacteristics = (0x1 << 0 | 0x0 << 2 | 0x1 << 3 |
				0x0 << 5 | 0x0 << 7),
	.bPwrOn2PwrGood = 10,
	.bHubContrCurrent = 0,
	.DeviceRemovable[0] = 0 << 1,
	.PortPwrCtrlMask[0] = 0xff,
};

/* port speed */
#define SCD_PORT_HS 0x0
#define SCD_PORT_FS 0x1
#define SCD_PORT_LS 0x2

/* PHY clock for port frame interval */
#define SCD_PHY_CLOCK_FSLS 30000
#define SCD_PHY_CLOCK_HS   (SCD_PHY_CLOCK_FSLS/8)


/* internal functions */
inline static bool is_port_connected(struct scd_port *port)
{
	return (port->state > SCD_PORT_S_DISCONNECTED);
}

enum usb_device_speed port_get_speed(struct scd_port *port)
{
	struct scd_hub *hub;
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 hprt;
	u8 speed;

	DBG(2, "<==\n");

	hub = port_to_hub(port);
	shost = hub_to_shost(hub);
	sdev = shost->sdev;

	hprt = SCD__READ(sdev, HPRT);
	speed = (hprt >> SCD_S_HPRT_PRTSPD) & SCD_M_HPRT_PRTSPD;

	if (speed == SCD_PORT_HS) {
		return USB_SPEED_HIGH;
	}
	else if (speed == SCD_PORT_FS) {
		return USB_SPEED_FULL;
	}
	else if (speed == SCD_PORT_LS) {
		return USB_SPEED_LOW;
	}

	DBG(2,"==>\n");
	return USB_SPEED_UNKNOWN;
}

static void port_frame_interval(struct scd_port *port, 
                                enum usb_device_speed speed)
{
	struct scd_hub *hub;
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 hfir  = SCD_PHY_CLOCK_FSLS;

	DBG(2, "<==\n");

	hub = port_to_hub(port);
	shost = hub_to_shost(hub);
	sdev = shost->sdev;

	if (speed == USB_SPEED_LOW || speed == USB_SPEED_FULL) {
		hfir = SCD_PHY_CLOCK_FSLS;
	}
	else if (speed == USB_SPEED_HIGH) {
		hfir = SCD_PHY_CLOCK_HS;
	}
	/* Default PHY clock is LS/FS */

	SCD__WRITE( sdev, HFIR, hfir );

	DBG(2,"==>\n");
}

static void port_clear_power(struct scd_port *port)
{
	struct scd_hub *hub;
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 gintmsk, hprt;
	int root_hub_port = 0;

	DBG(2, "<==\n");

	hub = port_to_hub(port);
	shost = hub_to_shost(hub);
	sdev = shost->sdev;

	/* Unsetting PortPower */
	hprt = SCD__READ(sdev, HPRT);
	hprt &= ~(SCD_B_HPRT_PRTOVRCURRCHNG |
		SCD_B_HPRT_PRTENCHNG |
		SCD_B_HPRT_PRTENA |
		SCD_B_HPRT_PRTCONNDET);	
	hprt &= ~(SCD_B_HPRT_PRTPWR);
	SCD__WRITE( sdev, HPRT, hprt );

	scd_device_power(sdev, false, root_hub_port);

	/* No chang HCFG.FSLSSupp */

	/* Mask GINTMSK.PrtInt */
	gintmsk = SCD__READ(sdev, GINTMSK);
	gintmsk &= ~(SCD_B_GINTMSK_PRTINTMSK);
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	DBG(2,"==>\n");
}

static void port_set_power(struct scd_port *port)
{
	struct scd_hub *hub;
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 gintmsk, gintsts, hcfg, hprt;
	int root_hub_port = 0;

	DBG(2, "<==\n");

	hub = port_to_hub(port);
	shost = hub_to_shost(hub);
	sdev = shost->sdev;

	/* Unmask GINTMSK.PrtInt */
	gintmsk = SCD__READ(sdev, GINTMSK);
	gintmsk |= SCD_B_GINTMSK_PRTINTMSK;
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	/* Setting HCFG.FSLSSupp to HS/FS/LS */
	gintsts = SCD__READ(sdev, GINTSTS);
	if(gintsts & SCD_B_GINTSTS_CURMOD) {
		hcfg = SCD__READ(sdev, HCFG);
		if (test_bit(SCD_DEVICE_ATTR_HS_ENABLE, &sdev->attributes)) {
			hcfg &= ~(SCD_B_HCFG_FSLSSUPP);
		}
		else {
			hcfg |= SCD_B_HCFG_FSLSSUPP;
		}
		SCD__WRITE(sdev, HCFG, hcfg);
	}

	/* Setting PortPower */
	hprt = SCD__READ(sdev, HPRT);
	hprt &= ~(SCD_B_HPRT_PRTOVRCURRCHNG |
		SCD_B_HPRT_PRTENCHNG |
		SCD_B_HPRT_PRTENA |
		SCD_B_HPRT_PRTCONNDET);	
	hprt |= SCD_B_HPRT_PRTPWR;
	SCD__WRITE(sdev, HPRT, hprt);
	scd_device_power(sdev, true, root_hub_port);

	DBG(2,"==>\n");
}

static int port_expire_reset(struct scd_timer *timer)
{
	struct scd_port *port = (struct scd_port *)timer->id;

	DBG(2, "<==\n");
	port_trans_state(port, SCD_PORT_T_RESET_END);

	DBG(2,"==>\n");
	return 0;
}

static void port_start_reset(struct scd_port *port)
{
	struct scd_hub *hub;
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 gintmsk, hprt;

	DBG(2, "<==\n");

	hub = port_to_hub(port);
	shost = hub_to_shost(hub);
	sdev = shost->sdev;

	hprt = SCD__READ(sdev, HPRT);

	/* Mask GINTMSK.PrtInt */
	gintmsk = SCD__READ(sdev, GINTMSK);
	gintmsk &= ~(SCD_B_GINTMSK_PRTINTMSK);
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	/* If it is necessary. */
	/* Driver transmits Resume of short time 
	   before the Reset transmission, if it is Suspend mode. */

	hprt &= ~(SCD_B_HPRT_PRTOVRCURRCHNG |
		SCD_B_HPRT_PRTENCHNG |
		SCD_B_HPRT_PRTENA |
		SCD_B_HPRT_PRTCONNDET);
	hprt |= SCD_B_HPRT_PRTRST;
	SCD__WRITE(sdev, HPRT, hprt);

	port->timer = scd_device_add_timer(sdev, SCD_RESET_TIME, 
		port_expire_reset, (void *)port);

	DBG(2,"==>\n");
}

static void port_stop_reset(struct scd_port *port)
{
	struct scd_hub *hub;
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 gintmsk, hprt;

	DBG(2, "<==\n");

	hub = port_to_hub(port);
	shost = hub_to_shost(hub);
	sdev = shost->sdev;

	hprt = SCD__READ(sdev, HPRT);
	hprt &= ~(SCD_B_HPRT_PRTOVRCURRCHNG |
		SCD_B_HPRT_PRTENCHNG |
		SCD_B_HPRT_PRTENA |
		SCD_B_HPRT_PRTCONNDET);
	hprt &= ~(SCD_B_HPRT_PRTRST);
	SCD__WRITE(sdev, HPRT, hprt);

	/* Unmask GINTMSK.PrtInt */
	gintmsk = SCD__READ(sdev, GINTMSK);
	gintmsk |= SCD_B_GINTMSK_PRTINTMSK;
	SCD__WRITE(sdev, GINTMSK, gintmsk);

	DBG(2,"==>\n");
}

static void port_set_suspend(struct scd_port *port)
{
	struct scd_hub *hub;
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 hprt, pcgctl;

	DBG(2, "<==\n");

	hub = port_to_hub(port);
	shost = hub_to_shost(hub);
	sdev = shost->sdev;

	/* Setting port susupend */
	hprt = SCD__READ(sdev, HPRT);
	hprt &= ~(SCD_B_HPRT_PRTOVRCURRCHNG |
		SCD_B_HPRT_PRTENCHNG |
		SCD_B_HPRT_PRTENA |
		SCD_B_HPRT_PRTCONNDET);
	hprt |= SCD_B_HPRT_PRTSUSP;
	SCD__WRITE(sdev, HPRT, hprt);

	/* Setting PHY clock stop */
	pcgctl = SCD__READ(sdev, PCGCCTL);
	pcgctl |= SCD_B_PCGCCTL_STOPPCLK;
	SCD__WRITE(sdev, PCGCCTL, pcgctl);

	DBG(2,"==>\n");
}

static int port_expire_resume(struct scd_timer *timer)
{
	struct scd_port *port = (struct scd_port *)timer->id;

	DBG(2, "<==\n");
	port_trans_state(port, SCD_PORT_T_RESUME_END);

	DBG(2,"==>\n");
	return 0;
}

static void port_start_resume(struct scd_port *port)
{
	struct scd_hub *hub;
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 hprt, pcgctl;

	DBG(2, "<==\n");

	hub = port_to_hub(port);
	shost = hub_to_shost(hub);
	sdev = shost->sdev;

	/* Setting PHY clock start */
	pcgctl = SCD__READ(sdev, PCGCCTL);
	pcgctl &= ~(SCD_B_PCGCCTL_STOPPCLK);
	SCD__WRITE(sdev, PCGCCTL, pcgctl);

	/* Setting port resume */
	hprt = SCD__READ(sdev, HPRT);
	hprt &= ~(SCD_B_HPRT_PRTOVRCURRCHNG |
		SCD_B_HPRT_PRTENCHNG |
		SCD_B_HPRT_PRTENA |
		SCD_B_HPRT_PRTCONNDET);
	hprt |= SCD_B_HPRT_PRTRES;
	SCD__WRITE(sdev, HPRT, hprt);

	port->timer = scd_device_add_timer(sdev, SCD_RESUME_TIME, 
		port_expire_resume, (void *)port);

	DBG(2,"==>\n");
}

static void port_stop_resume(struct scd_port *port)
{
	struct scd_hub *hub;
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 hprt;

	DBG(2, "<==\n");

	hub = port_to_hub(port);
	shost = hub_to_shost(hub);
	sdev = shost->sdev;

	/* Clear port resume */
	hprt = SCD__READ(sdev, HPRT);
	hprt &= ~(SCD_B_HPRT_PRTOVRCURRCHNG |
		SCD_B_HPRT_PRTENCHNG |
		SCD_B_HPRT_PRTENA |
		SCD_B_HPRT_PRTCONNDET);
	hprt &= ~(SCD_B_HPRT_PRTRES);
	SCD__WRITE(sdev, HPRT, hprt);

	DBG(2,"==>\n");
}

static inline void port_clear_c_flag(struct scd_port *port, u16 flag)
{
	DBG(2, "<==\n");
	port->status.wPortChange &= ~(cpu_to_le16(flag));
	DBG(2,"==>\n");
}

static void port_set_flag(struct scd_port *port, u16 flag)
{
	u16 le16_flag = cpu_to_le16(flag);

	DBG(2, "<==\n");

	if (!(le16_to_cpu(port->status.wPortStatus) & le16_flag) &&
	    flag <= USB_PORT_STAT_RESET) {
		port->status.wPortChange |= le16_flag;
	}
	port->status.wPortStatus |= le16_flag;

	DBG(2,"==>\n");
}

static void port_clear_flag(struct scd_port *port, u16 flag)
{
	u16 le16_flag = cpu_to_le16(flag);

	DBG(2, "<==\n");

	if ((le16_to_cpu(port->status.wPortStatus)) & le16_flag &&
	    flag <= USB_PORT_STAT_RESET) {
		port->status.wPortChange |= le16_flag;
	}
	port->status.wPortStatus &= ~le16_flag;

	DBG(2,"==>\n");
}

static bool port_check_flag(struct scd_port *port, u16 flag)
{
	u16 le16_flag = cpu_to_le16(flag);

	DBG(2, "<==\n");

	if (((le16_to_cpu(port->status.wPortStatus)) & le16_flag) || 
	    ((le16_to_cpu(port->status.wPortChange)) & le16_flag)) {
		return true;
	}
	DBG(2,"==>\n");

	return false;
}

static int port_trans_state(struct scd_port *port, u32 trigger)
{
	unsigned long flags;
	enum usb_device_speed speed;

	DBG(2, "<==\n");
	DBG(3, "port->state (exit) = %lu\n", (unsigned long)port->state);

	spin_lock_irqsave(&port->lock, flags);

	/* exit state */
	switch (port->state) {
	case SCD_PORT_S_RESETTING:
		switch (trigger) {
		case SCD_PORT_T_RESET_END:
		case SCD_PORT_T_CLEAR_PORT_ENABLE:
		case SCD_PORT_T_SET_PORT_RESET:
		case SCD_PORT_T_DETECT_DISCONNECT:
		case SCD_PORT_T_CLEAR_PORT_POWER:
			port_clear_flag(port, USB_PORT_STAT_RESET);
			port_stop_reset(port);
			break;
		}
		break;
	case SCD_PORT_S_RESUMING:
		switch (trigger) {
		case SCD_PORT_T_RESUME_END:
		case SCD_PORT_T_CLEAR_PORT_ENABLE:
		case SCD_PORT_T_SET_PORT_RESET:
		case SCD_PORT_T_DETECT_DISCONNECT:
		case SCD_PORT_T_CLEAR_PORT_POWER:
			port_stop_resume(port);
			break;
		}
		break;
	}

	/* transition state */
	if (trigger == SCD_PORT_T_CLEAR_PORT_POWER) {
		port->state = SCD_PORT_S_POWER_OFF;
		goto enter_state;
	}
	switch (port->state) {
	case SCD_PORT_S_POWER_OFF:
		if (trigger == SCD_PORT_T_SET_PORT_POWER) {
			port->state = SCD_PORT_S_DISCONNECTED;
			goto enter_state;
		}
		break;
	case SCD_PORT_S_DISCONNECTED:
		if (trigger == SCD_PORT_T_DETECT_CONNECT) {
			port->state = SCD_PORT_S_DISABLED;
			goto enter_state;
		}
		break;
	case SCD_PORT_S_DISABLED:
	case SCD_PORT_S_RESETTING:
	case SCD_PORT_S_ENABLED:
	case SCD_PORT_S_SUSPENDED:
	case SCD_PORT_S_RESUMING:
		if (trigger == SCD_PORT_T_DETECT_DISCONNECT) {
			port->state = SCD_PORT_S_DISCONNECTED;
			goto enter_state;
		}
		else if (trigger == SCD_PORT_T_SET_PORT_RESET) {
			port->state = SCD_PORT_S_RESETTING;
			goto enter_state;
		}
		else if (trigger == SCD_PORT_T_CLEAR_PORT_ENABLE) {
			port->state = SCD_PORT_S_DISABLED;
			goto enter_state;
		}
		switch (port->state) {
		case SCD_PORT_S_RESETTING:
			if (trigger == SCD_PORT_T_RESET_END) {
				port->state = SCD_PORT_S_ENABLED;
				goto enter_state;
			}
			break;
		case SCD_PORT_S_ENABLED:
			if (trigger == SCD_PORT_T_SET_PORT_SUSPEND) {
				port->state = SCD_PORT_S_SUSPENDED;
				goto enter_state;
			}
			break;
		case SCD_PORT_S_SUSPENDED:
			if (trigger == SCD_PORT_T_DETECT_REMOTEWAKEUP) {
				port->state = SCD_PORT_S_RESUMING;
				goto enter_state;
			}
			else if (trigger == SCD_PORT_T_CLEAR_PORT_SUSPEND) {
				port->state = SCD_PORT_S_RESUMING;
				goto enter_state;
			}
			break;
		case SCD_PORT_S_RESUMING:
			if (trigger == SCD_PORT_T_RESUME_END) {
				port->state = SCD_PORT_S_ENABLED;
				goto enter_state;
			}
			break;
		}
		break;
	}
	goto error_end;

	/* enter state */
enter_state:
	switch (port->state) {
	case SCD_PORT_S_POWER_OFF:
		port_clear_flag(port, USB_PORT_STAT_CONNECTION);
		port_clear_flag(port, USB_PORT_STAT_ENABLE);
		port_clear_flag(port, USB_PORT_STAT_SUSPEND);
		port_clear_power(port);
		port_clear_flag(port, USB_PORT_STAT_POWER);
		break;
	case SCD_PORT_S_DISCONNECTED:
		port_clear_flag(port, USB_PORT_STAT_CONNECTION);
		port_clear_flag(port, USB_PORT_STAT_ENABLE);
		port_clear_flag(port, USB_PORT_STAT_SUSPEND);
		port_clear_flag(port, USB_PORT_STAT_OVERCURRENT);
		port_set_power(port);
		port_set_flag(port, USB_PORT_STAT_POWER);
		break;
	case SCD_PORT_S_DISABLED:
		port_set_flag(port, USB_PORT_STAT_CONNECTION);
		port_clear_flag(port, USB_PORT_STAT_ENABLE);
		port_clear_flag(port, USB_PORT_STAT_SUSPEND);
		break;
	case SCD_PORT_S_RESETTING:
		port_clear_flag(port, USB_PORT_STAT_ENABLE);
		port_clear_flag(port, USB_PORT_STAT_SUSPEND);
		port_start_reset(port);
		port_set_flag(port, USB_PORT_STAT_RESET);
		break;
	case SCD_PORT_S_ENABLED:
		port_set_flag(port, USB_PORT_STAT_ENABLE);
		port_clear_flag(port, USB_PORT_STAT_SUSPEND);
		speed = port_get_speed(port);
		if (speed == USB_SPEED_LOW) {
			port_set_flag(port, USB_PORT_STAT_LOW_SPEED);
			port_clear_flag(port, USB_PORT_STAT_HIGH_SPEED);
		}
		else if (speed == USB_SPEED_HIGH) {
			port_clear_flag(port, USB_PORT_STAT_LOW_SPEED);
			port_set_flag(port, USB_PORT_STAT_HIGH_SPEED);
		}
		else { /* speed == USB_SPEED_FULL or other */
			port_clear_flag(port, USB_PORT_STAT_LOW_SPEED);
			port_clear_flag(port, USB_PORT_STAT_HIGH_SPEED);
		}
		break;
	case SCD_PORT_S_SUSPENDED:
		port_set_flag(port, USB_PORT_STAT_SUSPEND);
		port_set_suspend(port);
		break;
	case SCD_PORT_S_RESUMING:
		port_start_resume(port);
		break;
	default:
		goto error_end;
	}

	spin_unlock_irqrestore(&port->lock, flags);

	DBG(3, "port->state (enter) = %lu\n", (unsigned long)port->state);
	DBG(2,"==>\n");
	return 0;

error_end:
	spin_unlock_irqrestore(&port->lock, flags);

	DBG(3, "port->state (enter) = %lu\n", (unsigned long)port->state);
	DBG(2,"==>\n");
	return -EINVAL;
}

/* operation functions */

void scd_port_init(struct scd_port* port)
{
	DBG(2, "<==\n");
	spin_lock_init(&port->lock);
	port_trans_state(port, SCD_PORT_T_CLEAR_PORT_POWER);
	DBG(2,"==>\n");
}

void scd_port_notify_over_current(struct scd_port* port)
{
	DBG(2, "<==\n");
	port_set_flag(port, USB_PORT_STAT_OVERCURRENT);
	port_trans_state(port, SCD_PORT_T_CLEAR_PORT_POWER);
	DBG(2,"==>\n");
}

irqreturn_t scd_hub_irq(struct usb_hcd* hcd, u8 index)
{
	struct scd_host *shost = hcd_to_shost(hcd);
	struct scd_device *sdev = shost->sdev;
	struct scd_port *port = &shost->hub.port[0];
	int root_hub_port = 0;
	bool connected;
	irqreturn_t ret;
	u32 gintsts, hprt;
	u32 grstctl, gintmsk;

	DBG(2, "<==\n");

	ret = IRQ_NONE;

	/* Call side did spin_lock() */
	gintsts = SCD__READ(sdev, GINTSTS);
	if (gintsts & SCD_B_GINTSTS_WKUPINT) {
		DBG(3, "GINTSTS_WKUPINT\n");
		/* During construction */
		port_trans_state(port, SCD_PORT_T_DETECT_REMOTEWAKEUP);
		SCD__WRITE(sdev, GINTSTS, SCD_B_GINTSTS_WKUPINT);

		ret = IRQ_HANDLED;
		goto irq_end;
	}

	hprt = SCD__READ(sdev, HPRT);
	connected = is_port_connected(port);
	if (hprt & SCD_B_HPRT_PRTCONNDET) {
		DBG(3, "HPRT_PRTCONNDET %snotified\n",
		    hprt & SCD_B_HPRT_PRTCONNSTS && !connected ? "" : "NOT ");
		if (hprt & SCD_B_HPRT_PRTCONNSTS && !connected) {
			scd_device_connect(sdev, true, root_hub_port);
			port_trans_state(port, SCD_PORT_T_DETECT_CONNECT);
		}
		hprt &= ~(SCD_B_HPRT_PRTOVRCURRCHNG |
			SCD_B_HPRT_PRTENCHNG |
			SCD_B_HPRT_PRTENA |
			SCD_B_HPRT_PRTCONNDET);
		hprt |= SCD_B_HPRT_PRTCONNDET;
		SCD__WRITE(sdev, HPRT, hprt);

		ret = IRQ_HANDLED;
		goto irq_end;
	}
	if (hprt & SCD_B_HPRT_PRTENCHNG) {
		DBG(3, "HPRT_PRTENCHNG\n");
		if (hprt & SCD_B_HPRT_PRTENA) {
			enum usb_device_speed speed;
			speed = port_get_speed(port);
			port_frame_interval(port, speed);
		}
		if (!(hprt & SCD_B_HPRT_PRTENA) &&
		    (hprt & SCD_B_HPRT_PRTOVRCURRACT)) {
			if (connected) {
				port_trans_state(port, 
				    SCD_PORT_T_DETECT_DISCONNECT);
			}
			if(port_check_flag(port, USB_PORT_STAT_POWER)) {
				port_trans_state(port, 
				    SCD_PORT_T_CLEAR_PORT_POWER);
			}
		}
		hprt &= ~(SCD_B_HPRT_PRTOVRCURRCHNG |
			SCD_B_HPRT_PRTENCHNG |
			SCD_B_HPRT_PRTENA |
			SCD_B_HPRT_PRTCONNDET);
		hprt |= SCD_B_HPRT_PRTENCHNG;
		SCD__WRITE(sdev, HPRT, hprt);

		ret = IRQ_HANDLED;
		goto irq_end;
	}

	if (gintsts & SCD_B_GINTSTS_DISCONNINT) {
		u32 hcfg;
		u32 grxfsiz;
		u32 gnptxfsiz;
		u32 hptxfsiz;

		grstctl = SCD__READ(sdev, GRSTCTL);
		gintmsk = SCD__READ(sdev, GINTMSK);

		hcfg = SCD__READ(sdev, HCFG);
		grxfsiz = SCD__READ(sdev, GRXFSIZ);
		gnptxfsiz = SCD__READ(sdev, GNPTXFSIZ);
		hptxfsiz = SCD__READ(sdev, HPTXFSIZ);

		DBG(3, "GINTSTS_DISCONNINT %snotified\n",
		    connected ? "" : "NOT ");
		if (connected) {
			scd_device_connect(sdev, false, root_hub_port);
			port_trans_state(port, SCD_PORT_T_DETECT_DISCONNECT);
		}

		SCD__WRITE(sdev, GRSTCTL, grstctl);

		SCD__WRITE(sdev, HCFG, hcfg);
		SCD__WRITE(sdev, GRXFSIZ, grxfsiz);
		SCD__WRITE(sdev, GNPTXFSIZ, gnptxfsiz);
		SCD__WRITE(sdev, HPTXFSIZ, hptxfsiz);

		SCD__WRITE(sdev, GINTMSK, gintmsk);
		SCD__WRITE(sdev, GINTSTS, SCD_B_GINTSTS_DISCONNINT);

		ret = IRQ_HANDLED;
		goto irq_end;
	}

	if (hprt & SCD_B_HPRT_PRTOVRCURRCHNG) {
		DBG(3, "HPRT_PRTOVRCURRCHNG\n");
		if (hprt & SCD_B_HPRT_PRTOVRCURRACT) {
			DBG(1, "usb_scd VbusError\n");
			scd_device_vbuserror(sdev, root_hub_port);
		}
		hprt &= ~(SCD_B_HPRT_PRTOVRCURRCHNG |
			SCD_B_HPRT_PRTENCHNG |
			SCD_B_HPRT_PRTENA |
			SCD_B_HPRT_PRTCONNDET);
		hprt |= SCD_B_HPRT_PRTOVRCURRCHNG;
		SCD__WRITE(sdev, HPRT, hprt);

		ret = IRQ_HANDLED;
		goto irq_end;
	}

irq_end:

	DBG(2,"==>\n");
	return ret;
}

void scd_hub_overcurrent(struct usb_hcd *hcd, int port_num)
{
	struct scd_host *shost;
	struct scd_port *port;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	port = &shost->hub.port[0];

	if(port_check_flag(port, USB_PORT_STAT_POWER)) {
		port_trans_state(port, SCD_PORT_T_CLEAR_PORT_POWER);
	}

	DBG(2,"==>\n");
}

int scd_hub_status_data(struct usb_hcd *hcd, char *buf)
{
	struct scd_host *shost;
	struct scd_port *port;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	port = &shost->hub.port[0];

	if (port->status.wPortChange) {
		buf[0] = 1 << 1;
	}

	DBG(2,"==>\n");
	return 1;		/* return byte */
}

static int scd_hub_control_test(struct usb_hcd *hcd,
			enum usb_otg_control_test_mode test_mode)
{
	DBG(2, "<==\n");

	switch (test_mode) {
	case USB_OTG_TEST_MODE_H_PACKET:
		DBG(2, "set test mode : PACKET\n");
		break;
	case USB_OTG_TEST_MODE_H_J:
		DBG(2, "set test mode : J\n");
		break;
	case USB_OTG_TEST_MODE_H_K:
		DBG(2, "set test mode : K\n");
		break;
	case USB_OTG_TEST_MODE_H_SE0_NAK:
		DBG(2, "set test mode : SE0_NAK\n");
		break;
	case USB_OTG_TEST_MODE_H_HS:
		DBG(2, "set test mode : FORCE_ENABLE\n");
		break;
	default:
		break;
	}

	DBG(2, "==>\n");
	return 0;
}

int scd_hub_device_set_test_mode(struct usb_hcd *hcd,
			     enum usb_otg_control_test_mode test_mode)
{
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 hprt;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;
	hprt = SCD__READ(sdev, HPRT);
	hprt &= ~(SCD_M_HPRT_PRTTSTCTL << SCD_S_HPRT_PRTTSTCTL);

	switch (test_mode) {
	case USB_OTG_TEST_MODE_H_PACKET:
		hprt |= SCD_E_HPRT_PRTTSTCTL_PACKET << SCD_S_HPRT_PRTTSTCTL;
		break;
	case USB_OTG_TEST_MODE_H_K:
		hprt |= SCD_E_HPRT_PRTTSTCTL_K << SCD_S_HPRT_PRTTSTCTL;
		break;
	case USB_OTG_TEST_MODE_H_J:
		hprt |= SCD_E_HPRT_PRTTSTCTL_J << SCD_S_HPRT_PRTTSTCTL;
		break;
	case USB_OTG_TEST_MODE_H_SE0_NAK:
		hprt |= SCD_E_HPRT_PRTTSTCTL_SE0NAK << SCD_S_HPRT_PRTTSTCTL;
		break;
	case USB_OTG_TEST_MODE_H_HS:
		hprt |= SCD_E_HPRT_PRTTSTCTL_FORCE << SCD_S_HPRT_PRTTSTCTL;
		break;
	default:
		return -EINVAL;
	}

	hprt &= ~(SCD_B_HPRT_PRTOVRCURRCHNG |
		SCD_B_HPRT_PRTENCHNG |
		SCD_B_HPRT_PRTENA |
		SCD_B_HPRT_PRTCONNDET);	
	SCD__WRITE(sdev, HPRT, hprt);

	DBG(2, "==>\n");
	return 0;
}

static enum usb_otg_control_test_mode 
scd_hub_device_get_test_mode(struct usb_hcd *hcd)
{
	enum usb_otg_control_test_mode test_mode = 0;
	struct scd_host *shost;
	struct scd_device *sdev;
	u32 hprt;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;
	hprt = SCD__READ(sdev, HPRT);
	hprt &= (SCD_M_HPRT_PRTTSTCTL << SCD_S_HPRT_PRTTSTCTL);

	if (hprt == SCD_E_HPRT_PRTTSTCTL_PACKET << SCD_S_HPRT_PRTTSTCTL) {
		test_mode = USB_OTG_TEST_MODE_H_PACKET;
	} else if (hprt == SCD_E_HPRT_PRTTSTCTL_K << SCD_S_HPRT_PRTTSTCTL) {
		test_mode = USB_OTG_TEST_MODE_H_K;
	} else if (hprt == SCD_E_HPRT_PRTTSTCTL_J << SCD_S_HPRT_PRTTSTCTL) {
		test_mode = USB_OTG_TEST_MODE_H_J;
	} else if (hprt == SCD_E_HPRT_PRTTSTCTL_SE0NAK << SCD_S_HPRT_PRTTSTCTL) {
		test_mode = USB_OTG_TEST_MODE_H_SE0_NAK;
	} else if (hprt == SCD_E_HPRT_PRTTSTCTL_FORCE << SCD_S_HPRT_PRTTSTCTL) {
		test_mode = USB_OTG_TEST_MODE_H_HS;
	} else if (hprt) {
		/* error */
		test_mode = 0;
	}

	DBG(2, "==>\n");
	return test_mode;
}

enum usb_otg_control_test_mode
scd_hub_get_test_mode(struct usb_hcd *hcd)
{
	enum usb_otg_control_test_mode test_mode;

	DBG(2, "<==\n");
	test_mode = scd_hub_device_get_test_mode(hcd);

	DBG(2, "==>\n");
	return test_mode;
}

static int
scd_hub_set_test_mode(struct usb_hcd *hcd,
	      enum usb_otg_control_test_mode test_mode)
{
	int retval;

	DBG(2, "<==\n");
	scd_hub_control_test(hcd, test_mode);
	retval = scd_hub_device_set_test_mode(hcd, test_mode);

	DBG(2, "==>\n");
	return retval;
}

int scd_hub_control(struct usb_hcd *hcd, u16 typeReq, u16 wValue, u16 wIndex,
		    char *buf, u16 wLength)
{
	struct scd_host *shost;
	struct scd_device *sdev;
	struct scd_port *port;
	int status;
	unsigned long flags;
	u8 testmode;
	const u8 ports = SCD_NUM_HUB_PORT;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	sdev = shost->sdev;
	port = &shost->hub.port[0];

	spin_lock_irqsave(&shost->lock, flags);

	switch (typeReq) {
	case ClearHubFeature:
		switch (wValue) {
		case C_HUB_LOCAL_POWER:
		case C_HUB_OVER_CURRENT:
			/* none */
			break;
		default:
			goto error_end;
		}
		break;
	case ClearPortFeature:
		if (!wIndex || wIndex > ports)
			goto error_end;
		switch (wValue) {
		case USB_PORT_FEAT_ENABLE:
			status = port_trans_state(port,
						  SCD_PORT_T_CLEAR_PORT_ENABLE);
			break;
		case USB_PORT_FEAT_C_ENABLE:
			port_clear_c_flag(port, USB_PORT_STAT_C_ENABLE);
			break;
		case USB_PORT_FEAT_SUSPEND:
			status = port_trans_state(port,
						  SCD_PORT_T_CLEAR_PORT_SUSPEND);
			break;
		case USB_PORT_FEAT_C_SUSPEND:
			port_clear_c_flag(port, USB_PORT_STAT_C_SUSPEND);
			break;
		case USB_PORT_FEAT_POWER:
			status = port_trans_state(port,
						  SCD_PORT_T_CLEAR_PORT_POWER);
			break;
		case USB_PORT_FEAT_C_CONNECTION:
			port_clear_c_flag(port, USB_PORT_STAT_C_CONNECTION);
			break;
		case USB_PORT_FEAT_OVER_CURRENT:
			/* none */
			break;
		case USB_PORT_FEAT_C_OVER_CURRENT:
			port_clear_c_flag(port, USB_PORT_STAT_C_OVERCURRENT);
			break;
		case USB_PORT_FEAT_RESET:
			/* none */
			break;
		case USB_PORT_FEAT_C_RESET:
			port_clear_c_flag(port, USB_PORT_STAT_C_RESET);
			break;
		case USB_PORT_FEAT_INDICATOR:
			port_clear_flag(port, USB_PORT_STAT_INDICATOR);
			break;
		case USB_PORT_FEAT_LOWSPEED:
			/* none */
			break;
		case USB_PORT_FEAT_HIGHSPEED:
			/* none */
			break;
		default:
			goto error_end;
		}
		break;
	case GetHubDescriptor:
		memcpy(buf, &hub_desc, sizeof(hub_desc));
		break;
	case GetHubStatus:
		memset(buf, 0, 4);
		break;
	case GetPortStatus:
		if (!wIndex || wIndex > ports)
			goto error_end;
		memcpy(buf, &port->status, sizeof(port->status));
		break;
	case SetHubFeature:
		goto error_end;
	case SetPortFeature:
		if (!(wIndex & 0xff) || (wIndex & 0xff) > ports)
			goto error_end;
		switch (wValue) {
		case USB_PORT_FEAT_SUSPEND:
			status = port_trans_state(port,
						  SCD_PORT_T_SET_PORT_SUSPEND);
			break;
		case USB_PORT_FEAT_POWER:
			status = port_trans_state(port,
						  SCD_PORT_T_SET_PORT_POWER);
			break;
		case USB_PORT_FEAT_RESET:
			status = port_trans_state(port,
						  SCD_PORT_T_SET_PORT_RESET);
			break;
		case USB_PORT_FEAT_OVER_CURRENT:
			/* none */
			break;
		case USB_PORT_FEAT_INDICATOR:
			port_set_flag(port, USB_PORT_STAT_INDICATOR);
			break;
		case USB_PORT_FEAT_LOWSPEED:
			/* none */
			break;
		case USB_PORT_FEAT_HIGHSPEED:
			/* none */
			break;
		case USB_PORT_FEAT_TEST:
			switch (wIndex >> 8) {
			case 0x1: /* Test_J */
				testmode = USB_OTG_TEST_MODE_H_J;
				break;
			case 0x2: /* Test_K */
				testmode = USB_OTG_TEST_MODE_H_K;
				break;
			case 0x3: /* Test_SE0_NAK */
				testmode = USB_OTG_TEST_MODE_H_SE0_NAK;
				break;
			case 0x4: /* Test_Packet */
				testmode = USB_OTG_TEST_MODE_H_PACKET;
				break;
			case 0x5: /* Test_Force_Enable */
				testmode = USB_OTG_TEST_MODE_H_HS;
				break;
			default:
				goto error_end;
			}
			scd_hub_set_test_mode(hcd, testmode);
		}
		break;
	case ScdClearPortFeature:
		if (!(wIndex & 0xff) || (wIndex & 0xff) > ports)
			goto error_end;
		switch (wValue) {
		case SCD_PORT_FEAT_EP_NO_DMA:
			scd_hep_clear_atr(port, wIndex & 0xff00 >> 8,
					  SCD_HEP_ATR_NO_DMA);
			break;
		default:
			goto error_end;
		}
		break;
	case ScdSetPortFeature:
		if (!(wIndex & 0xff) || (wIndex & 0xff) > ports)
			goto error_end;
		switch (wValue) {
		case SCD_PORT_FEAT_EP_NO_DMA:
			/* Nothing to do */
			break;
		case SCD_PORT_FEAT_TEST_SE0:
			testmode = USB_OTG_TEST_MODE_H_SE0_NAK;
			goto set_testmode;
		case SCD_PORT_FEAT_TEST_J:
			testmode = USB_OTG_TEST_MODE_H_J;
			goto set_testmode;
		case SCD_PORT_FEAT_TEST_K:
			testmode = USB_OTG_TEST_MODE_H_K;
			goto set_testmode;
		case SCD_PORT_FEAT_TEST_PACKET:
			testmode = USB_OTG_TEST_MODE_H_PACKET;
			goto set_testmode;
		case SCD_PORT_FEAT_TEST_FORCE_ENABLE:
			testmode = USB_OTG_TEST_MODE_H_HS;
			goto set_testmode;
		case SCD_PORT_FEAT_TEST_SETUP:
			/* Nothing to do */
			break;
		case SCD_PORT_FEAT_TEST_IN:
			/* Nothing to do */
			break;
set_testmode:
			scd_hub_set_test_mode(hcd, testmode);
			break;
		default:
			goto error_end;
		}
		break;
	default:
		goto error_end;
	}
	spin_unlock_irqrestore(&shost->lock, flags);

	DBG(2,"==>\n");
	return 0;

error_end:
	spin_unlock_irqrestore(&shost->lock, flags);

	return -EPIPE;
}

int scd_hub_bus_suspend(struct usb_hcd *hcd)
{
	DBG(2, "<==\n");
	DBG(2,"==>\n");
	return 0;
}

int scd_hub_bus_resume(struct usb_hcd *hcd)
{
	struct scd_host *shost;
	struct scd_port *port;

	DBG(2, "<==\n");

	shost = hcd_to_shost(hcd);
	port = &shost->hub.port[0];

	port_trans_state(port, SCD_PORT_T_CLEAR_PORT_POWER);

	usb_hcd_resume_root_hub(hcd);

	DBG(2,"==>\n");
	return 0;
}
