/*--------------------------------------------------------------------------*
 * drivers/usb/scd/scd_info.c
 * 
 * procfs and first command parser
 * 
 * Copyright 2011 Sony Corporation
 *
 * This file is part of the HS-OTG Controller Driver SCD.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/interrupt.h>
#include <linux/device.h>

#include <asm/uaccess.h>
#include <linux/proc_fs.h>

#include "scd.h"
#include "scd_debug.h"
#include "scd_register.h"

#define PROC_WRITE_BUFSIZE	50

/* LinkĤʤǥ֥ȤŪ˳ݤ */
struct scd_info_control {
	struct scd_device *sdev;
	struct scd_control control;

	struct proc_dir_entry *proc_entry;
};

#ifdef SCD_STR_LOG_ON
#define MAX_LOG_BUF (1024*4)
#define MAX_TMP_LOG_BUF (128)
static char _log_buf[MAX_LOG_BUF];
static char _tmp_log_buf[MAX_TMP_LOG_BUF];
static u32 _log_idx = 0;
static DEFINE_RAW_SPINLOCK(_log_lock);

static _logcpy(char *p1, char *p2, int len)
{
	memcpy(p1, p2, len);
	_log_idx += len;
	if (_log_idx == MAX_LOG_BUF) {
		_log_idx = 0;
	}
}

int __printk(const char *fmt, ...)
{
	int ret;
	va_list arg;
	int len;
	int rem;
	int clen;
	char *p1;
	char *p2;

	__spin_lock(&_log_lock);

	va_start(arg, fmt);
	ret = vsnprintf(_tmp_log_buf, MAX_TMP_LOG_BUF, fmt, arg);
	va_end(arg);

	len = strnlen(_tmp_log_buf, MAX_TMP_LOG_BUF);
	rem = MAX_LOG_BUF - _log_idx;
	clen = min(len, rem);

	p1 = _log_buf + _log_idx;
	p2 = _tmp_log_buf;
	_logcpy(p1, p2, clen);

	if (len - clen > 0) {
		p1 = _log_buf;
		p2 = _tmp_log_buf + clen;
		_logcpy(p1, p2, len - clen);
	}

	__spin_unlock(&_log_lock);

	return ret;
}
EXPORT_SYMBOL(__printk);
#endif

static u8 scd_infoc_count;

static inline
struct scd_info_control *control_to_info(struct scd_control* control)
{
	return container_of(control, struct scd_info_control, control);
}

static int setup_sysdep(struct scd_device *sdev, const char *fmt, ...)
{
	va_list ap;
	int status;

	va_start(ap, fmt);

	status = sdev->ops->vmessage(sdev,SCD_CONTROL_TYPE_SYSDEP,
				     SCD_MESSAGE_TYPE_SETTING, fmt, ap);

	va_end(ap);

	return status;
}

/*
 * dump memory
 *
 * str_adr : start address
 * size    : size
 * mode    : 1->byte  2->half word  3->word
 */
int scd_mem_dump(struct scd_info_control *infoc, u32 str_adr, u32 size,
                 u8 mode)
{
        unsigned long top, col, end_adr, i;
        unsigned char cpod;
        unsigned char cnt_space[] = {1, 2, 4}; /* modeʸ */

        const char *prt_scale[3] = {
                "\n 0x%08X +0 +1 +2 +3 +4 +5 +6 +7 +8 +9 +A +B +C +D +E +F   0123456789ABCDEF\n",
                "\n 0x%08X   +0   +2   +4   +6   +8   +A   +C   +E   \n",
                "\n 0x%08X       +0       +4       +8       +C       \n"
            };

        if(2 < mode) {
                return -EINVAL;
        }

        end_adr = str_adr + size - 1;

        for (top = str_adr; top <= end_adr; top += 16) {
                /* 256ХOffset()ɽ */
                if (!((top - str_adr) & 0xff)) {
                        printk(prt_scale[mode], top);
                }
                
                /* Offsetɽ() */
                printk("+0x%08X ", (int)(top - str_adr));
                
                /* ǽ饤λϺǽɥ쥹ޤǤɽ */
                for (col = 0; (col < 16) && (top + col <= end_adr);
                     col += (0x01 << mode)) {
                        switch(mode) {
                            case 0:
                                printk("%02X ",
                                          (int)*(unsigned char *)
                                          ((top + col)));
                                break;
                            case 1:
                                printk("%04X ",
                                          (int)*(unsigned short *)
                                          ((top + col)));
                                break;
                            default: /* 2 */
                                printk("%08X ",
                                          (int)*(unsigned long *)
                                          ((top + col)));
                                break;
                        }
                }
                
                /* 16 byte ܿǤʤκǽζɽ */
                if (col != 16) {
                        for (; col < 16; col += (0x01 << mode)) {
                                for (i = 0; i < cnt_space[mode]; i++) {
                                        printk("  ");
                                }
                                /* ɽζڤ */
                                scd_print(infoc->sdev, " ");
                        }
                }
                
                /* byte dump ξϡʸɽ */
                if(mode == 0) {
                        scd_print(infoc->sdev, "  ");
                        for (col = 0; (col < 16) && (top + col <= end_adr);
                             col++) {
                                cpod = *(unsigned char *)(top + col);
                                if ((cpod >= 0x20) && (cpod <= 0x7e)) {
                                        scd_print(infoc->sdev, "%c", cpod);
                                } else {
                                        scd_print(infoc->sdev, ".");
                                }
                        }
                }
                
                scd_print(infoc->sdev, "\n");
        } /* for () */

        return 0;
}

static int proc_read(char *page,
		     char **start, off_t off, int count, int *eof, void *data)
{
#ifdef SCD_STR_LOG_ON
	memcpy(page, _log_buf + _log_idx, MAX_LOG_BUF - _log_idx);
	memcpy(page + MAX_LOG_BUF - _log_idx, _log_buf, _log_idx);

	*eof = 1;

	return MAX_LOG_BUF;
#else
	return 0;
#endif
}

static int proc_write(struct file *file, const char *buffer,
			  unsigned long count, void *data)
{
        int res;
	char *s;
	struct scd_info_control *infoc;
	struct scd_device *sdev;
	char param[PROC_WRITE_BUFSIZE];

	infoc = (struct scd_info_control *)data;
	sdev = infoc->sdev;

	/* ޥʸκǾͥå */
	count = min(count, (unsigned long)PROC_WRITE_BUFSIZE - 1);
        /* User֤饳ԡ */
	if (copy_from_user(param, buffer, count)) {
		res = -EFAULT;
		goto exit;
	}
	param[count] = '\0'; /* ʸü */
	s = strnchr(param, count, 0xa);	/* LF */
	if (s) {
		*s = '\0';
	}
	res = count;

#define CMPCMD(FMT, CMD) strncmp((FMT), (CMD), sizeof(CMD) - 1)
	if (!CMPCMD(param, "setphy")) {
		setup_sysdep(infoc->sdev, param);
		res = 0;
	}
	else if (!CMPCMD(param, ":")) {
		unsigned long base = (unsigned long)infoc->sdev->regs;
		switch (param[1]) {
		case 'h':	/* ":h" host regs */
			res = scd_mem_dump(infoc, base + SCD_R_HCFG,
					   SCD_R_HCCHAR(16) - SCD_R_HCFG, 2);
			break;
		case 'd':	/* ":d" device regs */
			res = scd_mem_dump(infoc, base + SCD_R_DCFG,
					   SCD_R_DOEPCTL(16) - SCD_R_DCFG, 2);
			break;
		case 'p':	/* ":p" power regs */
			res = scd_mem_dump(infoc, base + SCD_R_PCGCCTL, 0x4, 2);
			break;
		default:	/*  global regs */
			res = scd_mem_dump(infoc, base, SCD_R_GRXSTSP, 2);
			/* GRXSTSPRead POP뤿פϥåפ */
			res = scd_mem_dump(infoc, base + SCD_R_GRXSTSP + 0x4,
					   SCD_R_DIEPTXF(16) - SCD_R_GRXSTSP, 2);
			break;
		}
	}
	else if (!CMPCMD(param, "s:")) {
		sdev->ops->message(sdev, SCD_CONTROL_TYPE_SYSDEP,
				   SCD_MESSAGE_TYPE_SETTING, &param[2]);
		res = 0;
	}
	else if (!CMPCMD(param, "d:")) {
		sdev->ops->message(sdev, SCD_CONTROL_TYPE_DEVICE,
				   SCD_MESSAGE_TYPE_SETTING, &param[2]);
		res = 0;
	}
	else if (!CMPCMD(param, "g:")) {
		sdev->ops->message(sdev, SCD_CONTROL_TYPE_GADGET,
				   SCD_MESSAGE_TYPE_SETTING, &param[2]);
		res = 0;
	}
	else if (!CMPCMD(param, "h:")) {
		sdev->ops->message(sdev, SCD_CONTROL_TYPE_HCD,
				   SCD_MESSAGE_TYPE_SETTING, &param[2]);
		res = 0;
	}
	else {
		scd_error(infoc->sdev, "Command %s not implemented\n", param);
		res = 0;
	}

    exit:
	return res;
}

/*
 *	scd intf ops
 */

static struct scd_control *probe(struct scd_device *sdev)
{
	struct scd_info_control *infoc;

	infoc = (struct scd_info_control *)kzalloc(sizeof(*infoc), GFP_KERNEL);
	if (infoc == NULL) {
		printk("%s:ERR:Can't alloc memory for scd_info\n",
		       dev_name(sdev->dev));
		return NULL;
	}
	infoc->sdev = sdev;

	infoc->proc_entry = create_proc_entry(dev_name(sdev->dev),
					      S_IFREG | S_IRUGO | S_IWUSR, NULL);
	if (!infoc->proc_entry) {
		scd_error(infoc->sdev,
			  "Cannot create a vaild proc file entry\n");
		kfree(infoc);
		return NULL;
	}

	infoc->proc_entry->data = infoc;
	infoc->proc_entry->owner = THIS_MODULE;
	infoc->proc_entry->read_proc = proc_read;
	infoc->proc_entry->write_proc = proc_write;
	infoc->proc_entry->size = 0;

	printk("%s:INFO:Registered /proc/%s\n",
	       dev_name(sdev->dev), infoc->proc_entry->name);

#ifdef SCD_STR_LOG_ON
	memset(_log_buf, '\0', MAX_LOG_BUF);
#endif

	return &infoc->control;
}

static int remove(struct scd_control *control)
{
	struct scd_info_control *infoc;

	infoc = control_to_info(control);

	remove_proc_entry(infoc->proc_entry->name, NULL);
	kfree(infoc);

	return 0;
}

static int vmessage(struct scd_control *control, enum scd_message_type type,
		    const char *fmt, va_list ap)
{
	struct scd_info_control *infoc;
	const static char *type_string[] = {
		"UKN",		/* SCD_MESSAGE_TYPE_UNKNOWN = 0, */
		"ERR",		/* SCD_MESSAGE_TYPE_ERROR, */
		"WRN",		/* SCD_MESSAGE_TYPE_WARN, */
		"INF",		/* SCD_MESSAGE_TYPE_INFO, */
	};

	infoc = control_to_info(control);

	if (type < SCD_MESSAGE_TYPE_SETTING && type >= 0) {
		printk("%s:%s:", dev_name(infoc->sdev->dev), type_string[type]);
	}
	vprintk(fmt, ap);

	return 0;
}

static struct scd_intf info_intf = {
	.type = SCD_CONTROL_TYPE_INFO,
	.probe = probe,
	.remove = remove,

	.vmessage = vmessage,
};

/* ؿ - otgľܸƤФ */
int __init usb_scd_info_init(void)
{
	scd_infoc_count = 0;

	usb_scd_device_set_intf(&info_intf);

	return 0;
}

/* λؿ */
void __exit usb_scd_info_exit(void)
{
	usb_scd_device_unset_intf(&info_intf);
}
