/*--------------------------------------------------------------------------*
 * drivers/usb/scd/scd_sysdep_cxd41xx.c
 * 
 * System dependent definitions
 * 
 * Copyright 2011, 2012 Sony Corporation
 *
 * This file is part of the HS-OTG Controller Driver SCD.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *---------------------------------------------------------------------------*/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/io.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <mach/hardware.h>
#include <mach/hwtimer.h>
#include <mach/gpio.h>
#include <asm/irq.h>
#include <asm/sizes.h>
#include <asm/uaccess.h>
#include <asm/delay.h>


#include "scd.h"
#include "scd_debug.h"
#include "scd_register.h"

/************************** macro, variable definitions **********************/
#define REG_SET			(0x04)
#define REG_CLR			(0x08)
#define REG_STS			(0x0c)

/* CLKRST */
#define VA_CLKRST_XPS_EXT	(VA_CLKRST(32))
#define VA_CLKRST_XPS_EXTSET	(VA_CLKRST_XPS_EXT + REG_SET)
#define VA_CLKRST_XPS_EXTCLR	(VA_CLKRST_XPS_EXT + REG_CLR)
#define XPS_EXT_USBPHY		BIT(4)

/* PER_MISC(PER_CLKRST) */
#define VA_PERCLKRST		(VA_PERMISC + 0x1000)
#define PERCLKRST_USB_AHB	(0x1500)
#define VA_USBAHB_PS		(VA_PERCLKRST + PERCLKRST_USB_AHB)
#define VA_USBAHB_PSSET		(VA_USBAHB_PS + REG_SET)
#define VA_USBAHB_PSCLR		(VA_USBAHB_PS + REG_CLR)
#define VA_USBAHB_PSSTS		(VA_USBAHB_PS + REG_STS)
#define USB_AHBCLK		BIT(0)
#define PERCLKRST_USB_HRST	(0x1510)
#define VA_USBHRST_SOFTRST	(VA_PERCLKRST + PERCLKRST_USB_HRST)
#define VA_USBHRST_SOFTRSTSET	(VA_USBHRST_SOFTRST + REG_SET)
#define VA_USBHRST_SOFTRSTCLR	(VA_USBHRST_SOFTRST + REG_CLR)
#define HRST_SOFTRST		BIT(0)
#define PERCLKRST_USB_PRST	(0x1520)
#define VA_USBPRST_SOFTRST	(VA_PERCLKRST + PERCLKRST_USB_PRST)
#define VA_USBPRST_SOFTRSTSET	(VA_USBPRST_SOFTRST + REG_SET)
#define VA_USBPRST_SOFTRSTCLR	(VA_USBPRST_SOFTRST + REG_CLR)
#define PRST_SOFTRST		BIT(0)

/* PHY(MISCTOP) */
#define ADR_USB_PHY		(0xf3050000)
#define VA_USB_PHY		IO_ADDRESSP(ADR_USB_PHY)
#define USB_PHYRST		(0x0)
#define VA_USB_PHYRST		(VA_USB_PHY + USB_PHYRST)
#define VA_USB_PHYRSTSET	(VA_USB_PHYRST + REG_SET)
#define VA_USB_PHYRSTCLR	(VA_USB_PHYRST + REG_CLR)
#define PHYRST			BIT(0)
#define USB_CLKSEL		(0x10)
#define VA_USB_CLKSEL		(VA_USB_PHY + USB_CLKSEL)
#define VA_USB_CLKSELSET	(VA_USB_CLKSEL + REG_SET)
#define VA_USB_CLKSELCLR	(VA_USB_CLKSEL + REG_CLR)
#define USB_PHYTUNE		(0x20)
#define VA_USB_PHYTUNE		(VA_USB_PHY + USB_PHYTUNE)
#define VA_USB_PHYTUNESET	(VA_USB_PHYTUNE + REG_SET)
#define VA_USB_PHYTUNECLR	(VA_USB_PHYTUNE + REG_CLR)
#define USB_UNLIMIT		(0x30)
#define VA_USB_UNLIMIT		(VA_USB_PHY + USB_UNLIMIT)
#define VA_USB_UNLIMITSET	(VA_USB_UNLIMIT + REG_SET)
#define VA_USB_UNLIMITCLR	(VA_USB_UNLIMIT + REG_CLR)
#define USB_ID_EXT		(0x40)
#define VA_USB_ID_EXT		(VA_USB_PHY + USB_ID_EXT)

/* wait */
#define WAIT_US_T1_T2		(10)	/* us */
#define WAIT_US_T2_T3		(800)	/* us */
#define WAIT_T3_T4		(400)	/* ns : 12clock@30MHz(slowest clock) */
#define WAIT_T4_T5		(200)	/* ns :  6clock@30MHz(slowest clock) */

/* GPIO */
#define OVERCURRENT_GPIO_PORT	(0)
#define OVERCURRENT_GPIO_BIT	(2)
#define OVERCURRENT_GPIO_IRQ	IRQ_GPIO_FALL(OVERCURRENT_GPIO_BIT)

/* PORT */
#define PORT_0			(0)

/* command only sysdep */
struct command_set {
        char *str;
        int (*func)(const char *);
};

static const char phy_def_param[] = "setphy 00658008"; /* 侩 */

static struct scd_sysdep_control {
	struct scd_device *sdev;
	struct scd_control control;
        struct hwtimer timer;
        struct tasklet_struct event_timeout_tasklet;
        u32 phy_param;
} sysdepc;

static inline void phy_clock_on(void);
static inline void phy_clock_off(void);
static inline void ahb_clock_on(void);
static inline void ahb_clock_off(void);
static inline void clock_on(void);
static inline void clock_off(void);
static inline int set_pparam(u32 param);
static char *parse_2_hex(char *str, u32 *value, u8 fix_digit);
static int parse_pparam(const char *param, u32 *pparam);
static int set_phy_parameter(const char *param);
static int parse_cmd(const char *param, char **pstr);
static int setting(const char *fmt);
static irqreturn_t ovc_irq(int irq, void *data);
static inline
struct scd_sysdep_control *control_to_sysdep(struct scd_control* control);
static void timeout_tasklet(unsigned long data);
void timer_function(unsigned long data);
static struct scd_control *probe(struct scd_device *device);
static int remove(struct scd_control *control);
static int suspend(struct scd_control *control, pm_message_t state);
static int resume(struct scd_control *control);
static int start(struct scd_control *control);
static int stop(struct scd_control *control);
static int vmessage(struct scd_control *control, enum scd_message_type type,
		    const char *fmt, va_list ap);
static scd_usec get_time(struct scd_control *control);
static int set_timeout(struct scd_control *control, scd_usec timeout);
static int cancel_timeout(struct scd_control *control);
int power(struct scd_control *control, bool onoff, int port);
static void nop_release(struct device *dev);
int __init usb_scd_sysdep_init(void);
void __exit usb_scd_sysdep_exit(void);

struct command_set cmd_set[] = {
        {"setphy ",	set_phy_parameter},
};

#ifdef SCD_SYSDEP_TEST
#include "scd_sysdep_testintf_cxd41xx.h"
#endif /* SCD_SYSDEP_TEST */

/*----------------------------------------------------------------------*
 * name : phy_clock_on							*
 * desc : PHY clock on 						*
 *----------------------------------------------------------------------*/
static inline void phy_clock_on(void)
{
	u32 clkrst_xps_ext = readl(VA_CLKRST_XPS_EXT);

	if ((clkrst_xps_ext & XPS_EXT_USBPHY)) {
		DBG(2, "WARN : clkrst_xps_ext=%08x\n", (int)clkrst_xps_ext);
	}
        writel(XPS_EXT_USBPHY, VA_CLKRST_XPS_EXTSET);
}

/*----------------------------------------------------------------------*
 * name : phy_clock_off							*
 * desc : PHY clock off 						*
 *----------------------------------------------------------------------*/
static inline void phy_clock_off(void)
{
        writel(XPS_EXT_USBPHY, VA_CLKRST_XPS_EXTCLR);
}

/*----------------------------------------------------------------------*
 * name : ahb_clock_on							*
 * desc : AHB clock on 						*
 *----------------------------------------------------------------------*/
static inline void ahb_clock_on(void)
{
        u32 clk_ahb_status = readl(VA_USBAHB_PSSTS);

        if ((clk_ahb_status & USB_AHBCLK)) {
                DBG(2, "WARN : clk_ahb_status=%08x\n", (int)clk_ahb_status);
        }
        writel(USB_AHBCLK, VA_USBAHB_PSSET);
}

/*----------------------------------------------------------------------*
 * name : ahb_clock_off							*
 * desc : AHB clock off 						*
 *----------------------------------------------------------------------*/
static inline void ahb_clock_off(void)
{
        writel(USB_AHBCLK, VA_USBAHB_PSCLR);
}

/*----------------------------------------------------------------------*
 * name : clock_on							*
 * desc : clock on (HCLK/PHY)					*
 *----------------------------------------------------------------------*/
static inline void clock_on(void)
{
        u32 phy_rst = readl(VA_USB_PHYRST);
        u32 hrst_status = readl(VA_USBHRST_SOFTRST);
        u32 prst_status = readl(VA_USBPRST_SOFTRST);

        ahb_clock_on();
        phy_clock_on();

        udelay(WAIT_US_T1_T2);

	DBG(2, "READ phy_rst=%08x\n", phy_rst);
	if (!(phy_rst & PHYRST)) {
		DBG(2, "WARN : phy_rst=%08x\n", (int)phy_rst);
	}
        writel(PHYRST, VA_USB_PHYRSTCLR);

        udelay(WAIT_US_T2_T3);
        ndelay(WAIT_T3_T4);

        if ((prst_status & PRST_SOFTRST)) {
                DBG(2, "WARN : prst_status=%08x\n", (int)prst_status);
        }
        writel(PRST_SOFTRST, VA_USBPRST_SOFTRSTSET);

        ndelay(WAIT_T4_T5);

	if ((hrst_status & HRST_SOFTRST)) {
                DBG(2, "WARN : hrst_status=%08x\n", (int)hrst_status);
        }
        writel(HRST_SOFTRST, VA_USBHRST_SOFTRSTSET);
}

/*----------------------------------------------------------------------*
 * name : clock_off							*
 * desc : clock off (HCLK/PHY)					*
 *----------------------------------------------------------------------*/
static inline void clock_off(void)
{
        writel(HRST_SOFTRST, VA_USBHRST_SOFTRSTCLR);
        writel(PRST_SOFTRST, VA_USBPRST_SOFTRSTCLR);

        writel(PHYRST, VA_USB_PHYRSTSET);

        phy_clock_off();
        ahb_clock_off();
}

/*----------------------------------------------------------------------*
 * name : set_pparam							*
 * desc : phy parameter Υ쥸					*
 *----------------------------------------------------------------------*/
static inline int set_pparam(u32 param)
{
        writel(param, VA_USB_PHYTUNE);

        DBG(2, "PHYTUNE 0x%08x\n", (int)readl(VA_USB_PHYTUNE));

	return 0;
}

/*----------------------------------------------------------------------*
 * name : parse_2_hex							*
 * desc : 16 ʿ(8 )ؤ parse					*
 *----------------------------------------------------------------------*/
static char *parse_2_hex(char *str, u32 *value, u8 fix_digit)
{
	const char hexchars[] = "0123456789abcdef ";
	u8 i, num;

        if (8 < fix_digit) {
                return NULL;
        }

        /* ƬζФ */
        while (*str == ' ') str++;

        num = 0;
	*value = 0;
        do {
		for (i = 0; hexchars[i] != '\0'; i++) {
                        if (*str == hexchars[i]) {
                                if (*str != ' ') {
                                        *value = (*value << 4) + i;
                                        num++;
                                }
                                break;
			}
		}
                /* եޥåȰȿǧ */
		if (hexchars[i] == '\0') {
                        /*  [0-9a-f]ʳ */
                        ERR("format error(%d)\n", __LINE__);
                        return NULL;
                }
                if (num == 9) {
                        /* 9 (8ʾ) */
                        return NULL;
                }
                /* ڤʸǧ */
                if (*str == ' ') {
                        while (*++str == ' ') ;
                        break;
                }                
        } while (*(++str) != '\0');

        /* եޥåȰȿǧ */
        if ((fix_digit) && (num != fix_digit)) {
                        /* ʳ */
                        ERR("format error(%d)\n", __LINE__);
                        return NULL;
                }

	return str;
}

/*----------------------------------------------------------------------*
 * name : parse_pparam							*
 * desc : phy parameter  parse					*
 *----------------------------------------------------------------------*/
static int parse_pparam(const char *param, u32 *pparam)
{
        int err = 0;
	u32 data;
	char *pstr = (char *)param;

        *pparam = readl(VA_USB_PHYTUNE); /* ߤͤȤꤢ */

        pstr = parse_2_hex(pstr, &data, 8);
        if (pstr == NULL) {
                ERR("parse error. default param will be used.\n");
                
                err = -EINVAL;
        } else {
                DBG(2, "set phy_param: 0x%08x\n", (int)data);
                
                *pparam = data;
        }

        return err;
}

/*----------------------------------------------------------------------*
 * name : set_phy_parameter						*
 * desc : phy parameter 						*
 *----------------------------------------------------------------------*/
static int set_phy_parameter(const char *param)
{
        int err;
        u32 phy_param;

        err = parse_pparam(param, &phy_param);
        if (!err) {
                set_pparam(phy_param);
                sysdepc.phy_param = phy_param; /* resume Ѥݻ */
        }

        return err;
}

/*----------------------------------------------------------------------*
 * name : parse_cmd							*
 * desc : command  parse						*
 *----------------------------------------------------------------------*/
#define MAX_CMD_NUM		(sizeof(cmd_set)/sizeof(struct command_set))
#define CMD_STRING(x)		(cmd_set[(x)].str)
#define CMD_LENGTH(x)		(strlen(cmd_set[(x)].str))
static int parse_cmd(const char *param, char **pstr)
{
        int i;
        int command_number = MAX_CMD_NUM; /* ޥ԰פ */

        *pstr = (char *)param;
        for (i = 0; i < MAX_CMD_NUM; i++) {
                DBG(2, "i(%d):%s:\n", i, CMD_STRING(i));
                if (!strncmp(param, CMD_STRING(i), CMD_LENGTH(i))) {
                        /* פޥɤμʸ֤򥻥å */
                        *pstr = (char *)&param[CMD_LENGTH(i)];
                        DBG(2, "[%d]:%s:\n", i, CMD_STRING(i));
                        DBG(2, "    :%s:\n", pstr);
                        command_number = i;
                        break; /* ޥɰפǽλ */
                }
        }

        return command_number;
}

/*----------------------------------------------------------------------*
 * name : setting							*
 * desc : sysdep ѥåƥ󥰴ؿ					*
 *----------------------------------------------------------------------*/
static int setting(const char *fmt)
{
        int err;
        int cmd_num;
        char *pstr = "";

        cmd_num = parse_cmd(fmt, &pstr);

        if (unlikely(cmd_num == MAX_CMD_NUM)) {
                /* ץޥɤʤ */
#ifdef SCD_SYSDEP_TEST
                err = test_sysdep(fmt);
                if (err != NO_CMD) {
                        return err;
                }
#endif /* SCD_SYSDEP_TEST */
                err = -EINVAL;
                ERR("Command Error\n");
        } else {
                err = cmd_set[cmd_num].func(pstr);
        }

        return err;
}

/*----------------------------------------------------------------------*
 * name : ovc_irq							*
 * desc : over current ߥϥɥ					*
 *----------------------------------------------------------------------*/
static irqreturn_t ovc_irq(int irq, void *data)
{
        struct scd_sysdep_control *sysdepc;

        sysdepc = (struct scd_sysdep_control *)data;

        sysdepc->sdev->ops->overcurrent(sysdepc->sdev, PORT_0);

        return IRQ_HANDLED;
}

/*----------------------------------------------------------------------*
 * name : control_to_sysdep						*
 * desc : scd_sysdep_control ¤μؿ				*
 *----------------------------------------------------------------------*/
static inline
struct scd_sysdep_control *control_to_sysdep(struct scd_control* control)
{
	return container_of(control, struct scd_sysdep_control, control);
}

/*----------------------------------------------------------------------*
 * name : timeout_tasklet						*
 * desc : scd timer - ̤ؤ timer expire  tasklet ؿ		*
 *----------------------------------------------------------------------*/
static void timeout_tasklet(unsigned long data)
{
        struct scd_sysdep_control *sysdepc;
        sysdepc = (struct scd_sysdep_control *)data;

        sysdepc->sdev->ops->timeout(sysdepc->sdev);

        return ;
}

/*----------------------------------------------------------------------*
 * name : timer_function						*
 * desc : scd timer - timer expire  callback ؿ			*
 *----------------------------------------------------------------------*/
void timer_function(unsigned long data)
{
        struct scd_sysdep_control *sysdepc;

        sysdepc = (struct scd_sysdep_control *)data;

        tasklet_schedule(&sysdepc->event_timeout_tasklet);

        return ;
}

/*----------------------------------------------------------------------*
 * name : probe								*
 * desc : scd intf ؿ - probe						*
 *----------------------------------------------------------------------*/
static struct scd_control *probe(struct scd_device *device)
{
	sysdepc.sdev = device;

        hwtimer_init_timer(&sysdepc.timer, timer_function,
                           (unsigned long)&sysdepc, -1);

        tasklet_init(&sysdepc.event_timeout_tasklet,
                     timeout_tasklet,
                     (unsigned long)&sysdepc);

        /* clock  */
        clock_on();

        /* PHY Ĵ(ǥե) */
        setting(phy_def_param);

	return &sysdepc.control;
}

/*----------------------------------------------------------------------*
 * name : remove							*
 * desc : scd intf ؿ - remove					*
 *----------------------------------------------------------------------*/
static int remove(struct scd_control *control)
{
        tasklet_kill(&sysdepc.event_timeout_tasklet);

        /* cpu lock ֤Ǥ API ȯԲ */
        hwtimer_del_timer(&sysdepc.timer);

        /* clock  */
        clock_off();

        sysdepc.sdev = NULL;

	return 0;
}

/*----------------------------------------------------------------------*
 * name : suspend							*
 * desc : scd intf ؿ - suspend					*
 *----------------------------------------------------------------------*/
static int suspend(struct scd_control *control, pm_message_t state)
{
        struct scd_sysdep_control *sysdepc;

	sysdepc = control_to_sysdep(control);

        /* clock  */
        clock_off();

        return 0;
}

/*----------------------------------------------------------------------*
 * name : resume							*
 * desc : scd intf ؿ - resume					*
 *----------------------------------------------------------------------*/
static int resume(struct scd_control *control)
{
        struct scd_sysdep_control *sysdepc;

	sysdepc = control_to_sysdep(control);

        /* clock  */
        clock_on();

        /* PHY Ĵ */
        set_pparam(sysdepc->phy_param);

        return 0;
}

/*----------------------------------------------------------------------*
 * name : start								*
 * desc : scd intf ؿ - start						*
 *----------------------------------------------------------------------*/
static int start(struct scd_control *control)
{
	struct scd_sysdep_control *sysdepc;

	sysdepc = control_to_sysdep(control);

	return 0;
}

/*----------------------------------------------------------------------*
 * name : stop								*
 * desc : scd intf ؿ - stop						*
 *----------------------------------------------------------------------*/
static int stop(struct scd_control *control)
{
	struct scd_sysdep_control *sysdepc;

	sysdepc = control_to_sysdep(control);

	return 0;
}

/*----------------------------------------------------------------------*
 * name : vmessage							*
 * desc : message ؿ							*
 *----------------------------------------------------------------------*/
static int vmessage(struct scd_control *control, enum scd_message_type type,
		    const char *fmt, va_list ap)
{
        struct scd_sysdep_control *sysdepc;
        int err = 0;

	sysdepc = control_to_sysdep(control);

        switch (type) {
            case SCD_MESSAGE_TYPE_SETTING:
                err = setting(fmt);
                break;
            case SCD_MESSAGE_TYPE_ERROR:
            case SCD_MESSAGE_TYPE_WARN:
            case SCD_MESSAGE_TYPE_INFO:
            case SCD_MESSAGE_TYPE_UNKNOWN:
                /* message Ϥ info ή */
                err = sysdepc->sdev->ops->message(sysdepc->sdev,
                                                  SCD_CONTROL_TYPE_INFO,
                                                  type,
                                                  fmt);
            default:
                DBG(2, "Command %c not implemented\n", type);
                break;
        }

	return err;
}

/*----------------------------------------------------------------------*
 * name : get_time							*
 * desc : scd timer θ߻					*
 *----------------------------------------------------------------------*/
static scd_usec get_time(struct scd_control *control)
{
        struct scd_sysdep_control *sysdepc;
        unsigned long usec = 0;

	sysdepc = control_to_sysdep(control);

        usec = hwtimer_read_clock_cpu(sysdepc->timer.cpuid);

        return (scd_usec)usec;
}

/*----------------------------------------------------------------------*
 * name : set_timeout							*
 * desc : scd timer Υॢ					*
 *----------------------------------------------------------------------*/
static int set_timeout(struct scd_control *control, scd_usec timeout)
{
	struct scd_sysdep_control *sysdepc;

	sysdepc = control_to_sysdep(control);

        hwtimer_mod_timer(&sysdepc->timer, (unsigned long)timeout);

        return 0;
}

/*----------------------------------------------------------------------*
 * name : cancel_timeout						*
 * desc : scd timer Υ󥻥					*
 *----------------------------------------------------------------------*/
static int cancel_timeout(struct scd_control *control)
{
	struct scd_sysdep_control *sysdepc;

	sysdepc = control_to_sysdep(control);

        hwtimer_del_timer(&sysdepc->timer);

        return 0;
}

/*----------------------------------------------------------------------*
 * name : power								*
 * desc : scd intf ؿ - power						*
 *----------------------------------------------------------------------*/
int power(struct scd_control *control, bool onoff, int port)
{
        int err = 0;

        if (onoff) {
                /* gpio  */
                gpio_set_mode_bit(OVERCURRENT_GPIO_PORT,
                                  OVERCURRENT_GPIO_BIT,
                                  GPIOMODE_INPUT);
                
                err = request_irq(OVERCURRENT_GPIO_IRQ, ovc_irq, IRQF_DISABLED,
                                  "usb_ovc", (void *)&sysdepc);
                if (err < 0) {
                        ERR("ERROR(%d)\n", (int)err);
                }
        } else {
                free_irq(OVERCURRENT_GPIO_IRQ, (void *)&sysdepc);
        }

	return err;
}

static struct scd_sysdep_intf sysdep_intf = {
	.base_intf = {
		.type = SCD_CONTROL_TYPE_SYSDEP,
		.probe = probe,
		.remove = remove,
                .shutdown = NULL,
                .suspend = suspend,
                .suspend_late = NULL,
                .resume = resume,
                .resume_early = NULL,
		.start = start,
		.stop = stop,
		.vmessage = vmessage,
	},
	.set_timeout = set_timeout,
	.cancel_timeout = cancel_timeout,
	.get_time = get_time,
	.power = power,
};

/*
 * device
 */
/* ɥ쥹 */
#define ADR_START	(0xf0040000)
#define ADR_END  	ADR_START + SZ_256K
#define IRQ_NUM		IRQ_USB_BASE

static struct resource scd_resources[] = {
	[0] = {
		.start = ADR_START,
		.end = ADR_END,
		.flags = IORESOURCE_MEM,
	},
	[1] = {
		.name = "USB0",
		.start = IRQ_NUM,
		.end = IRQ_NUM,
		.flags = IORESOURCE_IRQ,
	},
};

static u64 scd_dma_mask = 0xffffffffUL;

/*----------------------------------------------------------------------*
 * name : nop_release							*
 * desc : ߡؿ							*
 *----------------------------------------------------------------------*/
static void nop_release(struct device *dev)
{
    DBG(2, "[%s] %s\n", __FUNCTION__, dev->bus_id);
}

struct platform_device scd_platform_device = {
	.name = "usb_scd",
	.id = 0,
	.dev = {
		.dma_mask = &scd_dma_mask,
		.coherent_dma_mask = 0xffffffffUL,
		.release = nop_release,
		},
	.num_resources = ARRAY_SIZE(scd_resources),
	.resource = scd_resources,
};

/*----------------------------------------------------------------------*
 * name : usb_scd_sysdep_init						*
 * desc : ؿ - otg ľܸƤФ				*
 *----------------------------------------------------------------------*/
int __init usb_scd_sysdep_init(void)
{
        int err;
	err = usb_scd_device_set_intf(&sysdep_intf.base_intf);
        if (err) {
                DBG(2, "%s(%d) error:%d", __FUNCTION__, __LINE__, err);
                return err;
        }

    	/* Platform device ȤϿ */
	platform_device_register(&scd_platform_device);

	return 0;
}

/*----------------------------------------------------------------------*
 * name : usb_scd_sysdep_exit						*
 * desc : λؿ							*
 *----------------------------------------------------------------------*/
void __exit usb_scd_sysdep_exit(void)
{
        int err;

    	/* Platform device ȤƤϿ */
        platform_device_unregister(&scd_platform_device);

	err = usb_scd_device_unset_intf(&sysdep_intf.base_intf);
        if (err) {
                DBG(2, "%s(%d) error:%d", __FUNCTION__, __LINE__, err);
        }
}
