/*--------------------------------------------------------------------------*
 * drivers/usb/scd/scd_xperi.c
 * 
 * Gadget event handlers
 * 
 * Copyright 2011 Sony Corporation
 *
 * This file is part of the HS-OTG Controller Driver SCD.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 *---------------------------------------------------------------------------*/
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/usb.h>
#include <linux/usb/ch9.h>
#include <linux/usb/gadget.h>
#include <linux/usb/scd/usb_otg_control.h>
#include <linux/usb/scd/scd_ioctl.h>
#include <linux/dma-mapping.h>
#include <linux/wait.h>

#include "scd.h"
#include "scd_version.h"
#include "scd_register.h"
#include "scd_gadget.h"
#include "scd_debug.h"
#include "scd_gfifo.h"

/*------------------------------------------------------------------------------
	Functions Declaration 
------------------------------------------------------------------------------*/
int scd_chg_devstate(int, struct scd_gadget *);
int scd_fmapinfo_init(struct scd_gadget *);
int scd_fifo_map_reset(struct scd_gadget *);
int scd_fifo_map_config(struct scd_gadget *, unsigned long);
int chk_flush_comp(struct scd_gadget *, u32);
u8	scd_g_get_devstate(struct scd_gadget *);
void scd_fparam_reset(struct scd_gadget *);
void scd_gadget_dtct_spd(struct scd_gadget *);
#ifdef SCD_POLL_ON
int scd_reset_poll_state(struct scd_gadget *, int);
#endif
static int fifo_map_calc_depth(struct scd_gadget *, u8);
static int fifo_map_mapping(struct scd_gadget *);
#ifndef SCD_CONFIG_MAP
static int get_cnt_epdir(struct scd_gadget *, u32 *, u32 *);
static int get_max_maxp(struct scd_gadget *, u32 *, int);
#endif
static int get_max_numbuf(struct scd_gadget *, u32 *, int);
static void init_fmapinfo(struct scd_gadget *);
static void init_fmapinfo2(struct scd_gadget *);
static void reset_clr_int(struct scd_gadget *);

extern int scd_chg_ep0stage(struct scd_gadget *, int);
extern void scd_g_init_ep0buf(struct scd_dmaep0_buf *);
extern void scd_gadget_flush_ep_requests(struct scd_gep *, const int);

extern int scd_g_fifo_flush_rx(struct scd_gadget *, u8, u8);

#ifdef SCD_POLL_ON
extern int set_gwq_flag(struct scd_gadget_wq *, int);
#endif
/* a buffer for EP0 DMA */
extern struct scd_dmaep0_buf	scd_dma_ep0buf;

/*------------------------------------------------------------------------------
	Functions Definition 
------------------------------------------------------------------------------*/

/*-------------------------------------------------------
	name	: scd_gadget_resume
	argment	: struct scd_gadget *sgadget
	return	: none
	description: to do after detecting a resume.
	date	: 2010/03/19
-------------------------------------------------------*/
void scd_gadget_resume(struct scd_gadget *sgadget)
{
	struct scd_device *sdev = sgadget->g_device;
	u8 rslt;
	u32 msk_val, reg_val;
	
	DBG(1, "<==\n");

	/* ٱ¿װθ */
	if(sgadget->bBusState == SCD_BUS_RESET){
		/* ڥɥơȽ */
		reg_val = SCD__READ(sdev, DSTS);
		if(!(reg_val&SCD_B_DSTS_SUSPSTS)){
			/* Skip in Non-Suspend */
			DBG(1,"skip suspend by %x \n",sgadget->bBusState);
			DBG(2,"==>\n");
			return;
		}
	}
	/* ǥХѹRESUME */
	rslt = scd_chg_devstate(SCD_GST_RESUME_EN, sgadget);
	if(rslt!=0){
		ERR("already state entered! RESUME \n");
		DBG(2,"==>\n");
		return;
	}
	/* ڥɥӥåȥꥢ */
	sgadget->bSuspended = false;
	sgadget->bBusState  = SCD_BUS_RESUME;
	
	if(sgadget->pGadgetDriver && sgadget->pGadgetDriver->resume){
		/* ̤إڥ */
#ifdef RW_LOCK_ON
		SCD_G_READ_LOCK(&sgadget->rwLock);
#endif
		sgadget->pGadgetDriver->resume(sgadget->pGadget);
#ifdef RW_LOCK_ON
		SCD_G_READ_UNLOCK(&sgadget->rwLock);
#endif
	}
	/* ڥɳߵġ쥸塼߶ػ */
	msk_val = SCD__READ(sdev, GINTMSK);
	msk_val = (msk_val&~SCD_B_GINTMSK_WKUPINTMSK)|SCD_B_GINTSTS_USBSUSP;
	SCD__WRITE(sdev, GINTMSK, msk_val);
	DBG(1, "==>\n");
}

/*-------------------------------------------------------
	name	: scd_gadget_suspend
	argment	: struct scd_gadget *sgadget
	return	: none
	description: to do after detecting a suspend.
	date	: 2010/03/19
-------------------------------------------------------*/
void scd_gadget_suspend(struct scd_gadget *sgadget)
{
	struct scd_device *sdev = sgadget->g_device;
	u8 rslt;
	u32 msk_val, reg_val;
	
	DBG(1, "<==\n");

	/* ٱ¿װθ */
	if((sgadget->bBusState == SCD_BUS_RESET)&&
		(sgadget->bBusState == SCD_BUS_RESUME)){
		/* ڥɥơȽ */
		reg_val = SCD__READ(sdev, DSTS);
		if(!(reg_val&SCD_B_DSTS_SUSPSTS)){
			/* Skip in Non-Suspend */
			DBG(1,"skip suspend by %x \n",sgadget->bBusState);
			DBG(2,"==>\n");
			return;
		}
	}
	/* ǥХѹSUSPEND */
	rslt = scd_chg_devstate(SCD_GST_SUS_EN, sgadget);
	if(rslt!=0){
		ERR("not change to the suspend state! INIT \n");
		DBG(2,"==>\n");
		return;
	}
	/* ڥɥӥåȥå */
	sgadget->bSuspended = true;
	sgadget->bBusState  = SCD_BUS_SUSPEND;
	
	if(sgadget->pGadgetDriver && sgadget->pGadgetDriver->suspend){
		/* ̤إڥ */
#ifdef RW_LOCK_ON
		SCD_G_READ_LOCK(&sgadget->rwLock);
#endif
		sgadget->pGadgetDriver->suspend(sgadget->pGadget);
#ifdef RW_LOCK_ON
		SCD_G_READ_UNLOCK(&sgadget->rwLock);
#endif
	}
	/* ڥɳ߶ػߡ쥸塼ߵ */
	msk_val = SCD__READ(sdev, GINTMSK);
	msk_val = (msk_val&~SCD_B_GINTSTS_USBSUSP)|SCD_B_GINTMSK_WKUPINTMSK;
	SCD__WRITE(sdev, GINTMSK, msk_val);
	DBG(1, "==>\n");
}

/*-------------------------------------------------------
	name	: scd_gadget_disconnect
	argment	: struct scd_gadget *sgadget
	return	: none
	description: indicate to the upper layer(gcore) 
			in case that disconnect is occured.
	date	: 2010/03/16
-------------------------------------------------------*/
void scd_gadget_disconnect(struct scd_gadget *sgadget)
{
	int i;

	DBG(1, "<==\n");
	
	if(sgadget->pGadgetDriver){
		for(i=0;i<sgadget->bEps+SCD_NUM_EP0;i++){
			/* եå&ꥯȴλ */
			scd_gadget_flush_ep_requests(&(sgadget->gep[i]), -ESHUTDOWN);
		}
		if(sgadget->pGadgetDriver&&sgadget->pGadgetDriver->disconnect){
#ifdef RW_LOCK_ON
			SCD_G_READ_LOCK(&sgadget->rwLock);
#endif
			sgadget->pGadgetDriver->disconnect(sgadget->pGadget);
#ifdef RW_LOCK_ON
			SCD_G_READ_UNLOCK(&sgadget->rwLock);
#endif
		}
	}	

	DBG(1,"==>\n");
}

/*-------------------------------------------------------
	name	: scd_gadget_reset
	argment	: struct scd_gadget *sgadget
	return	: none
	description: to do after detcting a USB bus reset 
	date	: 2010/07/13
-------------------------------------------------------*/
void scd_gadget_reset(struct scd_gadget *sgadget)
{
	struct scd_device* sdev = sgadget->g_device;
	u32 reg_val;
	int i;

	DBG(1, "<==\n");

	/* 쥸װꥢ */
	reset_clr_int(sgadget);

#ifdef SCD_POLL_ON	/*  */
	scd_reset_poll_state(sgadget, SCD_GQ_IDLE);
#endif	
	SCD_G_SPIN_LOCK(&sgadget->gLock);

	/* Set a Setup Count */
	reg_val = SCD_SUPCNT_VAL<<SCD_S_DOEPTSIZ0_SUPCNT;
	SCD__WRITE(sdev, DOEPTSIZ0, reg_val);
	
#ifdef WARN_RESET
	reg_val = SCD__READ(sdev, DOEPINT(SCD_G_EP0));
	DBG(2,"INFO  DOEPINT %x \n", reg_val);
	if(reg_val&SCD_B_DXEPINTN_SETUP){
		ERR("WARNING setup int delay !! \n");
		/* ˴롣 */
		reg_val = SCD_B_DXEPINTN_SETUP;
		SCD__WRITE(sdev,DOEPINT(SCD_G_EP0),reg_val);
	}
#endif	

	/* ǥХѹDFLT*/
	sgadget->bDevState		= SCD_GST_DFLT;
	/* ХѹRESET*/
	sgadget->bBusState		= SCD_BUS_RESET;
	sgadget->bSuspended		= false;
	sgadget->bSetAddr		= false;
	sgadget->bFifoCnfg		= false;
	sgadget->bTestMode		= false;
	sgadget->bTestModeNo	= USB_IOCTL_TESTMODE_NORMAL;
#ifdef CPU_TEST
	sgadget->bTransMode		= SCD_GADGET_TRANS_CPU;
#else
	sgadget->bTransMode		= SCD_GADGET_TRANS_DMA;
#endif
	/* ǥХɥ쥹ưʤΤǡSWꤹɬפ롣 */
	/* ɥ쥹 */
	reg_val = SCD__READ(sdev, DCFG);
	reg_val &= ~(SCD_M_DCFG_DEVADDR<<SCD_S_DCFG_DEVADDR);
	SCD__WRITE(sdev, DCFG, reg_val);

	sgadget->bDevAddr	= SCD_DFLT_DEVADDR;	/* default device address */

	SCD_G_SPIN_UNLOCK(&sgadget->gLock);

	/* FIFO mapping parameter reset */
	scd_fparam_reset(sgadget);

	SCD_G_SPIN_LOCK(&sgadget->gLock);

	/* ֽ */
	sgadget->gep[SCD_G_EP0].bState	  = SCD_EP_IDLE;
	sgadget->gep[SCD_G_EP0].bSubState = SCD_EPSUB_IDLE;
	/* lastbuf 󥿽 */
	sgadget->gep[SCD_G_EP0].lbuf.size  = 0;
	for(i=0; i<sgadget->bEps+SCD_NUM_EP0; i++){
		sgadget->gep[i].bState	  = SCD_EP_IDLE;
		sgadget->gep[i].bSubState = SCD_EPSUB_IDLE;
		/* lastbuf 󥿽 */
		sgadget->gep[i].lbuf.size  = 0;
	}
	/* ߥӥåȥޥ EP0 */
	reg_val = (1|1<<SCD_S_DAINTMSK_OUTEPMSK);
	SCD__WRITE(sdev, DAINTMSK, reg_val);
	
	/* EP0ctrl˾ SETUPԤ */
	scd_chg_ep0stage(sgadget, SCD_EP0_STAGE_SETUP);

	/* ߥӥåȥޥDOEPMSK
	   λSETUP */
	reg_val = (SCD_B_DOEPMSK_SETUPMSK|SCD_B_DOEPMSK_XFERCOMPLMSK);
	SCD__WRITE(sdev, DOEPMSK, reg_val);
	
	/* ߥӥåȥޥDIEPMSK
	   λॢ */
	reg_val = (SCD_B_DIEPMSK_TIMEOUTMSK|SCD_B_DIEPMSK_XFERCOMPLMSK);
	SCD__WRITE(sdev, DIEPMSK, reg_val);	
	
	/* ž⡼Ƚ */
	if(sgadget->bTransMode == SCD_GADGET_TRANS_DMA){

		/* EP0Хåե */
		scd_g_init_ep0buf(&scd_dma_ep0buf);
		/* Хɥ쥹 */

		SCD_G_SPIN_UNLOCK(&sgadget->gLock);

		/* DMAѥɥ쥹ޥåԥ */
		/* decode */
		scd_dma_ep0buf.dma_buf 
			= (void *)dma_map_single(sdev->dev, 
				scd_dma_ep0buf.buf, 
					sizeof(*(scd_dma_ep0buf.buf))*SCD_DMA_EP0_SZ, 
						DMA_BIDIRECTIONAL);
		SCD_G_SPIN_LOCK(&sgadget->gLock);

		/* DMAХåե */
		SCD__WRITE(sdev, DOEPDMA(SCD_G_EP0), scd_dma_ep0buf.dma_buf);
		
	}
	/* RxFlvl */
	reg_val = SCD__READ(sdev, GINTMSK);

	/* DMA⡼ RxFlvl̤ */
	reg_val &= ~SCD_B_GINTMSK_RXFLVLMSK;

	/* oepintmsk iepintmsk */
	reg_val |= SCD_B_GINTMSK_OEPINTMSK|SCD_B_GINTMSK_IEPINTMSK;

	/* enumdone, suspendߵġresume߶ػ */
	reg_val |= SCD_B_GINTMSK_USBSUSPMSK|SCD_B_GINTMSK_ENUMDONEMSK;
	reg_val &= ~SCD_B_GINTMSK_WKUPINTMSK;
	SCD__WRITE(sdev, GINTMSK, reg_val);

	SCD_G_SPIN_UNLOCK(&sgadget->gLock);
	DBG(1,"==>\n");
}

/*-------------------------------------------------------
	name	: scd_gadget_dtct_spd
	argment	: struct scd_gadget *sgadget
	return	: none
	description: to do after finishing a chirp 
	date	: 2010/05/06
-------------------------------------------------------*/
void scd_gadget_dtct_spd(struct scd_gadget *sgadget)
{
	struct scd_device* sdev = sgadget->g_device;
	u32 reg_val;
	u32 maxp0;
	u8	num;
	int rslt;
	
	DBG(1, "<==\n");

	SCD_G_SPIN_LOCK(&sgadget->gLock);
	
	/* ®پ */
	reg_val = SCD__READ(sdev, DSTS);
	num = (reg_val >> SCD_S_DSTS_ENUMSPD)&SCD_M_DSTS_ENUMSPD;
	
	/* ®Ƚ */
	switch(num){
	case SCD_E_DSTS_ENUMSPD_HS:
		/* HS */
		sgadget->pGadget->speed = USB_SPEED_HIGH;
		maxp0 = SCD_E_DIEPCTL0_MPS_64B;
		DBG(3,"speed HS \n");
		break;
	case SCD_E_DSTS_ENUMSPD_FS:
		/* FS maxp64Byte */
		sgadget->pGadget->speed = USB_SPEED_FULL;
		maxp0 = SCD_E_DIEPCTL0_MPS_64B;
		DBG(3,"speed FS \n");
		break;
	case SCD_E_DSTS_ENUMSPD_LS_PCLK48MHZ:
	case SCD_E_DSTS_ENUMSPD_FS_PCLK48MHZ:
		ERR("non-support speed:%x\n",num);
		sgadget->pGadget->speed = USB_SPEED_UNKNOWN;
		maxp0 = 0;	
		return;	/* ꤨʤ顼֤ǤΤǡȴ롣 */
	}
	/* MAXP0 */
	reg_val = SCD__READ(sdev, DIEPCTL0);
	reg_val &= ~SCD_M_DIEPCTL0_MPS;
	reg_val |= maxp0&SCD_M_DIEPCTL0_MPS;
	SCD__WRITE(sdev, DIEPCTL0, reg_val);

	SCD_G_SPIN_UNLOCK(&sgadget->gLock);

	/* FIFO mapping(DFLTsetting) */
	rslt = scd_fifo_map_reset(sgadget);
	if(rslt != 0){
		ERR("mapping error in a reset");
	}

	if (sgadget->pGadgetDriver) {
		/* ̤reset */   
		scd_gadget_disconnect(sgadget);
	}

	SCD_G_SPIN_LOCK(&sgadget->gLock);

	/* ž⡼Ƚ */
	if(sgadget->bTransMode == SCD_GADGET_TRANS_DMA){
		/* EPEna */
		reg_val = SCD__READ(sdev, DOEPCTL0);
		reg_val |= SCD_B_DOEPCTL0_EPENA;
		SCD__WRITE(sdev, DOEPCTL0, reg_val);
	}else{
		/* RxFlvl */
		reg_val = SCD__READ(sdev, GINTMSK);
		/* CPU⡼ RxFlvl߻ */
		reg_val |= SCD_B_GINTMSK_RXFLVLMSK;
		SCD__WRITE(sdev, GINTMSK, reg_val);
	}

	SCD_G_SPIN_UNLOCK(&sgadget->gLock);

	DBG(1,"==>\n");
}
/*-------------------------------------------------------
	name	: scd_chg_devstate
	argment	: int enter, struct scd_gadget *sgadget
	return	: int
	description: DevState State Machine
	date	: 2010/03/19
-------------------------------------------------------*/
int scd_chg_devstate(int enter, struct scd_gadget *sgadget){

	int rslt=0;
	unsigned long rwflags;

	DBG(1,"<== DevState %x from %d \n", enter,sgadget->bDevState);

#ifdef RW_LOCK_ON
	SCD_G_READ_LOCK_IRQ(&sgadget->rwLock, rwflags);
#endif

	switch(enter){
	/*-- PULLUPOFF --*/
	case SCD_GST_PULLUPOFF_EN:
		switch(sgadget->bDevState){
		case SCD_GST_INIT:
			DBG(2,"already state entered! PULLUPOFF in %x\n", sgadget->bDevState);
			rslt = 0;	/* but normal */
			break;
		case SCD_GST_PULLUPON:
		case SCD_GST_DFLT:
		case SCD_GST_ADDR:
		case SCD_GST_CNFG:
		case SCD_GST_PULLUPON_SUS:
		case SCD_GST_DFLT_SUS:
		case SCD_GST_ADDR_SUS:
		case SCD_GST_CNFG_SUS:
			/* ѹ */
			sgadget->bDevState = SCD_GST_INIT;
			rslt = 0;
			break;
		default:
			rslt = -EINVAL;
		}
		break;
		
	/*-- PULLUPON --*/
	case SCD_GST_PULLUPON_EN:
		switch(sgadget->bDevState){
		case SCD_GST_PULLUPON:
		case SCD_GST_PULLUPON_SUS:
			DBG(2,"already state entered! PULLUPON in %x\n", sgadget->bDevState);
			rslt = 0;	/* but normal */
			break;
		case SCD_GST_INIT:
			/* ѹ */
			sgadget->bDevState = SCD_GST_PULLUPON;
			rslt = 0;
			break;
		case SCD_GST_DFLT:
			DBG(2,"already PULLUPON state %x\n", sgadget->bDevState);
			rslt = 0;	/* but normal */
			break;				
		case SCD_GST_ADDR:
		case SCD_GST_CNFG:
		case SCD_GST_DFLT_SUS:
		case SCD_GST_ADDR_SUS:
		case SCD_GST_CNFG_SUS:
			DBG(1, "already PULLUPON state %x\n", sgadget->bDevState);
			rslt = SCD_DEVST_SKIP; /* local return code */
			break;			
		default:
			rslt = -EINVAL;	
		}
		break;
		
	/*-- RESET --*/
	case SCD_GST_RESET_EN:
		switch(sgadget->bDevState){
		case SCD_GST_DFLT:
			DBG(2,"already state entered! RESET in %x\n", sgadget->bDevState);
			rslt = 0;	/* but normal */
			break;
		case SCD_GST_PULLUPON:
		case SCD_GST_ADDR:
		case SCD_GST_CNFG:
		case SCD_GST_PULLUPON_SUS:
		case SCD_GST_DFLT_SUS:
		case SCD_GST_ADDR_SUS:
		case SCD_GST_CNFG_SUS:
			/* ѹ */
			sgadget->bDevState = SCD_GST_DFLT;
			rslt = 0;
			break;
		default:
			rslt = -EINVAL;
		}
		break;

	/*-- ADDRESS --*/
	case SCD_GST_ADDR:
		switch(sgadget->bDevState){
		case SCD_GST_ADDR:
			DBG(2,"already state entered! ADDR in %x\n", sgadget->bDevState);
			rslt = 0;	/* but normal */
			break;
		case SCD_GST_DFLT:
		case SCD_GST_CNFG:
		case SCD_GST_DFLT_SUS:
			/* ѹ */
			sgadget->bDevState = SCD_GST_ADDR;
			rslt = 0;
			break;
		default:
			rslt = -EINVAL;
		}
		break;
		
	/*-- CNFG --*/
	case SCD_GST_CNFG:
		switch(sgadget->bDevState){
		case SCD_GST_CNFG:
			DBG(2,"new config entered! CNFG in %x\n", sgadget->bDevState);
			rslt = 0;	/* but normal */
			break;
		case SCD_GST_ADDR:
			/* ѹ */
			sgadget->bDevState = SCD_GST_CNFG;
			rslt = 0;
			break;
		default:
			rslt = -EINVAL;
		}
		break;

	/*-- SUSPEND --*/
	case SCD_GST_SUS_EN:
		switch(sgadget->bDevState){
		case SCD_GST_PULLUPON:
			/* ѹ */
			sgadget->bDevState = SCD_GST_PULLUPON_SUS;
			rslt = 0;
			break;
		case SCD_GST_DFLT:
			/* ѹ */
			sgadget->bDevState = SCD_GST_DFLT_SUS;
			rslt = 0;
			break;
		case SCD_GST_ADDR:
			/* ѹ */
			sgadget->bDevState = SCD_GST_ADDR_SUS;
			rslt = 0;
			break;
		case SCD_GST_CNFG:
			/* ѹ */
			sgadget->bDevState = SCD_GST_CNFG_SUS;
			rslt = 0;
			break;
		case SCD_GST_PULLUPON_SUS:
		case SCD_GST_DFLT_SUS:
		case SCD_GST_ADDR_SUS:
		case SCD_GST_CNFG_SUS:
			DBG(2,"already state entered! SUS in %x\n", sgadget->bDevState);
			rslt = 0;	/* but normal */
			break;
		default:
			rslt = -EINVAL;
		}
		break;

	/*-- RESUME  --*/
	case SCD_GST_RESUME_EN:
		switch(sgadget->bDevState){
		case SCD_GST_PULLUPON_SUS:
			/* ѹ */
			sgadget->bDevState = SCD_GST_PULLUPON;
			rslt = 0;
			break;
		case SCD_GST_DFLT_SUS:
			/* ѹ */
			sgadget->bDevState = SCD_GST_DFLT;
			rslt = 0;
			break;
		case SCD_GST_ADDR_SUS:
			/* ѹ */
			sgadget->bDevState = SCD_GST_ADDR;
			rslt = 0;
			break;
		case SCD_GST_CNFG_SUS:
			/* ѹ */
			sgadget->bDevState = SCD_GST_CNFG;
			rslt = 0;
			break;
		default:
			rslt = -EINVAL;
		}
		break;
	default:
		rslt = -EINVAL;
	}
	DBG(2,"==> State :%d\n", sgadget->bDevState);
#ifdef RW_LOCK_ON
	SCD_G_READ_UNLOCK_IRQ(&sgadget->rwLock, rwflags);
#endif
	return rslt;
}

/*-------------------------------------------------------
	name	: scd_fifo_map_config
	argment	: struct scd_gadget *sgadget
			  unsigned long param
	return	: int
	description: FIFO mapping on a configuration
			 (only tx fifo)
		϶ͭFIFOΤmaxp1024ȤƸ
		Ƥ롣usbХꥻåȻޤFIFO_RESETˤ
		Ԥ롣ϸ̤ǽǤ롣
		âܴؿ뤵٤ˡȥ
		ʬʤˡEPФ׻FIFOmappingԤ
		ΥХåե̿ϰ(fparam->bType)Ѥ롣
	date	: 2010/07/01
-------------------------------------------------------*/
int scd_fifo_map_config(struct scd_gadget *sgadget, unsigned long param)
{
	struct scd_device *sdev = sgadget->g_device;
	struct scd_fifo_each *ep_param = sgadget->fparam.ep_param;
	struct usb_ioctl_fifo_param *ifparam;
	int rslt=0;
	u32 reg_val;
	u8	epnum;
	unsigned long flags;

	DBG(2,"<==\n");

	SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);

	ifparam = (struct usb_ioctl_fifo_param *)param;
	
	/* FIFOmapping */
	switch(ifparam->bEpAdr&USB_ENDPOINT_DIR_MASK){
	case USB_DIR_IN:
		if(sgadget->fparam.cnt < sgadget->bEps+SCD_NUM_EP0){
			/* Υѥ᡼EP¤ξ˥åȤ롣 */
			/* set_configurationǤγEPϻĤ롣*/
			ep_param[sgadget->fparam.cnt].bEpAddr	= ifparam->bEpAdr;
			ep_param[sgadget->fparam.cnt].wSize		= ifparam->wSize;
			ep_param[sgadget->fparam.cnt].wNumbuf	= ifparam->bType;
			sgadget->fparam.cnt++;
		}else{
			ERR("ERROR over a number of eps \n");
			DBG(2,"==>\n");
			SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);
			return -EINVAL;
		}
		SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);

		/* FIFO depth׻(size׻) */
		fifo_map_calc_depth(sgadget, true);

		SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);

		/* INλΤ߹Ԥ*/
		/* FIFO mapping(sizeaddress) */
		fifo_map_mapping(sgadget);
		
		/*
		 *	FIFO Flush
		 */
			
		/* 
		 * FIFO եåֹEP 
		 */
		reg_val = SCD__READ(sdev, GRSTCTL);
		reg_val &= ~(SCD_M_GRSTCTL_TXFNUM<<SCD_S_GRSTCTL_TXFNUM);
		reg_val |= SCD_GRSTCTL_FFLUSH_NUM_ALL<<SCD_S_GRSTCTL_TXFNUM;
			
		/* 
		* FIFO եå 
		*/
		reg_val |= SCD_B_GRSTCTL_TXFFLSH;
		SCD__WRITE(sdev, GRSTCTL, reg_val);

		SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);

		/* FLUSHλԤ */
		rslt = chk_flush_comp(sgadget, SCD_B_GRSTCTL_TXFFLSH);
		if(rslt){
			DBG(2,"==>\n");
			return -EPERM;
		}

		SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);

		/* 
	 	 * FIFO եå 
		 */
		reg_val = SCD__READ(sdev, GRSTCTL);
		reg_val |= SCD_B_GRSTCTL_RXFFLSH;
		SCD__WRITE(sdev, GRSTCTL, reg_val);

		SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);

		/* FLUSHλǧ */
		rslt = chk_flush_comp(sgadget, SCD_B_GRSTCTL_RXFFLSH);
		if(rslt){
			DBG(2,"==>\n");
			return -EPERM;
		}

		/* EPֹ */
		epnum = ifparam->bEpAdr&USB_ENDPOINT_NUMBER_MASK;
		DBG(2,"epnum:%x\n",epnum);

		/* EP active */
		/* EP0ʳbus-resetˤƥꥢ롣 */
		reg_val = SCD__READ(sdev, DIEPCTL(epnum));
		reg_val |= SCD_B_DXEPCTLN_USBACTEP;
		SCD__WRITE(sdev, DIEPCTL(epnum), reg_val);
		
		break;
	case USB_DIR_OUT:
		if(sgadget->fparam.cnt < sgadget->bEps+SCD_NUM_EP0){
			/* Υѥ᡼EP¤ξ˥åȤ롣 */
			/* set_configurationǤγEPϻĤ롣*/
			ep_param[sgadget->fparam.cnt].bEpAddr	= ifparam->bEpAdr;
			ep_param[sgadget->fparam.cnt].wSize		= ifparam->wSize;
			ep_param[sgadget->fparam.cnt].wNumbuf	= ifparam->bType;
			sgadget->fparam.cnt++;
		}else{
			ERR("ERROR over a number of eps \n");
			DBG(2,"==>\n");
			SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);
			return -EINVAL;
		}
		/* EPֹ */
		epnum = ifparam->bEpAdr&USB_ENDPOINT_NUMBER_MASK;
		DBG(2,"epnum:%x\n",epnum);

		/* EP active */
		/* EP0ʳbus-resetˤƥꥢ롣 */
		reg_val = SCD__READ(sdev, DOEPCTL(epnum));
		reg_val |= SCD_B_DXEPCTLN_USBACTEP;
		SCD__WRITE(sdev,DOEPCTL(epnum), reg_val);

		DBG(2,"active out ep \n");
		break;
	}
	SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);

	DBG(2,"==>\n");
	return 0;
}

/*-------------------------------------------------------
	name	: scd_fifo_map_reset
	argment	: struct scd_gadget *sgadget
	return	: int
	description: FIFO mapping on a bus reset/configuration
			this FIFO RESET process will be also done at first
			at a Configutration.
			locked inside
			this may be called by ioctl.
	date	: 2010/10/25
-------------------------------------------------------*/
int scd_fifo_map_reset(struct scd_gadget *sgadget)
{
	u32 reg_val;
	int	rslt;
	
	struct scd_device *sdev = sgadget->g_device;

	DBG(2,"<==\n");
	
	/* FIFO depth׻(size׻) */
	fifo_map_calc_depth(sgadget, false);

	/* FIFO mapping(sizeaddress) */
	fifo_map_mapping(sgadget);
	
	/*
	 *	FIFO Flush
	 */

	/* 
	 * FIFO եåֹEP 
	 */
	reg_val = SCD__READ(sdev, GRSTCTL);
	reg_val &= ~(SCD_M_GRSTCTL_TXFNUM<<SCD_S_GRSTCTL_TXFNUM);
	reg_val |= SCD_GRSTCTL_FFLUSH_NUM_ALL<<SCD_S_GRSTCTL_TXFNUM;
	
	/* 
	 * FIFO եå 
	 */
	reg_val |= SCD_B_GRSTCTL_TXFFLSH;
	SCD__WRITE(sdev, GRSTCTL, reg_val);
	
	/* FLUSHλԤ */
	rslt = chk_flush_comp(sgadget, SCD_B_GRSTCTL_TXFFLSH);
	if(rslt){
		/* 顼ʥ롼Ǥ롣 */
		DBG(2,"==>\n");
		return rslt;
	}
	
	/* 
	 * FIFO եå 
	 */
	reg_val = SCD__READ(sdev, GRSTCTL);
	reg_val |= SCD_B_GRSTCTL_RXFFLSH;
	SCD__WRITE(sdev, GRSTCTL, reg_val);
	
	/* FLUSHλǧ */
	rslt = chk_flush_comp(sgadget, SCD_B_GRSTCTL_RXFFLSH);
	if(rslt){
		/* 顼ʥ롼Ǥ롣 */
		DBG(2,"==>\n");
		return rslt;
	}
	DBG(2,"==>\n");
	return 0;
}
/*-------------------------------------------------------
	name	: scd_fmapinfo_init
	argment	: struct scd_gadget *sgadget
	return	: int
	description: caluclation each FIFO depth init
	date	: 2010/05/06
-------------------------------------------------------*/
int scd_fmapinfo_init(struct scd_gadget *sgadget)
{
	struct scd_device *sdev = sgadget->g_device;
	struct scd_gadget_fmapinfo *fminfo = scd_gadget_fminfo; 
	unsigned long flags;

	u32	reg_val;
	int	i;

	DBG(2,"<==\n");

	SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);

	/*  */
	init_fmapinfo(sgadget);
	
	/*
	 *	HWѥ᡼ѥ쥸ͼ
	 *	OTG_DFIFO_DEPTH
	 */
	reg_val = SCD__READ(sdev, GHWCFG3);
	fminfo->all_dfifo_dep 
		= (reg_val>>SCD_S_GHWCFG3_DFIFODEPTH)&SCD_M_GHWCFG3_DFIFODEPTH;
	
	/* OTG_RX_DFIFO_DEPTH饤Ȳǽ쥸Τ
	   Ʊ¤ȤʤOTG_DFIFO_DEPTH
	 */
	fminfo->all_rx_dfifo_dep = fminfo->all_dfifo_dep;
	
	/*	
	 *	FIFOdepth
	 *	OTG_TX_DINEP_DFIFO_DEPTH(ޥ)
	 */
	/*	for EP0	
	 */
	(fminfo->max_tx_dfifo_dep)[0] = SCD_OTG_TX_DPERIO_DFIFO_DEPTH;

	/* for EP[i] 
	 */
	for(i=1; i<sgadget->bEps+1;i++){
		(fminfo->max_tx_dfifo_dep)[i] = SCD_OTG_TX_DPERIO_DFIFO_DEPTH;
	}
	/* OTG_NUM_IN_EPS 
	 */
	reg_val = SCD__READ(sdev, GHWCFG4);
	fminfo->num_iep_max	=(reg_val>>SCD_S_GHWCFG4_INEPS)&SCD_M_GHWCFG4_INEPS;

	/* ȥž 
	 */
	fminfo->num_ctrl_ep = 1;	/* fixed 1 now!! */

	SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);
	DBG(2,"==>\n");
	return 0;
}
/*-------------------------------------------------------
	name	: fifo_map_calc_depth
	argment	: struct scd_gadget *sgadget
	return	: int
	description: caluclation each FIFO depth
	date	: 2010/07/01
-------------------------------------------------------*/
static int fifo_map_calc_depth(struct scd_gadget *sgadget, u8 is_map)
{
	struct scd_gadget_fmapinfo *fminfo = scd_gadget_fminfo; 
	struct scd_gadget_fdepth *fdepth = &sgadget->fdepth;
	struct scd_fifo_each *ep_param = sgadget->fparam.ep_param;
	u32 rx_dep;
	int	i, rslt;
	unsigned long flags;

	DBG(2,"<==\n");

	SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);

	/*  */
	init_fmapinfo2(sgadget);
	
	/* EPEP0 
	 */
#ifdef SCD_CONFIG_MAP
	fminfo->num_oep = sgadget->bEps;	/* 4 */
#else
	/* ĥ */
	rslt = get_cnt_epdir(sgadget, &fminfo->num_oep, &fminfo->num_iep);
	if(rslt){
		SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);
		DBG(2,"==>\n");
		return rslt;
	}
#endif
	DBG(2, "num_oep:%x num_iep:%x \n",fminfo->num_oep, fminfo->num_iep);

	/* maxpμʶͭFIFOǤΤߡ 
	 */
#ifdef SCD_CONFIG_MAP
	/* Rx FIFOresetˤΤꤹ롣 */
	fminfo->rx_maxp_dep = SCD_GRX_MAXFIFOSZ; /* ͭFIFO1024 */
	DBG(2, "max_maxp :%x at a config \n",fminfo->rx_maxp_dep);
#else
	if(is_map){
		/* at a Configuration */
		fminfo->rx_maxp_dep = SCD_GRX_MAXFIFOSZ; /* ͭFIFO1024 */
		DBG(2, "max_maxp :%x at a config \n",fminfo->rx_maxp_dep);
	}else{
		/* at a Bus Reset */
		rslt = get_max_maxp(sgadget, &fminfo->rx_maxp_dep, USB_DIR_OUT);
		if(rslt){
			SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);
			DBG(2,"==>\n");
			return rslt;
		}
		DBG(2, "max_maxp :%x at a busreset \n",fminfo->rx_maxp_dep);
	}
#endif

	/* ѥХåե̸Ƥ 
	 */
	rslt = get_max_numbuf(sgadget, &(fminfo->rxc), USB_DIR_OUT);
	if(rslt){
		SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);
		DBG(2,"==>\n");
		return rslt;
	}
	DBG(2, "rxc:%x \n",fminfo->rxc);

	/* ꥵ׻ 
	 */
	if(fminfo->rx_maxp_dep%4!=0){
		/* aligned to DWORD */
		rx_dep = fminfo->rx_maxp_dep + 4 -(fminfo->rx_maxp_dep%4);	
	}else{
		rx_dep = fminfo->rx_maxp_dep;
	}
	/* ͽη׻ѿϵǽ;갷 */
	fdepth->rx_fifosz =	(4*fminfo->num_ctrl_ep) +6 + fminfo->rxc*(1+rx_dep/4) 
						+ 1 + (2*fminfo->num_oep);

	DBG(2, "rx_fifosz:%x \n",fdepth->rx_fifosz);

	/* EP0 Хåե= */
	fdepth->tx_fifosz[SCD_G_EP0] = ep_param[SCD_G_EP0].wSize/4;
	
	DBG(2, "tx_fifosz[0]:%x \n",fdepth->tx_fifosz[SCD_G_EP0]);
	/* ȡ̤׻ 
	 */
	fminfo->chk_dep = fdepth->rx_fifosz;
	
	/*
	 *  calculating each tx fifo depth
	 */

	for(i=1; i<fminfo->num_iep_max+SCD_NUM_EP0; i++){
		DBG(2, "num [%x] \n",i);

		switch((ep_param[i].bEpAddr)&USB_ENDPOINT_DIR_MASK){
		case USB_DIR_IN:
			/* ׵Ĺ(req_dep)׻롣 
			 */
			fminfo->req_dep = ep_param[i].wSize*ep_param[i].wNumbuf/4;
			/* ͤĶʤ褦ˡǾͤ򤹤롣 
			 */
			DBG(2, "fminfo->req_dep %x \n",fminfo->req_dep);

			fminfo->req_dep = min(fminfo->req_dep, fminfo->max_tx_dfifo_dep[i]);
			DBG(2, "fminfo->req_dep min %x \n",fminfo->req_dep);

			/* ȡ̤׻
			 */
			fminfo->chk_dep += fminfo->req_dep;

			DBG(2, "fminfo->chk_dep %x \n",fminfo->chk_dep);

			/* ȡȽ
			 */
			if((fdepth->rx_fifosz + fminfo->chk_dep) > fminfo->all_dfifo_dep){

				DBG(3, "rx_fifosz %x chk_dep %x all_dfifo_dep %x\n",fdepth->rx_fifosz, fminfo->chk_dep, fminfo->all_dfifo_dep);
				/* ΰ襪С  
				 */
				ERR("ERROR no memory - fifo map total size over!! \n");
				DBG(2,"==>\n");
				SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);
				return -ENOMEM;
			}
			/* ꥵ
			 */
			fdepth->tx_fifosz[i] = fminfo->req_dep;
			break;
		case USB_DIR_OUT:
			/* ꥵ
			 */
			fdepth->tx_fifosz[i] = 0;	/* Τ */
			DBG(2, "tx_fifosz %x \n",fdepth->tx_fifosz[i]);
			break;
		}
	}
	SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);

	DBG(2,"==>\n");
	return 0;
}

/*-------------------------------------------------------
	name	: init_fmapinfo
	argment	: struct scd_gadget *sgadget
	return	: void
	description: 
	date	: 2010/05/06
-------------------------------------------------------*/
static void init_fmapinfo(struct scd_gadget *sgadget)
{
	int i;

	DBG(2,"<==\n");
	
	scd_gadget_fminfo->all_dfifo_dep	= 0;
	scd_gadget_fminfo->all_rx_dfifo_dep	= 0;
	for(i=0; i<sgadget->bEps+1; i++){
		scd_gadget_fminfo->max_tx_dfifo_dep[i]	= 0;
	}
	scd_gadget_fminfo->req_dep			= 0;
	scd_gadget_fminfo->chk_dep			= 0;
	scd_gadget_fminfo->chk_dep2			= 0;
	scd_gadget_fminfo->num_ctrl_ep		= 0;
	scd_gadget_fminfo->num_oep			= 0;
	scd_gadget_fminfo->num_iep_max		= 0;
	scd_gadget_fminfo->num_iep			= 0;
	scd_gadget_fminfo->rx_maxp_dep		= 0;
	scd_gadget_fminfo->rxc				= 0;

	DBG(2,"==>\n");

}

/*-------------------------------------------------------
	name	: init_fmapinfo2
	argment	: struct scd_gadget *sgadget
	return	: void
	description: 
	date	: 2010/04/13
-------------------------------------------------------*/
static void init_fmapinfo2(struct scd_gadget *sgadget)
{
	DBG(2,"<==\n");

	scd_gadget_fminfo->req_dep			= 0;
	scd_gadget_fminfo->chk_dep			= 0;
	scd_gadget_fminfo->chk_dep2			= 0;
	scd_gadget_fminfo->num_oep			= 0;
	scd_gadget_fminfo->num_iep			= 0;
	scd_gadget_fminfo->rxc				= 0;

	DBG(2,"==>\n");
}
#ifndef SCD_CONFIG_MAP
/*-------------------------------------------------------
	name	: get_cnt_epdir
	argment	: struct scd_gadget *sgadget, u32 *onum, u32 *inum
	return	: int
	description: calculation a total number of each EP direction
	date	: 2010/03/25
-------------------------------------------------------*/
static int get_cnt_epdir(struct scd_gadget *sgadget, u32 *onum, u32 *inum)
{
	struct scd_fifo_each *ep_param;
	u32 o_num, i_num;
	int i;

	DBG(2,"<==\n");
	
	if(sgadget->bEps==0) {
		ERR("EPs is 0! \n");
		return -EINVAL;
	}
	ep_param = sgadget->fparam.ep_param;
	
	for(i=1, o_num=0, i_num=0; i<sgadget->bEps+1; i++){
		/* EP direction check */
		switch(ep_param[i].bEpAddr&USB_ENDPOINT_DIR_MASK){
		case USB_DIR_OUT:
			o_num++;
			break;
		case USB_DIR_IN:
			i_num++;
			break;
		}
	}
	/* set result */
	if(onum) *onum = o_num;
	if(inum) *inum = i_num;

	DBG(2,"==>\n");

	return 0;
}
/*-------------------------------------------------------
	name	: get_max_maxp
	argment	: struct scd_gadget *sgadget, u32 *maxsz, int dir
	return	: int
	description: calculation the max size of maxp size
	date	: 2010/03/25
-------------------------------------------------------*/
static int get_max_maxp(struct scd_gadget *sgadget, u32 *maxsz, int dir)
{
	struct scd_fifo_each *ep_param;
	u16 max, tmp;
	int i;

	DBG(2,"<==\n");
	
	if(sgadget->bEps==0){
		ERR("EPs is 0! \n");
		DBG(2,"==>\n");
		return -EINVAL;
	}
	if((dir!=USB_DIR_OUT)&&(dir!=USB_DIR_IN)){
		ERR("EP direction is invalid! \n");
		DBG(2,"==>\n");
		return -EINVAL;
	}
	ep_param = sgadget->fparam.ep_param;
	
	/* must include ep0 */
	for(i=0, max=0, tmp=0; i<sgadget->bEps+1; i++){
		if((ep_param[i].bEpAddr&USB_ENDPOINT_DIR_MASK)==dir){
			tmp = ep_param[i].wSize;
			if(max<tmp){
				/* change the max value */
				max = tmp;
			}
		}
	}
	/* set result */
	if(maxsz) *maxsz = max;

	DBG(2,"==>\n");	
	return 0;
}
#endif
/*-------------------------------------------------------
	name	: get_max_numbuf
	argment	: struct scd_gadget *sgadget, u32 *maxsz, int dir
	return	: int
	description: comparison the number of a buf 
	date	: 2010/03/25
-------------------------------------------------------*/
static int get_max_numbuf(struct scd_gadget *sgadget, u32 *maxsz, int dir)
{
	struct scd_fifo_each *ep_param;
	u16 max, tmp;
	int i;

	DBG(2,"<==\n");

	if(sgadget->bEps==0){
		ERR("EPs is 0! \n");
		DBG(2,"==>\n");
		return -EINVAL;
	}
	if((dir!=USB_DIR_OUT)&&(dir!=USB_DIR_IN)){
		ERR("EP direction is invalid! \n");
		DBG(2,"==>\n");
		return -EINVAL;
	}
	ep_param = sgadget->fparam.ep_param;
	
	/* must include ep0 */
	for(i=0, max=0, tmp=0; i<sgadget->bEps+1; i++){
		if((ep_param[i].bEpAddr&USB_ENDPOINT_DIR_MASK)==dir){
			tmp = ep_param[i].wNumbuf;
			if(max<tmp){
				/* change the max value */
				max = tmp;
			}
		}
	}
	/* set result */
	if(maxsz) *maxsz = max;

	DBG(2,"==>\n");	
	return 0;
}
/*-------------------------------------------------------
	name	: fifo_map_mapping
	argment	: struct scd_gadget *sgadget
	return	: int
	description: setting each FIFO parametaer
	date	: 2010/05/06
-------------------------------------------------------*/
static int fifo_map_mapping(struct scd_gadget *sgadget)
{
	struct scd_device *sdev = sgadget->g_device;
	struct scd_gadget_fdepth *fdepth;
	struct scd_fifo_each *ep_param;
	u32	reg_val;
	u8	i;
	u16 addr;

	DBG(2,"<==\n");

	fdepth	 = &sgadget->fdepth;
	ep_param = sgadget->fparam.ep_param;
	
	/* 
	 * EPFIFO
	 */
	reg_val = (fdepth->rx_fifosz&SCD_M_GRXFSIZ_RXFDEP) << SCD_S_GRXFSIZ_RXFDEP;
	SCD__WRITE(sdev, GRXFSIZ, reg_val);
	
#ifdef SCD_CONFIG_MAP
	/* ɽ */
	reg_val = SCD__READ(sdev, GRXFSIZ);
	DBG(2,"shared FIFO GRXFIZ 0x%x\n",reg_val);
#endif
	/* 
	 * EP0FIFO 
	 */
	reg_val 
	 = (fdepth->tx_fifosz[SCD_G_EP0]&SCD_M_GNPTXFSIZ_INEPTXF0DEP)<<SCD_S_GNPTXFSIZ_INEPTXF0DEP;
	/* 
	 * EP0FIFOɥ쥹
	 */
	addr = fdepth->rx_fifosz&SCD_M_GNPTXFSIZ_INEPTXF0STADDR;
	reg_val |= addr;
	/* 
	 * EP0 쥸饤
	 */
	SCD__WRITE(sdev, GNPTXFSIZ, reg_val);
	
	/*  EP1 */
	for(i=1; i<sgadget->bEps+1;i++){
		/* EPiFIFO
		 */
		reg_val = (fdepth->tx_fifosz)[i]<<SCD_S_DIEPTXFN_IEPNTXFDEP;
		/* 
		 * EPiFIFOɥ쥹
		 */
		addr += (fdepth->tx_fifosz)[i-1]&SCD_M_DIEPTXFN_INEPNTXFSTADDR;
		reg_val |= addr;
		/* 
		 * EPi 쥸饤
		 */
		SCD__WRITE(sdev, DIEPTXF(i), reg_val);
#ifdef SCD_CONFIG_MAP
		/* ɽ */
		reg_val = SCD__READ(sdev, DIEPTXF(i));
		DBG(2,"tx fifo DIEPTXF(%d) 0x%x\n",i, reg_val);
#endif

	}
	DBG(2,"==>\n");

	return 0;
}

/*-------------------------------------------------------
	name	: chk_flush_comp
	argment	: struct scd_gadget *sgadget, u32 f_bit
	return	: int
	description: check a FIFO flush completion
	date	: 2010/05/06
-------------------------------------------------------*/
int chk_flush_comp(struct scd_gadget *sgadget, u32 f_bit)
{
	struct scd_device *sdev = sgadget->g_device;
	u32 reg_val;
	int rslt=0;
	scd_usec start, end;
	int cnt=0;

	DBG(2,"<==\n");
	
	/* bit */
	switch(f_bit){
	case SCD_B_GRSTCTL_TXFFLSH:
	case SCD_B_GRSTCTL_RXFFLSH:
		break;
	default:
		ERR("FIFO Flush non-support bit! \n");
		DBG(2,"==>\n");
		return -EINVAL;
	}
	
	switch(sgadget->pGadget->speed){
	case USB_SPEED_HIGH:
		/* FLUSHλǧ */
		start = scd_device_get_time(sdev);
		while(1){
			reg_val = SCD__READ(sdev, GRSTCTL);
			if(!(reg_val&f_bit)){
				/* loop exit */
				rslt = 0;
				break;
			}
			udelay(SCD_FLUSH_HS_UDELAY);
			cnt++;
			if(cnt>=SCD_FLUSH_HS_CNT){
				end = scd_device_get_time(sdev);
				/* Ϥ絬֤timeout */
				if (end - start > SCD_FLUSH_HS_LOOP_OUT) {
					ERR("FIFO FLUSH failed timeout\n");
					rslt = -ETIME;
					break;
				}
			}
		}
		break;
	case USB_SPEED_FULL:
		/* FLUSHλǧ */
		start = scd_device_get_time(sdev);
		while(1){
			reg_val = SCD__READ(sdev, GRSTCTL);
			if(!(reg_val&f_bit)){
				/* loop exit */
				rslt = 0;
				break;
			}
			udelay(SCD_FLUSH_FS_UDELAY);
			cnt++;
			if(cnt>=SCD_FLUSH_FS_CNT){
				end = scd_device_get_time(sdev);
				/* Ϥ絬֤timeout */
				if (end - start > SCD_FLUSH_FS_LOOP_OUT) {
					ERR("FIFO FLUSH failed timeout\n");
					rslt = -ETIME;
					break;
				}
			}
		}
		break;
	default:
		rslt = -EINVAL;
	}
	
	DBG(2,"==>\n");
	return rslt;
}
/*-------------------------------------------------------
	name	: scd_g_get_devstate()
	argment	: struct scd_gadget *sgadget
	return	: u8
	description: Get DevState State
	date	: 2010/03/26
-------------------------------------------------------*/
u8	scd_g_get_devstate(struct scd_gadget *sgadget)
{
	unsigned long rwflags;
	u8 state;

	DBG(2,"<==>\n");

#ifdef RW_LOCK_ON
	SCD_G_READ_LOCK_IRQ(&sgadget->rwLock, rwflags);
#endif
	state = sgadget->bDevState;

#ifdef RW_LOCK_ON
	SCD_G_READ_UNLOCK_IRQ(&sgadget->rwLock, rwflags);
#endif
	return state;
}

/*-------------------------------------------------------
	name	: scd_fparam_reset()
	argment	: struct scd_gadget *sgadget
	return	: u8
	description: fifo mapping parameter
			  called when every FIFO reset.
	date	: 2010/04/21
-------------------------------------------------------*/
void scd_fparam_reset(struct scd_gadget *sgadget)
{
	struct scd_device *sdev = sgadget->g_device;
	u32 reg_val;
	u8 num;
	unsigned long flags;

	DBG(2,"<==\n");

	SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);

	/* ƤOUT-EPФNAKӥåȤ */
	
	/**** EP0 */
	reg_val = SCD__READ(sdev, DOEPCTL0);
	reg_val |= SCD_B_DOEPCTL0_SNAK;
	SCD__WRITE(sdev, DOEPCTL0, reg_val);

	/* FIFO mapping */
	sgadget->fparam.ep_param[SCD_G_EP0].bEpAddr = 0;
	sgadget->fparam.ep_param[SCD_G_EP0].wSize 	= SCD_EP0_MAXPSZ;
	sgadget->fparam.ep_param[SCD_G_EP0].wNumbuf	= SCD_FIFO0_NUMBUF;

	/**** EP0ʳ */
	for(num=1; num<sgadget->bEps+1; num++){
		reg_val = SCD__READ(sdev, DOEPCTL(num));
		reg_val |= SCD_B_DXEPCTLN_SNAK;
		SCD__WRITE(sdev, DOEPCTL(num), reg_val);
		
		/* FIFO mapping */
		sgadget->fparam.ep_param[num].bEpAddr 	= 0;
		sgadget->fparam.ep_param[num].wSize 	= 0;
		sgadget->fparam.ep_param[num].wNumbuf	= SCD_FIFO_NUMBUF;
		
		/* ֽ */
		sgadget->gep[num].bState	= SCD_EP_IDLE;
		sgadget->gep[num].bSubState	= SCD_EPSUB_IDLE;
	}
	/* ꥫ󥿥å */
	/* EP0ʳdummy椨FIFO_MAPѤEP0ΤߤͭȤ롣
	 */
	sgadget->fparam.cnt	= SCD_SKIP_EP0;	/* possible to set ep info*/
	SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);
	DBG(2,"==>\n");
}

/*-------------------------------------------------------
	name	: reset_clr_int()
	argment	: struct scd_gadget *sgadget
	return	: none
	description: clear register interrupt bits for gadget 
	date	: 2010/05/07
-------------------------------------------------------*/
static void reset_clr_int(struct scd_gadget *sgadget)
{
	struct scd_device *sdev=sgadget->g_device;
	u32 clr_val;
	int i;
	unsigned long flags;

	SCD_G_SPIN_LOCK_IRQ(&sgadget->gLock, flags);
	/* GINTSTSװꥢ */
	clr_val = 	SCD_B_GINTSTS_ENUMDONE|
				SCD_B_GINTSTS_USBSUSP|
				SCD_B_GINTSTS_ERLYSUSP|
				SCD_B_GINTSTS_WKUPINT|
				SCD_B_GINTSTS_RESETDET|
				SCD_B_GINTSTS_INCOMPISOO|
				SCD_B_GINTSTS_INCOMPISOI|
				SCD_B_GINTSTS_ISOOUTDROP|
				SCD_B_GINTSTS_GOUTNAKEFF|
				SCD_B_GINTSTS_GINNAKEFF|
				SCD_B_GINTSTS_EOPF|
				SCD_B_GINTSTS_SOF;
	SCD__WRITE(sdev, GINTSTS, clr_val);

	/* DOEPINT(0)װꥢ */
	clr_val = 	SCD_B_DXEPINTN_NYETINTRPT|
				SCD_B_DXEPINTN_NAKINTRPT|
				SCD_B_DXEPINTN_BBLEERRINTRPT|
				SCD_B_DXEPINTN_AHBERR|
				SCD_B_DXEPINTN_EPDISBLD|
				SCD_B_DXEPINTN_XFERCOMPL;
	clr_val &=	~SCD_B_DXEPINTN_BACK2BACKSETUP;	/* 0ꥢ */

	SCD__WRITE(sdev, DOEPINT(SCD_G_EP0), clr_val);

	/* DIEPINT(0)װꥢ */
	clr_val = 	SCD_B_DXEPINTN_BBLEERRINTRPT|
				SCD_B_DXEPINTN_TXFEMP|
				SCD_B_DXEPINTN_INEPNAKEFF|
				SCD_B_DXEPINTN_INTKNTXFEMP|
				SCD_B_DXEPINTN_TIMEOUT|
				SCD_B_DXEPINTN_AHBERR|
				SCD_B_DXEPINTN_EPDISBLD|
				SCD_B_DXEPINTN_XFERCOMPL;

	SCD__WRITE(sdev, DIEPINT(SCD_G_EP0), clr_val);

	/* DOEPINTװꥢ */
	clr_val = 	SCD_B_DXEPINTN_NYETINTRPT|
				SCD_B_DXEPINTN_NAKINTRPT|
				SCD_B_DXEPINTN_BBLEERRINTRPT|
				SCD_B_DXEPINTN_AHBERR|
				SCD_B_DXEPINTN_EPDISBLD|
				SCD_B_DXEPINTN_XFERCOMPL;

	for(i=1;i<sgadget->bEps+1;i++){
		SCD__WRITE(sdev, DOEPINT(i), clr_val);
	}
	/* DIEPINTװꥢ */
	clr_val = 	SCD_B_DXEPINTN_NAKINTRPT|
				SCD_B_DXEPINTN_BBLEERRINTRPT|
				SCD_B_DXEPINTN_TXFEMP|
				SCD_B_DXEPINTN_INEPNAKEFF|
				SCD_B_DXEPINTN_INTKNTXFEMP|
				SCD_B_DXEPINTN_AHBERR|
				SCD_B_DXEPINTN_EPDISBLD|
				SCD_B_DXEPINTN_XFERCOMPL;

	for(i=1;i<sgadget->bEps+1;i++){
		SCD__WRITE(sdev, DIEPINT(i), clr_val);
	}
	SCD_G_SPIN_UNLOCK_IRQ(&sgadget->gLock, flags);
}


#ifdef SCD_POLL_ON
/*-------------------------------------------------------
	name	: scd_reset_poll_state
	argment	: struct scd_gadget *sgadget
			  int state
	return	: none
	description: reset state flag for polling check
	date	: 2010/05/17
-------------------------------------------------------*/
int scd_reset_poll_state(struct scd_gadget *sgadget, int state)
{
	struct scd_gep *gep;
	struct scd_gadget_wq *gevent;
	int i, rslt=0;

	DBG(2,"<==\n");
	for(i=0;i<sgadget->bEps+SCD_NUM_EP0;i++){
		gep = &sgadget->gep[i];
		gevent = &gep->wq;
		rslt = set_gwq_flag(gevent, state);
		if(rslt){
			ERR("error can't change to %x \n",state);
		}
	}
	DBG(2,"==>\n");
	return 0;
}
#endif	
