/*--------------------------------------------------------------------------*
 * drivers/usb/scd/ups_cxd41xx.c
 * 
 * USB port selector functions
 * 
 * Copyright 2011 Sony Corporation
 *
 * This file is part of the USB port selecter driver UPS.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *---------------------------------------------------------------------------*/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/io.h>
#include <linux/platform_device.h>
#include <linux/udif/types.h>
#include <asm/mach/irq.h>
#include <mach/hardware.h>
#include <mach/gpio.h>
#include <mach/gpio_port.h>

#include <linux/usb.h>
#include <linux/usb/scd/usb_otg_control.h>
#include <linux/usb/scd/ups.h>

#include "scd_otg.h"

#define DEVICE_NAME	"ups"
#define IRQ_NAME_SIZE	16

#define DRIVER_VERSION	"0.02"
#define DRIVER_AUTHOR	"Sony Corporation"
#define DRIVER_DESC	DEVICE_NAME " - USB port selector for cxd41xx"

#define UPS_DEFAULT_PORT	0 /* 0..UP_MAX_PORT_NUM */
#define FILTER_INTERVAL		10 /* [ms] */
#define FILTER_OUT_COUNT	3 /* filter time = interval * out_count */
#define CID_STAT_UNKNOWN	-1 /* not a-dev, not b-dev */

#ifndef CHAR_BIT
#define CHAR_BIT		8
#endif
/************************** macro, variable definitions ***********************/
#define REG_SET			(0x04)
#define REG_CLR			(0x08)
#define REG_STS			(0x0c)

/* PHY(MISCTOP) */
#define ADR_USB_PHY		(0xf3050000)
#define USB_ID_EXT		(0x40)
#define VA_USB_ID_EXT		IO_ADDRESSP(ADR_USB_PHY + USB_ID_EXT)
#define USB_ID_EXT_BIT		0x1

struct ups_device_port {
	u32 cid_check_count;
	s32 last_cid;
	s32 ref_cid;
	u32 vbus_check_count;
	enum usb_otg_vbus_stat last_vbus;
	enum usb_otg_vbus_stat ref_vbus;
};

static struct ups_device {
	struct platform_device *pdev;
	struct usb_otg_control *otgc;
	struct ups_port_descriptor *desc;
	struct ups_port_descriptor *saved_desc;
	struct usb_otg_control_ops *ops;
	struct timer_list detector;

	unsigned long sflags;

	u8 current_port;
	struct ups_device_port port[UPS_MAX_PORT_NUM];
} _ups_device;

/*
 *	ե饰
 */
/* ACTIVE : ʣݡͭư */
#define UPS_SFLAG_ACTIVE		0
/* PSUED_ACTIVE : scdˤϱƶͿupsΤstart/stop祻å */
#define UPS_SFLAG_PSEUD_ACTIVE		1
/* SUSPEND : ڥ */
#define UPS_SFLAG_SUSPENDED		2
/* USABLE : ѤƤIOݡȤΥӥå */
#define UPS_SFLAG_USABLE_IO_BASE	3
#define UPS_SFLAG_USABLE_CID(PN)	((PN) * 2 + UPS_SFLAG_USABLE_IO_BASE)
#define UPS_SFLAG_USABLE_VBUS(PN)	(UPS_SFLAG_USABLE_CID(PN) + 1)

static DEFINE_RWLOCK(ups_lock);

static irqreturn_t det_vbuscid(int irq, void *param);

inline struct ups_device *otgc_to_upsd(struct usb_otg_control *otgc)
{
	return &_ups_device;
}

/*
 *	OTG API Hooks
 */

static int get_port_info(struct usb_otg_control *control,
			 struct usb_otg_control_port_info *info)
{
	struct ups_device *upsd;

	read_lock(&ups_lock);

	upsd = otgc_to_upsd(control);

	if (upsd->desc) {
		info->nr = upsd->desc->port_num;
		info->current_port = upsd->current_port;
	}
	else {
		/* ǥץϿϥݡȿ0Ȥ֤ */
		info->nr = 0;
		info->current_port = 0;
	}

	read_unlock(&ups_lock);

	return 0;
}

static int select_port(struct usb_otg_control *control, unsigned int pn)
{
	int status;
	struct ups_device *upsd;

	status = 0;

	read_lock(&ups_lock);

	upsd = otgc_to_upsd(control);

	if (upsd->desc == NULL || upsd->desc->port_num == 1) {
		/* ǥץ̤ϿϿ줿ǥץΥݡȿ */
		/* ĤΤߤξϡ⤻ｪλ */
		goto select_port_end;
	}

	if (pn & ~0x1) {
		/* upsGPIO1bitǤʤᡢ */
		/* 1bit(0 or 1)ʾξ票顼Ȥ */
		status = -EINVAL;
		goto select_port_end;
	}

	status = gpio_set_data_bit(upsd->desc->select.gpio_port,
				   upsd->desc->select.gpio_bit, pn);
	if (status < 0) {
		goto select_port_end;
	}
	upsd->current_port = pn;

select_port_end:
	read_unlock(&ups_lock);

#define UPS_CID_FILTER_DELAY 35 /* CIDե륿35[ms]Ԥ */
	if (status == 0) {
		msleep(UPS_CID_FILTER_DELAY);
	}

	return status;
}

/* ioγߥե꡼ */
static void free_double_edged_irqs(struct ups_device *upsd,
				   unsigned int rise_irqn,
				   unsigned int fall_irqn)
{
	free_irq(rise_irqn, upsd);
	free_irq(fall_irqn, upsd);
}

/* ioξåߥꥯȤ */
static int request_double_edged_irqs(struct ups_device *upsd,
				     unsigned int rise_irqn,
				     unsigned int fall_irqn)
{
	int status;

	status = request_irq(rise_irqn, det_vbuscid, IRQF_DISABLED,
			     dev_name(&upsd->pdev->dev), upsd);
	if (status < 0) {
		goto request_double_edged_irqs_end;
	}
	status = request_irq(fall_irqn, det_vbuscid, IRQF_DISABLED,
			     dev_name(&upsd->pdev->dev), upsd);
	if (status < 0) {
		free_irq(rise_irqn, upsd);
		goto request_double_edged_irqs_end;
	}

request_double_edged_irqs_end:
	return status;
}

#define GPIO_PORT_0_MAX_BIT	16
#define GPIO_PORT_1_MAX_BIT	6

inline unsigned int get_irqoffset(struct ups_io *io)
{
	/* ü */
	/* ߥޥåפߥեåȤϼ̤ */
	/*   port0[0..15]:port1[0..6]:port0[16] */
	return (io->gpio_bit ==  GPIO_PORT_0_MAX_BIT ? 23 :
		(io->gpio_port << 4) + io->gpio_bit);
}

static void disable_io(struct ups_device *upsd, struct ups_io *io)
{
	unsigned int irqoffset;

	switch (io->type) {
	case UPS_IO_TYPE_MISC:
		free_double_edged_irqs(upsd, IRQ_USB_ID_RISE, IRQ_USB_ID_FALL);
		break;
	case UPS_IO_TYPE_GPIO:
	case UPS_IO_TYPE_GPIOX:
		irqoffset = get_irqoffset(io);
		free_double_edged_irqs(upsd,
				       IRQ_GPIO_RISE(irqoffset),
				       IRQ_GPIO_FALL(irqoffset));
		break;
	case UPS_IO_TYPE_HFIX:
	case UPS_IO_TYPE_LFIX:
	default:
		break;
	}
}

static int enable_io(struct ups_device *upsd, struct ups_io *io)
{
	int status;
	unsigned int irqoffset;

	status = 0;

	switch (io->type) {
	case UPS_IO_TYPE_MISC:
		status = request_double_edged_irqs(upsd,
						   IRQ_USB_ID_RISE,
						   IRQ_USB_ID_FALL);
		break;
	case UPS_IO_TYPE_GPIO:
	case UPS_IO_TYPE_GPIOX:
		irqoffset = get_irqoffset(io);
		status = request_double_edged_irqs(upsd,
						   IRQ_GPIO_RISE(irqoffset),
						   IRQ_GPIO_FALL(irqoffset));
		break;
	case UPS_IO_TYPE_HFIX:
	case UPS_IO_TYPE_LFIX:
	default:
		break;
	}

	return status;
}

static int stop_control(struct usb_otg_control *control)
{
	int status;
	unsigned char i;
	struct ups_device *upsd;

	status = 0;

	read_lock(&ups_lock);

	upsd = otgc_to_upsd(control);

	if (upsd->ops->stop_control &&
	    !test_bit(UPS_SFLAG_PSEUD_ACTIVE, &upsd->sflags)) {
		status = upsd->ops->stop_control(control);
		if (status < 0) {
			goto stop_control_err;
		}
		clear_bit(UPS_SFLAG_ACTIVE, &upsd->sflags);
	}

	if (upsd->desc == NULL || upsd->desc->port_num == 1) {
		goto stop_control_err;
	}

	for (i = 0; i < upsd->desc->port_num; i++) {
		/*  */
		if (test_bit(UPS_SFLAG_USABLE_CID(i), &upsd->sflags)) {
			disable_io(upsd, &upsd->desc->port[i].cid);
		}
		clear_bit(UPS_SFLAG_USABLE_CID(i), &upsd->sflags);
		if (test_bit(UPS_SFLAG_USABLE_VBUS(i), &upsd->sflags)) {
			disable_io(upsd, &upsd->desc->port[i].vbus);
		}
		clear_bit(UPS_SFLAG_USABLE_VBUS(i), &upsd->sflags);
	}

	del_timer(&upsd->detector);

stop_control_err:
	read_unlock(&ups_lock);

	return status;
}

static int start_control(struct usb_otg_control *control)
{
	int status;
	unsigned char i;
	struct ups_device *upsd;
	bool can_select;

	status = 0;
	can_select = true;

	read_lock(&ups_lock);

	upsd = otgc_to_upsd(control);

	if (upsd->desc == NULL || upsd->desc->port_num == 1 ||
	    test_bit(UPS_SFLAG_ACTIVE, &upsd->sflags)) {
		/* ǥץ̤ϿϿ줿ǥץΥݡȿ */
		/* ĤΤߡ뤤ϴstart_controlѤߤξϡ*/
		/* start_controlƽλ */
		can_select = false;
		goto start_control_end;
	}

	for (i = 0; i < upsd->desc->port_num; i++) {
		/* CID */
		upsd->port[i].cid_check_count = 0;
		upsd->port[i].last_cid = upsd->port[i].ref_cid =
			CID_STAT_UNKNOWN;

		/* Vbus */
		upsd->port[i].vbus_check_count = 0;
		upsd->port[i].last_vbus = upsd->port[i].ref_vbus =
			USB_OTG_VBUS_STAT_UNKNOWN;

		/* ߳ */
		status = enable_io(upsd, &upsd->desc->port[i].cid);
		if (status < 0) {
			dev_alert(&upsd->pdev->dev, "enable_io cid[%d] (%d)\n",
				  i, status);
			goto start_control_err2;
		}
		set_bit(UPS_SFLAG_USABLE_CID(i), &upsd->sflags);
		status = enable_io(upsd, &upsd->desc->port[i].vbus);
		if (status < 0) {
			dev_alert(&upsd->pdev->dev, "enable_io vbus[%d] (%d)\n",
				  i, status);
			goto start_control_err2;
		}
		set_bit(UPS_SFLAG_USABLE_VBUS(i), &upsd->sflags);
	}

start_control_end:
	read_unlock(&ups_lock);

	if (upsd->ops->start_control &&
	    !test_bit(UPS_SFLAG_PSEUD_ACTIVE, &upsd->sflags)) {
		status = upsd->ops->start_control(control);
		if (status < 0) {
			goto start_control_err0;
		}
	}

	if (can_select) {
		status = select_port(control, UPS_DEFAULT_PORT);
		if (status < 0) {
			goto start_control_err1;
		}
	}

	set_bit(UPS_SFLAG_ACTIVE, &upsd->sflags);

	/*  */
	mod_timer(&upsd->detector, jiffies);

	return 0;

start_control_err2:
	read_unlock(&ups_lock);
start_control_err1:
	stop_control(control);
start_control_err0:
	return status;
}

static struct usb_otg_control_ops ops_hooks = {
	.get_port_info = get_port_info,
	.select_port = select_port,
	.start_control = start_control,
	.stop_control = stop_control,
};

/* CID, VbusʳscdΥ٥Ȥ򥹥롼 */
static int notify_hook(union usb_otg_event *event)
{
	int status;
	struct ups_device *upsd;

	status = 0;
	/* otgc_to_upsdؿؤΰǤotg_coreϻȽʤᡢ*/
	/* NULLͿ */
	upsd = otgc_to_upsd(NULL);

	if (upsd->otgc &&
	    event->type != USB_OTG_EVENT_TYPE_CID &&
	    event->type != USB_OTG_EVENT_TYPE_VBUS) {
		status = usb_otg_core_notify(upsd->otgc->otg_core, event);
	}

	return status;
}

/*
 *	CID Vbus detector
 */

static irqreturn_t det_vbuscid(int irq, void *param)
{
	struct ups_device *upsd;

	upsd = (struct ups_device *)param;

	if (timer_pending(&upsd->detector) ||
	    !test_bit(UPS_SFLAG_ACTIVE, &upsd->sflags)) {
		goto det_vbuscid_end;
	}
	mod_timer(&upsd->detector, jiffies);

det_vbuscid_end:
	return IRQ_HANDLED;
}

bool notify_cid(struct ups_device *upsd, u32 pn)
{
	int status;
	unsigned int gpio_cid;
	union usb_otg_event otg_event;
	struct ups_device_port *upsp;
	struct ups_io *cid;
	int current_cid;
	bool checking;

	cid = &upsd->desc->port[pn].cid;
	switch (cid->type) {
	case UPS_IO_TYPE_HFIX:
		current_cid = 1;
		break;
	case UPS_IO_TYPE_LFIX:
		current_cid = 0;
		break;
	case UPS_IO_TYPE_MISC:
		current_cid = readl(VA_USB_ID_EXT) & USB_ID_EXT_BIT ? 1 : 0;
		break;
	case UPS_IO_TYPE_GPIO:
		status = gpio_get_data_bit(cid->gpio_port, cid->gpio_bit,
					   &gpio_cid);
		if (status != UDIF_ERR_OK) {
			return false;
		}
		current_cid = (int)gpio_cid;
		break;
	case UPS_IO_TYPE_GPIOX:
		status = gpio_get_data_bit(cid->gpio_port, cid->gpio_bit,
					   &gpio_cid);
		if (status != UDIF_ERR_OK) {
			return false;
		}
		current_cid = gpio_cid ? 0 : 1;
		break;
	default:
		return false;
	}

	upsp = &upsd->port[pn];
	checking = true;

	if (current_cid) {
		/* A-Plugǻ¨ */
		if (current_cid != upsp->last_cid) {
			upsp->last_cid = current_cid;
			/* CID OTG٥ȥʥ */
			otg_event.type = USB_OTG_EVENT_TYPE_CID;
			otg_event.cid.value = current_cid;
			otg_event.cid.port = pn;
			usb_otg_core_notify(upsd->otgc->otg_core, &otg_event);
		}
		upsp->cid_check_count = 0;
		checking = false;
	} else if (current_cid != upsp->last_cid) {
		if (current_cid == upsp->ref_cid) {
			/* A-Plug³ե륿 */
			upsp->cid_check_count++;
			if (upsp->cid_check_count >= FILTER_OUT_COUNT) {
				upsp->last_cid = current_cid;
				upsp->cid_check_count = 0;
				/* CID OTG٥ȥʥ */
				otg_event.type = USB_OTG_EVENT_TYPE_CID;
				otg_event.cid.value = current_cid;
				otg_event.cid.port = pn;
				usb_otg_core_notify(upsd->otgc->otg_core,
						    &otg_event);
				checking = false;
			}
		}
		else {
			upsp->cid_check_count = 0;
		}
	} else {
		/* CIDѲ̵ϥ󥿥ꥻå */
		upsp->cid_check_count = 0;
		checking = false;
	}
	upsp->ref_cid = current_cid;

	return checking;
}

bool notify_vbus(struct ups_device *upsd, unsigned int pn)
{
	int status;
	unsigned int data;
	union usb_otg_event otg_event;
	struct ups_device_port *upsp;
	struct ups_io *io;
	enum usb_otg_vbus_stat current_vbus;
	enum usb_otg_vbus_stat notify_vbus;
	bool checking;

	io = &upsd->desc->port[pn].vbus;
	switch (io->type) {
	case UPS_IO_TYPE_HFIX:
		current_vbus = USB_OTG_VBUS_STAT_SESS;
		break;
	case UPS_IO_TYPE_GPIO:
	case UPS_IO_TYPE_GPIOX:
		status = gpio_get_data_bit(io->gpio_port, io->gpio_bit, &data);
		if (status != UDIF_ERR_OK) {
			return false;
		}
		if (io->type == UPS_IO_TYPE_GPIOX) {
			data = !data;
		}
		current_vbus = (data ?
			USB_OTG_VBUS_STAT_SESS : USB_OTG_VBUS_STAT_OFF);
		break;
	default:
		return false;
	}

	upsp = &upsd->port[pn];
	checking = true;

	if (upsp->last_vbus > current_vbus) {
		/* ǻ¨ */
		/* Vbus OTG٥ȥʥ */
		otg_event.type = USB_OTG_EVENT_TYPE_VBUS;
		notify_vbus = upsp->last_vbus;
		do {
			notify_vbus--;
			otg_event.vbus.vbus_stat = notify_vbus;
			otg_event.vbus.port = pn;
			usb_otg_core_notify(upsd->otgc->otg_core, &otg_event);
		} while (notify_vbus > current_vbus);
		/* åΤVbusͤlast_vbusسǼ */
		upsp->last_vbus = notify_vbus;
		upsp->vbus_check_count = 0;
		checking = false;
	} else if (upsp->last_vbus < current_vbus) {
		if (current_vbus == upsp->ref_vbus) {
			/* ³ե륿 */
			upsp->vbus_check_count++;
			if (upsp->vbus_check_count >= FILTER_OUT_COUNT) {
				/* Vbus OTG٥ȥʥ */
				otg_event.type = USB_OTG_EVENT_TYPE_VBUS;
				notify_vbus = upsp->last_vbus;
				do {
					notify_vbus++;
					otg_event.vbus.vbus_stat = notify_vbus;
					otg_event.vbus.port = pn;
					usb_otg_core_notify(upsd->otgc->otg_core,
							    &otg_event);
				} while (notify_vbus < current_vbus);
				/* åΤVbusͤlast_vbusسǼ */
				upsp->last_vbus = notify_vbus;
				upsp->vbus_check_count = 0;
				checking = false;
			}
		} else {
			upsp->vbus_check_count = 0;
		}
	} else {
		/* VbusѲ̵ϥ󥿥ꥻå */
		upsp->vbus_check_count = 0;
		checking = false;
	}
	upsp->ref_vbus = current_vbus;

	return checking;
}

/*
 * ؿdetector_handler
 * CID/Vbusƻؿ(ޥϥɥ)
 */

static void detector_handler(unsigned long ptr)
{
	unsigned char i;
	bool checking_vbuscid;
	struct ups_device *upsd;

	upsd = (struct ups_device *)ptr;

	checking_vbuscid = false;

	for (i = 0; i < upsd->desc->port_num; i++) {
		checking_vbuscid |= notify_cid(upsd, i);
		checking_vbuscid |= notify_vbus(upsd, i);
	}

	/* νԤ줿stop_controlACTIVEˤʤä */
	/* ΥޡưϹԤʤ */
	if (!checking_vbuscid ||
	    !test_bit(UPS_SFLAG_ACTIVE, &upsd->sflags)) {
		return;
	}

	mod_timer(&upsd->detector, jiffies + FILTER_INTERVAL * HZ / 1000);
}

/*
 *	UPS APIs
 */
/* ̾OTG control */
struct usb_otg_control *ups_lookup_control_by_name(const char* name)
{
	return scd_otg_lookup_control_by_name(name);
}
EXPORT_SYMBOL(ups_lookup_control_by_name);

static int open_io(struct ups_device *upsd, struct ups_io *io)
{
	int status;

	status = 0;

	switch (io->type) {
	case UPS_IO_TYPE_GPIO:
	case UPS_IO_TYPE_GPIOX:
		if (io->gpio_port != GPIO_PORT_0 &&
		    io->gpio_port != GPIO_PORT_1) {
			/* ü */
			/* port0 or port1Τ߲߳ǽΤᡢ*/
			/* ¾ΥݡȤꤵ줿¸ߤʤݡȤ */
			/* ꤷȤƥ顼 */
			status = -ENOENT;
			goto open_io_end;
		}
		if ((io->gpio_port == GPIO_PORT_0
		     && io->gpio_bit > GPIO_PORT_0_MAX_BIT) ||
		    (io->gpio_port == GPIO_PORT_1
		     && io->gpio_bit > GPIO_PORT_1_MAX_BIT)) {
			status = -EBADF;
			goto open_io_end;
		}
		status = gpio_set_mode_bit(io->gpio_port, io->gpio_bit,
					   GPIOMODE_INPUT);
		if (status != UDIF_ERR_OK) {
			status = -EIO;
			goto open_io_end;
		}
		break;
	case UPS_IO_TYPE_MISC:
	case UPS_IO_TYPE_HFIX:
	case UPS_IO_TYPE_LFIX:
		break;
	default:
		status = -EIO;
		goto open_io_end;
	}

open_io_end:
	return status;
}

static void close_io(struct ups_device *upsd, struct ups_io *io)
{
	switch (io->type) {
	case UPS_IO_TYPE_MISC:
	case UPS_IO_TYPE_GPIO:
	case UPS_IO_TYPE_GPIOX:
	case UPS_IO_TYPE_HFIX:
	case UPS_IO_TYPE_LFIX:
	default:
		break;
	}
}

void ups_unregist_descriptor(struct usb_otg_control *control)
{
	int status;
	unsigned long flags;
	unsigned char i;
	struct ups_device *upsd;

	status = 0;

	upsd = otgc_to_upsd(control);
	if (upsd == NULL || upsd->desc == NULL) {
		return;
	}

	if (test_bit(UPS_SFLAG_ACTIVE, &upsd->sflags)) {
		/* PSEUD_ACTIVEե饰ΩƤƤ*/
		/* scd:stop_controlƤӽФ򤷡upsΤstop */
		set_bit(UPS_SFLAG_PSEUD_ACTIVE, &upsd->sflags);
		stop_control(control);
	}

	/* ǥץϿ˥ޡưʤȤݾڤ */
	del_timer_sync(&upsd->detector);

	write_lock_irqsave(&ups_lock, flags);

	upsd->otgc = NULL;

	for (i = 0; i < upsd->desc->port_num; i++) {
		close_io(upsd, &upsd->desc->port[i].cid);
		close_io(upsd, &upsd->desc->port[i].vbus);
	}

	scd_otg_hook_notify(control, NULL);
	scd_otg_hook_ops(control, NULL);

	upsd->desc = NULL;
	upsd->ops = NULL;

	write_unlock_irqrestore(&ups_lock, flags);
}
EXPORT_SYMBOL(ups_unregist_descriptor);

int ups_regist_descriptor(struct usb_otg_control *control,
			  struct ups_port_descriptor *desc)
{
	int status;
	unsigned long flags;
	unsigned char i;
	struct ups_device *upsd;

	if (desc == NULL) {
		return -EINVAL;
	}
	if (desc->port_num > UPS_MAX_PORT_NUM) {
		return -E2BIG;
	}

	write_lock_irqsave(&ups_lock, flags);

	upsd = otgc_to_upsd(control);
	if (upsd == NULL) {
		status = -ENODEV;
		goto ups_regist_descriptor_err1;
	}

	/* ǥץϿѤߤξ */
	if (upsd->desc != NULL) {
		status = -EALREADY;
		goto ups_regist_descriptor_err1;
	}

	/* ups insmodscdưξ */
	if (usb_otg_control_get_mode(control) != USB_OTG_CONTROL_STOP) {
		status = -EINPROGRESS;
		goto ups_regist_descriptor_err1;
	}
	/* ľưˤؤ餺ǥץϿ줿 */
	if (test_bit(UPS_SFLAG_ACTIVE, &upsd->sflags)) {
		status = -EBUSY;
		goto ups_regist_descriptor_err1;
	}

	clear_bit(UPS_SFLAG_PSEUD_ACTIVE, &upsd->sflags);
	upsd->otgc = control;

	for (i = 0; i < desc->port_num; i++) {
		status = open_io(upsd, &desc->port[i].cid);
		if (status < 0) {
			goto ups_regist_descriptor_err0;
		}
		status = open_io(upsd, &desc->port[i].vbus);
		if (status < 0) {
			goto ups_regist_descriptor_err0;
		}
	}

	status = gpio_set_mode_bit(desc->select.gpio_port,
				   desc->select.gpio_bit, GPIOMODE_OUTPUT_L);
	if (status != UDIF_ERR_OK) {
		status = -EIO;
		goto ups_regist_descriptor_err0;
	}

	upsd->desc = desc;

	status = scd_otg_hook_notify(control, notify_hook);
	if (status < 0) {
		goto ups_regist_descriptor_err0;
	}

	upsd->ops = scd_otg_hook_ops(control, &ops_hooks);
	if (upsd->ops == NULL) {
		status = -EINVAL;
		goto ups_regist_descriptor_err0;
	}
	/* upsѰդƤʤOTG APIؿscdδؿ򤽤Τޤ޻Ѥ */
	for (i = 0; i < sizeof(ops_hooks) / sizeof(void *); i++) {
		if (((void **)&ops_hooks)[i] == NULL) {
			((void **)&ops_hooks)[i] = ((void **)upsd->ops)[i];
		}
	}

ups_regist_descriptor_err1:
	write_unlock_irqrestore(&ups_lock, flags);
	goto ups_regist_descriptor_end;
ups_regist_descriptor_err0:
	write_unlock_irqrestore(&ups_lock, flags);
	ups_unregist_descriptor(control);
ups_regist_descriptor_end:
	return status;
}
EXPORT_SYMBOL(ups_regist_descriptor);

/*
 *	platform bus operations
 */
static int probe(struct platform_device *pdev)
{
	struct ups_device *upsd;

	upsd = &_ups_device;	/* device */

	platform_set_drvdata(pdev, upsd);

	init_timer(&upsd->detector);
	upsd->detector.function = detector_handler;
	upsd->detector.data = (unsigned long)upsd;

	return 0;
}

static int remove(struct platform_device *pdev)
{
	struct ups_device *upsd;

	upsd = (struct ups_device *)platform_get_drvdata(pdev);

	ups_unregist_descriptor(upsd->otgc);

	platform_set_drvdata(pdev, NULL);

	return 0;
}

static int suspend(struct platform_device *pdev, pm_message_t message)
{
	struct ups_device *upsd;

	upsd = (struct ups_device *)platform_get_drvdata(pdev);

	upsd->saved_desc = upsd->desc;
	if (test_bit(UPS_SFLAG_ACTIVE, &upsd->sflags)) {
		set_bit(UPS_SFLAG_PSEUD_ACTIVE, &upsd->sflags);
	}

	ups_unregist_descriptor(upsd->otgc);

	set_bit(UPS_SFLAG_SUSPENDED, &upsd->sflags);

	return 0;
}

static int resume(struct platform_device *pdev)
{
	struct ups_device *upsd;

	upsd = (struct ups_device *)platform_get_drvdata(pdev);

	if (upsd->saved_desc) {
		ups_regist_descriptor(upsd->otgc, upsd->saved_desc);
		upsd->saved_desc = NULL;
	}
	if (test_bit(UPS_SFLAG_PSEUD_ACTIVE, &upsd->sflags)) {
		start_control(upsd->otgc);
		clear_bit(UPS_SFLAG_PSEUD_ACTIVE, &upsd->sflags);
	}

	clear_bit(UPS_SFLAG_SUSPENDED, &upsd->sflags);

	return 0;
}

static struct platform_driver ups_platform_driver = {
	.probe = probe,
	.remove = remove,
	.suspend = suspend,
	.resume = resume,
	.driver = {
		.name = DEVICE_NAME,
		.owner = THIS_MODULE,
	},
};

/*
 *	module initialize and exit
 */
int __init ups_module_init(void)
{
    	/* Platform deviceȤϿ */
	_ups_device.pdev = platform_device_register_simple(DEVICE_NAME, -1,
							   NULL, 0);
	if (_ups_device.pdev == NULL) {
		err("cannot register device\n");
		return -ENODEV;
	}

	return platform_driver_register(&ups_platform_driver);
}
module_init(ups_module_init);

void __exit ups_module_exit(void)
{
	platform_driver_unregister(&ups_platform_driver);

    	/* Platform deviceȤϿ */
	platform_device_unregister(_ups_device.pdev);
}
module_exit(ups_module_exit);

MODULE_DESCRIPTION(DRIVER_DESC DRIVER_VERSION);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_LICENSE("GPL");
