/*
 *  linux/fs/proc/stack_usage.c
 *
 *  proc interface for CONFIG_DEBUG_STACK_USAGE
 *
 *  Copyright 2010 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/pid_namespace.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/sched.h>

#define PROC_NAME "stack_usage"

static void *stackusage_start(struct seq_file *m, loff_t *pos)
{
	pid_t pid = (pid_t)m->private;

	if (0 == pid) {
		return NULL;
	}
	return pos;
}

static void *stackusage_next(struct seq_file *m, void *v, loff_t *pos)
{
	pid_t pid = (pid_t)m->private;

	(*pos)++;
	if (pid > 0) {
		pid++;
		m->private = (void *)pid;
		return pos;
	} else {
		return NULL;
	}
}

static void stackusage_stop(struct seq_file *m, void *v)
{
}

static int stackusage_show(struct seq_file *m, void *v)
{
	pid_t pid = (pid_t)m->private;
	struct task_struct *task;
	char comm[sizeof(task->comm)];
	unsigned long free;
#ifdef CONFIG_IRQSTACK
	extern int show_irqstack_usage(struct seq_file *m, void *v);
#endif
	rcu_read_lock();
	task = pid_task(find_ge_pid(pid, current->nsproxy->pid_ns),
			PIDTYPE_PID);
	if (task) {
		get_task_struct(task);
	}
	rcu_read_unlock();

	if (task) {
		pid = task->pid;
		free = stack_not_used(task);
		get_task_comm(comm, task);
		put_task_struct(task);
		seq_printf(m, "%5d %-20.20s %5lu\n", pid, comm, free);
	} else {
#ifdef CONFIG_IRQSTACK
		show_irqstack_usage(m, v);
#endif /* CONFIG_IRQSTACK */
		pid = 0; /* EOF */
	}
	m->private = (void *)pid;
	return 0;
}

static const struct seq_operations stackusage_ops = {
	.start	= stackusage_start,
	.next	= stackusage_next,
	.stop	= stackusage_stop,
	.show	= stackusage_show,
};

static int stackusage_open(struct inode *inode, struct file *filp)
{
	int ret;
	struct seq_file *m;

	ret = seq_open(filp, &stackusage_ops);
	if (ret) {
		return ret;
	}
	m = (struct seq_file *)filp->private_data;
	m->private = (void *)1;
	return 0;
}

static const struct file_operations proc_stackusage_ops = {
	.open	= stackusage_open,
	.read	= seq_read,
	.llseek	= seq_lseek,
	.release = seq_release,
};

static int __init stackusage_init(void)
{
	proc_create(PROC_NAME, S_IRUSR, NULL, &proc_stackusage_ops);
        return 0;
}
module_init(stackusage_init);
