/* 2010-06-17: File added and changed by Sony Corporation */
/*
 *  lib/ssboot_eval.c
 *
 *  Copyright 2005-2008 Sony Corporation
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 *  02110-1301, USA.
 */
#include <linux/suspend.h>
#include <linux/mm.h>
#include <linux/mmzone.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/sysfs.h>
#include <linux/pm.h>
#include <linux/vmalloc.h>
#include <linux/dma-mapping.h>
#include <linux/kobject.h>
#include <asm/page.h>
#include <asm/sections.h>
#include <asm/uaccess.h>
#include <asm/io.h>

/* image creation mode */
#define IMG_MODE_NONE	0
#define IMG_MODE_MIN	1
#define IMG_MODE_MAX	2
#define IMG_MODE_NUM	3

/* count phase */
#define COUNT_START	0
#define COUNT_DOING	1
#define COUNT_END	2

#define ssboot_attr(_name)			\
static struct kobj_attribute _name##_attr = {	\
	.attr	= {				\
		.name = __stringify(_name),	\
		.mode = 0644,			\
	},					\
	.show	= _name##_show,			\
	.store	= _name##_store,		\
}

/* section list */
typedef struct img_section {
	u_int32_t virt;
	u_int32_t phys;
	u_int32_t len;
} img_section_t;

extern struct kset power_subsys;
static unsigned long cur_phys, cur_virt, cur_sect;
static unsigned long num_sect, num_pages, order_sect;
static unsigned int  img_mode = IMG_MODE_MIN;
static img_section_t *img_sect;

typedef int (*count_cb_t)(resource_size_t phys, unsigned long virt,
			  unsigned long state);

static unsigned long
is_free_region(unsigned long pfn)
{
	struct page *page = pfn_to_page(pfn);
        struct zone *zone = page_zone(page);
	struct list_head *curr;
	int order;
	int type;

	for_each_migratetype_order(order, type) {
		list_for_each(curr,
			      &zone->free_area[order].free_list[type]) {
			if (page == list_entry(curr, struct page, lru)) {
				return (1 << order);
			}
		}
	}

        return 0;
}

static u_int32_t
is_saveable(unsigned long pfn, unsigned long *num_free)
{
	extern char __nosave_begin[];
	extern char __nosave_end[];
	unsigned long pfn_stext, pfn_etext;
	unsigned long pfn_srodata, pfn_erodata;
	unsigned long pfn_snosave, pfn_enosave;
	unsigned long pfn_end;

	pfn_stext   =            __pa(&_stext)         >> PAGE_SHIFT;
	pfn_etext   = PAGE_ALIGN(__pa(&_etext))        >> PAGE_SHIFT;
	pfn_srodata =            __pa(&__start_rodata) >> PAGE_SHIFT;
	pfn_erodata = PAGE_ALIGN(__pa(&__end_rodata))  >> PAGE_SHIFT;
	pfn_snosave =            __pa(&__nosave_begin) >> PAGE_SHIFT;
	pfn_enosave = PAGE_ALIGN(__pa(&__nosave_end))  >> PAGE_SHIFT;
	pfn_end     = PAGE_ALIGN(__pa(&_end))          >> PAGE_SHIFT;

	*num_free = 0;

	/* sanity check */
	if (!pfn_valid(pfn)) {
		return 0;
	}

	/* kernel nosave -> not saved */
	if (pfn >= pfn_snosave && pfn < pfn_enosave) {
		return 0;
	}

	/* smaller than kernel start address -> not saved */
	if (pfn < pfn_stext) {
		return 0;
	}

	/* kernel text/data/bss -> saved */
	if (pfn <= pfn_end) {
		return 1;
	}

	/* free pages -> not saved */
	*num_free = is_free_region(pfn);
	if (*num_free > 0) {
		return 0;
	}

	/* all other pages -> saved */
	return 1;
}

static int
count_pages(count_cb_t callback)
{
	struct zone *zone;
	unsigned long pfn;
	unsigned long virt;
	unsigned long phys;
	unsigned long free = 0;
	int ret;

	/* prepare */
	ret = callback(0, 0, COUNT_START);
	if (ret < 0) {
		return ret;
	}

	/* count pages */
	for_each_zone(zone) {
		if (is_highmem(zone)) {
			continue;
		}

		for (pfn = zone->zone_start_pfn;
		     pfn < zone->zone_start_pfn + zone->spanned_pages; pfn++) {
			ret = is_saveable(pfn, &free);
			if (ret == 0) {
				/* skip free pages */
				if (free > 0) {
					pfn += free - 1;
				}
			} else {
				/* page is saved */
				virt = (unsigned long)page_address(pfn_to_page(pfn));
				phys = virt_to_phys((void *)virt);
				ret = callback(phys, virt, COUNT_DOING);
				if (ret < 0) {
					return ret;
				}
			}
		}
	}

	/* finalize */
	ret = callback(0, 0, COUNT_END);
	if (ret < 0) {
		return ret;
	}

	return 0;
}

static void
free_section_list(void)
{
	if (img_sect) {
		free_pages((unsigned long)img_sect, order_sect);
		img_sect = NULL;
	}
}

static int
alloc_section_list(resource_size_t phys, unsigned long virt, int state)
{
	switch (state) {
	case COUNT_START:
		cur_phys = -1UL;
		cur_virt = -1UL;
		num_sect = 0;
		break;
	case COUNT_DOING:
		if (phys != (cur_phys + PAGE_SIZE) ||
		    virt != (cur_virt + PAGE_SIZE)) {
			num_sect++;
		}
		cur_phys = phys;
		cur_virt = virt;
		break;
	case COUNT_END:
		/* free old section list if allocated */
		free_section_list();

		/* allocate section list */
		order_sect = (PAGE_ALIGN(num_sect * sizeof(img_section_t))
			      >> PAGE_SHIFT);
		order_sect = get_bitmask_order(order_sect - 1);
		img_sect = (img_section_t *)__get_free_pages(GFP_ATOMIC,
							     order_sect);
		if(img_sect == NULL){
			printk(KERN_ERR "ssboot: cannot allocate section list\n");
			return -ENOMEM;
		}
		break;
	default:
		BUG();
		return -1;
	}

	return 0;
}

static int
fill_section_list(resource_size_t phys, unsigned long virt, int state)
{
	switch (state) {
	case COUNT_START:
		cur_phys = -1UL;
		cur_virt = -1UL;
		cur_sect = -1UL;
		num_sect = 0;
		num_pages = 0;
		break;
	case COUNT_DOING:
		if (phys != (cur_phys + PAGE_SIZE) ||
		    virt != (cur_virt + PAGE_SIZE)) {
			num_sect++;
			cur_sect++;
			img_sect[cur_sect].virt = virt;
			img_sect[cur_sect].phys = phys;
			img_sect[cur_sect].len = 0;
		}
		img_sect[cur_sect].len += PAGE_SIZE;
		cur_phys = phys;
		cur_virt = virt;
		num_pages++;
		break;
	case COUNT_END:
		break;
	default:
		BUG();
		return -1;
	}

	return 0;
}

static int show_section_list(void)
{
	int i;
	int ret = 0;

	/* estimate sections and allocate memory */
	ret = count_pages((count_cb_t)alloc_section_list);
	if(ret < 0) {
		return ret;
	}

	/* create section list */
	ret = count_pages((count_cb_t)fill_section_list);
	if (ret < 0) {
		return ret;
	}

	/* dump section list */
	for (i = 0; i < num_sect; i++) {
		printk("ssboot: Section #%04d: 0x%08x - 0x%08x (%5ld pages)\n",
		       i,
		       img_sect[i].virt,
		       img_sect[i].virt + img_sect[i].len - 1,
		       img_sect[i].len / PAGE_SIZE);
	}
	printk("ssboot: Total: %ld pages (%ldMB)\n",
	       num_pages, num_pages * PAGE_SIZE / 1024 / 1024);

	return 0;
}

#define NR_SHRINK_PAGES	0x100000

static void free_memory(void)
{
	unsigned long num = 0;

	/* free memory */
	if (img_mode == IMG_MODE_MIN) {
		printk("Freeing memory...");
		num = shrink_all_memory(NR_SHRINK_PAGES);
		printk("done (%ldMB)\n", num * 4 /1024);
	}
}

/*
 * sysfs interfaces
 */
struct kobject *ssboot_kobj;

static ssize_t eval_show(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
	return 0;
}

static ssize_t eval_store(struct kobject *kobj, struct kobj_attribute *attr,
			  const char *buf, size_t n)
{
	free_memory();

	local_irq_disable();
	show_section_list();
	local_irq_enable();

	return n;
}
ssboot_attr(eval);

static ssize_t mode_show(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
	int ret;

	switch(img_mode) {
	case IMG_MODE_NONE:
		ret = sprintf(buf, "%u (None)\n", img_mode);
		break;
	case IMG_MODE_MIN:
		ret = sprintf(buf, "%u (Min)\n", img_mode);
		break;
	case IMG_MODE_MAX:
		ret = sprintf(buf, "%u (Max)\n", img_mode);
		break;
	default:
		ret = -EINVAL;
	}
	return ret;
}

static ssize_t mode_store(struct kobject *kobj, struct kobj_attribute *attr,
			  const char *buf, size_t n)
{
	unsigned char val;

	val = *buf - '0';

	if (val < IMG_MODE_NUM) {
		img_mode = val;
		return n;
	}
	printk(KERN_ERR "ssboot: invalid mode: %d\n", val);

	return -EINVAL;
}
ssboot_attr(mode);

static struct attribute *ssboot_attrs[] = {
	&eval_attr.attr,
	&mode_attr.attr,
	NULL,
};

static struct attribute_group ssboot_attr_group = {
	.attrs = ssboot_attrs,
};

static int __init ssboot_eval_init(void)
{
	int ret;

	ssboot_kobj = kobject_create_and_add("ssboot", NULL);
	if (!ssboot_kobj) {
		printk(KERN_WARNING "ssboot kobject create failed \n");
		return -EINVAL;
	}

	ret = sysfs_create_group(ssboot_kobj, &ssboot_attr_group);
	if (ret < 0) {
		return ret;
	}

	return 0;
}
core_initcall(ssboot_eval_init);
